/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

	File   		:	decoup.c

	Date		:	18 01 1994
	Version		:	1

	Auteur      :   BEAURIN oliver

	Description :   Premier essais de decoupage d'une netlist en colonne.

\*********************************************************************/
/*
  first just for statistics
  08 01 94 : First implementation.
  17 01 94 : Vdd and vss bug fixed.
  21 01 94 : connnector file implementation.
  24 01 94 : implementation with x +- width >> 1 for cell stats.
  24 01 94 : lofigchain implementation for connectors statistics.
  26 01 94 : first implementation of decoup process
             idea : use statistics for open vertical channel ... (just for see)
  10 02 94 : make one cell per columns
  14 03 94 : subdivided make_channel on three functions
  14 03 94 : add one param : max cells per colums.
  05 04 94 : add one option : sort each line of cells
  10 06 94 : add banner
  10 06 94 : supress debug options : all weight, sort, horizontal channel
  29 09 94 : first implementation in dpp
*/

/*
   Hyp : - all cells in dplib have all there transparences
         - the placement don't generate SYM_X cells
         _ ref_name : con_name + _nnn
         _ all cells must have the same height
*/

/*
   Bug : - phcol have multi vdd and vss with same name !!!
         - Verif sur les connecteurs : reference ou vrai connecteur
*/

/*---------------------------------------------------------\
                     les includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include MUT_H
#include MLO_H
#include MPH_H
#include APR_H

#include GFP_H

#include "dpr.h"
#include "placer_decoup.h"

/*---------------------------------------------------------\
                     Les defines
\---------------------------------------------------------*/

#define BASE_CON_NAME     "new_con_name"
#define DEBUG_VAR         "DEBUG_DECOUP"
#define DECOUP_NAME       "decoup"

#define MAX(x,y) ( ( ( x ) < ( y ) ) ? ( y ) : ( x ) )

/*---------------------------------------------------------\
                     getchainlenght
\---------------------------------------------------------*/
long getchainlenght( pt_chainphcol )
chain_list *pt_chainphcol;
{
    chain_list *ptchain;
    long        nbcolumns;

    for( nbcolumns = 0, ptchain = pt_chainphcol;
         ptchain;
         nbcolumns++,   ptchain = ptchain->NEXT );

    return nbcolumns;

} /* end of getchainlenght */

/*---------------------------------------------------------\
                     found_vdd_width
\---------------------------------------------------------*/
unsigned long found_vdd_width( figname )
char *figname;
{
    phfig_list *pt_model;
    phcon_list *pt_phcon;
    
    pt_model = getphfig( figname, 'P' );

    for( pt_phcon = pt_model->PHCON; pt_phcon; pt_phcon = pt_phcon->NEXT )
    {
        if( isvdd( pt_phcon->NAME )
          &&( ( pt_phcon->ORIENT == EAST ) || ( pt_phcon->ORIENT == WEST ) ) )
        {
            return pt_phcon->WIDTH;
         }
    }
    return( 8L * SCALE_X );
} /* end of found_vdd_width */

/*---------------------------------------------------------\
                     found_vss_width
\---------------------------------------------------------*/
unsigned long found_vss_width( figname )
char *figname;
{
    phfig_list *pt_model;
    phcon_list *pt_phcon;
    
    pt_model = getphfig( figname, 'P' );

    for( pt_phcon = pt_model->PHCON; pt_phcon; pt_phcon = pt_phcon->NEXT )
    {
        if( isvss( pt_phcon->NAME )
          &&( ( pt_phcon->ORIENT == EAST ) || ( pt_phcon->ORIENT == WEST ) ) )
        {
            return pt_phcon->WIDTH;
         }
    }
    return( 8L * SCALE_X );
} /* end of found_vss_width */

/*---------------------------------------------------------\
                     found_height_of_cell
\---------------------------------------------------------*/
static unsigned long found_height_of_cell( cell_name )
char *cell_name;
{
    phfig_list *pt_model;
    unsigned long height;

    pt_model = getphfig( cell_name, 'P' );
	height = ( pt_model->YAB2 - pt_model->YAB1 ) / SCALE_X;

	return( height );

} /* fin de found_height_of_cell */

/*---------------------------------------------------------\
                     found_width_of_cell
\---------------------------------------------------------*/
static unsigned long found_width_of_cell( cell_name )
char *cell_name;
{
    phfig_list *pt_model;
    unsigned long width;

    pt_model = getphfig( cell_name, 'P' );
    width = ( pt_model->XAB2 - pt_model->XAB1 ) / SCALE_X;

    return( width );

} /* fin de found_width_of_cell */

/*---------------------------------------------------------\
                     check_cells_height
\---------------------------------------------------------*/
void check_cells_height( ptlofig )
lofig_list *ptlofig;
{
    phfig_list *phfig;
    loins_list *first_loins;
    loins_list *loins;
    long        first_height;
    long        height;

    first_loins = ptlofig->LOINS;

    if( ! first_loins )
        return;

    phfig        = getphfig( first_loins->FIGNAME, 'P' );
    first_height = phfig->YAB2 - phfig->YAB1;

    for( loins = first_loins->NEXT; loins; loins = loins->NEXT )
    {
        phfig  = getphfig( loins->FIGNAME, 'P' );
        height = phfig->YAB2 - phfig->YAB1;

        if( first_height != height )
        {
            fprintf( stderr,
                     "*** %s error *** : the models %s and %s don't have the same height.\n",
                     DECOUP_NAME, first_loins->FIGNAME, loins->FIGNAME );
            exit( 1 );
        }
    }
} /* fin de check_cells_height */

/*---------------------------------------------------------\
                     compute_cells_size
\---------------------------------------------------------*/
unsigned long compute_cells_size( pt_place )
placement_fig *pt_place;
{
    cell_list     *pt_cell;
    long           iCell;
    unsigned long  height;
    unsigned long  size;

    size = 0L;
	height = found_height_of_cell( pt_place->Rows[ 0 ].tail->ins->FIGNAME );

    for( pt_cell = pt_place->Cells, iCell = 0;
         iCell < pt_place->NbCells;
         iCell++, pt_cell++ )
    {
        size += height * pt_cell->width;
    }

    return size;
} /* fin de compute_cells_size */

/*---------------------------------------------------------\
                     compute_cols_size
\---------------------------------------------------------*/
unsigned long compute_cols_size( pt_headchainphcol )
chain_list *pt_headchainphcol;
{
    phfig_list    *pt_phcol;
    chain_list    *pt_chainphcol;
    unsigned long  size;

    size = 0L;

    for( pt_chainphcol = (chain_list *)pt_headchainphcol;
         pt_chainphcol;
         pt_chainphcol = pt_chainphcol->NEXT )
    {
        pt_phcol = (phfig_list *)pt_chainphcol->DATA;

       size += ( pt_phcol->YAB2 - pt_phcol->YAB1) / SCALE_X
              *( pt_phcol->XAB2 - pt_phcol->XAB1) / SCALE_X;
    }

    return size;

} /* fin de compute_cols_size */

/*---------------------------------------------------------\
                     compute_init_place_size
\---------------------------------------------------------*/
unsigned long compute_init_place_size( pt_place )
placement_fig *pt_place;
{
    long           iRow;
    unsigned long  height;
    unsigned long  size;
    unsigned long  max_width;
    unsigned long  width;

    size = 0L;
	height = found_height_of_cell( pt_place->Rows[ 0 ].tail->ins->FIGNAME );

    max_width = 0L;
    for( iRow = 0; iRow < pt_place->NbRows; iRow++)
    {
        width     = pt_place->Rows[ iRow ].length;
        max_width = ( max_width > width ) ? max_width : width;
    }

    size = max_width * height * pt_place->NbRows;

    return size;
} /* fin de compute_init_place_size */

/*---------------------------------------------------------\
                     fill_structure
\---------------------------------------------------------*/
static void fill_structure( pt_place )
placement_fig *pt_place;
{
    long iCell;
    long iRow;
    long width;
    cell_list *pt_cell;
    cell_list *pt_search_cell;

    for( iRow = 0; iRow < pt_place->NbRows; iRow++)
    {
        pt_place->Rows[ iRow ].head = NULL;
        pt_place->Rows[ iRow ].tail = NULL;
    }

    for( pt_cell = pt_place->Cells, iCell = 0;
         iCell < pt_place->NbCells;
         iCell++, pt_cell++ )
    {
        /* fill the lofig_list's USER field */
/*         pt_cell->ins->USER = (struct ptype *)pt_cell; */

        /* make a good placement struct */
        iRow = pt_cell->row;

        if( pt_place->Rows[ iRow ].head == NULL )
        {
            pt_cell->next               = NULL;
            pt_cell->prev               = NULL;
            pt_place->Rows[ iRow ].head = pt_cell;
            pt_place->Rows[ iRow ].tail = pt_cell;
        }
        else if( pt_place->Rows[ iRow ].head->x > pt_cell->x )
        {
            pt_cell->next               = pt_place->Rows[ iRow ].head;
            pt_cell->prev               = NULL;
            pt_place->Rows[ iRow ].head = pt_cell;
        }
        else
        {
            for( pt_search_cell = pt_place->Rows[ iRow ].head;
                 pt_search_cell->next && ( pt_search_cell->next->x < pt_cell->x );
                 pt_search_cell = pt_search_cell->next );

            if( ( pt_search_cell->next ) && ( pt_search_cell->next->x < pt_cell->x ) )
            {
                pt_cell->next               = pt_search_cell->next;
                pt_cell->prev               = pt_search_cell;
                pt_search_cell->next->prev  = pt_cell;
                pt_search_cell->next        = pt_cell;
            }
            else
            {
                pt_cell->next               = pt_search_cell->next;
                pt_cell->prev               = pt_search_cell;
                pt_search_cell->next        = pt_cell;
                pt_place->Rows[ iRow ].tail = pt_cell;
            }
        }
    }

    for( iRow = 0; iRow < pt_place->NbRows; iRow++)
    {
        width = 0;
        for( pt_cell = pt_place->Rows[ iRow ].head;
             pt_cell;
             pt_cell = pt_cell->next )
        {
            width += pt_cell->width;
        }
        pt_place->Rows[ iRow ].length = width;
    }

} /* fin de fill_structure */

/*---------------------------------------------------------\
                     get_place_fig
\---------------------------------------------------------*/
placement_fig *get_place_fig( fig_name, pt_lofig )
char *fig_name;
lofig_list *pt_lofig;
{
    phfig_list    *pt_phfig;
    phins_list    *pt_phins;
    placement_fig *pt_place;
    long NbRows;
    long NbCells;
    long iCells;
    long width;
    long height;

    /* alloc struct */
    pt_place = (placement_fig *) mbkalloc ( (long) sizeof( placement_fig ) );

    /* get ph fig */
    pt_phfig = getphfig( fig_name, 'A' );

    /* if no instance, error */
    if( ! pt_phfig->PHINS )
    {
        fprintf( stderr,
                 "*** %s error *** : placement fig does not have any instance.\n",
                 DECOUP_NAME );
        exit( 1 );
    }

    /* found slice's number */
    height = found_height_of_cell( pt_phfig->PHINS->FIGNAME ) * SCALE_X;
    NbRows = ( pt_phfig->YAB2 - pt_phfig->YAB1 ) / height;
    pt_place->NbRows = NbRows;

    /* alloc Rows array, it will be fill after ... ( in fill_structure ) */
    pt_place->Rows = (row_elt *) mbkalloc ( (long) NbRows * sizeof( row_elt ) );

    /* counting cells */
    for( NbCells = 0, pt_phins =  pt_phfig->PHINS;
         pt_phins;
         NbCells++, pt_phins = pt_phins->NEXT );
    pt_place->NbCells = NbCells;

    /* alloc Cells array */
    pt_place->Cells = (cell_list *) mbkalloc ( (long) NbCells * sizeof( cell_list ) );

    /* Filling Cells array */
    for( iCells = 0, pt_phins =  pt_phfig->PHINS;
         iCells < NbCells;
         iCells++, pt_phins = pt_phins->NEXT )
    {
        width = found_width_of_cell( pt_phins->FIGNAME );

        if( height != ( found_height_of_cell( pt_phins->FIGNAME ) * SCALE_X ) )
        {
            fprintf( stderr,
                     "*** %s error *** : placement fig does not have instances with same heights.\n",
                     DECOUP_NAME );
            exit( 1 );
        }

        pt_place->Cells[ iCells ].width = width;
        pt_place->Cells[ iCells ].x     = pt_phins->XINS / SCALE_X + ( width >> 1 );
        pt_place->Cells[ iCells ].y     = pt_phins->YINS;
        pt_place->Cells[ iCells ].row   = pt_phins->YINS / height;
        pt_place->Cells[ iCells ].ins   = getloins( pt_lofig, pt_phins->INSNAME );
    }

    return( pt_place );

} /* fin de get_place_fig */

/*---------------------------------------------------------\
                     found_ideal_width
\---------------------------------------------------------*/
long found_ideal_width( pt_place, cells_per_col )
placement_fig *pt_place;
long           cells_per_col;
{
    cell_list  *pt_cell;
    long   ideal_width;
    long   iRow;
    long   iCell;
    long **col_width;
    long   width;
    long   sum_width;
    long   average_width;

	col_width = (long **) mbkalloc( (long) pt_place->NbRows * sizeof( long * ) );
    for( iRow = 0; iRow < pt_place->NbRows; iRow++ )
        col_width[ iRow ] = (long *) mbkalloc ( (long) cells_per_col * sizeof( long ) );

    sum_width = 0;
    for( iRow = 0; iRow < pt_place->NbRows; iRow++ )
    {
        for( iCell = 0; iCell < cells_per_col; iCell++ )
        {
            col_width[ iRow ][ iCell ] = 0; 
        }
        width   = 0;
        for( pt_cell = pt_place->Rows[iRow].head, iCell = 0;
             iCell < cells_per_col;
             iCell++ )
        {
            if( pt_cell )
            {
                width += pt_cell->width;
                pt_cell = pt_cell->next;
            }
            col_width[ iRow ][ iCell ] = width; 
        }
        sum_width += width;
    }

    average_width = sum_width / pt_place->NbRows;
    ideal_width   = average_width;

    for( iRow = 0; iRow < pt_place->NbRows; iRow++ )
    {
        if( col_width[ iRow ][ 0 ] > average_width )
        {
            ideal_width = MAX( ideal_width, col_width[ iRow ][ 0 ] );
            continue;
        }

        for( iCell = 1;
             ( iCell < cells_per_col ) && ( col_width[ iRow ][ iCell ] < average_width );
             iCell++ );

        if( iCell == cells_per_col )
            continue;

        if( ( col_width[ iRow ][ iCell ] - average_width ) < ( average_width - col_width[ iRow ][ iCell - 1 ] ) )
            ideal_width = MAX( ideal_width, col_width[ iRow ][ iCell ] );
    }

    for( iRow = 0; iRow < pt_place->NbRows; iRow++ )
        mbkfree( (char *)col_width[ iRow ] );
	mbkfree( (char *)col_width );

    return( ideal_width );

} /* fin de found_ideal_width */

/*---------------------------------------------------------\
                     create_phcol
\---------------------------------------------------------*/
static chain_list *create_phcol( pt_place, pt_lofig, cells_per_col )
placement_fig *pt_place;
lofig_list    *pt_lofig;
long           cells_per_col;
{
    phfig_list *pt_phcol;
    phins_list *pt_phins;
    chain_list *pt_chainphcol;
    cell_list  *pt_cell;
	unsigned long height;
    long col;
    long iRow;
	long ideal_width;
    long col_maxwidth;
    long col_width;
    long nb_cell;
    int finished;
    char *col_name      = NULL;

    col = 0;
    pt_chainphcol = NULL;

    /* found height of cells */
	height = found_height_of_cell( pt_place->Rows[ 0 ].tail->ins->FIGNAME );

    /* alloc col_name ( + 9 : col_nnn_ ... \0 ) */
    col_name = (char *)mbkalloc( (long) sizeof( char ) * ( strlen( pt_lofig->NAME ) + 9 ) );

    /* first, make phcol without alim, con and trans ( just placement ) */
    do
    {
        /* making name of the col */
        if( col > 999 )
        {
            fprintf( stderr,
                     "*** %s error *** : too much cols ( > 999 ).\n",
                     DECOUP_NAME );
            exit( 1 );
        }
        sprintf( col_name, "col_%ld_%s", col, pt_lofig->NAME );
        col ++;

        /* create phcol and locol */
        pt_phcol = addphfig( namealloc( col_name ) );

        finished = TRUE;
        col_maxwidth = 0;

        ideal_width = found_ideal_width( pt_place, cells_per_col );

        /* making cols */
        for( iRow = 0; iRow < pt_place->NbRows; iRow++)
        {
            pt_cell   = pt_place->Rows[iRow].head;
            col_width = 0;
            nb_cell   = 0;

            for( ; pt_cell; pt_cell = pt_cell->next )
            {
                if( ( col_width + pt_cell->width ) > ideal_width )
                    break;
                if( nb_cell == cells_per_col )
                    break;

                nb_cell ++;

                pt_phins = addphins( pt_phcol,
                                     pt_cell->ins->FIGNAME,
                                     pt_cell->ins->INSNAME,
                                     ( iRow & 1 ) ? SYM_Y :  NOSYM,
                                     col_width * SCALE_X,
                                     iRow * height * SCALE_X );

/*                 pt_phins->USER = (struct ptype *)pt_cell->ins; */

                finished = FALSE;

                col_width   += pt_cell->width;
                col_maxwidth = MAX( col_maxwidth, col_width );

                /* all other cells are already placed */
                pt_place->Rows[ iRow ].head = pt_cell->next;
            }
        }

        /* if is the last phcol, so delete it (empty) */
        if( col_maxwidth == 0 )
        {
            delphfig( col_name );
        }
        else
        {
            /* chain the new phcol */
            pt_chainphcol = addchain( pt_chainphcol, (void *)pt_phcol );

            /* making abutment box of cols */
            pt_phcol->XAB1 = 0;
            pt_phcol->YAB1 = 0; 
            pt_phcol->XAB2 = col_maxwidth * SCALE_X;
            pt_phcol->YAB2 = pt_place->NbRows * height * SCALE_X; 
        }

    } while( !finished );

    mbkfree( (char *)col_name );

	return( pt_chainphcol );

} /* fin de create_phcol */

/*---------------------------------------------------------\
                     make_col
\---------------------------------------------------------*/
static chain_list *make_col( pt_place, pt_lofig, cells_per_col )
placement_fig *pt_place;
lofig_list    *pt_lofig;
long           cells_per_col;
{
    chain_list *pt_headchainphcol;
    chain_list *pt_chain;
    phfig_list *pt_phfig;
    phfig_list *pt_phcol;
    long        width;
    long        height;
    long        cells_height;
    long        y;
    long        vdd_width;
    long        vss_width;

	/* fisrt, create ph_col */
    pt_headchainphcol = reverse( create_phcol( pt_place, pt_lofig, cells_per_col ) );

    pt_phfig = addphfig( pt_lofig->NAME );

    width  = 0L;
    height = 0L;

    for( pt_chain = pt_headchainphcol; pt_chain; pt_chain=pt_chain->NEXT )
    {
        pt_phcol = (phfig_list *)pt_chain->DATA;

        addphins( pt_phfig, pt_phcol->NAME, pt_phcol->NAME, NOSYM, width, 0L );

        width += pt_phcol->XAB2 - pt_phcol->XAB1;
        height = MAX( height, pt_phcol->YAB2 - pt_phcol->YAB1 );
    }

    defab( pt_phfig, 0L, 0L, width, height );

    rflattenphfig( pt_phfig, NO, YES );

    savephfig( pt_phfig );

    return( pt_headchainphcol );

} /* fin de make_col */

/*---------------------------------------------------------\
                     free_phcol
\---------------------------------------------------------*/
void free_phcol( pt_chainphcol )
chain_list *pt_chainphcol;
{
    chain_list *pt_chain;
    phfig_list *pt_phcol;

    for( pt_chain = pt_chainphcol; pt_chain; pt_chain=pt_chain->NEXT )
    {
        pt_phcol = (phfig_list *)pt_chain->DATA;

        delphfig( pt_phcol->NAME );
    }

    freechain( pt_chainphcol );

} /* end of free_phcol */

/*---------------------------------------------------------\
                     glue_logic_placer
\---------------------------------------------------------*/
long glue_logic_placer( netlist_name, rows, cells_per_col, nb_ite, place_fig_name, ptplacecon )
char *netlist_name;
long  rows;
long  cells_per_col;
long  nb_ite;
char *place_fig_name;
PlaceConList *ptplacecon;
{
    unsigned long weight_x      = 0;
    unsigned long weight_y      = 6;
    unsigned long weight_rows   = 5;
    unsigned long weight_con    = 6;

	unsigned long cells_size    = 0;
	unsigned long init_size     = 0;
	unsigned long cols_size;

    ulofig_glue *glue_parameters;
    ptype_list  *ptype;

    long newrows;
    long nbcolumns;
    char debug_mode;

    lofig_list    *pt_lofig;
    chain_list    *pt_chainphcol;
    placement_fig *pt_place;

    debug_mode = ( getenv( DEBUG_VAR ) != NULL );

    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> Loading netlist.\n" );
        fflush( stdout );
    }

    pt_lofig = getlofig( netlist_name, 'A' );

    ptype = getptype( pt_lofig->USER, (long) C_GfpUT_UserLoFigGlue );

    if( ptype != NULL )
    {
        glue_parameters = (ulofig_glue *)( ptype->DATA );

        cells_per_col = glue_parameters->CPC;
        nb_ite        = glue_parameters->ITER;
        newrows       = glue_parameters->WIDE;

        if( newrows > rows )
        {
            fprintf( stderr, "*** decoup error *** : rows number indicate in dpr file is highter than datapath height.\n" );
            exit( 1 );
        }
        rows = newrows;
    }

    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> Symbolic placement\n"                  );
        fprintf( stdout, "\n"                                            );
        fprintf( stdout, "        netlist        : %s\n", netlist_name   );
        fprintf( stdout, "        iterations     : %ld\n", nb_ite        );
        fprintf( stdout, "        rows           : %ld\n", rows          );
        fprintf( stdout, "        cells_per_cols : %ld\n", cells_per_col );
        if( debug_mode )
        {
        fprintf( stdout, "        weight_x       : %ld\n", weight_x      );
        fprintf( stdout, "        weight_y       : %ld\n", weight_y      );
        fprintf( stdout, "        weight_rows    : %ld\n", weight_rows   );
        fprintf( stdout, "        weight_con     : %ld\n", weight_con    );
        }
        fprintf( stdout, "\n"                                            );
    }

    if( cells_per_col < 1 )
    {
        fprintf( stderr,
                 "*** %s error *** : Cells per columns must be > 0.\n",
                 DECOUP_NAME );
        exit( 1 );
    }

    if( rows < 0 )
    {
        fprintf( stderr,
                 "*** %s error *** : rows number must be > 0.\n",
                 DECOUP_NAME );
        exit( 1 );
    }

    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> Checking cells height.\n" );
        fflush( stdout );
    }

    check_cells_height( pt_lofig );

    if( ( verbose_mode == VERBOSE_ALL ) && ( pt_lofig->LOINS ) )
    {
        fprintf( stdout, "     -> Number of track per slice : %ld.\n",
                 ( found_height_of_cell( pt_lofig->LOINS->FIGNAME ) - 15 ) / 5 + 1 );
        fflush( stdout );
    }

    pt_lofig = getlofig( netlist_name, 'A' );
    if( place_fig_name == NULL )
    {
        if( verbose_mode == VERBOSE_ALL )
        {
            fprintf( stdout, "     -> placing.\n" );
            fprintf( stdout, "\n"                );
            fflush( stdout );
        }

        pt_place = Place( pt_lofig,
                          nb_ite,
                          rows,
                          weight_x,
                          weight_y,
                          weight_rows,
                          weight_con,
                          ptplacecon );

        if( pt_place == NULL )
        {
            fprintf( stderr,
                     "*** %s error *** : placement fig is NULL : placement function problem.\n",
                     DECOUP_NAME );
            exit( 1 );
        }

        if( verbose_mode == VERBOSE_ALL )
        {
            fprintf( stdout, "\n" );
            fflush( stdout );
        }

    }
    else
    {
        if( verbose_mode == VERBOSE_ALL )
        {
            fprintf( stdout, "     -> getting placement fig.\n" );
            fflush( stdout );
        }

        pt_place = get_place_fig( place_fig_name, pt_lofig );
    }

    rows = pt_place->NbRows;

    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> initialising structure.\n" );
        fflush( stdout );
    }

    fill_structure( pt_place );

    if( ( verbose_mode == VERBOSE_ALL ) && ( debug_mode ) )
    {
        fprintf( stdout, "     -> computing cells size : " );
        fflush( stdout );
        cells_size = compute_cells_size( pt_place );
        fprintf( stdout, "%ld\n", cells_size );

        fprintf( stdout, "     -> computing init place size : " );
        fflush( stdout );
        init_size = compute_init_place_size( pt_place );
        fprintf( stdout, "%ld\n", init_size );
        fflush( stdout );

    }

    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> making channel.\n" );
        fflush( stdout );
    }

    delphfig( netlist_name );
    pt_chainphcol = make_col( pt_place, pt_lofig, cells_per_col );

    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> nb columns :" );
        fflush( stdout );
    }

    nbcolumns = getchainlenght( pt_chainphcol );

    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, " %ld.\n", nbcolumns );
        fflush( stdout );
    }

    if( ( verbose_mode == VERBOSE_ALL ) && ( debug_mode ) )
    {
        fprintf( stdout, "     -> computing cols size : " );
        fflush( stdout );
        cols_size = compute_cols_size( pt_chainphcol );
        fprintf( stdout, "%ld\n", cols_size );

        fprintf( stdout, "     -> initial size lost %ld for %d%% more size.\n",
                               init_size - cells_size,
                               (int)(100.0 * ((double)(init_size)/(double)cells_size - 1.0) ) );
        fprintf( stdout, "     -> total size lost %ld for %d%% more size.\n",
                               cols_size - cells_size,
                               (int)(100.0 * ((double)(cols_size)/(double)cells_size - 1.0) ) );
        fprintf( stdout, "     -> partial size lost %ld for %d%% more size.\n",
                               cols_size - init_size,
                               (int)(100.0 * ((double)(cols_size)/(double)init_size - 1.0) ) );
        fflush( stdout );
    }

    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> Done.\n" );
        fflush( stdout );
    }

    free_phcol( pt_chainphcol );

    return nbcolumns;

} /* fin de glue_logic_placer */

