/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

	File   		:	router_debug.c

	Date		:	16 03 1994
	Version		:	1

	Auteur      :   BEAURIN oliver

	Description :   debug functions for dpr router structure.

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include MLO_H
#include MPH_H

#include "router_dpr.h"
#include "router_util.h"

/*---------------------------------------------------------\
                     affdprpileleft
\---------------------------------------------------------*/
 void affdprpileleft(fsortie,ptacces)
 FILE *fsortie;
 ACCESS *ptacces;

 {
 for (;ptacces;ptacces=ptacces->NEXT)
     {
     if (ptacces->POS)
        fprintf(fsortie
                ,"  |_(ORIENT,X,Y,POS.CON.NAME.DONE)  :(%c, %ld, %ld, %s, %s) \n"
                ,ptacces->ORIENT
                ,ptacces->POS->X
                ,ptacces->POS->Y
                ,ptacces->POS->CON->NAME
                ,ptacces->DONE?"TRUE":"FALSE");
     else
        fprintf(fsortie
                ,"  |_(ORIENT,X,Y,obstacle)  :(%c) \n"
                ,ptacces->ORIENT);
     }
 }

/*---------------------------------------------------------\
                     affdprpile
\---------------------------------------------------------*/
 void affdprpile(fsortie,ptacces)
 FILE *fsortie;
 ACCESS *ptacces;

 {
 for (;ptacces;ptacces=ptacces->PREV)
     {
     if (ptacces->POS)
        fprintf(fsortie
                ,"  |_(ORIENT,X,Y,POS.CON.NAME.DONE)  :(%c, %ld, %ld, %s, %s) \n"
                ,ptacces->ORIENT
                ,ptacces->POS->X
                ,ptacces->POS->Y
                ,ptacces->POS->CON->NAME
                ,ptacces->DONE?"TRUE":"FALSE");
     else
        fprintf(fsortie
                ,"  |_(ORIENT,X,Y,obstacle)  :(%c) \n"
                ,ptacces->ORIENT);
     }
 }

/*---------------------------------------------------------\
                     affdprnoeud
\---------------------------------------------------------*/
 void affdprnoeud(fsortie,ptheadnoeud)
 FILE  *fsortie;
 NODE *ptheadnoeud;

{
 NODE *ptnoeud;
 long   j=0;
 long   i=0;

 for (;ptheadnoeud;ptheadnoeud=ptheadnoeud->NEXT)
 {
     fprintf(fsortie,"|________Grille de la colonne %ld \n ",j++);
     j=j++;
     i=0;
     for (ptnoeud=ptheadnoeud;ptnoeud;ptnoeud=ptnoeud->DPR_UP)
     {
         fprintf(fsortie,"  |________Pile de la ligne %ld \n ",i++);
         fprintf(fsortie,"  |(Noeud = %ld, ",ptnoeud);
         if (ptnoeud->NEXT)
            fprintf(fsortie,"Next = %ld, ",ptnoeud->NEXT);
         else
            fprintf(fsortie,"Next = (Nil), ");
         if (ptnoeud->PREV)
            fprintf(fsortie,"Pred = %ld, ",ptnoeud->PREV);
         else
            fprintf(fsortie,"Pred = (Nil), ");
         if (ptnoeud->DPR_UP)
            fprintf(fsortie,"Up = %ld, ",ptnoeud->DPR_UP);
         else
            fprintf(fsortie,"Up = (Nil), ");
         if (ptnoeud->DPR_DWN)
            fprintf(fsortie,"Dwn = %ld)\n ",ptnoeud->DPR_DWN);
         else
            fprintf(fsortie,"Dwn = (Nil))\n");
         fprintf(fsortie,"  |__Pile du sommet RIGHT = %ld \n",ptnoeud->DPR_RIGHT);
         if (ptnoeud->DPR_RIGHT)
            affdprpile(fsortie,ptnoeud->DPR_RIGHT);
         else
            fprintf(fsortie,"  |______Pile Vide RIGHT = (Nil)\n");
         fprintf(fsortie,"  |__Pile du sommet LEFT = %ld \n",ptnoeud->DPR_LEFT);
         if (ptnoeud->DPR_LEFT)
            affdprpileleft(fsortie,ptnoeud->DPR_LEFT);
         else
            fprintf(fsortie,"  |______Pile Vide LEFT = (Nil)\n");
         fprintf(fsortie,"  |_______________________________________\n");
     }
 }
}

/*---------------------------------------------------------\
                     info
\---------------------------------------------------------*/
 void info(ptcsig,node_l,pos_dwn,node_r,pos_up)
 chain_list   *ptcsig;
 NODE        *node_l;
 POSSIBILITY  *pos_dwn;
 NODE        *node_r;
 POSSIBILITY  *pos_up;
 {

  printf("On assure la continuite du SIGNAL %s,\n",((SIGNAL*)ptcsig->DATA)->NAME);
  printf("  entre la possibilite X=%ld en ligne %s,\n"
         ,pos_dwn->X
         ,((COLUMN *)pos_dwn->CON->COL->DATA)->NAME);
  printf("     et la possibilite X=%ld en ligne %s.\n"
         ,pos_up->X
         ,((COLUMN *)pos_up->CON->COL->DATA)->NAME);
  return;
 }

/*---------------------------------------------------------\
                     affcon
\---------------------------------------------------------*/
 void affcon(fsortie,ptcon)
 FILE      *fsortie;
 CONNECTOR *ptcon;
 {
 SIGNAL  *ptsig;
 char    *ps;

 if (ptcon!=NULL)
    {
    if (ptcon->INS)
       ps = ((INSTANCE *)ptcon->INS->DATA)->INSNAME;
    else ps = "$" ;
    fprintf ( fsortie ,
      "|	|__(INSNAME).CON.NAME____________________: (%s), %s \n",ps, ptcon->NAME );
    fprintf ( fsortie ,
      "|	|__CON.(TYPE,NBPOSS)   : ( %c, %ld ) \n", ptcon->TYPE, ptcon->NBPOSS );
    fprintf ( fsortie ,
      "|	|__CON.(MINPOS,MAXPOS) : ( %ld, %ld ) \n", ptcon->MINPOS, ptcon->MAXPOS );
    if (ptcon->SIG)
       {
       ptsig = (SIGNAL *)ptcon->SIG->DATA;
       fprintf ( fsortie ,
        "|	|__CON.Signal.(INDEX,NAME)  :( %ld, %s) \n",ptsig->INDEX, ptsig->NAME );
       }
    else fprintf ( fsortie ,"|	|__CON.Signal         : (vide)  \n");
    fprintf ( fsortie ,"|	|___________________________________\n");
    }
 else fprintf(fsortie ," pointeur VIDE \n");
 }
/*---------------------------------------------------------\
                     aff_SCR_con
\---------------------------------------------------------*/
void aff_SCR_con( fsortie, ptcon )
 FILE      *fsortie;
 ConnectorList *ptcon;
{
    fprintf( fsortie, "-> %ld (%ld)\n",ptcon->ConName,ptcon->Mark );
} /* end of aff_SCR_con */
/*---------------------------------------------------------\
                     aff_channel
\---------------------------------------------------------*/
void aff_channel(fsortie,channel)
 FILE      *fsortie;
 CHANNEL   *channel;
{
    ConnectorList *ptcon;

    fprintf( fsortie, "Connectors north : \n" );
    ptcon = channel->NORTHCON;
    while( ptcon )
    {
        aff_SCR_con(fsortie,ptcon);
        ptcon=ptcon->NextCon;
    }
    fprintf( fsortie, "Connectors south : \n" );
    ptcon = channel->SOUTHCON;
    while( ptcon )
    {
        aff_SCR_con(fsortie,ptcon);
        ptcon=ptcon->NextCon;
    }
    fprintf( fsortie, "Connectors west : \n" );
    ptcon = channel->WESTCON;
    while( ptcon )
    {
        aff_SCR_con(fsortie,ptcon);
        ptcon=ptcon->NextCon;
    }
    fprintf( fsortie, "Connectors east : \n" );
    ptcon = channel->EASTCON;
    while( ptcon )
    {
        aff_SCR_con(fsortie,ptcon);
        ptcon=ptcon->NextCon;
    }
    
} /* end of aff_channel */

