/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_hierarch.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include  <stdio.h>
#include  MPH_H
#include  MUT_H

#include  "dpr.h"
#include  "strecher_support.h"
#include  "strecher_intv.h"
#include  "strecher_allowed.h"
#include  "strecher_supply.h"
#include  "strecher_phcut.h"
#include  "strecher_dtdupl.h"
#include  "strecher_phtalu.h"
#include  "strecher_hierarch.h"
#include  "strecher_strech.h"
#include  "strecher_phcma.h"

/*---------------------------------------------------------\
                     defines
\---------------------------------------------------------*/
#define        fgetc(stream)    /* fgetc(stream); */

/* Constantes internes au module. */
#define        MAXTPSTATE        250

/*---------------------------------------------------------\
                     structures
\---------------------------------------------------------*/
/* Type 'state_type', element de base de la pile des etats */
typedef struct
{
    struct phfig   *ptfig_father;       /* Modele du pere.           */
    struct phins   *ptins_son;  /* Prochaine instance fille. */
} state_type;

/* La structure 'lastpos_phfig' permet de faire des mises a plat
   successives des coordonnees d'un modele. */
typedef struct
{
    long            XFLAT, YFLAT, XAB1, YAB1, XAB2, YAB2;
    char            ATRANSF, RTRANSF;
} lastpos_phfig;

/* La structure 'modelepos_phfig' permet de remmettre un
   modele a sa position initiale ( avant mise a plat ). */
typedef struct
{
    long            XMODL, YMODL;
} modelepos_phfig;

/*---------------------------------------------------------\
                     locals
\---------------------------------------------------------*/
static long     sp_state;
static state_type stack_state[MAXTPSTATE];

/*---------------------------------------------------------\
                     reset_stack_state
\---------------------------------------------------------*/
static void     reset_stack_state()
{
    sp_state = 0L;
}

/*---------------------------------------------------------\
                     push_state
\---------------------------------------------------------*/
static void     push_state(ptfig_father, ptins_son)
phfig_list     *ptfig_father;
phins_list     *ptins_son;
{
    stack_state[sp_state].ptfig_father = ptfig_father;
    stack_state[sp_state].ptins_son = ptins_son;

    sp_state++;
}

/*---------------------------------------------------------\
                     pop_state
\---------------------------------------------------------*/
static void     pop_state(pptfig_father, pptins_son)
phfig_list    **pptfig_father;
phins_list    **pptins_son;
{
    sp_state--;

    *pptins_son = stack_state[sp_state].ptins_son;
    *pptfig_father = stack_state[sp_state].ptfig_father;
}


/*---------------------------------------------------------\
                     orientflat
\---------------------------------------------------------*/
static void     orientflat(oriout, ori, trsf)
char           *oriout, ori, trsf;
{
    /* Transformation X --> -X               */
    /* ( inverse les connecteurs EST/OUEST ) */
    if ((trsf == SYM_X) || (trsf == SYMXY))
        switch (ori)
        {
        case NORTH:
            *oriout = ori;
            break;
        case SOUTH:
            *oriout = ori;
            break;
        case EAST:
            *oriout = WEST;
            break;
        case WEST:
            *oriout = EAST;
            break;
        default:
            fflush( stdout );
            fprintf( stderr, "*** strecher error *** : Connector orientation invalid.\n" );
            exit( 1 );
        }

    /* Transformation Y --> -Y               */
    /* ( inverse les  connecteurs NORD/SUD ) */
    if ((trsf == SYM_Y) || (trsf == SYMXY))
        switch (ori)
        {
        case NORTH:
            *oriout = SOUTH;
            break;
        case SOUTH:
            *oriout = NORTH;
            break;
        case EAST:
            *oriout = ori;
            break;
        case WEST:
            *oriout = ori;
            break;
        default:
            fflush( stdout );
            fprintf( stderr, "*** strecher error *** : Connector orientation invalid.\n" );
            exit( 1 );
        }
}

/*---------------------------------------------------------\
                     transflat
\---------------------------------------------------------*/
static char     transflat(transf_father, transf_son)
char            transf_father, transf_son;
{
#define        LOCAL_NOSYM         0
#define        LOCAL_SYM_X         1    /* 2 puissance 0. */
#define        LOCAL_SYM_Y         2    /* 2 puissance 1. */
#define        LOCAL_SYMXY       ( LOCAL_SYM_X|LOCAL_SYM_Y )    /* 3 */

    char            transf_result = -1;

    /* Recodage de la transformation du pere. */
    switch (transf_father)
    {
    case NOSYM:
        transf_result = LOCAL_NOSYM;
        break;
    case SYM_X:
        transf_result = LOCAL_SYM_X;
        break;
    case SYM_Y:
        transf_result = LOCAL_SYM_Y;
        break;
    case SYMXY:
        transf_result = LOCAL_SYMXY;
        break;
    }

    /* Recodage de la transformation du fils et combinaison avec celle du pere. */
    switch (transf_son)
    {
    case NOSYM:
        break;
    case SYM_X:
        transf_result ^= LOCAL_SYM_X;
        break;
    case SYM_Y:
        transf_result ^= LOCAL_SYM_Y;
        break;
    case SYMXY:
        transf_result ^= LOCAL_SYMXY;
        break;
    }

    /* Decodage de la transformation resultante. */
    switch (transf_result)
    {
    case LOCAL_NOSYM:
        return (NOSYM);
    case LOCAL_SYM_X:
        return (SYM_X);
    case LOCAL_SYM_Y:
        return (SYM_Y);
    case LOCAL_SYMXY:
        return (SYMXY);
    }

    return LOCAL_NOSYM;
#undef        LOCAL_NOSYM
#undef        LOCAL_SYM_X
#undef        LOCAL_SYM_Y
#undef        LOCAL_SYMXY
}

/*---------------------------------------------------------\
                     alloc_structpos
\---------------------------------------------------------*/
static void     alloc_structpos(ptfig)
phfig_list     *ptfig;
{
    ptype_list     *ptuser;
    lastpos_phfig  *plast;
    modelepos_phfig *pmodl;

    /* Recherche/creation dans la liste chainee. */
    ptuser = get_ls_data(ptfig, C_SUPP_ULASTPOS, F_SUPP_UCREAT);

    if (ptuser->DATA == (void *) NULL)
    {
        /* Allocation et initialisation de la structure. */
        plast = (lastpos_phfig *) mbkalloc(
                                           sizeof(lastpos_phfig));
        plast->XAB1 = ptfig->XAB1;
        plast->YAB1 = ptfig->YAB1;
        plast->XAB2 = ptfig->XAB2;
        plast->YAB2 = ptfig->YAB2;
        plast->ATRANSF = NOSYM;
        plast->RTRANSF = NOSYM;

        ptuser->DATA = (void *) plast;
    }

    /* Recherche ( et creation si necessaire ) de la structure contenant la position initiale du modele. */
    ptuser = get_ls_data(ptfig, C_SUPP_UMODLPOS, F_SUPP_UCREAT);

    if (ptuser->DATA == (void *) NULL)
    {
        /* Allocation et initialisation de la structure. */
        pmodl = (modelepos_phfig *) mbkalloc(
                                             sizeof(modelepos_phfig));

        pmodl->XMODL = ptfig->XAB1;
        pmodl->YMODL = ptfig->YAB1;

        ptuser->DATA = (void *) pmodl;
    }
}


/*---------------------------------------------------------\
                     dupmodlpos
\---------------------------------------------------------*/
void           *dupmodlpos(a_ptfig_src, a_ptfig_dest)
phfig_list     *a_ptfig_src;
phfig_list     *a_ptfig_dest;
{
#define  access_modelepos  ( (modelepos_phfig*)(ptuser->DATA) )

    ptype_list     *ptuser;
    modelepos_phfig *pmodl;

    /* Recuperation de la structure d'origine. */
    if ((ptuser = get_ls_data(a_ptfig_src, C_SUPP_UMODLPOS, FALSE))
        == (ptype_list *) NULL)
        return ((void *) NULL);

    if (ptuser->DATA == (void *) NULL)
        return ((void *) NULL);

    /* Duplication de la structure. */
    pmodl = (modelepos_phfig *) mbkalloc(sizeof(modelepos_phfig));

    pmodl->XMODL = access_modelepos->XMODL;
    pmodl->YMODL = access_modelepos->YMODL;

    return ((void *) pmodl);

#undef  access_modelepos
} /* end of dupmodlpos */


void           *dupflatpos(a_ptfig_src, a_ptfig_dest)
phfig_list     *a_ptfig_src;
phfig_list     *a_ptfig_dest;
{
#define  access_lastpos  ( (lastpos_phfig*)(ptuser->DATA) )

    ptype_list     *ptuser;
    lastpos_phfig  *plast;

    /* Recuperation de la structure d'origine. */
    if ((ptuser = get_ls_data(a_ptfig_src, C_SUPP_ULASTPOS, FALSE))
        == (ptype_list *) NULL)
        return ((void *) NULL);

    if (ptuser->DATA == (void *) NULL)
        return ((void *) NULL);

    /* Duplication de la structure. */
    plast = (lastpos_phfig *) mbkalloc(sizeof(lastpos_phfig));

    plast->XFLAT = access_lastpos->XFLAT;
    plast->YFLAT = access_lastpos->YFLAT;
    plast->XAB1 = access_lastpos->XAB1;
    plast->YAB1 = access_lastpos->YAB1;
    plast->XAB2 = access_lastpos->XAB2;
    plast->YAB2 = access_lastpos->YAB2;
    plast->ATRANSF = access_lastpos->ATRANSF;
    plast->RTRANSF = access_lastpos->RTRANSF;

    return ((void *) plast);

#undef  access_lastpos
}

/*---------------------------------------------------------\
                     flatpos_ab
\---------------------------------------------------------*/
static void     flatpos_ab(ptfig, xflat, yflat, transf)
phfig_list     *ptfig;
long            xflat, yflat;
char            transf;
{
#define    access_lastpos    ( (lastpos_phfig*)ptuser->DATA )

    ptype_list     *ptuser;
    long            swap;

    /* Verification de l'existence des imformations de retour. */
    if ((ptuser = get_ls_data(ptfig, C_SUPP_ULASTPOS, FALSE)) == (ptype_list *) NULL)
    {
        alloc_structpos(ptfig);

        ptuser = get_ls_data(ptfig, C_SUPP_ULASTPOS, FALSE);
    }

    access_lastpos->XFLAT = xflat;
    access_lastpos->YFLAT = yflat;
    access_lastpos->XAB1 = ptfig->XAB1;
    access_lastpos->YAB1 = ptfig->YAB1;
    access_lastpos->XAB2 = ptfig->XAB2;
    access_lastpos->YAB2 = ptfig->YAB2;
    access_lastpos->RTRANSF = transflat(access_lastpos->ATRANSF, transf);
    access_lastpos->ATRANSF = transf;

    /* Mise a plat des deux angles de l'abutment-box. */
    /* AT : il peut y avoir inversion des angles.     */
    xyflat(&ptfig->XAB1, &ptfig->YAB1,
           ptfig->XAB1, ptfig->YAB1,
           xflat, yflat,
           access_lastpos->XAB1, access_lastpos->YAB1,
           access_lastpos->XAB2, access_lastpos->YAB2,
           access_lastpos->RTRANSF);

    xyflat(&ptfig->XAB2, &ptfig->YAB2,
           ptfig->XAB2, ptfig->YAB2,
           xflat, yflat,
           access_lastpos->XAB1, access_lastpos->YAB1,
           access_lastpos->XAB2, access_lastpos->YAB2,
           access_lastpos->RTRANSF);

    /* Correction de l'inversion des angles en X. */
    if ((access_lastpos->RTRANSF == SYM_X)
        || (access_lastpos->RTRANSF == SYMXY))
    {
        swap = ptfig->XAB1;
        ptfig->XAB1 = ptfig->XAB2;
        ptfig->XAB2 = swap;
    }

    /* Correction de l'inversion des angles en Y. */
    if ((access_lastpos->RTRANSF == SYM_Y)
        || (access_lastpos->RTRANSF == SYMXY))
    {
        swap = ptfig->YAB1;
        ptfig->YAB1 = ptfig->YAB2;
        ptfig->YAB2 = swap;
    }

#undef    access_lastpos
}


/*---------------------------------------------------------\
                     flatpos_ins
\---------------------------------------------------------*/
static void     flatpos_ins(ptfig)
phfig_list     *ptfig;
{
#define access_lastpos (   (lastpos_phfig*)ptuser->DATA )

    phins_list     *ptins;
    phfig_list     *ptfig_ins;
    ptype_list     *ptuser;

    /* Recuperation des informations de position du pere. */
    ptuser = get_ls_data(ptfig, C_SUPP_ULASTPOS, FALSE);
    if (ptuser == (ptype_list *) NULL)
    {
        fflush( stdout );
        fprintf( stderr,
                 "*** strecher bugs *** : Unable to flat %s's instances (missing C_SUPP_ULASTPOS ptype).\n",
                 ptfig->NAME);
        exit( 1 );
    }

    /* Boucle de mise a plat des instances. */
    for (ptins = ptfig->PHINS; ptins != (phins_list *) NULL;
         ptins = ptins->NEXT)
    {                           /* Rupture de l'indentation. */
        ptfig_ins = getphfig(ptins->FIGNAME, 'A');

        /* Mise a plat des coordonnees de l'instance */
        /* ( par rapport aux coordonnees du pere )   */
        xyflat(&ptins->XINS, &ptins->YINS,
               ptins->XINS, ptins->YINS,
               access_lastpos->XFLAT, access_lastpos->YFLAT,
               access_lastpos->XAB1, access_lastpos->YAB1,
               access_lastpos->XAB2, access_lastpos->YAB2,
               access_lastpos->RTRANSF);

        /* Transformation du pere, par rapport au plus haut niveau. */
        switch (access_lastpos->RTRANSF)
        {
        case NOSYM:
            break;
        case SYM_X:
            /* Soustraction de la largeur de l'abutment-box. */
            /* ( repassage a l'angle inferieur gauche )      */
            ptins->XINS -= (ptfig_ins->XAB2 - ptfig_ins->XAB1);
            break;
        case SYM_Y:
            /* Soustraction de la hauteur de l'abutment-box. */
            ptins->YINS -= (ptfig_ins->YAB2 - ptfig_ins->YAB1);
            break;
        case SYMXY:
            /* Soustraction de la hauteur et de la largeur de l'abutment-box. */
            ptins->XINS -= (ptfig_ins->XAB2 - ptfig_ins->XAB1);
            ptins->YINS -= (ptfig_ins->YAB2 - ptfig_ins->YAB1);
            break;
        default:
            fflush( stdout );
            fprintf( stderr,
                     "*** strecher error *** : Transformation of instance %s not implemented.\n",
                     ptins->INSNAME );
            exit( 1 );
        }

        /* Composition avec la transformation du fils. */
        ptins->TRANSF = transflat(access_lastpos->RTRANSF, ptins->TRANSF);

    }                           /* Retour a l'indentation normale. */

#undef    access_lastpos
}

/*---------------------------------------------------------\
                     flatpos_seg
\---------------------------------------------------------*/
static void     flatpos_seg(ptfig)
phfig_list     *ptfig;
{
#define    access_lastpos    ( (lastpos_phfig*)ptuser->DATA )

    ptype_list     *ptuser;
    phseg_list     *pseg;
    long            swap;

    /* Recuperation des informations de mise a plat. */
    if ((ptuser = get_ls_data(ptfig, C_SUPP_ULASTPOS, FALSE)) == (ptype_list *) NULL)
    {
        fflush( stdout );
        fprintf( stderr, "*** strecher bugs *** : Unable to flat %s because structure C_SUPP_ULASTPOS missing.\n",
                 ptfig->NAME );
        fprintf( stderr, "                        I'am sorry.\n" );
        exit( 1 );
    }

    /* Boucle sur tous les segments de la figure. */
    for (pseg = ptfig->PHSEG; pseg != (phseg_list *) NULL;
         pseg = pseg->NEXT)
    {
        /* Mise a plat de ( X1, Y1 ). */
        xyflat(&pseg->X1, &pseg->Y1,
               pseg->X1, pseg->Y1,
               access_lastpos->XFLAT, access_lastpos->YFLAT,
               access_lastpos->XAB1, access_lastpos->YAB1,
               access_lastpos->XAB2, access_lastpos->YAB2,
               access_lastpos->RTRANSF);

        /* Mise a plat de ( X2, Y2). */
        xyflat(&pseg->X2, &pseg->Y2,
               pseg->X2, pseg->Y2,
               access_lastpos->XFLAT, access_lastpos->YFLAT,
               access_lastpos->XAB1, access_lastpos->YAB1,
               access_lastpos->XAB2, access_lastpos->YAB2,
               access_lastpos->RTRANSF);

        /* Correction de l'inversion des angles en X. */
        if ((access_lastpos->RTRANSF == SYM_X)
            || (access_lastpos->RTRANSF == SYMXY))
        {
            swap = pseg->X1;
            pseg->X1 = pseg->X2;
            pseg->X2 = swap;
        }

        /* Correction de l'inversion des angles en Y. */
        if ((access_lastpos->RTRANSF == SYM_Y)
            || (access_lastpos->RTRANSF == SYMXY))
        {
            swap = pseg->Y1;
            pseg->Y1 = pseg->Y2;
            pseg->Y2 = swap;
        }
    }

#undef    access_lastpos
}

/*---------------------------------------------------------\
                     flatpos_via
\---------------------------------------------------------*/
static void     flatpos_via(ptfig)
phfig_list     *ptfig;
{
#define    access_lastpos    ( (lastpos_phfig*)ptuser->DATA )

    ptype_list     *ptuser;
    phvia_list     *pvia;

    /* Recuperation des informations de mise a plat. */
    if ((ptuser = get_ls_data(ptfig, C_SUPP_ULASTPOS, FALSE)) == (ptype_list *) NULL)
    {
        fflush( stdout );
        fprintf( stderr, "*** strecher bugs *** : Unable to flat %s because structure C_SUPP_ULASTPOS missing.\n",
                 ptfig->NAME );
        fprintf( stderr, "                        I'am sorry.\n" );
        exit( 1 );
    }

    /* Boucle sur tous les VIAs de la figure. */
    for (pvia = ptfig->PHVIA; pvia != (phvia_list *) NULL; pvia = pvia->NEXT)
        /* Mise a plat de ( XVIA, YVIA ). */
        xyflat(&pvia->XVIA, &pvia->YVIA,
               pvia->XVIA, pvia->YVIA,
               access_lastpos->XFLAT, access_lastpos->YFLAT,
               access_lastpos->XAB1, access_lastpos->YAB1,
               access_lastpos->XAB2, access_lastpos->YAB2,
               access_lastpos->RTRANSF);

#undef    access_lastpos
}

/*---------------------------------------------------------\
                     flatpos_ref
\---------------------------------------------------------*/
static void     flatpos_ref(ptfig)
phfig_list     *ptfig;
{
#define    access_lastpos    ( (lastpos_phfig*)ptuser->DATA )

    ptype_list     *ptuser;
    phref_list     *pref;

    /* Recuperation des informations de mise a plat. */
    if ((ptuser = get_ls_data(ptfig, C_SUPP_ULASTPOS, FALSE)) == (ptype_list *) NULL)
    {
        fflush( stdout );
        fprintf( stderr, "*** strecher bugs *** : Unable to flat %s because structure C_SUPP_ULASTPOS missing.\n",
                 ptfig->NAME );
        fprintf( stderr, "                        I'am sorry.\n" );
        exit( 1 );
    }

    /* Boucle sur tous les VIAs de la figure. */
    for (pref = ptfig->PHREF; pref != (phref_list *) NULL;
         pref = pref->NEXT)
        /* Mise a plat de ( XVIA, YVIA ). */
        xyflat(&pref->XREF, &pref->YREF,
               pref->XREF, pref->YREF,
               access_lastpos->XFLAT, access_lastpos->YFLAT,
               access_lastpos->XAB1, access_lastpos->YAB1,
               access_lastpos->XAB2, access_lastpos->YAB2,
               access_lastpos->RTRANSF);

#undef    access_lastpos
}


/*---------------------------------------------------------\
                     flatpos_con
\---------------------------------------------------------*/
static void     flatpos_con(ptfig)
phfig_list     *ptfig;
{
#define    access_lastpos    ( (lastpos_phfig*)ptuser->DATA )

    ptype_list     *ptuser;
    phcon_list     *pcon;

    /* Recuperation des informations de mise a plat. */
    if ((ptuser = get_ls_data(ptfig, C_SUPP_ULASTPOS, FALSE)) == (ptype_list *) NULL)
    {
        fflush( stdout );
        fprintf( stderr, "*** strecher bugs *** : Unable to flat %s because structure C_SUPP_ULASTPOS missing.\n",
                 ptfig->NAME );
        fprintf( stderr, "                        I'am sorry.\n" );
        exit( 1 );
    }

    /* Boucle sur tous les VIAs de la figure. */
    for (pcon = ptfig->PHCON; pcon != (phcon_list *) NULL;
         pcon = pcon->NEXT)
    {
        /* Mise a plat de ( XCON, YCON ). */
        xyflat(&pcon->XCON, &pcon->YCON,
               pcon->XCON, pcon->YCON,
               access_lastpos->XFLAT, access_lastpos->YFLAT,
               access_lastpos->XAB1, access_lastpos->YAB1,
               access_lastpos->XAB2, access_lastpos->YAB2,
               access_lastpos->RTRANSF);
        orientflat(&pcon->ORIENT,
                   pcon->ORIENT,
                   access_lastpos->RTRANSF);
    }

#undef    access_lastpos
}


/*---------------------------------------------------------\
                     flatpos_intvx
\---------------------------------------------------------*/
static intv_list *flatpos_intvx(ptintv, ptlast)
intv_list      *ptintv;
lastpos_phfig  *ptlast;
{
    intv_list      *pintv;
    intv_list      *pintv_head;
    long            y_coord, swap;

    /* Boucle sur tous les intervalles de la liste. */
    for (pintv = ptintv; pintv != (intv_list *) NULL;
         pintv = pintv->NEXT)
    {
        y_coord = 0;

        /* Mise a plat de la borne inferieure. */
        xyflat(&pintv->INF, &y_coord,
               pintv->INF, y_coord,
               ptlast->XFLAT, ptlast->YFLAT,
               ptlast->XAB1, ptlast->YAB1,
               ptlast->XAB2, ptlast->YAB2,
               ptlast->RTRANSF);

        /* Mise a plat de la borne superieure. */
        xyflat(&pintv->SUP, &y_coord,
               pintv->SUP, y_coord,
               ptlast->XFLAT, ptlast->YFLAT,
               ptlast->XAB1, ptlast->YAB1,
               ptlast->XAB2, ptlast->YAB2,
               ptlast->RTRANSF);

        /* Correction de l'inversion des angles en X. */
        if ((ptlast->RTRANSF == SYM_X) || (ptlast->RTRANSF == SYMXY))
        {
            swap = pintv->INF;
            pintv->INF = pintv->SUP;
            pintv->SUP = swap;
        }
    }
    if ((ptlast->RTRANSF == SYM_X) || (ptlast->RTRANSF == SYMXY))
    {
        pintv_head = ptintv;
        pintv = ptintv->NEXT;
        pintv_head->NEXT = NULL;
        while( pintv )
        {
            ptintv = pintv;
            pintv = pintv->NEXT;
            ptintv->NEXT = pintv_head;
            pintv_head = ptintv;
        }
        ptintv = pintv_head;
    }
    return ptintv;
} /* end of flatpos_intvx */

/*---------------------------------------------------------\
                     flatpos_talu
\---------------------------------------------------------*/
static void     flatpos_talu(ptfig)
phfig_list     *ptfig;
{
#define  access_lastpos  ( (lastpos_phfig*)ptuser_lastpos->DATA )

    ptype_list     *ptuser, *ptuser_lastpos;
    phtalu_list    *ptalu;
    long            y_coord, swap;

    /* Recuperation des information de mise a plat. */
    if ((ptuser_lastpos = get_ls_data(ptfig, C_SUPP_ULASTPOS, FALSE))
                      == (ptype_list *) NULL)
        return;

    /* Recuperation des structure contenant les transparences. */
    if ((ptuser = get_ls_data(ptfig, C_SUPP_UTALU2, FALSE)) != (ptype_list *) NULL)
        /* Boucle sur les transparences ALU2. */
        for (ptalu = (phtalu_list *) ptuser->DATA;
             ptalu != (phtalu_list *) NULL;
             ptalu = ptalu->NEXT)
        {
            /* Mise a plat de ( X1, YTALU ). */
            xyflat(&ptalu->X1, &ptalu->YTALU,
                   ptalu->X1, ptalu->YTALU,
                   access_lastpos->XFLAT, access_lastpos->YFLAT,
                   access_lastpos->XAB1, access_lastpos->YAB1,
                   access_lastpos->XAB2, access_lastpos->YAB2,
                   access_lastpos->RTRANSF);
            /* Mise a plat de ( X2, YTALU ). */
            y_coord = 0;
            xyflat(&ptalu->X2, &y_coord,
                   ptalu->X2, y_coord,
                   access_lastpos->XFLAT, access_lastpos->YFLAT,
                   access_lastpos->XAB1, access_lastpos->YAB1,
                   access_lastpos->XAB2, access_lastpos->YAB2,
                   access_lastpos->RTRANSF);
            /* Correction de l'inversion des angles en X. */
            if ((access_lastpos->RTRANSF == SYM_X) || (access_lastpos->RTRANSF == SYMXY))
            {
                swap = ptalu->X1;
                ptalu->X1 = ptalu->X2;
                ptalu->X2 = swap;
            }

            ptalu->INTV = flatpos_intvx(ptalu->INTV, ptuser_lastpos->DATA);
        }

    /* Recuperation des structure contenant les transparences. */
    if ((ptuser = get_ls_data(ptfig, C_SUPP_UTALU1, FALSE)) != (ptype_list *) NULL)
        /* Boucle sur les transparences ALU1. */
        for (ptalu = (phtalu_list *) ptuser->DATA;
             ptalu != (phtalu_list *) NULL;
             ptalu = ptalu->NEXT)
        {
            /* Mise a plat de ( X1, YTALU ). */
            xyflat(&ptalu->X1, &ptalu->YTALU,
                   ptalu->X1, ptalu->YTALU,
                   access_lastpos->XFLAT, access_lastpos->YFLAT,
                   access_lastpos->XAB1, access_lastpos->YAB1,
                   access_lastpos->XAB2, access_lastpos->YAB2,
                   access_lastpos->RTRANSF);
            /* Mise a plat de ( X2, YTALU ). */
            y_coord = 0;
            xyflat(&ptalu->X2, &y_coord,
                   ptalu->X2, y_coord,
                   access_lastpos->XFLAT, access_lastpos->YFLAT,
                   access_lastpos->XAB1, access_lastpos->YAB1,
                   access_lastpos->XAB2, access_lastpos->YAB2,
                   access_lastpos->RTRANSF);
            /* Correction de l'inversion des angles en X. */
            if ((access_lastpos->RTRANSF == SYM_X) || (access_lastpos->RTRANSF == SYMXY))
            {
                swap = ptalu->X1;
                ptalu->X1 = ptalu->X2;
                ptalu->X2 = swap;
            }

            ptalu->INTV = flatpos_intvx(ptalu->INTV, ptuser_lastpos->DATA);
        }

#undef  access_lastpos
}

/*---------------------------------------------------------\
                     flatpos_duplcut
\---------------------------------------------------------*/
static phcut_list *flatpos_duplcut(ptfig, ptcut, xflat, yflat, atransf)
phfig_list     *ptfig;
phcut_list     *ptcut;
long            xflat;
long            yflat;
char            atransf;
{
#define  access_modlpos  ( (modelepos_phfig*)(ptuser_modlpos->DATA) )

    ptype_list     *ptuser_modlpos;
    phcut_list     *pcut, *pthead;
    long            dxab, dyab, x_coord, y_coord;

    /* Recuperations des structures de mise a plat. ( position initiale du modele ) */
    if ((ptuser_modlpos = get_ls_data(ptfig, C_SUPP_UMODLPOS, FALSE)) == (ptype_list *) NULL)
        return (dup_ls_phcut(ptcut));

    /* Taille de l'abutment-box, quelquesoit la transformation actuelle du modele elle doit etre inchangee. */
    dxab = ptfig->XAB2 - ptfig->XAB1;
    dyab = ptfig->YAB2 - ptfig->YAB1;

    /* Mise a plat des lignes de coupe. */
    pthead = (phcut_list *) NULL;

    for (pcut = ptcut; pcut != (phcut_list *) NULL; pcut = pcut->NEXT)
    {
        x_coord = 0;

        xyflat(&x_coord, &y_coord,
               x_coord, pcut->YCUT,
               xflat, yflat,
               access_modlpos->XMODL, access_modlpos->YMODL,
               access_modlpos->XMODL + dxab, access_modlpos->YMODL + dyab,
               atransf);

        pthead = addophcut(pthead, y_coord, pcut->PITCHNB);
    }

    return (pthead);

#undef  access_modlpos
}

/*---------------------------------------------------------\
                     flatpos_fig
\---------------------------------------------------------*/
static void     flatpos_fig(ptfig, xflat, yflat, atransf)
phfig_list     *ptfig;
long            xflat;
long            yflat;
char            atransf;
{

    flatpos_ab(ptfig, xflat, yflat, atransf);
    flatpos_ins(ptfig);
    flatpos_seg(ptfig);
    flatpos_via(ptfig);
    flatpos_con(ptfig);
    flatpos_ref(ptfig);
    flatpos_talu(ptfig);
}

/*---------------------------------------------------------\
                     unflatpos_fig
\---------------------------------------------------------*/
static void     unflatpos_fig(ptfig)
phfig_list     *ptfig;
{
#define  access_modelepos  ( (modelepos_phfig*)(ptuser->DATA) )

    ptype_list     *ptuser;

    if ((ptuser = get_ls_data(ptfig, C_SUPP_UMODLPOS, FALSE)) == (ptype_list *) NULL)
        return;

    flatpos_fig(ptfig, access_modelepos->XMODL, access_modelepos->YMODL, NOSYM);

#undef  access_modelepos
}

/*---------------------------------------------------------\
                     unflatpos_duplcut
\---------------------------------------------------------*/
static phcut_list *unflatpos_duplcut(ptfig, ptcut)
phfig_list     *ptfig;
phcut_list     *ptcut;
{
#define  access_modlpos  ( (modelepos_phfig*)(ptuser_modlpos->DATA) )
#define  access_lastpos  ( (  lastpos_phfig*)(ptuser_lastpos->DATA) )

    ptype_list     *ptuser_modlpos, *ptuser_lastpos;
    phcut_list     *pcut, *pthead;
    long            x_coord, y_coord;
    char            rtransf;

    /* Recuperations des structures de mise a plat. */
    if ((ptuser_modlpos = get_ls_data(ptfig, C_SUPP_UMODLPOS, FALSE)) == (ptype_list *) NULL)
        return NULL;

    if ((ptuser_lastpos = get_ls_data(ptfig, C_SUPP_ULASTPOS, FALSE)) == (ptype_list *) NULL)
        return NULL;

    /* Calcul de la transformation relative. */
    rtransf = transflat(access_lastpos->ATRANSF, NOSYM);

    /* Calcul des positions des differentes lignes de coupe. */
    pthead = (phcut_list *) NULL;

    for (pcut = ptcut; pcut != (phcut_list *) NULL; pcut = pcut->NEXT)
    {
        x_coord = 0;

        xyflat(&x_coord, &y_coord,
               x_coord, pcut->YCUT,
               access_modlpos->XMODL, access_modlpos->YMODL,
               ptfig->XAB1, ptfig->YAB1,
               ptfig->XAB2, ptfig->YAB2,
               rtransf);

        pthead = addophcut(pthead, y_coord, pcut->PITCHNB);
    }

    return (pthead);

#undef  access_modlpos
#undef  access_lastpos
}

/*---------------------------------------------------------\
                     get_hr_phfig
\---------------------------------------------------------*/
void            get_hr_phfig(ptdupl_father)
dtdupl_list    *ptdupl_father;
{
    phfig_list     *ptfig_father, *ptfig_son;
    phins_list     *ptins_son;

    /* Chargement de la figure de plus haut niveau hierarchique. */
    ptdupl_father->PHFIG = getphfig(ptdupl_father->FIGNAME, 'A');
    ptdupl_father->FLAGS |= F_TDUPL_PHFIG;

    /* donnees de demarrage. */
    ptfig_father = ptdupl_father->PHFIG;

    /* Informations de flatten du pere. */
    flatpos_ab(ptfig_father, ptfig_father->XAB1, ptfig_father->YAB1, NOSYM);

    /* Assure la condition d'arret de la boucle. */
    reset_stack_state();
    push_state((phfig_list *) NULL, (phins_list *) NULL);

    /* Examun des eventuelles instances filles. */
    for (ptins_son = ptfig_father->PHINS; ptfig_father != (phfig_list *) NULL;)
    {
        /* Si la liste des instances est epuisee, retour au(x) niveau(x) superieur(s). */
        if (ptins_son == (phins_list *) NULL)
        {
            pop_state(&ptfig_father, &ptins_son);
            continue;
        }

        /* Chargement du modele de l'instance. */
        ptfig_son = getphfig(ptins_son->FIGNAME, 'A');

        /* Si la liste des instances n'est pas vide, passage au niveau inferieur. */
        if (ptfig_son->PHINS != (phins_list *) NULL)
        {
            push_state(ptfig_father, ptins_son->NEXT);

            ptfig_father = ptfig_son;
            ptins_son = ptfig_son->PHINS;
        }
        else
            ptins_son = ptins_son->NEXT;
    }
}

/*---------------------------------------------------------\
                     extract_hr_allowed
\---------------------------------------------------------*/
void            extract_hr_allowed(ptdupl_father, must_cross)
dtdupl_list    *ptdupl_father;
int             must_cross;
{
#define  access_ls_allowed ( (intv_list*)(ptuser_father->DATA) )

    phfig_list     *ptfig_father, *ptfig_son;
    phins_list     *ptins_son;
    ptype_list     *ptuser_father;
    phcut_list     *check_phcut;

    /* donnees de demarrage. */
    ptfig_father = ptdupl_father->PHFIG;

    /* Extraction des lignes d'alimentations. */
    extract_ls_supply(ptfig_father);

    modify_ls_allowed(ptfig_father, ptfig_father, must_cross);

    /* Assure la condition d'arret de la boucle. */
    reset_stack_state();
    push_state((phfig_list *) NULL, (phins_list *) NULL);

    /* Examun des eventuelles instances filles. */
    for (ptins_son = ptfig_father->PHINS;
         ptfig_father != (phfig_list *) NULL;)
    {
        /* Si la liste des instances est epuisee, retour au(x) niveau(x) superieur(s). */
        if (ptins_son == (phins_list *) NULL)
        {
            pop_state(&ptfig_father, &ptins_son);
            continue;
        }

        /* Recherche du modele de l'instance. */
        ptfig_son = getphfig(ptins_son->FIGNAME, 'A');

        /* Mise a plat des coordonnees du modele. */
        flatpos_fig(ptfig_son, ptins_son->XINS,
                    ptins_son->YINS,
                    ptins_son->TRANSF);

        /* Prise en compte des obstacles. */
        modify_ls_allowed(ptdupl_father->PHFIG, ptfig_son, must_cross);

        /* Si la liste des instances n'est pas vide, passage au niveau inferieur. */
        if (ptfig_son->PHINS != (phins_list *) NULL)
        {
            push_state(ptfig_father, ptins_son->NEXT);

            ptfig_father = ptfig_son;
            ptins_son = ptfig_son->PHINS;
        }
        else
            ptins_son = ptins_son->NEXT;
    }

    /* Normalisation du gabarit des coupes. */
    norm_allowed(ptdupl_father->PHFIG);

    ptuser_father = get_ls_data(ptdupl_father->PHFIG, C_SUPP_UALLOWED, FALSE);
    check_phcut = check_ls_phcut(ptdupl_father->PHCUT, access_ls_allowed);
    free_ls_phcut(ptdupl_father->PHCUT);
    ptdupl_father->PHCUT = check_phcut;

#undef    access_ls_allowed
}

/*---------------------------------------------------------\
                     alloc_hr_dtdupl
\---------------------------------------------------------*/
void            alloc_hr_dtdupl(ptdupl_father)
dtdupl_list    *ptdupl_father;
{
    phfig_list     *ptfig_father, *ptfig_son;
    phins_list     *ptins_son;
    phcut_list     *pcut_son, *pcut_f_son, *pcut;
    dtdupl_list    *ptdupl_son;
    long            flags, new_dupl;

    /* donnees de demarrage. */
    ptdupl_father->PHFIG = getphfig(ptdupl_father->FIGNAME, 'A');
    ptfig_father = ptdupl_father->PHFIG;

    /* Duplication de la figure physique. */
    ptdupl_father->PHDUP = addphfig(ptdupl_father->DUPNAME);

    dupphfig(ptdupl_father->PHFIG, ptdupl_father->PHDUP);
    dupuser(ptdupl_father->PHFIG, ptdupl_father->PHDUP);

    ptdupl_father->FLAGS |= F_TDUPL_PHDUP;
    ptfig_father = ptdupl_father->PHDUP;

    /* Assure la condition d'arret de la boucle. */
    reset_stack_state();
    push_state((phfig_list *) NULL, (phfig_list *) NULL);

    /* Examun des eventuelles instances filles. */
    for (ptins_son = ptfig_father->PHINS; ptfig_father != (phfig_list *) NULL;)
    {
        /* Si la liste des instances est epuisee, retour au(x) niveau(x) superieur(s). */
        if (ptins_son == (phins_list *) NULL)
        {
            pop_state(&ptfig_father, &ptins_son);
            continue;
        }

        /* Recherche du modele de l'instance. */
        ptfig_son = getphfig(ptins_son->FIGNAME, 'A');

        /* Mise a plat des coordonnees du modele. */
        flatpos_fig(ptfig_son, ptins_son->XINS, ptins_son->YINS, ptins_son->TRANSF);

        /* Reperage des lignes de coupes du bloc traversant le modele de l'instance fille. */
        pcut_son = (phcut_list *) NULL;
        for (pcut = ptdupl_father->PHCUT; pcut != (phcut_list *) NULL; pcut = pcut->NEXT)
            if ((pcut->YCUT <= ptfig_son->YAB2) && (pcut->YCUT > ptfig_son->YAB1))
                pcut_son = addophcut(pcut_son, pcut->YCUT, pcut->PITCHNB);

        new_dupl = FALSE;

        /* Traitement de duplication. */
        if (pcut_son != (phcut_list *) NULL)
        {
            /* Expression des lignes de coupes dans le repere initial du repere
               (toutes les lignes de coupes des duplicatas sont conservees dans ce repere.
            */
            pcut_f_son = unflatpos_duplcut(ptfig_son, pcut_son);
            free_ls_phcut(pcut_son);

            /* Creation, si necessaire, d'un nouveau duplicata. */
            if ((ptdupl_son = getdtdupl(ptfig_son->NAME, pcut_f_son)) == (dtdupl_list *) NULL)
            {
                new_dupl = TRUE;

                /* Caracterisation initiale du bloc. */
                flags = (ptfig_son->PHINS == (phins_list *) NULL) ? F_TDUPL_TERMINAL : FALSE;

                /* Creation d'un nouveau duplicata. */
                ptdupl_son = allocdtdupl(ptfig_son->NAME, pcut_f_son, flags);
                nameddtdupl(ptdupl_son, "*", pcut_f_son);

                /* Duplication de la figure physique. */
                ptdupl_son->PHFIG = ptfig_son;
                ptdupl_son->PHDUP = addphfig(ptdupl_son->DUPNAME);

                dupphfig(ptdupl_son->PHFIG, ptdupl_son->PHDUP);
                dupuser(ptdupl_son->PHFIG, ptdupl_son->PHDUP);

                ptdupl_son->FLAGS |= F_TDUPL_PHDUP;
            }
            else
                /* Mise a plat des coordonnees du duplicata. ( pour poursuivre correctement le processus de mise a plat hierarchique ). */
                flatpos_fig(ptdupl_son->PHDUP, ptins_son->XINS,
                            ptins_son->YINS,
                            ptins_son->TRANSF);

            ptins_son->FIGNAME = ptdupl_son->DUPNAME;
            ptfig_son = ptdupl_son->PHDUP;

            /* unflatpos_fig( ptfig_son); */
        }

        /* Si la liste des instances n'est pas vide, passage au niveau inferieur. */
        if ((new_dupl == TRUE) && (ptfig_son->PHINS != (phins_list *) NULL))
        {
            push_state(ptfig_father, ptins_son->NEXT);

            ptfig_father = ptfig_son;
            ptins_son = ptfig_son->PHINS;
        }
        else
            ptins_son = ptins_son->NEXT;
    }

}

/*---------------------------------------------------------\
                     strech_hr_dtdupl
\---------------------------------------------------------*/
void            strech_hr_dtdupl(ptdupl_father, must_cross)
dtdupl_list    *ptdupl_father;
int             must_cross;
{
#define  access_supply   ( (  supply_list*)(ptuser1_father->DATA) )
#define  access_lastpos  ( (lastpos_phfig*)(ptuser2_father->DATA) )

    dtdupl_list    *ptdupl_son;
    supply_list    *psupply;
    phfig_list     *ptfig_father, *ptfig_son;
    phins_list     *ptins_son;
    ptype_list     *ptuser1_father, *ptuser2_father;
    phcut_list     *pcut, *pflat_cut;
    long            y_sum;

    /* donnees de demarrage. */
    ptfig_father = ptdupl_father->PHDUP;

    /* Cas ou le duplicata a deja ete streche. */
    if ((ptdupl_father->FLAGS & F_TDUPL_STRECH) || (ptdupl_father->PHDUP == (phfig_list *) NULL))
        return;

    /* Remise du bloc en position initiale. */
    unflatpos_fig(ptfig_father);

    ptuser1_father = get_ls_data(ptdupl_father->PHFIG, C_SUPP_USUPPLY, FALSE);
    ptuser2_father = get_ls_data(ptfig_father, C_SUPP_ULASTPOS, FALSE);

    /* Verification de l'extraction. */
    if (access_supply == (void *) NULL)
    {
        fflush( stdout );
        fprintf( stderr,
                 "*** strecher error *** : Alimentation connectors not found in  %s.\n",
                 ptdupl_father->PHFIG->NAME );
        exit( 1 );
    }

    /* Mise a plat des lignes de coupes. */
    pflat_cut = flatpos_duplcut(ptfig_father,
                                ptdupl_father->PHCUT,
                                access_lastpos->XFLAT,
                                access_lastpos->YFLAT,
                                access_lastpos->ATRANSF);

    /* Boucle sur la liste des lignes de coupes. */
    for (y_sum = 0,
         pcut = pflat_cut; pcut != (phcut_list *) NULL;
         pcut = pcut->NEXT)
    {
        /* Recuperation des alims encadrant la coupe. */
        psupply = getnsupply(access_supply, pcut->YCUT);
        if ((psupply == (supply_list *) NULL) || (psupply->NEXT == (supply_list *) NULL))
        {
            fflush( stdout );
            fprintf( stderr,
                     "*** strecher error *** : Alimentation connectors not in slice position in  %s.\n",
                     ptdupl_father->PHFIG->NAME );
            exit( 1 );
        }

        /* Streching. */
        y_sum += strech_cell(ptfig_father, psupply, pcut->YCUT + y_sum, pcut->PITCHNB, must_cross);
    }
    ptdupl_father->FLAGS |= F_TDUPL_STRECH;

    /* Assure la condition d'arret de la boucle. */
    reset_stack_state();
    push_state((phfig_list *) NULL, (phins_list *) NULL);

    /* Examun des eventuelles instances filles. */
    for (ptins_son = ptfig_father->PHINS;
         ptfig_father != (phfig_list *) NULL;)
    {
        /* Si la liste des instances est epuisee, retour au(x) niveau(x) superieur(s). */
        if (ptins_son == (phins_list *) NULL)
        {
            pop_state(&ptfig_father, &ptins_son);
            continue;
        }

        /* Recherche du modele de l'instance. */
        ptfig_son = getphfig(ptins_son->FIGNAME, 'A');

        /* Mise a plat des coordonnees du modele. */
        flatpos_fig(ptfig_son, ptins_son->XINS,
                    ptins_son->YINS,
                    ptins_son->TRANSF);

        /* Traitement si le duplicata n'a pas deja ete streche. */
        if (((ptdupl_son = getdtdupl_byphdup(ptfig_son)) != (dtdupl_list *) NULL)
            && !(ptdupl_son->FLAGS & F_TDUPL_STRECH))
        {
            /* Boucle sur la liste des lignes de coupes. */
            for (y_sum = 0,
                 pcut = pflat_cut; pcut != (phcut_list *) NULL;
                 pcut = pcut->NEXT)
            {
                /* Toutes les lignes secantes ont etes prises en compte (triees dans l'ordre croissant). */
                if (pcut->YCUT + y_sum > ptfig_son->YAB2)
                    break;

                /* Lignes de coupes "inferieures". */
                if (pcut->YCUT + y_sum <= ptfig_son->YAB1)
                {
                    y_sum += pcut->PITCHNB * SX_PITCH_TALU2;
                    continue;
                }

                /* Recuperation des alims encadrant la coupe. */
                psupply = getnsupply(access_supply, pcut->YCUT);
                if ((psupply == (supply_list *) NULL) || (psupply->NEXT == (supply_list *) NULL))
                {
                    fflush( stdout );
                    fprintf( stderr,
                             "*** strecher error *** : Alimentation connectors not in slice position in  %s.\n",
                             ptdupl_father->PHFIG->NAME );
                    exit( 1 );
                }

                /* Streching. */
                y_sum += strech_cell(ptfig_son, psupply,
                                     pcut->YCUT + y_sum,
                                     pcut->PITCHNB,must_cross);
            }

            ptdupl_son->FLAGS |= F_TDUPL_STRECH;
        }


        /* Si la liste des instances n'est pas vide, passage au niveau inferieur. */
        if (ptfig_son->PHINS != (phins_list *) NULL)
        {
            push_state(ptfig_father, ptins_son->NEXT);

            ptfig_father = ptfig_son;
            ptins_son = ptfig_son->PHINS;
        }
        else
            ptins_son = ptins_son->NEXT;
    }

    /* Streching des lignes de coupes. */
    strech_supply(ptdupl_father->PHDUP, pflat_cut);

    /* Effacement de la liste des coupes mises a plat. */
    free_ls_phcut(pflat_cut);

#undef    access_supply
#undef    access_lastpos
}

/*---------------------------------------------------------\
                     alloc_hr_talu
\---------------------------------------------------------*/
void            alloc_hr_talu(ptdupl_father)
dtdupl_list    *ptdupl_father;
{
#define  access_supply   ( (  supply_list*)(ptuser1_father->DATA) )
#define  access_lastpos  ( (lastpos_phfig*)(ptuser2_father->DATA) )

    dtdupl_list    *ptdupl_son;
    supply_list    *psupply;
    phfig_list     *ptfig_father, *ptfig_son;
    phins_list     *ptins_son;
    ptype_list     *ptuser1_father, *ptuser2_father;
    phcut_list     *pcut, *pflat_cut;
    long            y_norm = 0, y_sum;

    /* donnees de demarrage. */
    ptfig_father = ptdupl_father->PHDUP;

    /* Cas ou les gabarits ont deja ete alloues. */
    if ((ptdupl_father->FLAGS & F_TDUPL_TALLOC)
        || (ptdupl_father->PHDUP == (phfig_list *) NULL))
        return;

    /* Remise du bloc en position initiale. */
    unflatpos_fig(ptfig_father);

    ptuser1_father = get_ls_data(ptdupl_father->PHFIG,
                                 C_SUPP_USUPPLY,
                                 FALSE);
    ptuser2_father = get_ls_data(ptdupl_father->PHDUP,
                                 C_SUPP_ULASTPOS,
                                 FALSE);

    /* Verification de l'extraction. */
    if (access_supply == (void *) NULL)
    {
        fflush( stdout );
        fprintf( stderr,
                 "*** strecher bugs *** : Unable to flat %s's instances (missing C_SUPP_ULASTPOS ptype).\n",
                 ptdupl_father->PHDUP->NAME);
        exit( 1 );
    }

    /* Mise a plat des lignes de coupes. */
    pflat_cut = flatpos_duplcut(ptfig_father,
                                ptdupl_father->PHCUT,
                                access_lastpos->XFLAT,
                                access_lastpos->YFLAT,
                                access_lastpos->ATRANSF);

    /* Boucle sur la liste des lignes de coupes. */
    for (y_sum = 0,
         pcut = pflat_cut; pcut != (phcut_list *) NULL;
         pcut = pcut->NEXT)
    {
        /* Recuperation des alims encadrant la coupe. */
        psupply = getnsupply(access_supply, pcut->YCUT);
        if ((psupply == (supply_list *) NULL)
            || (psupply->NEXT == (supply_list *) NULL))
        {
            fflush( stdout );
            fprintf( stderr,
                     "*** strecher error *** : Alimentation connectors not in slice position in  %s.\n",
                     ptdupl_father->PHFIG->NAME );
            exit( 1 );
        }

        /* Normalisation de l'ordonnee de coupe. ( en vue d'ajouter les transparences ) */
        y_norm = ycut_norm(psupply->YSUPPLY,
                           psupply->TYPE,
                           psupply->NEXT->YSUPPLY,
                           psupply->NEXT->TYPE,
                           pcut->YCUT);
        y_norm += y_sum;

        y_sum += pcut->PITCHNB * SX_PITCH_TALU2;

        /* Si le duplicata est celui d'un bloc ou d'une cellule feuille alors on lui ajoute des transparences. */
        alloc_ls_talu(ptfig_father, y_norm, pcut->PITCHNB,
                      F_PHTALU_CMA | F_PHTALU_TALU2);

    }
    ptdupl_father->FLAGS |= F_TDUPL_TALLOC;

    /* Assure la condition d'arret de la boucle. */
    reset_stack_state();
    push_state((phfig_list *) NULL, (phins_list *) NULL);

    /* Examun des eventuelles instances filles. */
    for (ptins_son = ptfig_father->PHINS;
         ptfig_father != (phfig_list *) NULL;)
    {
        /* Si la liste des instances est epuisee, retour au(x) niveau(x) superieur(s). */
        if (ptins_son == (phins_list *) NULL)
        {
            pop_state(&ptfig_father, &ptins_son);
            continue;
        }

        /* Recherche du modele de l'instance. */
        ptfig_son = getphfig(ptins_son->FIGNAME, 'A');

        /* Mise a plat des coordonnees du modele. */
        flatpos_fig(ptfig_son, ptins_son->XINS,
                    ptins_son->YINS,
                    ptins_son->TRANSF);

        /* Traitement si le duplicata n'a pas deja ete streche. */
        if (((ptdupl_son = getdtdupl_byphdup(ptfig_son))
             != (dtdupl_list *) NULL)
            && !(ptdupl_son->FLAGS & F_TDUPL_TALLOC))
        {
            /* Boucle sur la liste des lignes de coupes. */
            for (y_sum = 0,
                 pcut = pflat_cut; pcut != (phcut_list *) NULL;
                 pcut = pcut->NEXT)
            {
                /* Toutes les lignes secantes ont etes prises en compte (triees dans l'ordre croissant). */
                if (pcut->YCUT + y_sum > ptfig_son->YAB2)
                    break;

                /* Lignes de coupes "inferieures". */
                if (pcut->YCUT + y_sum <= ptfig_son->YAB1)
                {
                    y_sum += pcut->PITCHNB * SX_PITCH_TALU2;
                    continue;
                }

                if (ptfig_son->PHINS == (phins_list *) NULL)
                {
                    /* Recuperation des alims encadrant la coupe. */
                    psupply = getnsupply(access_supply, pcut->YCUT);
                    if ((psupply == (supply_list *) NULL) || (psupply->NEXT == (supply_list *) NULL))
                    {
                        fflush( stdout );
                        fprintf( stderr,
                                 "*** strecher error *** : Alimentation connectors not in slice position in  %s.\n",
                                 ptdupl_father->PHFIG->NAME );
                        exit( 1 );
                    }

                    /* Normalisation de l'ordonnee de coupe. ( en vue d'ajouter les transparences ) */
                    y_norm = ycut_norm(psupply->YSUPPLY,
                                       psupply->TYPE,
                                       psupply->NEXT->YSUPPLY,
                                       psupply->NEXT->TYPE,
                                       pcut->YCUT);
                    y_norm += y_sum;
                }

                /* Streching. */
                y_sum += pcut->PITCHNB * SX_PITCH_TALU2;

                /* Si le duplicata est celui d'une cellule feuille alors on lui ajoute des transparences. */
                if (ptfig_son->PHINS == (phins_list *) NULL)
                    alloc_ls_talu(ptfig_son, y_norm, pcut->PITCHNB, F_PHTALU_CMA | F_PHTALU_TALU2);
            }

            ptdupl_son->FLAGS |= F_TDUPL_TALLOC;
        }


        /* Si la liste des instances n'est pas vide, passage au niveau inferieur. */
        if (ptfig_son->PHINS != (phins_list *) NULL)
        {
            push_state(ptfig_father, ptins_son->NEXT);

            ptfig_father = ptfig_son;
            ptins_son = ptfig_son->PHINS;
        }
        else
            ptins_son = ptins_son->NEXT;
    }

    /* Effacement de la liste des coupes mises a plat. */
    free_ls_phcut(pflat_cut);

#undef    access_supply
#undef    access_lastpos
} /* end of alloc_hr_talu */

/*---------------------------------------------------------\
                     extract_hr_cma
\---------------------------------------------------------*/
void            extract_hr_cma(ptdupl_father)
dtdupl_list    *ptdupl_father;
{
#define  access_supply  ( (supply_list*)(ptuser_father->DATA) )

    dtdupl_list    *pdupl_son;
    phfig_list     *ptfig_father, *ptfig_son;
    phins_list     *ptins_son;
    ptype_list     *ptuser_father;

    /* donnees de demarrage. */
    ptfig_father = ptdupl_father->PHDUP;

    ptuser_father = get_ls_data(ptfig_father, C_SUPP_USUPPLY, FALSE);

    /* Verification de l'extraction. */
    if (access_supply == (void *) NULL)
    {
        fflush( stdout );
        fprintf( stderr,
                 "*** strecher error *** : Alimentation connectors not found in  %s.\n",
                 ptdupl_father->PHFIG->NAME );
        exit( 1 );
    }

    /* Extraction des CMAs du pere. ( en ne considerant que les "ref_con" ) */
    extract_ls_cmaref(ptfig_father, access_supply);

    /* Assure la condition d'arret de la boucle. */
    reset_stack_state();
    push_state((phfig_list *) NULL, (phins_list *) NULL);

    /* Examun des eventuelles instances filles. */
    for (ptins_son = ptfig_father->PHINS;
         ptfig_father != (phfig_list *) NULL;)
    {
        /* Si la liste des instances est epuisee, retour au(x) niveau(x) superieur(s). */
        if (ptins_son == (phins_list *) NULL)
        {
            pop_state(&ptfig_father, &ptins_son);
            continue;
        }

        /* Recherche du modele de l'instance. */
        ptfig_son = getphfig(ptins_son->FIGNAME, 'A');

        /* Mise a plat des coordonnees du modele. */
        flatpos_fig(ptfig_son, ptins_son->XINS,
                    ptins_son->YINS,
                    ptins_son->TRANSF);

        /* Cas ou les CMAs du duplicata on deja etes extraits, ou la cellule n'est pas "feuille". */
        if (((pdupl_son = getdtdupl_byphdup(ptfig_son)) != (dtdupl_list *) NULL)
            && !(pdupl_son->FLAGS & F_TDUPL_CMA)
            && (pdupl_son->FLAGS & F_TDUPL_TERMINAL))
        {
            /* Extraction des CMAs. */
            extract_ls_cma(pdupl_son->PHDUP, access_supply);

            pdupl_son->FLAGS |= F_TDUPL_CMA;
        }

        /* Traitement uniquement dans le cas des cellules feuilles : recherche des CMAs du fils coincidant avec des "ref_con" du pere. */
        (void) link_ls_cma(ptdupl_father->PHDUP,
                           ptfig_son,
                           ptins_son->XINS,
                           ptins_son->YINS,
                           ptins_son->TRANSF);

        /* Si la liste des instances n'est pas vide, passage au niveau inferieur. */
        if (ptfig_son->PHINS != (phins_list *) NULL)
        {
            push_state(ptfig_father, ptins_son->NEXT);

            ptfig_father = ptfig_son;
            ptins_son = ptfig_son->PHINS;
        }
        else
            ptins_son = ptins_son->NEXT;
    }
}

/*---------------------------------------------------------\
                     modify_ln_talu
\---------------------------------------------------------*/
void            modify_ln_talu()
{
    dtdupl_list    *pdupl;

    /* Parcourt de la liste des duplicatas. */
    for (pdupl = TABLE_DUPL; pdupl != (dtdupl_list *) NULL;
         pdupl = pdupl->NEXT)
    {
        /* Cas ou les gabarits du duplicata on deja etes retailles, ou la cellule n'est pas "feuille". */
        if ((pdupl->FLAGS & F_TDUPL_TALU) || !(pdupl->FLAGS & F_TDUPL_TERMINAL))
            continue;

        /* Retaillage des gabarits. */
        modify_ls_talu(pdupl->PHDUP, pdupl->PHDUP, F_PHTALU_TALU2 | F_PHTALU_CMA);

        pdupl->FLAGS |= F_TDUPL_TALU;
    }

} /* end of modify_ln_talu */

/*---------------------------------------------------------\
                     modify_hr_talu
\---------------------------------------------------------*/
void            modify_hr_talu(ptdupl_father)
dtdupl_list    *ptdupl_father;
{
    phfig_list     *ptfig_father, *ptfig_son;
    phins_list     *ptins_son;

    ptdupl_father->FLAGS |= F_TDUPL_TALU;

    /* donnees de demarrage. */
    ptfig_father = ptdupl_father->PHDUP;
    modify_ls_talu(ptfig_father, ptfig_father, F_PHTALU_TALU2 | F_PHTALU_CMA);

    /* Assure la condition d'arret de la boucle. */
    reset_stack_state();
    push_state((phfig_list *) NULL, (phins_list *) NULL);

    /* Examun des eventuelles instances filles. */
    for (ptins_son = ptfig_father->PHINS;
         ptfig_father != (phfig_list *) NULL;)
    {
        /* Si la liste des instances est epuisee, retour au(x) niveau(x) superieur(s). */
        if (ptins_son == (phins_list *) NULL)
        {
            pop_state(&ptfig_father, &ptins_son);
            continue;
        }

        /* Recuperation du modele de l'instance. */
        ptfig_son = getphfig(ptins_son->FIGNAME, 'A');

        /* Mise a plat des coordonnees du modele. */
        flatpos_fig(ptfig_son, ptins_son->XINS,
                    ptins_son->YINS,
                    ptins_son->TRANSF);

        /* Prise en compte des composants du fils. */
        modify_ls_talu(ptfig_father, ptfig_son,
                       F_PHTALU_TALU2 | F_PHTALU_CMA);

        /* Si la liste des instances n'est pas vide, passage au niveau inferieur. */
        if (ptfig_son->PHINS != (phins_list *) NULL)
        {
            push_state(ptfig_father, ptins_son->NEXT);

            ptfig_father = ptfig_son;
            ptins_son = ptfig_son->PHINS;
        }
        else
            ptins_son = ptins_son->NEXT;
    }

} /* end of modify_hr_talu */

/*---------------------------------------------------------\
                     add_hr_through
\---------------------------------------------------------*/
void            add_hr_through(ptdupl_father)
dtdupl_list    *ptdupl_father;
{
#define  access_supply  ( (supply_list*)(ptuser_father->DATA) )

    dtdupl_list    *ptdupl_son;
    phfig_list     *ptfig_father, *ptfig_son;
    phins_list     *ptins_son;
    ptype_list     *ptuser_father;

    /* Le travail a deja ete effectue. */
    if (ptdupl_father->FLAGS & F_TDUPL_THROUGH)
        return;

    ptdupl_father->FLAGS |= F_TDUPL_THROUGH;

    /* donnees de demarrage. */
    ptfig_father = ptdupl_father->PHDUP;

    ptuser_father = get_ls_data(ptfig_father, C_SUPP_USUPPLY, FALSE);

    /* Verification de l'extraction. */
    if (access_supply == (void *) NULL)
    {
        fflush( stdout );
        fprintf( stderr,
                 "*** strecher error *** : Alimentation connectors not found in  %s.\n",
                 ptdupl_father->PHFIG->NAME );
        exit( 1 );
    }

    add_ls_through(ptfig_father, access_supply, F_PHTALU_TALU2 | F_PHTALU_CMA);

    /* Assure la condition d'arret de la boucle. */
    reset_stack_state();
    push_state((phfig_list *) NULL, (phins_list *) NULL);

    /* Examun des eventuelles instances filles. */
    for (ptins_son = ptfig_father->PHINS;
         ptfig_father != (phfig_list *) NULL;)
    {
        /* Si la liste des instances est epuisee, retour au(x) niveau(x) superieur(s). */
        if (ptins_son == (phins_list *) NULL)
        {
            pop_state(&ptfig_father, &ptins_son);
            continue;
        }

        /* Recuperation du modele de l'instance. */
        ptfig_son = getphfig(ptins_son->FIGNAME, 'A');

        /* Mise a plat des coordonnees du modele. */
        flatpos_fig(ptfig_son, ptins_son->XINS,
                    ptins_son->YINS,
                    ptins_son->TRANSF);

        if (((ptdupl_son = getdtdupl_byphdup(ptfig_son)) != (dtdupl_list *) NULL)
            && !(ptdupl_son->FLAGS & F_TDUPL_THROUGH))
        {
            /* Pose des transparences. */
            add_ls_through(ptfig_son, access_supply, F_PHTALU_TALU2 | F_PHTALU_CMA);
            check_ls_through(ptfig_son);

            ptdupl_son->FLAGS |= F_TDUPL_THROUGH;
        }

        /* Exportation des "ref_" du fils. */
        export_ls_through(ptdupl_father->PHDUP, ptfig_son);

        /* Si la liste des instances n'est pas vide, passage au niveau inferieur. */
        if (ptfig_son->PHINS != (phins_list *) NULL)
        {
            push_state(ptfig_father, ptins_son->NEXT);

            ptfig_father = ptfig_son;
            ptins_son = ptfig_son->PHINS;
        }
        else
            ptins_son = ptins_son->NEXT;
    }

    /* Verification de l'ensemble des "ref_con" ajoutes. */
    check_ls_through(ptdupl_father->PHDUP);

    ptdupl_father->FLAGS |= F_TDUPL_THROUGH;

}

/*---------------------------------------------------------\
                     clean_ln_through
\---------------------------------------------------------*/
void clean_ln_through()
{
    dtdupl_list    *pdupl;

    /* Parcourt de la liste des duplicatas. */
    for (pdupl = TABLE_DUPL; pdupl != (dtdupl_list *) NULL; pdupl = pdupl->NEXT)
        /* Destruction des "ref_". */
        if ( (pdupl->PHDUP) && ( !(pdupl->FLAGS & F_TDUPL_CLEAN )) )
        {
            clean_ls_through(pdupl->PHDUP);
            pdupl->FLAGS |= F_TDUPL_CLEAN;
        }
} /* end of clean_ln_through */

/*---------------------------------------------------------\
                     save_ln_dtdupl
\---------------------------------------------------------*/
void save_ln_dtdupl()
{
    dtdupl_list    *pdupl;

    /* Parcourt de la liste des duplicatas. */
    for (pdupl = TABLE_DUPL; pdupl != (dtdupl_list *) NULL;
         pdupl = pdupl->NEXT)
    {
        if (pdupl->PHDUP)
        {
            unflatpos_fig(pdupl->PHDUP);
            savephfig(pdupl->PHDUP);
        }
    }

} /* end of save_ln_dtdupl */
