/*------------------------------------------------------------\
|                                                             |
| Tool    :                   DREAL                           |
|                                                             |
| File    :                  Create.c                         |
|                                                             |
| Author  :                Jacomme Ludovic                    |
|                                                             |
| Date    :                  28.03.95                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <Xm/Xm.h>
# include <Xm/PushBG.h>
# include <Xm/CascadeBG.h>
 
# include MUT_H
# include MPH_H
# include RDS_H
# include RPR_H
# include RFM_H
# include GRD_H
# include GMX_H
# include GTB_H
# include GSB_H
# include GMC_H

# include "GMC_create.h"
# include "GMC_dialog.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                           Rectangle                         |
|                                                             |
\------------------------------------------------------------*/

  char  DrealRectangleLayer = 0;
  char *DrealRectangleName  = (char *)NULL;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      DrealCreateFigureRds                   |
|                                                             |
\------------------------------------------------------------*/

void DrealCreateFigureRds()

{
  if ( DrealFigureRds == (rdsfig_list *)NULL )
  {
    rdsbegin();

    DrealNewFigure( DREAL_DEFAULT_FIGURE_NAME );

    DrealChangeTopLevelTitle( DrealFigureRds->NAME );

    rdsend();
  }
}  

/*------------------------------------------------------------\
|                                                             |
|                     DrealCreateRectangleRds                 |
|                                                             |
\------------------------------------------------------------*/

void DrealCreateRectangleRds( LambdaX1, LambdaY1, LambdaX2, LambdaY2 )

   long LambdaX1;
   long LambdaY1;
   long LambdaX2;
   long LambdaY2;
{
  rdsrec_list *Rectangle;
  long         Swap;

  rdsbegin();

  if ( ( LambdaX1 != LambdaX2 ) &&
       ( LambdaY1 != LambdaY2 ) )
  {
    if ( LambdaX1 > LambdaX2 )
    {
      Swap = LambdaX1; LambdaX1 = LambdaX2; LambdaX2 = Swap;
    }

    if ( LambdaY1 > LambdaY2 )
    {
      Swap = LambdaY1; LambdaY1 = LambdaY2; LambdaY2 = Swap;
    }
  }
  else
  {
    DrealErrorMessage( DrealMainWindow, "This rectangle is too small !" );

    rdsend();
    return;
  }

  DrealCreateFigureRds();

  Rectangle = DrealAddRectangle( DrealRectangleName,
                                 DrealRectangleLayer,
                                 LambdaX1 * RDS_PHYSICAL_GRID,
                                 LambdaY1 * RDS_PHYSICAL_GRID,
                                 ( LambdaX2 - LambdaX1 ) * RDS_PHYSICAL_GRID,
                                 ( LambdaY2 - LambdaY1 ) * RDS_PHYSICAL_GRID );

  DrealAddUndo();
  DrealAddUndoRec( Rectangle );

  DrealDisplayRectangle( Rectangle );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                   Dreal Create Abutmentbox                  |
|                                                             |
\------------------------------------------------------------*/

void DrealCreateAbutmentBoxRds( LambdaX1, LambdaY1, LambdaX2, LambdaY2 )

  long LambdaX1;
  long LambdaY1;
  long LambdaX2;
  long LambdaY2;
{
  rdsrec_list *Rectangle;
  long         Swap;

  rdsbegin();

  if ( ( LambdaX1 != LambdaX2 ) &&
       ( LambdaY1 != LambdaY2 ) )
  {
    if ( LambdaX1 > LambdaX2 )
    {
      Swap = LambdaX1; LambdaX1 = LambdaX2; LambdaX2 = Swap;
    }

    if ( LambdaY1 > LambdaY2 )
    {
      Swap = LambdaY1; LambdaY1 = LambdaY2; LambdaY2 = Swap;
    }
  }  
  else
  {
    DrealErrorMessage( DrealMainWindow, "The abutment box must be rectangular !" );

    rdsend();
    return;
  }

  DrealCreateFigureRds();

  Rectangle = DrealAddAbox( LambdaX1 * RDS_PHYSICAL_GRID,
                            LambdaY1 * RDS_PHYSICAL_GRID,
                            ( LambdaX2 - LambdaX1 ) * RDS_PHYSICAL_GRID,
                            ( LambdaY2 - LambdaY1 ) * RDS_PHYSICAL_GRID );

  DrealAddUndo();
  DrealAddUndoRec( Rectangle );

  DrealDisplayRectangle( Rectangle );
  
  DrealChangeEditMode( DREAL_EDIT_MEASURE,
                       DrealPromptEditMeasure );

  rdsend();
}
