/*------------------------------------------------------------\
|                                                             |
| Tool    :                   DREAL                           |
|                                                             |
| File    :                  Select.c                         |
|                                                             |
| Author  :                Jacomme Ludovic                    |
|                                                             |
| Date    :                  28.03.95                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include MUT_H
# include MPH_H
# include RDS_H
# include RPR_H
# include RFM_H
# include GRD_H
# include GSB_H

# include "GRD_window.h"
# include "GRD_select.h"

/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  drealselect *DrealHeadSelect  = (drealselect *)NULL;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                       Alloc Functions                       |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        DrealAllocSelect                     |
|                                                             |
\------------------------------------------------------------*/

drealselect *DrealAllocSelect()
{
  return((drealselect *)rdsalloc(sizeof(drealselect), 1));
}

/*------------------------------------------------------------\
|                                                             |
|                        Free Functions                       |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         DrealFreeSelect                     |
|                                                             |
\------------------------------------------------------------*/

void DrealFreeSelect( FreeSelect )

   drealselect *FreeSelect;
{
  rdsfree((char *)FreeSelect, sizeof(drealselect));
}

/*------------------------------------------------------------\
|                                                             |
|                       DrealAcceptRectangle                  |
|                                                             |
\------------------------------------------------------------*/
 
void DrealAcceptRectangle( Rectangle )
 
  rdsrec_list *Rectangle;
{
  rdsbegin();
 
  SetDrealAccepted( Rectangle );

  rdsend();
}

/*------------------------------------------------------------\ 
|                                                             | 
|                       DrealRejectRectangle                  | 
|                                                             |
\------------------------------------------------------------*/ 

void DrealRejectRectangle( Rectangle ) 

  rdsrec_list *Rectangle;
{ 
  rdsbegin();

  ClearDrealAccepted( Rectangle ); 

  rdsend();
} 

/*------------------------------------------------------------\
|                                                             |
|                       DrealSelectRectangle                  |
|                                                             |
\------------------------------------------------------------*/

void DrealSelectRectangle( Rectangle )

  rdsrec_list *Rectangle;
{
  rdsbegin();

  SetDrealSelected( Rectangle );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                      DrealUnselectRectangle                 |
|                                                             |
\------------------------------------------------------------*/

void DrealUnselectRectangle( Rectangle )

  rdsrec_list *Rectangle;
{
  rdsbegin();

  ClearDrealSelected( Rectangle );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                       DrealDeleteRectangle                  |
|                                                             |
\------------------------------------------------------------*/

void DrealDeleteRectangle( Rectangle )

  rdsrec_list *Rectangle;
{
  rdsbegin();

  SetDrealDeleted( Rectangle );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                     DrealUndeleteRectangle                  |
|                                                             |
\------------------------------------------------------------*/

void DrealUndeleteRectangle( Rectangle )

  rdsrec_list *Rectangle;
{
  rdsbegin();

  ClearDrealDeleted( Rectangle );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                       DrealAddSelect                        |
|                                                             |
\------------------------------------------------------------*/

void DrealAddSelect( Rectangle )

  rdsrec_list *Rectangle;
{
  drealselect *Select;

  rdsbegin();

  Select = DrealAllocSelect();

  Select->RECTANGLE = Rectangle;
  Select->NEXT      = DrealHeadSelect;
  DrealHeadSelect   = Select;

  DrealSelectRectangle( Rectangle );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                       DrealDelSelect                        |
|                                                             |
\------------------------------------------------------------*/

void DrealDelSelect()
{
  drealselect    *DelSelect;
  drealselect    *Select;

  rdsbegin();

  Select          = DrealHeadSelect;
  DrealHeadSelect = (drealselect *)NULL;

  while( Select != (drealselect *)NULL )
  {
    DrealRejectRectangle( Select->RECTANGLE );

    DelSelect  = Select;
    Select     = Select->NEXT;
    
    DrealFreeSelect( DelSelect );
  }

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                       DrealPurgeSelect                      |
|                                                             |
\------------------------------------------------------------*/

void DrealPurgeSelect()
{
  drealselect    *DelSelect;
  drealselect    *Select;
  drealselect   **Previous;

  rdsbegin();

  Previous = &DrealHeadSelect;
  Select   =  DrealHeadSelect;

  while( Select != (drealselect *)NULL )
  {
    if ( ! IsDrealAccepted( Select->RECTANGLE ) )
    {
      DelSelect  = Select;
      Select     = Select->NEXT;
      *Previous  = Select;
    
      DrealFreeSelect( DelSelect );
    }
    else
    {
      Previous = &Select->NEXT;
      Select   = Select->NEXT;
    }
  }

  rdsend();
}
