/*------------------------------------------------------------\
|                                                             |
| Tool    :                   DREAL                           |
|                                                             |
| File    :                  cursor.c                         |
|                                                             |
| Authors :              Jacomme Ludovic                      |
|                                                             |
| Date    :                  28.03.95                         |
|                                                             |
\------------------------------------------------------------*/

/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <X11/cursorfont.h>
# include <X11/Intrinsic.h>
# include RDS_H
# include GTB_H
# include "GTB_cursor.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  static DrealMouseCursor DrealMouseCursorArray [ DREAL_MAX_CURSOR ] =

         {
           { XC_left_ptr      , True  },
           { XC_watch         , True  },
           { XC_pirate        , True  },
           { XC_cross_reverse , True  },
           { 0                , True  }

         };

  static char  DrealCursorMaskOr [ 8 ] = 

  {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
  };

  static char  DrealCursorMaskAnd [ 8 ] =

  {
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
  };

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      DrealSetMouseCursor                    |
|                                                             |
\------------------------------------------------------------*/

void DrealSetMouseCursor( MainWidget, CursorType )

     Widget MainWidget;
     char   CursorType;
{
  Display          *DisplayId;
  XEvent            MouseEvent;
  Widget            TopShell;
  Widget            ScanShell;
  Window            TopWindow;
  Window            ScanWindow;
  XWindowAttributes WindowAtt;
  XtAppContext      Context;
  XColor            White;
  XColor            Black;
  XColor            ForgetIt;
  Colormap          ColorMap;
  Pixmap            MaskOr;
  Pixmap            MaskAnd;
  Cursor            NewCursor;

  rdsbegin();

  ScanShell = MainWidget;

  while ( ! XtIsShell( ScanShell ) )
  {
    ScanShell = XtParent( ScanShell );
  }

  TopShell = ScanShell;

  while ( ! XtIsTopLevelShell( TopShell ) )
  {
    TopShell = XtParent( TopShell );
  }

  XFlush(XtDisplay (TopShell));

  DisplayId = XtDisplay( ScanShell );

  TopWindow  = XtWindow( TopShell  );
  ScanWindow = XtWindow( ScanShell );

  if ( DrealMouseCursorArray[ CursorType ].MAKE == True )
  {
    if ( DrealMouseCursorArray[ CursorType ].CURSOR != 0 )
    {
      NewCursor = XCreateFontCursor( DisplayId, 
                                     DrealMouseCursorArray[ CursorType ].CURSOR );
    }
    else
    {
      ColorMap = DefaultColormapOfScreen ( XtScreen( MainWidget ) );

      XAllocNamedColor( DisplayId, ColorMap, "black", &Black, &ForgetIt );
      XAllocNamedColor( DisplayId, ColorMap, "white", &White, &ForgetIt );

      MaskOr = 

        XCreatePixmapFromBitmapData( DisplayId,
                                     XtWindow( MainWidget ),
                                     DrealCursorMaskOr,
                                     8, 8,
                                     Black.pixel,
                                     White.pixel, 1 );
      MaskAnd =

        XCreatePixmapFromBitmapData( DisplayId,
                                     XtWindow( MainWidget ),
                                     DrealCursorMaskAnd,
                                     8, 8,
                                     Black.pixel,
                                     White.pixel, 1 );

      NewCursor = XCreatePixmapCursor( DisplayId, 
                                       MaskAnd, MaskOr,
                                       &Black, &White, 0, 0 );
    }

    DrealMouseCursorArray[ CursorType ].CURSOR = NewCursor;
    DrealMouseCursorArray[ CursorType ].MAKE   = False;
  }

  XDefineCursor( XtDisplay( TopShell ), 
                 XtWindow( MainWidget ),
                 DrealMouseCursorArray[ CursorType ].CURSOR );

  Context = XtWidgetToApplicationContext( TopShell );

  if ( XtIsRealized( ScanShell ) && XtIsRealized( TopShell ) )
  {
    while ( ( XGetWindowAttributes( DisplayId, ScanWindow, &WindowAtt ) ) &&
            ( WindowAtt.map_state != IsViewable                         ) )
    {
      if ( ( XGetWindowAttributes( DisplayId, TopWindow, &WindowAtt ) ) && 
           ( WindowAtt.map_state != IsViewable                        ) ) 

          break;

      XtAppNextEvent( Context, &MouseEvent );
      XtDispatchEvent( &MouseEvent );
    }
  }

  XmUpdateDisplay( TopShell );

  rdsend();
}
