/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


 
/*
 * %% \textrm{This File Looks Nicer with\textsf{ LGrind} and \LaTeXe.}       %%
 * %|                                                                        |%
 * %| $Id: FpGenMBK.c,v 1.4 1999/10/21 13:50:10 jpc Exp $       |%
 * %|                                                                        |%
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |          Alliance CAD System 3.2b                                   | |%
 * %| |     F p G e n   P a c k a g e   1.10                               | |%
 * %| |                                                                    | |%
 * %| |  Copyright(c) 95-97, MASI, CAO-VLSI Team                           | |%
 * %| |                                                                    | |%
 * %| |  Author    :            Jean-Paul CHAPUT                           | |%
 * %| |  E-mail    :        cao-vlsi@masi.ibp.fr                           | |%
 * %| | ================================================================== | |%
 * %| |  Module    :        "FpGenMBK.c"                                   | |%
 * %| | ****************************************************************** | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 */


# include  <stdio.h>
# include  <string.h>
# include  <sys/stat.h>

# include  "GenericGN.h"
# include  "mbkgen.h"

# ifndef  __GaAs__
   /* Compilation with CMOS full customs generators. */
#  include   GRF_H
#  include   GBS_H
#  include   RSA_H
#  include   GAM_H
# else
   /* Compilation with GaAs full customs generators. */
#  include   GRAGA_H
#  include   GSARA_H
#  include   GZARA_H
#  include   GCLARA_H
#  include   GBAT_H
# endif

# include  "FpGenMBK.h"
# include  "FpGenMGN.h"


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |             External Structures and Variables                  | |%
 *   %| \----------------------------------------------------------------/ |%
 *   %|                                                                    |%
 *   %| All those variables must be declared `extern' in "FpGenMBK.h".     |%
 */

       FpgenVGlob_t fpgen_Glob = {                0L, /* ModuleId  */
                                   (losig_list*)NULL, /* SigVdd    */
                                   (losig_list*)NULL, /* SigVss    */
                                   (locon_list*)NULL, /* ConVdd    */
                                   (locon_list*)NULL, /* ConVss    */
                           C_FgnMGN_ConModeEXPLICIT , /* ConMode   */
                                                  0L, /* WidthDP   */
                           C_FgnMGN_LSBIndexZero    , /* Msb0      */
                                         (char*)NULL, /* CurrFunc  */
                                   (lofig_list*)NULL, /* CurrLoFig */
                                                  0L, /* CallDefLo */
                                                  0L, /* CallSavLo */
                           C_FgnMGN_TraceModeOff    , /* TraceMode */
                                                  0L,
                  "\n*** FpGen Context ***:Initialisation step.",
                                   (lofig_list*)NULL  /* Old LoFig */
                                 };


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                        Internal Constants                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

# define        IC_FpMBK_MaxLenString             16384

# define        IC_FpMBK_TypeRfgFlattOptim        0
# define        IC_FpMBK_TypeRfgFlattCompat       1
# define        IC_FpMBK_TypeRfgComplete          3
# define        IC_FpMBK_TypeRfgR0Normal          0
# define        IC_FpMBK_TypeRfgR0Zero            1
# define        IC_FpMBK_TypeRfgEdgeFall          1L
# define        IC_FpMBK_TypeRfgEdgeRise          2L
# define        IC_FpMBK_RfgMinWords              2
# define        IC_FpMBK_RfgMaxWords            256

# define        IF_FpMBK_GenericMFModel           1
# define        IF_FpMBK_GenericMFWidth           2
# define        IF_FpMBK_GenericMFSlice           4
# define        IF_FpMBK_GenericMFDrive           8
# define        IF_FpMBK_GenericMFMask           16
# define        IF_FpMBK_GenericMFEncod          32
# define        IF_FpMBK_GenericMFWords          64
# define        IF_FpMBK_GenericMFT0            128
# define        IF_FpMBK_GenericMFT1            256
# define        IF_FpMBK_GenericMFN             512


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |               Internal Structures and Variables                    | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/* Structure for all Macro-Function specifics informations. */
typedef struct {
        char  NameMF[  32];  /* Name of the Macro-Function.        */
        char  NameRoot[32];  /* Root used to build the model name. */
        long  Function;      /* Generic generator function code.   */
  ptype_list *TabPModel;     /* List of generated models.          */
  } MacroDB_t;


/* Declaration of internal variables of module. */
static MacroDB_t  MacroDB[C_FgnMBK_FunctionNB] = {
    {"DP_INV"      ,  "inv"    ,C_GfpGN_FunctionInv      ,(ptype_list*)NULL},
    {"DP_BUFF"     , "buff"    ,C_GfpGN_FunctionBuff     ,(ptype_list*)NULL},
    {"DP_NAND2"    , "nand2"   ,C_GfpGN_FunctionNand2    ,(ptype_list*)NULL},
    {"DP_NAND3"    , "nand3"   ,C_GfpGN_FunctionNand3    ,(ptype_list*)NULL},
    {"DP_AND2"     ,  "and2"   ,C_GfpGN_FunctionAnd2     ,(ptype_list*)NULL},
    {"DP_AND3"     ,  "and3"   ,C_GfpGN_FunctionAnd3     ,(ptype_list*)NULL},
    {"DP_NOR2"     , "nor2"    ,C_GfpGN_FunctionNor2     ,(ptype_list*)NULL},
    {"DP_NOR3"     , "nor3"    ,C_GfpGN_FunctionNor3     ,(ptype_list*)NULL},
    {"DP_OR2"      ,  "or2"    ,C_GfpGN_FunctionOr2      ,(ptype_list*)NULL},
    {"DP_OR3"      ,  "or3"    ,C_GfpGN_FunctionOr3      ,(ptype_list*)NULL},
    {"DP_XNOR2"    , "xnor2"   ,C_GfpGN_FunctionXnor2    ,(ptype_list*)NULL},
    {"DP_XOR2"     ,  "xor2"   ,C_GfpGN_FunctionXor2     ,(ptype_list*)NULL},
    {"DP_NAND2MASK","nand2mask",C_GfpGN_FunctionNand2mask,(ptype_list*)NULL},
    {"DP_NOR2MASK" , "nor2mask",C_GfpGN_FunctionNor2mask ,(ptype_list*)NULL},
    {"DP_XNOR2MASK","xnor2mask",C_GfpGN_FunctionXnor2mask,(ptype_list*)NULL},
    {"DP_CONST"    ,"const"    ,C_GfpGN_FunctionConst    ,(ptype_list*)NULL},
    {"DP_NBUSE"    ,"nbuse"    ,C_GfpGN_FunctionNbuse    ,(ptype_list*)NULL},
    {"DP_BUSE"     , "buse"    ,C_GfpGN_FunctionBuse     ,(ptype_list*)NULL},
    {"DP_NMUX2CS"  ,"nmux2cs"  ,C_GfpGN_FunctionNmux2cs  ,(ptype_list*)NULL},
    {"DP_MUX2CS"   , "mux2cs"  ,C_GfpGN_FunctionMux2cs   ,(ptype_list*)NULL},
    {"DP_MUX3CD"   , "mux3cd"  ,C_GfpGN_FunctionMux3cd   ,(ptype_list*)NULL},
    {"DP_MUX3CS"   , "mux3cs"  ,C_GfpGN_FunctionMux3cs   ,(ptype_list*)NULL},
    {"DP_MUX4CS"   , "mux4cs"  ,C_GfpGN_FunctionMux4cs   ,(ptype_list*)NULL},
    {"DP_MUX5CS"   , "mux5cs"  ,C_GfpGN_FunctionMux5cs   ,(ptype_list*)NULL},
    {"DP_MUX6CS"   , "mux6cs"  ,C_GfpGN_FunctionMux6cs   ,(ptype_list*)NULL},
    {"DP_NUL"      , "nul"     ,C_GfpGN_FunctionNul      ,(ptype_list*)NULL},
    {"DP_ADD2S"    , "add2s"   ,C_GfpGN_FunctionAdd2s    ,(ptype_list*)NULL},
    {"DP_ADD2F"    , "add2f"   ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_ADSB2F"   ,"adsb2f"   ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_SHIFT"    ,"shift"    ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_NPLATCH1" , "npl1"    ,C_GfpGN_FunctionNpl1     ,(ptype_list*)NULL},
    {"DP_PLATCH1"  ,  "pl1"    ,C_GfpGN_FunctionPl1      ,(ptype_list*)NULL},
    {"DP_PMSNW"    , "pmsnw"   ,C_GfpGN_FunctionPmsnw    ,(ptype_list*)NULL},
    {"DP_PMSNWQ"   , "pmsnwq"  ,C_GfpGN_FunctionPmsnwq   ,(ptype_list*)NULL},
    {"DP_PMS"      , "pms"     ,C_GfpGN_FunctionPms      ,(ptype_list*)NULL},
    {"DP_PMSQ"     , "pmsq"    ,C_GfpGN_FunctionPmsq     ,(ptype_list*)NULL},
    {"DP_PMST"     , "pmst"    ,C_GfpGN_FunctionPmst     ,(ptype_list*)NULL},
    {"DP_PMSTQ"    , "pmstq"   ,C_GfpGN_FunctionPmstq    ,(ptype_list*)NULL},
    {"DP_PDFFNW"   ,"pdffnw"   ,C_GfpGN_FunctionPdffnw   ,(ptype_list*)NULL},
    {"DP_PDFFNWQ"  ,"pdffnwq"  ,C_GfpGN_FunctionPdffnwq  ,(ptype_list*)NULL},
    {"DP_PDFF"     ,"pdff"     ,C_GfpGN_FunctionPdff     ,(ptype_list*)NULL},
    {"DP_PDFFR"    ,"pdffr"    ,C_GfpGN_FunctionPdffr    ,(ptype_list*)NULL},
    {"DP_PDFFQ"    ,"pdffq"    ,C_GfpGN_FunctionPdffq    ,(ptype_list*)NULL},
    {"DP_PDFFT"    ,"pdfft"    ,C_GfpGN_FunctionPdfft    ,(ptype_list*)NULL},
    {"DP_PDFFUT"   ,"pdffut"   ,C_GfpGN_FunctionPdffut   ,(ptype_list*)NULL},
    {"DP_PDFFRT"   ,"pdffrt"   ,C_GfpGN_FunctionPdffrt   ,(ptype_list*)NULL},
    {"DP_PDFFTQ"   ,"pdfftq"   ,C_GfpGN_FunctionPdfftq   ,(ptype_list*)NULL},
    {"DP_RFG1CC"   , "rfg1cc"  ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFG2CC"   , "rfg2cc"  ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFG1FC"   , "rfg1fc"  ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFG2FC"   , "rfg2fc"  ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFG1C0"   , "rfg1c0"  ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFG2C0"   , "rfg2c0"  ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFG1F0"   , "rfg1f0"  ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFG2F0"   , "rfg2f0"  ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFGU1CC"  , "rfgu1cc" ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFGU2CC"  , "rfgu2cc" ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFGU1FC"  , "rfgu1fc" ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFGU2FC"  , "rfgu2fc" ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFGU1C0"  , "rfgu1c0" ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFGU2C0"  , "rfgu2c0" ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFGU1F0"  , "rfgu1f0" ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_RFGU2F0"  , "rfgu2f0" ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_MULT"     , "mult"    ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL},
    {"DP_MULTP"    , "multp"   ,C_GfpGN_FunctionUnknow   ,(ptype_list*)NULL}
    };


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |                       Macros Definitions                       | |%
 *   %| \----------------------------------------------------------------/ |%
 */

#define  mc_InCatalog(s)  (incatalog(s) ? TRUE : FALSE)


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                  Internal Functions Declarations                   | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

static       long  CheckInsName     __FP(( char *aName ));
static chain_list *FgnMBK_NameToSig __FP(( chain_list *aPtChain,
                                                 char *aSigName,
                                                 long  aSigType));
static chain_list *FgnMBK_MatchImplicitConn
                                    __FP((       char *aInstcName,
                                           chain_list *aChainConName,
                                           chain_list *aChainSigName,
                                           lofig_list *aPtLoFig,
                                                 long *aPtSigCount ));
static       void  SaveMBKEnv       __FP(( void ));
static       void  RestoreMBKEnv    __FP(( void ));


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                     Functions Definitions                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/*  Function  :  'FgnMBK_DefLoFig'.                                         */
/* ======================================================================== */
/*  MBK function associated to "DP_DEFLOFIG" (module "FpgenMGN").           */

extern void  FgnMBK_DefLoFig()
{
  static char  NameCATAL[6] = "CATAL";


  if ( fpgen_Glob.CallDefLo == 1L ) 
  {
    struct stat  DirStat;


    /* Check the value of MBK_SEPAR. */
    if ( SEPAR != '.' )
      GfpUT_PrintWarng( fpgen_Glob.ModuleId, 
                        W_FgnMBK_DEFLOMbkSepar,
                        fpgen_Glob.CurrFunc,
                        SEPAR );

    /* Check the value of MBK_IN_PH and MBK_OUT_PH. */
    if ( strcmp( IN_PH, OUT_PH ) )
      GfpUT_PrintWarng( fpgen_Glob.ModuleId, 
                        W_FgnMBK_DEFLOMbkInOutPH,
                        fpgen_Glob.CurrFunc,
                        IN_PH,
                        OUT_PH );


    fpgen_Glob.OLD_WORKL = (char *)mbkalloc(
                             ( 14 + strlen( WORK_LIB ) ) * sizeof( char ) );

    sprintf( fpgen_Glob.OLD_WORKL, "MBK_WORK_LIB=%s", WORK_LIB );


    /* Search the directory stat. */
    if (    ( UtEnv_Environ.FPGEN_LIB != (char *)NULL )
         && ( strcmp( UtEnv_Environ.FPGEN_LIB, WORK_LIB ) ) )
    {
      if ( stat( UtEnv_Environ.FPGEN_LIB, &DirStat ) )
      {
        FgnMBK_ErrorHeader( "DP_DEFLOFIG" );
        FgnMBK_ErrorPrintF( " Can\'t stat FpGen library:\n  \"%ls\".\n",
                            UtEnv_Environ.FPGEN_LIB );
        FgnMBK_ErrorTrap( SIGTERM );
      }

      if ( !S_ISDIR( DirStat.st_mode ) )
      {
        FgnMBK_ErrorHeader( "DP_DEFLOFIG" );
        FgnMBK_ErrorPrintF( " FpGen library is not a directory:\n  \"%ls\".\n",
                            UtEnv_Environ.FPGEN_LIB );
        FgnMBK_ErrorTrap( SIGTERM );
      }

      fpgen_Glob.FPGEN_LIB = (char *)mbkalloc(
              ( 14 + strlen( UtEnv_Environ.FPGEN_LIB ) ) * sizeof( char ) );

      fpgen_Glob.NameCATAL = NameCATAL;

      sprintf( fpgen_Glob.FPGEN_LIB,
               "MBK_WORK_LIB=%s",
               UtEnv_Environ.FPGEN_LIB );
    }
    else
    {
      fpgen_Glob.FPGEN_LIB = (char *)mbkalloc(
                   ( strlen(fpgen_Glob.OLD_WORKL ) + 1 ) * sizeof( char ) );

      fpgen_Glob.NameCATAL = CATAL;

      strcpy( fpgen_Glob.FPGEN_LIB, fpgen_Glob.OLD_WORKL );
    }
  }


  fpgen_Glob.CurrLofig = WORK_LOFIG;

  /* Declaration of signals and terminals :
   *             - "vdd"
   *             - "vss"
   */

  /* Creation of SIGNAL Vdd. */
  num_index += 1;
  fpgen_Glob.SigVdd = addlosig( WORK_LOFIG,
                                num_index,
                                addchain( (chain_list *)NULL,
                                                (void *)"vdd" ),
                                EXTERNAL );

  /* Creation of SIGNAL Vss. */
  num_index += 1;
  fpgen_Glob.SigVss = addlosig( WORK_LOFIG,
                                num_index,
                                addchain( (chain_list *)NULL,
                                                (void *)"vss" ),
                                EXTERNAL );


  /* Opening CATAL of FpGen Library for updating.
   * We add all the new column models.
   */
  if ( ( fpgen_Glob.FileCATAL
           = fopen( UtBas_PrintF( "%s/%s",
                      fpgen_Glob.FPGEN_LIB + 13,
                      fpgen_Glob.NameCATAL ),
                    "a+" ) ) == (FILE *)NULL )
  {
    FgnMBK_ErrorHeader( "DP_DEFLOFIG" );
    FgnMBK_ErrorPrintF(
      " Can\'t open CATAL of FpGen library:\n  \"%s/%s\".\n",
      fpgen_Glob.FPGEN_LIB + 13,
      fpgen_Glob.NameCATAL );
    FgnMBK_ErrorTrap( SIGTERM );
  }
  else
  {
    /* If the CATAL file is to be created, we must fill it by at least
     * one line to avoid errors wen 'mbkenv' will read it.
     */
    fprintf( fpgen_Glob.FileCATAL, "%-50s  C\n", "FpGen___________________" );
    fflush(  fpgen_Glob.FileCATAL );
  }
}


/*  Function  :  'FgnMBK_SavLoFig'.                               */
/* -------------------------------------------------------------- */
/*  MBK function associated to "DP_SAVLOFIG" (module "FpgenMGN").
 *  Perform the following checks :
 *      - Look for the power supplies "vdd" and "vss".
 *      - Check the specific data-path hierarchical consistency.
 *                                                (inactivated)
 */

extern void  FgnMBK_SavLoFig()
{ 
    /* Look for the VDD power supplie. */
    if (Dut_GetLoConBySig(WORK_LOFIG,fpgen_Glob.SigVss) == (locon_list*)NULL)
        GfpUT_PrintError(fpgen_Glob.ModuleId,
                       E_FgnMBK_SAVLOMissVSS,
                         fpgen_Glob.CurrFunc,
                            WORK_LOFIG->NAME);


    /* Look for the VSS power supplies. */
    if (Dut_GetLoConBySig(WORK_LOFIG,fpgen_Glob.SigVdd) == (locon_list*)NULL)
        GfpUT_PrintError(fpgen_Glob.ModuleId,
                       E_FgnMBK_SAVLOMissVDD,
                         fpgen_Glob.CurrFunc,
                            WORK_LOFIG->NAME);


    /* Check the hierarchical consistency. */
    UtHir_HierarchCheck(WORK_LOFIG);

    /* Close CATAL of FpGen Library. */
    fclose( fpgen_Glob.FileCATAL );
}


/*  Function  :  'FgnMBK_LoConPower'.                             */
/* -------------------------------------------------------------- */
/*  MBK function associated to "DP_LOCON" (module "FpgenMGN").    */

extern long  FgnMBK_LoConPower(aConName)
    char *aConName;
{
    losig_list *PowerSig = (losig_list*)NULL;
          char *NameVSS  = namealloc("vss");
          char *NameVDD  = namealloc("vdd");

    /* Look for VSS power supply connector names. */
    if (isvss(aConName))
    {
        /* The name of the power supply must be "vss". */
        if (namealloc(aConName) != NameVSS)
            GfpUT_PrintError(fpgen_Glob.ModuleId,
                 E_FgnMBK_LOCONPOWERBadNamePower,
                             fpgen_Glob.CurrFunc,
                                WORK_LOFIG->NAME,
                                        aConName);
        else
            PowerSig = fpgen_Glob.SigVss;
    }

    /* Look for VSS power supply connector names. */
    if (isvdd(aConName))
    {
        /* The name of the power supply must be "vss". */
        if (namealloc(aConName) != NameVDD)
            GfpUT_PrintError(fpgen_Glob.ModuleId,
                 E_FgnMBK_LOCONPOWERBadNamePower,
                             fpgen_Glob.CurrFunc,
                                WORK_LOFIG->NAME,
                                        aConName);
        else
            PowerSig = fpgen_Glob.SigVdd;
    }


    /* If no power supply connector has been recognized, return false. */
    if (PowerSig == (losig_list*)NULL) return(FALSE);

    /* Check for a previously added power connector. */
    if (Dut_GetLoConBySig(WORK_LOFIG, PowerSig) != (locon_list*)NULL)
        GfpUT_PrintError(fpgen_Glob.ModuleId,
                E_FgnMBK_LOCONPOWERDuplicate,
                         fpgen_Glob.CurrFunc,
                            WORK_LOFIG->NAME,
                                    aConName);

    /* Add the power connector. */
    (void)addlocon(WORK_LOFIG,
                   aConName  ,
                   PowerSig  ,
                   IN        );

    return(TRUE);
}


/*  Fonction  :  'FgnMBK_Import'.                                           */
/* ------------------------------------------------------------------------ */
/*  MBK function associated to "DP_IMPORT" (module "FpgenMGN").             */

extern void  FgnMBK_Import( ModelName, InstName, aStdArgs )
     char *ModelName;
     char *InstName;
  va_list  aStdArgs;
{
  lofig_list *PtLoFig;


  /* If the function have been call before DP_DEFLOFIG :
   *     - Initialisation of error data base.
   *     - Process error : "no previous call to DP_DEFLOFIG".
   */
  if (fpgen_Glob.CallDefLo == 0)
  {
    FgnMBK_Init();
    GfpUT_PrintError( fpgen_Glob.ModuleId,
                      E_FgnMBK_SharedMissDefLo,
                      "DP_IMPORT" );
  }


  FgnMBK_TraceHeader();
  FgnMBK_TracePrintF( "%20s : DP_IMPORT.\n", "Macro-Function" );


  /* Recuperation of fixed arguments :
   *     - Model name (ModelName).
   *     - Instance name (InstName).
   */
  FgnMBK_TracePrintF( "%20s : ", "(char *)ModelName" );

  FgnMBK_TracePrintF( "(0x%08lX) ", ModelName );
  FgnMBK_TracePrintF(  "\"%s\".\n", ModelName );

  FgnMBK_TracePrintF( "%20s : ", "(char *)InstName" );

  FgnMBK_TracePrintF( "(0x%08lX) ", InstName );
  FgnMBK_TracePrintF(  "\"%s\".\n", InstName );

  (void)CheckInsName( InstName );


  /* Load the logical model. */
  PtLoFig = UtHir_LoadLoFig( ModelName );

  /* Instanciate the logical figure. */
  FgnMBK_xvLoIns( ModelName, InstName, (chain_list*)NULL, aStdArgs );

  FgnMBK_TraceFlush();
}


/*  Function  :  'FgnMBK_Flatten'.                                */
/* -------------------------------------------------------------- */
/*  MBK function associated to "DP_FLATTEN" (module "FpgenMGN").  */

extern void  FgnMBK_Flatten()
{ UtHir_HierarchFlattenLo(WORK_LOFIG); }


/*  Function  :  'FgnMBK_GenericMF'.                                        */
/* ------------------------------------------------------------------------ */
/*  Generic Macro-Function for all regular operators.                       */

extern void  FgnMBK_GenericMF( aIdentMF, InstName, aStdArgs )
     long  aIdentMF;
     char *InstName;
  va_list  aStdArgs;
{
#define  access_ModelLofig  ((lofig_list*)(ModelPtype->DATA))

  chain_list *ChainCTRLConn = (chain_list*)NULL;
  ptype_list *ModelPtype;
        char *ModelName;
        long  Width;
        long  Slice;
        char *StrMask;
        long  *XlMask;
        long     Drive;
        long  MaxDrive;
        long  DecWidth    = 0L;
        long  EncWidth    = 0L;
        long  RSA_TypeCsa;
        long  RSA_TypeSub;
        long  RSA_TypeAsb;
        long  RFG_BusNB   = 0L;
        long  RFG_Stuck;
        long  RFG_Edge;
        long  RFG_Type;
        long  GAM_T0;
        long  GAM_T1;
        long  GAM_N;
        long  FlagProcess = 0L;

        char *TmpName;
        char  LoFigMode;


  FgnMBK_TraceHeader();
  FgnMBK_TracePrintF( "%20s : ", "Macro-Function" );

  switch( aIdentMF )
  {
    case C_FgnMBK_FunctionInv:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveInv;
      break;

    case C_FgnMBK_FunctionBuff:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveBuff;
      break;

    case C_FgnMBK_FunctionAnd2:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveAnd2;
      break;

    case C_FgnMBK_FunctionAnd3:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveAnd3;
      break;

    case C_FgnMBK_FunctionNand2:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveNand2;
      break;

    case C_FgnMBK_FunctionNand3:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveNand3;
      break;

    case C_FgnMBK_FunctionOr2:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveOr2;
      break;

    case C_FgnMBK_FunctionOr3:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveOr3;
      break;

    case C_FgnMBK_FunctionNor2:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveNor2;
      break;

    case C_FgnMBK_FunctionNor3:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFDrive;
      MaxDrive    =    C_GfpGN_DriveNor3;
      break;

    case C_FgnMBK_FunctionXnor2:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice;
      MaxDrive    =    C_GfpGN_DriveXnor2;
      break;

    case C_FgnMBK_FunctionXor2:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice;
      MaxDrive    =    C_GfpGN_DriveXor2;
      break;

    case C_FgnMBK_FunctionNand2mask:
    case C_FgnMBK_FunctionNor2mask:
    case C_FgnMBK_FunctionXnor2mask:
    case C_FgnMBK_FunctionConst:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFMask;
      break;

    case C_FgnMBK_FunctionNbuse:
    case C_FgnMBK_FunctionBuse:
    case C_FgnMBK_FunctionNmux2cs:
    case C_FgnMBK_FunctionMux2cs:
    case C_FgnMBK_FunctionMux3cd:
    case C_FgnMBK_FunctionMux3cs:
    case C_FgnMBK_FunctionMux4cs:
    case C_FgnMBK_FunctionMux5cs:
    case C_FgnMBK_FunctionMux6cs:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice
                    | IF_FpMBK_GenericMFEncod;
      break;

    case C_FgnMBK_FunctionAdd2s:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice;
      break;

    case C_FgnMBK_FunctionAdd2f:

      FlagProcess = 0L;
      RSA_TypeCsa = 0L;
      RSA_TypeSub = 0L;
      RSA_TypeAsb = 0L;
      break;

    case C_FgnMBK_FunctionAdsb2f:

      FlagProcess = 0L;
      RSA_TypeCsa = 0L;
      RSA_TypeSub = 0L;
      RSA_TypeAsb = 1L;
      break;

    case C_FgnMBK_FunctionShift:

      FlagProcess = 0L;
      break;

    case C_FgnMBK_FunctionNul:
    case C_FgnMBK_FunctionNpl1:
    case C_FgnMBK_FunctionPl1:
    case C_FgnMBK_FunctionPmsnw:
    case C_FgnMBK_FunctionPmsnwq:
    case C_FgnMBK_FunctionPms:
    case C_FgnMBK_FunctionPmsq:
    case C_FgnMBK_FunctionPmst:
    case C_FgnMBK_FunctionPmstq:
    case C_FgnMBK_FunctionPdffnw:
    case C_FgnMBK_FunctionPdffnwq:
    case C_FgnMBK_FunctionPdff:
    case C_FgnMBK_FunctionPdffr:
    case C_FgnMBK_FunctionPdffq:
    case C_FgnMBK_FunctionPdfft:
    case C_FgnMBK_FunctionPdffut:
    case C_FgnMBK_FunctionPdffrt:
    case C_FgnMBK_FunctionPdfftq:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFSlice;
      break;

    case C_FgnMBK_FunctionRfg2cc:
    case C_FgnMBK_FunctionRfgu2cc:

      RFG_BusNB  += 1;

    case C_FgnMBK_FunctionRfg1cc:
    case C_FgnMBK_FunctionRfgu1cc:

      RFG_Stuck   =   IC_FpMBK_TypeRfgR0Normal;
      RFG_Type    =   IC_FpMBK_TypeRfgComplete;
      FlagProcess =   IF_FpMBK_GenericMFWords
                    | IF_FpMBK_GenericMFEncod;
      break;

    case C_FgnMBK_FunctionRfg2fc:
    case C_FgnMBK_FunctionRfgu2fc:

      RFG_BusNB  += 1;

    case C_FgnMBK_FunctionRfg1fc:
    case C_FgnMBK_FunctionRfgu1fc:

      RFG_Stuck   = IC_FpMBK_TypeRfgR0Normal;
      RFG_Type    = IC_FpMBK_TypeRfgFlattOptim;
      FlagProcess = IF_FpMBK_GenericMFWords;
      break;

    case C_FgnMBK_FunctionRfg2c0:
    case C_FgnMBK_FunctionRfgu2c0:

      RFG_BusNB  += 1;

    case C_FgnMBK_FunctionRfg1c0:
    case C_FgnMBK_FunctionRfgu1c0:

      RFG_Stuck   =   IC_FpMBK_TypeRfgR0Zero;
      RFG_Type    =   IC_FpMBK_TypeRfgComplete;
      FlagProcess =   IF_FpMBK_GenericMFWords
                    | IF_FpMBK_GenericMFEncod;
      break;

    case C_FgnMBK_FunctionRfg2f0:
    case C_FgnMBK_FunctionRfgu2f0:

      RFG_BusNB  += 1;

    case C_FgnMBK_FunctionRfg1f0:
    case C_FgnMBK_FunctionRfgu1f0:

      RFG_Stuck   = IC_FpMBK_TypeRfgR0Zero;
      RFG_Type    = IC_FpMBK_TypeRfgFlattOptim;
      FlagProcess = IF_FpMBK_GenericMFWords;
      break;

    case C_FgnMBK_FunctionMult:

      GAM_T0      = 0;
      GAM_T1      = 0;
      GAM_N       = 0;
      FlagProcess = IF_FpMBK_GenericMFWidth;
      break;

    case C_FgnMBK_FunctionMultp:

      FlagProcess =   IF_FpMBK_GenericMFWidth
                    | IF_FpMBK_GenericMFT0
                    | IF_FpMBK_GenericMFT1
                    | IF_FpMBK_GenericMFN;
      break;

    default:

      FgnMBK_ErrorHeader( "GenericMF" );
      FgnMBK_ErrorPrintF( " Invalid Macro-Function code %ld.\n",
                           aIdentMF );
      FgnMBK_ErrorTrap( SIGTERM );
  }


  RFG_Edge = IC_FpMBK_TypeRfgEdgeFall;

  switch( aIdentMF )
  {
    case C_FgnMBK_FunctionRfgu2cc:
    case C_FgnMBK_FunctionRfgu1cc:
    case C_FgnMBK_FunctionRfgu2fc:
    case C_FgnMBK_FunctionRfgu1fc:
    case C_FgnMBK_FunctionRfgu2c0:
    case C_FgnMBK_FunctionRfgu1c0:
    case C_FgnMBK_FunctionRfgu2f0:
    case C_FgnMBK_FunctionRfgu1f0:

      RFG_Edge = IC_FpMBK_TypeRfgEdgeRise;
      break;
  }


  fpgen_Glob.CurrFunc = MacroDB[ aIdentMF ].NameMF;
  FgnMBK_TracePrintF( "%s.\n", MacroDB[aIdentMF].NameMF );


  /* If the function have been call before DP_DEFLO :
   *     - Initialisation of error data base.
   *     - Process error : "no previous call to DP_DEFLO".
   */
  if (fpgen_Glob.CallDefLo == 0)
  {
    FgnMBK_Init();
    GfpUT_PrintError( fpgen_Glob.ModuleId,
                      E_FgnMBK_SharedMissDefLo,
                      MacroDB[ aIdentMF ].NameMF );
  }


  /* Read the instance name (always done). */
  FgnMBK_TracePrintF( "%20s : ", "(char *)InstName" );

  FgnMBK_TracePrintF( "(0x%08lX) ", InstName );
  FgnMBK_TracePrintF(  "\"%s\".\n", InstName );

  (void)CheckInsName( InstName );


  if (FlagProcess & IF_FpMBK_GenericMFWidth)
  {
    /* Read and check the 'Width' argument. */
    FgnMBK_TracePrintF( "%20s : ", "(long)Width" );

    if ((Width = va_arg( aStdArgs, long )) == C_FgnMGN_DEFAULTWidth)
      Width = fpgen_Glob.WidthDP;

    FgnMBK_TracePrintF( "(0x--------)  %ld.\n", Width );

    if (   (Width < 1                 )
        || (Width > fpgen_Glob.WidthDP))
    {
      GfpUT_PrintError( fpgen_Glob.ModuleId,
                        E_FgnMBK_GenericMFWidth, 
                        MacroDB[ aIdentMF ].NameMF,
                        InstName,
                        Width );
    }
  }


  if (FlagProcess & IF_FpMBK_GenericMFSlice)
  {
    /* Read and check the 'Slice' argument. */
    FgnMBK_TracePrintF( "%20s : ", "(long)Slice" );

    if ((Slice = va_arg( aStdArgs, long )) == C_FgnMGN_DEFAULTSlice)
      Slice = 0;

    FgnMBK_TracePrintF( "(0x--------)  %ld.\n"    , Slice );

    if (   (Slice       < 0                 )
        || (Slice+Width > fpgen_Glob.WidthDP))
    {
      GfpUT_PrintError( fpgen_Glob.ModuleId,
                        E_FgnMBK_GenericMFSlice, 
                        MacroDB[ aIdentMF ].NameMF,
                        InstName,
                        Slice );
    }
  }


  /* Read the and check the 'Drive' argument (if needed). */
  if (FlagProcess & IF_FpMBK_GenericMFDrive)
  {
    FgnMBK_TracePrintF( "%20s : ", "(long)Drive" );

    Drive = va_arg(aStdArgs, long);

    FgnMBK_TracePrintF( "(0x--------)  %ld.\n"    , Drive );

    if (   (Drive < 1       )
        || (Drive > MaxDrive))
    {
       GfpUT_PrintError( fpgen_Glob.ModuleId,
                         E_FgnMBK_GenericMFDrive, 
                         MacroDB[ aIdentMF ].NameMF,
                         InstName,
                         Drive );
    }
  }
  else
  {
    Drive = 1;
  }


  if (FlagProcess & IF_FpMBK_GenericMFMask)
  {
    /* Read the and check the 'Mask' argument. */
    FgnMBK_TracePrintF( "%20s : ", "(char *)Mask" );

    StrMask = va_arg( aStdArgs, char* );

    FgnMBK_TracePrintF( "(0x%08X) " , StrMask );
    FgnMBK_TracePrintF(  "\"%s\".\n", StrMask );

    /* Check the syntax of the string. */
    (void)GfpUT_XlStrCheck( StrMask );

    XlMask = GfpUT_XlAlloc();

    /* Convert the string into an extended long. */
    GfpUT_XlStrToXl( StrMask, XlMask );
  }


  if (FlagProcess & IF_FpMBK_GenericMFWords)
  /* Read and check the 'Words' argument. */
  {
    FgnMBK_TracePrintF( "%20s : ", "(long)RegNB" );

    DecWidth = va_arg( aStdArgs, long );

    FgnMBK_TracePrintF( "(0x--------)  %ld.\n", DecWidth );

    if (   (DecWidth < IC_FpMBK_RfgMinWords)
        || (DecWidth > IC_FpMBK_RfgMaxWords))
    {
        GfpUT_PrintError( fpgen_Glob.ModuleId,
                          E_FgnMBK_GenericMFWords, 
                          MacroDB[ aIdentMF ].NameMF,
                          InstName,
                          DecWidth );
    }
  }


  if (FlagProcess & IF_FpMBK_GenericMFT0)
  /* Read and check the 'T0' argument, for GAM/AMG. */
  {
    FgnMBK_TracePrintF( "%20s : ", "(long)T0" );

    GAM_T0 = va_arg( aStdArgs, long );

    FgnMBK_TracePrintF( "(0x--------)  %ld.\n", GAM_T0 );

    GAM_T0 = ( GAM_T0 == 0 ) ? 0 : 1;
  }


  if (FlagProcess & IF_FpMBK_GenericMFT1)
  /* Read and check the 'T1' argument, for GAM/AMG. */
  {
    FgnMBK_TracePrintF( "%20s : ", "(long)T1" );

    GAM_T1 = va_arg( aStdArgs, long );

    FgnMBK_TracePrintF( "(0x--------)  %ld.\n", GAM_T1 );

    GAM_T1 = ( GAM_T1 == 0 ) ? 0 : 1;
  }


  if (FlagProcess & IF_FpMBK_GenericMFN)
  /* Read and check the 'N' argument, for GAM/AMG. */
  {
    FgnMBK_TracePrintF( "%20s : ", "(long)N" );

    GAM_N = va_arg( aStdArgs, long );

    FgnMBK_TracePrintF( "(0x--------)  %ld.\n", GAM_N );
  }


  /* Build an explicit model name. */
  switch( aIdentMF )
  {
    case C_FgnMBK_FunctionInv:
    case C_FgnMBK_FunctionBuff:
    case C_FgnMBK_FunctionNand2:
    case C_FgnMBK_FunctionNand3:
    case C_FgnMBK_FunctionAnd2:
    case C_FgnMBK_FunctionAnd3:
    case C_FgnMBK_FunctionNor2:
    case C_FgnMBK_FunctionNor3:
    case C_FgnMBK_FunctionOr2:
    case C_FgnMBK_FunctionOr3:
    case C_FgnMBK_FunctionXnor2:
    case C_FgnMBK_FunctionXor2:

      ModelName = UtBas_PrintF( "%s_%ldx%ldx%ldx%ld%c_cl",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                Width,
                                Slice,
                                Drive,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l') );

      break;

    case C_FgnMBK_FunctionNand2mask:
    case C_FgnMBK_FunctionNor2mask:
    case C_FgnMBK_FunctionXnor2mask:

      ChainCTRLConn = addchain( ChainCTRLConn, "msk" );

    case C_FgnMBK_FunctionConst:

      ModelName = UtBas_PrintF( "%s%s_%ldx%ldx%ld%c_cl",
                                MacroDB[ aIdentMF ].NameRoot,
                                GfpUT_XlToStr( XlMask,
                                               fpgen_Glob.WidthDP,
                                               C_GfpUT_BaseHexadecimal ),
                                fpgen_Glob.WidthDP,
                                Width,
                                Slice,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l') );

      break;

    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux3cd:

      DecWidth += 1;

    case C_FgnMBK_FunctionBuse:
    case C_FgnMBK_FunctionNbuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:

      DecWidth += 2;

      /* Compute the wide of the encoded selection bus. */
      for( EncWidth = 1 ; (DecWidth - 1) >> (EncWidth + 1) ; EncWidth++ );

      /* Create the name of the implicit CTRL terminals. */
      if (FlagProcess & IF_FpMBK_GenericMFEncod)
      {
        if (EncWidth > 1)
        {
          TmpName = UtBas_PrintF( "sel[%ld:0]", EncWidth - 1 );
          ChainCTRLConn = addchain( ChainCTRLConn,
                                    (void*)namealloc( TmpName ) );
        }
        else
        {
          ChainCTRLConn = addchain( ChainCTRLConn, "sel" );
        }
      }
      else
      {
        TmpName = UtBas_PrintF( "sel[%ld:0]", DecWidth - 1 );
        ChainCTRLConn = addchain( ChainCTRLConn,
                                  (void*)namealloc( TmpName ) );
      }

      /* Build an explicit model name. */
      ModelName = UtBas_PrintF( "%s_%ldx%ldx%ld%c_cl",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                Width,
                                Slice,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l') );
      break;

    case C_FgnMBK_FunctionAdsb2f:

      ChainCTRLConn = addchain(ChainCTRLConn,  "sub");

    case C_FgnMBK_FunctionAdd2f:
    case C_FgnMBK_FunctionAdd2s:

      ChainCTRLConn = addchain( ChainCTRLConn, "over" );
      ChainCTRLConn = addchain( ChainCTRLConn, "cout" );

      ModelName = UtBas_PrintF( "%s_%ldx%ldx%ld%c_bk",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                fpgen_Glob.WidthDP,
                                0,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l')   );
      break;

    case C_FgnMBK_FunctionShift:

    if ( UtTec_Family() == C_UtTec_FamilyAsGa )
    {
      ChainCTRLConn = addchain( ChainCTRLConn,  "shl" );
      ChainCTRLConn = addchain( ChainCTRLConn,  "shr" );
    }
    else
    {
      ChainCTRLConn = addchain( ChainCTRLConn, "left" );
      ChainCTRLConn = addchain( ChainCTRLConn,  "rot" );
      ChainCTRLConn = addchain( ChainCTRLConn,  "ext" );
    }

      ModelName = UtBas_PrintF( "%s_%ldx%ldx%ld%c_bk",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                fpgen_Glob.WidthDP,
                                0,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l') );
      break;

    case C_FgnMBK_FunctionNpl1:
    case C_FgnMBK_FunctionPl1:

      ChainCTRLConn = addchain( ChainCTRLConn, "ck" );

    case C_FgnMBK_FunctionNul:

      ModelName = UtBas_PrintF( "%s_%ldx%ldx%ld%c_cl",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                Width,
                                Slice,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l') );
      break;

    case C_FgnMBK_FunctionPmst:
    case C_FgnMBK_FunctionPmstq:

      ChainCTRLConn = addchain( ChainCTRLConn, "scin"  );
      ChainCTRLConn = addchain( ChainCTRLConn, "scout" );
      ChainCTRLConn = addchain( ChainCTRLConn,  "test" );

    case C_FgnMBK_FunctionPms:
    case C_FgnMBK_FunctionPmsq:

      ChainCTRLConn = addchain( ChainCTRLConn,   "wen" );

    case C_FgnMBK_FunctionPmsnw:
    case C_FgnMBK_FunctionPmsnwq:

      ChainCTRLConn = addchain( ChainCTRLConn,   "ckm" );
      ChainCTRLConn = addchain( ChainCTRLConn,   "cks" );

      ModelName = UtBas_PrintF( "%s_%ldx%ldx%ld%c_cl",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                Width,
                                Slice,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l') );
      break;

    case C_FgnMBK_FunctionPdfft:
    case C_FgnMBK_FunctionPdffut:
    case C_FgnMBK_FunctionPdfftq:

      ChainCTRLConn = addchain( ChainCTRLConn, "scin"  );
      ChainCTRLConn = addchain( ChainCTRLConn,  "test" );

    case C_FgnMBK_FunctionPdff:
    case C_FgnMBK_FunctionPdffq:

      ChainCTRLConn = addchain( ChainCTRLConn,   "wen" );

    case C_FgnMBK_FunctionPdffnw:
    case C_FgnMBK_FunctionPdffnwq:

      ChainCTRLConn = addchain( ChainCTRLConn,    "ck" );

      ModelName = UtBas_PrintF( "%s_%ldx%ldx%ld%c_cl",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                Width,
                                Slice,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l') );

      break;

    case C_FgnMBK_FunctionPdffrt:

      ChainCTRLConn = addchain( ChainCTRLConn, "scin"  );
      ChainCTRLConn = addchain( ChainCTRLConn,  "test" );

    case C_FgnMBK_FunctionPdffr:

      ChainCTRLConn = addchain( ChainCTRLConn,   "wen" );
      ChainCTRLConn = addchain( ChainCTRLConn,    "ck" );

      ModelName = UtBas_PrintF( "%s_%ldx%ldx%ld%c_cl",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                Width,
                                Slice,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l') );

      break;

    case C_FgnMBK_FunctionRfg1cc:
    case C_FgnMBK_FunctionRfg2cc:
    case C_FgnMBK_FunctionRfg1fc:
    case C_FgnMBK_FunctionRfg2fc:
    case C_FgnMBK_FunctionRfg1c0:
    case C_FgnMBK_FunctionRfg2c0:
    case C_FgnMBK_FunctionRfg1f0:
    case C_FgnMBK_FunctionRfg2f0:
    case C_FgnMBK_FunctionRfgu2cc:
    case C_FgnMBK_FunctionRfgu1cc:
    case C_FgnMBK_FunctionRfgu2fc:
    case C_FgnMBK_FunctionRfgu1fc:
    case C_FgnMBK_FunctionRfgu2c0:
    case C_FgnMBK_FunctionRfgu1c0:
    case C_FgnMBK_FunctionRfgu2f0:
    case C_FgnMBK_FunctionRfgu1f0:

      /* Create the name of the implicit CTRL terminals. */
      if (FlagProcess & IF_FpMBK_GenericMFEncod)
      {
        ChainCTRLConn = addchain( ChainCTRLConn, "we"   );
        ChainCTRLConn = addchain( ChainCTRLConn, "ck_m" );

        /* Compute the wide of the encoded R/W address bus. */
        for( EncWidth = 1 ; (DecWidth - 1) >> (EncWidth + 1) ; EncWidth++ );

        /* Write address bus ("ad_w[MSB:LSB]"). */
        TmpName = UtBas_PrintF( "ad_w[%ld:0]", EncWidth );
        ChainCTRLConn = addchain( ChainCTRLConn,
                                  namealloc( TmpName ) );

        /* First read address bus ("ad_r_a[MSB:LSB]"). */
        TmpName = UtBas_PrintF( "ad_r_a[%ld:0]", EncWidth );
        ChainCTRLConn = addchain( ChainCTRLConn,
                                  namealloc( TmpName ) );

        if (RFG_BusNB > 0)
        {
          /* Second read address bus ("ad_r_b[MSB:LSB]"). */
          TmpName = UtBas_PrintF( "ad_r_b[%ld:0]", EncWidth );
          ChainCTRLConn = addchain( ChainCTRLConn,
                                    namealloc( TmpName ) );
        }
      }
      else
      {
        ChainCTRLConn = addchain( ChainCTRLConn, "w_r_a" );
        if (RFG_BusNB > 0)
        {
          ChainCTRLConn = addchain( ChainCTRLConn, "w_r_b" );
        }

        /* Write address bus ("w[MSB:LSB]"). */
        TmpName = UtBas_PrintF( "w[%ld:0]", DecWidth );
        ChainCTRLConn = addchain( ChainCTRLConn,
                                  namealloc( TmpName ) );

        /* First read address bus ("r_a[MSB:LSB]"). */
        TmpName = UtBas_PrintF( "r_a[%ld:0]", DecWidth );
        ChainCTRLConn = addchain( ChainCTRLConn,
                                  namealloc( TmpName ) );

        if (RFG_BusNB > 0)
        {
          /* Second read address bus ("r_b[MSB:LSB]"). */
          TmpName = UtBas_PrintF( "r_b[%ld:0]", DecWidth );
          ChainCTRLConn = addchain( ChainCTRLConn,
                                    namealloc( TmpName ) );
        }
      }

      ModelName = UtBas_PrintF( "%s_%ldx%ldx%ldx%ld%c_bk",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                fpgen_Glob.WidthDP,
                                0,
                                DecWidth,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l')   );
      break;

    case C_FgnMBK_FunctionMultp:

      ChainCTRLConn = addchain( ChainCTRLConn,    "ck" );
      ChainCTRLConn = addchain( ChainCTRLConn,    "sd" );
      ChainCTRLConn = addchain( ChainCTRLConn,    "st" );
      ChainCTRLConn = addchain( ChainCTRLConn, "scin"  );
      ChainCTRLConn = addchain( ChainCTRLConn, "scout" );

    case C_FgnMBK_FunctionMult:

      ModelName = UtBas_PrintF( "%s_%ldx%ldx%c%c%ld%c_bk",
                                MacroDB[ aIdentMF ].NameRoot,
                                fpgen_Glob.WidthDP,
                                Width,
                                (GAM_T0 == 1) ? 'i' : 'n',
                                (GAM_T1 == 1) ? 'o' : 'n',
                                GAM_N,
                                ((fpgen_Glob.Msb0) ? 'm' : 'l')   );
      break;
  }
  ModelName = namealloc( ModelName );


  /* Search in list of already generated models. */
  for( ModelPtype  = MacroDB[ aIdentMF ].TabPModel ;
       ModelPtype != (ptype_list*)NULL ;
       ModelPtype  =  ModelPtype->NEXT )
    if (access_ModelLofig->NAME == ModelName) break;

  /* Generation, if necessary of model. */
  if (ModelPtype == (ptype_list*)NULL)
  {
    FgnMBK_TracePrintF(
      "%20s : \"%s\" : Generated on memory ]\n", "[ Model", ModelName );


    /* Save the current logical figure. */
    SaveMBKEnv();

    /* Add the newly generated model in head of list. */
    MacroDB[ aIdentMF ].TabPModel = addptype( MacroDB[ aIdentMF ].TabPModel,
                                              C_GfpUT_UserModelName,
                                              (void*)NULL );

    ModelPtype = MacroDB[ aIdentMF ].TabPModel;

    switch( aIdentMF )
    {
      case C_FgnMBK_FunctionInv:
      case C_FgnMBK_FunctionBuff:
      case C_FgnMBK_FunctionNand2:
      case C_FgnMBK_FunctionNand3:
      case C_FgnMBK_FunctionAnd2:
      case C_FgnMBK_FunctionAnd3:
      case C_FgnMBK_FunctionNor2:
      case C_FgnMBK_FunctionNor3:
      case C_FgnMBK_FunctionOr2:
      case C_FgnMBK_FunctionOr3:
      case C_FgnMBK_FunctionXnor2:
      case C_FgnMBK_FunctionXor2:

        (void)GfpGN_GenericDispatch( MacroDB[ aIdentMF ].Function,
                                     ModelName,
                                     fpgen_Glob.WidthDP,
                                     Width,
                                     Slice,
                                     Drive,
                                     fpgen_Glob.Msb0,
                                     TRUE,
                                     TRUE,
                                     TRUE,
                                     TRUE,
                                     FALSE );
        break;

      case C_FgnMBK_FunctionNand2mask:
      case C_FgnMBK_FunctionNor2mask:
      case C_FgnMBK_FunctionXnor2mask:
      case C_FgnMBK_FunctionConst:

        (void)GfpGN_GenericDispatch( MacroDB[ aIdentMF ].Function,
                                     ModelName,
                                     fpgen_Glob.WidthDP,
                                     Width,
                                     Slice,
                                     XlMask,
                                     fpgen_Glob.Msb0,
                                     TRUE,
                                     TRUE,
                                     TRUE,
                                     TRUE,
                                     FALSE );
        break;

      case C_FgnMBK_FunctionNbuse:
      case C_FgnMBK_FunctionBuse:
      case C_FgnMBK_FunctionNmux2cs:
      case C_FgnMBK_FunctionMux2cs:
      case C_FgnMBK_FunctionMux3cd:
      case C_FgnMBK_FunctionMux3cs:
      case C_FgnMBK_FunctionMux4cs:
      case C_FgnMBK_FunctionMux5cs:
      case C_FgnMBK_FunctionMux6cs:
      case C_FgnMBK_FunctionAdd2s:
      case C_FgnMBK_FunctionNpl1:
      case C_FgnMBK_FunctionPl1:
      case C_FgnMBK_FunctionPmsnw:
      case C_FgnMBK_FunctionPmsnwq:
      case C_FgnMBK_FunctionPms:
      case C_FgnMBK_FunctionPmsq:
      case C_FgnMBK_FunctionPmst:
      case C_FgnMBK_FunctionPmstq:
      case C_FgnMBK_FunctionPdffnw:
      case C_FgnMBK_FunctionPdffnwq:
      case C_FgnMBK_FunctionPdff:
      case C_FgnMBK_FunctionPdffr:
      case C_FgnMBK_FunctionPdffq:
      case C_FgnMBK_FunctionPdfft:
      case C_FgnMBK_FunctionPdffut:
      case C_FgnMBK_FunctionPdffrt:
      case C_FgnMBK_FunctionPdfftq:

        (void)GfpGN_GenericDispatch( MacroDB[ aIdentMF ].Function,
                                     ModelName,
                                     fpgen_Glob.WidthDP,
                                     Width,
                                     Slice,
                                     fpgen_Glob.Msb0,
                                     TRUE,
                                     TRUE,
                                     TRUE,
                                     TRUE,
                                     FALSE );
        break;

      case C_FgnMBK_FunctionNul:

#       ifdef  __GaAs__

/*      if ( UtTec_Family() == C_UtTec_FamilyAsGa ) */
        {
          (void)zara(
                    fpgen_Glob.WidthDP,  /* Size of words in bits.      */
                             ModelName,  /* Model name.                 */
                      (int)       TRUE,  /* Layout view.                */
                      (int)       TRUE,  /* VHDL behavioral view.       */
                      (int)      FALSE,  /* Datasheet.                  */
                      (int)      FALSE,  /* Patterns.                   */
                      (int)      FALSE,  /* Icon.                       */
                      (int)         1L,  /* virtual connectors.         */
                      (int)         0L   /* LSB have index ZERO.        */
                   );
        }
/*      else */

#       else

        {
          (void)GfpGN_GenericDispatch( MacroDB[ aIdentMF ].Function,
                                       ModelName,
                                       fpgen_Glob.WidthDP,
                                       Width,
                                       Slice,
                                       fpgen_Glob.Msb0,
                                       TRUE,
                                       TRUE,
                                       TRUE,
                                       TRUE,
                                       FALSE );
        }

#       endif

        break;

      case C_FgnMBK_FunctionAdd2f:
      case C_FgnMBK_FunctionAdsb2f:

#       ifdef  __GaAs__

/*      if ( UtTec_Family() == C_UtTec_FamilyAsGa ) */
        {
          (void)clara(
                    fpgen_Glob.WidthDP,  /* Size of words in bits.      */
                             ModelName,  /* Model name.                 */
                      (int)       TRUE,  /* Layout view.                */
                      (int)       TRUE,  /* VHDL behavioral view.       */
                      (int)      FALSE,  /* Datasheet.                  */
                      (int)      FALSE,  /* Patterns.                   */
                      (int)      FALSE,  /* Icon.                       */
                      (int)         0L,  /* No carry in.                */
                      (int)         1L,  /* Carry out pin.              */
                      (int)RSA_TypeAsb,  /* Adder/substractor.          */
                      (int)         1L,  /* "over" on the north side.   */
                      (int)         1L,  /* virtual connectors.         */
                      (int)         0L   /* LSB have index ZERO.        */
                   );
        }
/*      else */

#       else

        {
          (void)rsa(
                    fpgen_Glob.WidthDP,  /* Size of words in bits.      */
                                    0L,  /* No carry in.                */
                                    0L,  /* No intermediate carry.      */
                                    1L,  /* "over" on the north side.   */
                                    0L,  /* direct output.              */
                           RSA_TypeCsa,  /* Two operand adder.          */
                           RSA_TypeSub,  /* Substractor.                */
                           RSA_TypeAsb,  /* Adder/substractor.          */
                                    0L,  /* ovr     : BULL option.      */
                                    0L,  /* Deci    : BULL option.      */
                                    0L,  /* Stretch : BULL option.      */
                                    0L,  /* Stretch2: BULL option.      */
                                    0L,  /* LSB have index ZERO.        */
                                    1L,  /* virtual connectors.         */
                                  TRUE,  /* Layout view.                */
                                 FALSE,  /* Icon.                       */
                                  TRUE,  /* VHDL behavioral view.       */
                                 FALSE,  /* Patterns.                   */
                                 FALSE,  /* Datasheet.                  */
                                 FALSE,  /* Outline.                    */
                                  TRUE,  /* Connector netlist.          */
                                 FALSE,  /* SPICE performances.         */
                             ModelName   /* Model name.                 */
                   );
        }

#       endif

        break;

      case C_FgnMBK_FunctionShift:

#       ifdef  __GaAs__

/*      if ( UtTec_Family() == C_UtTec_FamilyAsGa ) */
        {
          (void)sara(
                    fpgen_Glob.WidthDP,  /* Size of words in bits.      */
                             ModelName,  /* Model name.                 */
                            (int) TRUE,  /* Layout view.                */
                            (int) TRUE,  /* VHDL behavioral view.       */
                            (int)FALSE,  /* Datasheet.                  */
                            (int)FALSE,  /* Patterns.                   */
                            (int)FALSE,  /* Icon.                       */
                            (int) TRUE,  /* Arithmetical shift allowed. */
                            (int)   1L,  /* virtual connectors.         */
                            (int)   0L   /* LSB have index ZERO.        */
                   );
        }
/*      else */

#       else

        {
          (void)bsg(
                    fpgen_Glob.WidthDP,  /* Size of words in bits.      */
                                    1L,  /* virtual connectors.         */
                                    0L,  /* LSB have index ZERO.        */
                             ModelName,  /* Model name.                 */
                                  TRUE,  /* Layout view.                */
                                 FALSE,  /* Icon.                       */
                                  TRUE,  /* VHDL behavioral view.       */
                                 FALSE,  /* Patterns.                   */
                                 FALSE,  /* Datasheet.                  */
                                 FALSE,  /* Outline.                    */
                                  TRUE   /* Connector netlist.          */
                   );
        }

#       endif

        break;

      case C_FgnMBK_FunctionRfg1cc:
      case C_FgnMBK_FunctionRfg2cc:
      case C_FgnMBK_FunctionRfg1fc:
      case C_FgnMBK_FunctionRfg2fc:
      case C_FgnMBK_FunctionRfg1c0:
      case C_FgnMBK_FunctionRfg2c0:
      case C_FgnMBK_FunctionRfg1f0:
      case C_FgnMBK_FunctionRfg2f0:
      case C_FgnMBK_FunctionRfgu2cc:
      case C_FgnMBK_FunctionRfgu1cc:
      case C_FgnMBK_FunctionRfgu2fc:
      case C_FgnMBK_FunctionRfgu1fc:
      case C_FgnMBK_FunctionRfgu2c0:
      case C_FgnMBK_FunctionRfgu1c0:
      case C_FgnMBK_FunctionRfgu2f0:
      case C_FgnMBK_FunctionRfgu1f0:

#       ifdef  __GaAs__

/*      if ( UtTec_Family() == C_UtTec_FamilyAsGa ) */
        {
          (void)raga(
             (int)       RFG_BusNB + 1,  /* Number of read buses.      */
            (long)            DecWidth,  /* Number of words.           */
            (long)  fpgen_Glob.WidthDP,  /* Size of words in bits.     */
             (int)                TRUE,  /* Layout view.               */
             (int)               FALSE,  /* Data-sheet.                */
             (int)                TRUE,  /* VHDL behavioral view.      */
             (int)               FALSE,  /* Patterns.                  */
             (int)               FALSE,  /* Icon.                      */
             (int)( 1 - RFG_Type / 2 ),  /* Type of register.          */
             (int)           RFG_Stuck,  /* Normal register 0.         */
             (int)                TRUE,  /* Write Enable active hight. */
             (int)                  1L,  /* D flip-flops.              */
             (int)                  0L,  /* Multi-access connectors.   */
             (int)                  0L,  /* Msb0 on slice 0.           */
                             ModelName   /* Model name.                */
                   );
        }
/*      else */

#       else

        {
          (void)rfg(
                              DecWidth,  /* Number of words.          */
                    fpgen_Glob.WidthDP,  /* Size of words in bits.    */
                             RFG_BusNB,  /* Number of read buses.     */
                             RFG_Edge ,  /* D flip-flops.             */
                                    0L,  /* Non-inverted outputs.     */
                                    0L,  /* Normal power.             */
                                    1L,  /* Write enable.             */
                                    0L,  /* Write enable active high. */
                             RFG_Stuck,  /* Normal register 0.        */
                                    0L,  /* Multi-access connectors.  */
                                    0L,  /* Msb0 on slice 0.          */
                             RFG_Type ,  /* Type of register.         */
                             ModelName,  /* Model name.               */
                                  TRUE,  /* Layout view.              */
                                 FALSE,  /* Outline.                  */
                                  TRUE,  /* VHDL behavioral view.     */
                                 FALSE,  /* Patterns.                 */
                                 FALSE,  /* Data-sheet.               */
                                  TRUE,  /* Connector netlist.        */
                                 FALSE   /* Icon.                     */
                   );
        }

#       endif

        break;

      case C_FgnMBK_FunctionMult:
      case C_FgnMBK_FunctionMultp:

#       ifdef  __GaAs__

/*      if ( UtTec_Family() == C_UtTec_FamilyAsGa ) */
        {
          (void)bat(
                         fpgen_Glob.WidthDP,  /* Size of multiplicand.      */
                                     Width ,  /* Size of multiplicator.     */
                                (int)  TRUE,  /* Layout view.               */
                                (int)  TRUE,  /* VHDL behavioral view.      */
                                (int) FALSE,  /* Patterns.                  */
                                (int) FALSE,  /* Data-sheet.                */
                                (int) FALSE,  /* Icon.                      */
                                (int)     0,  /* Msb0 on slice 0.           */
                                (int)     1,  /* Multi-access connectors.   */
                                  ModelName   /* Model name.                */
                   );
        }
/*      else */

#       else

        {
          (void)amg(
                    (int)fpgen_Glob.WidthDP,  /* Size of multiplicand.      */
                                (int)Width ,  /* Size of multiplicator.     */
                                (int)GAM_T0,  /* Input  pipe line register. */
                                (int)GAM_T1,  /* Output pipe line register. */
                                (int)GAM_N ,  /* Array  pipe line register. */
                                (int)  TRUE,  /* Layout view.               */
                                (int)  TRUE,  /* VHDL behavioral view.      */
                                (int) FALSE,  /* Data-sheet.                */
                                (int) FALSE,  /* Icon.                      */
                                (int) FALSE,  /* Patterns.                  */
                                (int)     1,  /* Multi-access connectors.   */
                                (int)     0,  /* Msb0 on slice 0.           */
                                  ModelName   /* Model name.                */
                   );
        }

#       endif

        break;

    }

    /* Recover the logical figure of model asked for. */ 
    switch( MacroDB[ aIdentMF ].Function )
    {
      default:
      case C_GfpGN_FunctionNul:

        LoFigMode = ( UtTec_Family() == C_UtTec_FamilyAsGa ) ? 'P' : 'A';

        break;

      case C_GfpGN_FunctionUnknow:

        LoFigMode = 'P';

        break;
    }
    ModelPtype->DATA = (void *)getlofig( ModelName, LoFigMode );

    /* Add the newly generated model to the CATAL file of the FpGen
     * library.
     */
    if (!mc_InCatalog( ModelName ))
    {
      fprintf( fpgen_Glob.FileCATAL, "%-50s  C\n", ModelName );
    }

    /* Restore current logical figure. */
    RestoreMBKEnv();
  }
  else
  {
    FgnMBK_TracePrintF(
      "%20s : \"%s\" : Found on memory ]\n", "[ Model", ModelName );
  }


  /* Instanciation of the newly created model. */
  FgnMBK_xvLoIns( ModelName,
                  InstName,
                  ChainCTRLConn,
                  aStdArgs );

  /* Attempt to delete the list of implicit connectors name. */
  if (ChainCTRLConn != (chain_list*)NULL)
  {
    freechain( ChainCTRLConn );
  }


  FgnMBK_TraceFlush();

  
#undef  access_ModelLofig
}


/*  Function  :  'FgnMBK_Init'.                                             */
/* ------------------------------------------------------------------------ */
/*  Adds to the main error table all errors and warnings belongings to the
 *  functions of module "fpgen". This function is normally called by
 *  'DP_DEFLO' at is only first call.
 */

extern void  FgnMBK_Init()
{
  static long  FirstCall = TRUE;


  if (FirstCall == TRUE)
  {
    /* Attribution of a module identifier. */
    fpgen_Glob.ModuleId = GfpUT_GetModuleId( "FpGen", F_GfpUT_ModuleAdd );


    mbkenv();


    /* Trap SIGSEGV signal. */
    if (signal( SIGSEGV, FgnMBK_ErrorTrap ) == SIG_ERR)
    {
      FgnMBK_ErrorHeader( "Unable to trap signal SIGSEGV.\n" );
      FgnMBK_ErrorTrap( SIGSEGV );
    }


    if (signal( SIGTERM, FgnMBK_ErrorTrap ) == SIG_ERR)
    {
      FgnMBK_ErrorHeader( "Unable to trap signal SIGTERM.\n" );
      FgnMBK_ErrorTrap( SIGTERM );
    }


    /* Set error mode to 'kill'. */
    UtBas_SetErrorMode( C_UtBas_ErrorModeKill );


    /* Errors belonging to function DP_DEFLOFIG.
     */
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_DEFLOCall,
    "[%s]:Model \"%s\":Previous figure not saved by DP_SAVLOFIG.");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_DEFLOWidthMax,
    "[%s]:Model \"%s\":Invalid MaxWidth:%ld (maximum is %ld).");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_DEFLOWidthEven,
    "[%s]:Model \"%s\":Even width (%ld) are not supported yet.");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_DEFLOLsbPos,
    "[%s]:Model \"%s\":Invalid lsbpos:%ld.");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_DEFLOBadType,
    "[%s]:Model \"%s\":Invalid data-path type:'%c'.");

    /* Warnings belonging to function DP_DEFLOFIG.
     */
    GfpUT_AddWarng(fpgen_Glob.ModuleId, W_FgnMBK_DEFLOMbkSepar,
    "[%s]:MBK_SEPAR(%c) different from \".\".");
    GfpUT_AddWarng(fpgen_Glob.ModuleId, W_FgnMBK_DEFLOMbkInOutPH,
    "[%s]:MBK_IN_PH(%s) must be equal to MBK_OUT_PH(%s).");
           

    /* Errors belonging to function DP_SAVLOFIG.
     */
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_SAVLOCall,
    "[%s]:No data-path figure to save, use DP_DEFLOFIG.");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_SAVLOMissVDD,
    "[%s]:Model \"%s\":Power supply connector VDD is missing.");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_SAVLOMissVSS,
    "[%s]:Model \"%s\":Power supply connector VSS is missing.");


    /* Errors belonging to function DP_LOCON.
     */
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_LOCONPOWERDuplicate,
    "[%s]:Model \"%s\":Redeclaration of power connector \"%s\".");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_LOCONPOWERBadNamePower,
"[%s]:Model \"%s\":Invalid power name \"%s\", \"vdd\" or \"vss\" only.");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_LOCONBadSide,
    "[%s]:Model \"%s\":Invalid side '%c' for connector \"%s\".");


    /* Errors belonging to function DP_CONMODE.
     */
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_CONMODEUnknow,
    "[%s]:Unknow CTRL terminal mode:%ld.");


    /* Errors shared by various functions.
     */
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_SharedMissDefLo,
    "[%s]:Missing previous call to DP_DEFLOFIG.");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_SharedBusName,
    "[%s]:Invalid bus name \"%s\".");


    /* Errors belonging to the 'FgnMBK_GenericMF' function.
     */
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_GenericMFIdent,
    "[%s]:Bad \"IdentMF\":%ld.");
    GfpUT_AddError( fpgen_Glob.ModuleId, E_FgnMBK_GenericMFMissName,
    "[%s]:Missing instance name.");
    GfpUT_AddError( fpgen_Glob.ModuleId, E_FgnMBK_GenericMFWidth,
    "[%s]:Instance \"%s\":Width out of range:%ld.");
    GfpUT_AddError( fpgen_Glob.ModuleId, E_FgnMBK_GenericMFSlice,
    "[%s]:Instance \"%s\":Slice out of range:%ld.");
    GfpUT_AddError( fpgen_Glob.ModuleId, E_FgnMBK_GenericMFDrive,
    "[%s]:Instance \"%s\":Drive out of range:%ld.");
    GfpUT_AddError( fpgen_Glob.ModuleId, E_FgnMBK_GenericMFWords,
    "[%s]:Instance \"%s\":Words out of range:%ld.");


    /* Errors belonging to function 'FgnMBK_NameToSig'.
     */
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_NameToSigCtrlData,
    "[%s]:Invalid connexion CTRL-DATA on signal:\"%s\".");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_NameToSigCtrlCtrl,
    "[%s]:Invalid connexion CTRL-CTRL on signal \"%s\".");


    /* Errors belonging to function 'FgnMBK_MatchImplicitConn'.
     */
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_MatchInterfNoMoreCon,
    "[%s]:Unable to match interface, more signals than connectors.");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_MatchInterfNoMoreSig,
    "[%s]:Unable to match interface, more connectors than signals.");


    /* Errors belonging to function 'FgnGLU_ImportGlue'.
     */
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_ImportGlueCPC,
    "[DP_GLUE]:Instance \"%s\":CellPerCol negative:%ld.");
    GfpUT_AddError(fpgen_Glob.ModuleId, E_FgnMBK_ImportGluePlaceFunc,
    "[DP_GLUE]:Instance \"%s\":Symbolic placement failed.");


    FirstCall = FALSE;
  }
}


/*  Fonction  :  'CheckInsName'.                                            */
/* ------------------------------------------------------------------------ */
/*  Check if the string is a valid instance name, i.e. do not contain any
 *  of the following chararacters '[', ']', '(', ')', '.', ' '.
 */

static long  CheckInsName( aName )
  char *aName;
{
  /* Table of the forbidden characters. */
  char *Table = "[](). ";

  char *PtCh;
  long  Index, i;


  for( Index = 0 ; aName[ Index ] != (char)0 ; Index++ )
  {
    for( PtCh = Table ; *PtCh != (char)0 ; PtCh++ )
    {
      if (aName[ Index ] == *PtCh)
      {
        FgnMBK_ErrorHeader( "CheckInsName" );
        FgnMBK_ErrorPrintF( "Invalid character \'%c\' in instance name:",
                            *PtCh );
        FgnMBK_ErrorPrintF( "\nFpGen:\"%s\".\nFpGen:", aName );

        for( i = 0 ; i < Index + 1 ; i++ )
          FgnMBK_ErrorPrintF( "%c", '_' );

        FgnMBK_ErrorPrintF( "%c", '?' );

        FgnMBK_ErrorTrap( SIGTERM );
      }
    }
  }
}


/*  Function  :  'FgnMBK_BusToChain'                              */
/* -------------------------------------------------------------- */
/*      Transform a bus signal name to a chained list of sample
 *  signal names (in the MBK internal format). Then add it in
 *  front of the list "pthead". For check purpose it gives the
 *  bus wide in the long integer pointed by "pcount".
 */

static chain_list *FgnMBK_BusToChain(pthead,busname,pcount)
    chain_list *pthead;
          char *busname;
          long *pcount;
{
    chain_list *BusChainName;
    chain_list *TmpChain;

          char  StringBuffer[C_SUPP_MBKSZNAME];
          char  PinName[     C_SUPP_MBKSZNAME];
          long  OrigIndex;
          long  ExtrIndex;
          long   TmpIndex;

    BusChainName = (chain_list*)NULL;

    /* Decod bus name. */
    if (!bus_decod(busname                 ,
                   StringBuffer,&OrigIndex ,
                                &ExtrIndex ))
        GfpUT_PrintError(fpgen_Glob.ModuleId,
                      E_FgnMBK_SharedBusName,
                         fpgen_Glob.CurrFunc,
                                     busname);

    /* Creation of signal names associated to the bus. */
    if (OrigIndex > ExtrIndex)
    {
        for(TmpIndex = OrigIndex; TmpIndex >= ExtrIndex;
                                  TmpIndex--           )
        {
            /* Name of bus pin. */
            (void)sprintf( PinName     ,
                          "%s %ld"     ,
                           StringBuffer,
                           TmpIndex    );

            /* Addition to the pin list. */
            BusChainName = addchain( BusChainName        ,
                                (void*)namealloc(PinName));
        }
        /* Signal count. */
        *pcount = OrigIndex - ExtrIndex +1;
    }
    else
    {
        for(TmpIndex = OrigIndex; TmpIndex <= ExtrIndex;
                                  TmpIndex++           )
        {
            /* Name of bus pin. */
            (void)sprintf( PinName     ,
                          "%s %ld"     ,
                           StringBuffer,
                           TmpIndex    );

            /* Addition a la liste des pattes. */
            BusChainName = addchain( BusChainName       ,
                                (void*)namealloc(PinName));
        }
        /* Signal count. */
        *pcount = ExtrIndex - OrigIndex +1;
    }
    
    /* Add the list of pin bus to the front of the list. */
    for( TmpChain        = BusChainName     ;
         TmpChain->NEXT != (chain_list*)NULL; 
         TmpChain        = TmpChain->NEXT   );

    TmpChain->NEXT  = pthead;
    pthead = BusChainName;

    return( BusChainName);
}


/*  Function  :  'FgnMBK_NameToSig'                               */
/* -------------------------------------------------------------- */
/*  Add to the front of the MBK signal list pointed by "aPtChain"
 *  a new signal named "aSigName". A pointer to the new
 *  'chain_list' is returned.
 *    First the function search though the list of signals to
 *  find if it have been already created. If not found the
 *  function add a new signal.
 */

static chain_list *FgnMBK_NameToSig( aPtChain, aNameSig, aTypeSig )
  chain_list *aPtChain;
        char *aNameSig;
        long  aTypeSig;
{
    chain_list *TmpChainName;
    chain_list  *PtChainNew;
    losig_list *TmpSig;
    ptype_list *PtPtype;
          char *TmpName;


    TmpName = namealloc((char*)(aNameSig));
        
    /* Go through the list of existing signals. */
    for(TmpSig  = WORK_LOFIG->LOSIG;
        TmpSig != (losig_list*)NULL;
        TmpSig  = TmpSig->NEXT     )
    {
        /* Go through the list of signal names. */
        for(TmpChainName  = TmpSig->NAMECHAIN ;
            TmpChainName != (chain_list*)NULL ;
            TmpChainName  = TmpChainName->NEXT)
            if ((char*)(TmpChainName->DATA) == TmpName)
                break;
            
        /* The name of the signal has been found. */
        if (TmpChainName != (chain_list*)NULL) break;
    }

    /* If not found : create it. */
    if (TmpSig == (losig_list*)NULL)
    {
        num_index += 1;
        TmpSig = addlosig(WORK_LOFIG                 ,
                           num_index                 ,
                          addchain((chain_list*)NULL,
                                   (void*)TmpName   ),
                          INTERNAL );
    }

    /* Addition of signal to the list. */
    PtChainNew = addchain(aPtChain, (void*)TmpSig);
    PtPtype    = getptype(TmpSig->USER, C_GfpUT_UserSigType);
        
    /* Check if the signal is typed. */
    if (PtPtype == (ptype_list*)NULL)
        TmpSig->USER = addptype(TmpSig->USER       ,
                                C_GfpUT_UserSigType,
                                    (void*)aTypeSig);
    else
    {

        /* Check for invalid connexions CTRL-DATA. */
        if (PtPtype->DATA != (void*)aTypeSig)
            GfpUT_PrintError(fpgen_Glob.ModuleId,
                      E_FgnMBK_NameToSigCtrlData,
                             fpgen_Glob.CurrFunc,
                                         TmpName);

        /* Check for invalid connexions CTRL-CTRL. */
        if (   (PtPtype->DATA == (void*)C_FgnMBK_TypeSigCTRL)
            && (  aTypeSig    ==        C_FgnMBK_TypeSigCTRL))
            GfpUT_PrintError(fpgen_Glob.ModuleId,
                      E_FgnMBK_NameToSigCtrlCtrl,
                             fpgen_Glob.CurrFunc,
                                         TmpName);
    }

    return( PtChainNew);
}


/*  Function  :  'FgnMBK_MatchImplicitConn'                       */
/* -------------------------------------------------------------- */
/*  Read the locon list 'ptcon' and look for supplies connectors
 *  or implicit connectors. If one of those connectors is found
 *  a corresponding signal name is inserted at the right place in
 *  the signal name list 'ptchain'. 
 *    In the case of an implicit connector (i.e. a hidden control
 *  connector) it is added to the interface of the current logi-
 *  cal figure with a new signal.
 */

static chain_list *FgnMBK_MatchImplicitConn(aInstcName   ,
                                            aChainConName,
                                            aChainSigName,
                                            aPtLoFig     ,
                                            aPtSigCount  )
          char *aInstcName;     /* Instance name.               */
    chain_list *aChainConName;  /* List of CTRL terminals name. */
    chain_list *aChainSigName;
    lofig_list *aPtLoFig;
          long *aPtSigCount;
{
          char  *TmpName;
          char  *SigName;
          long   SigNameIndex;
          long   SigType;
    locon_list  *PtCon;
    chain_list  *PtChainSigName;
    chain_list  *PtChainCon;
    chain_list  *PtChainSig = (chain_list*)NULL;

    
    FgnMBK_TracePrintF( "*** FpGen xvLoins ***:MBK terminals assigments\n" );

    PtChainSigName = aChainSigName;
    /* Initialize the number of added signals. */
    *aPtSigCount = 0;

    for(PtCon  = aPtLoFig->LOCON  ;
        PtCon != (locon_list*)NULL;
        PtCon  =  PtCon->NEXT     )
    {
        switch(GfpUT_TypeConName(PtCon->NAME))
        {
            case C_GfpUT_TypeConVdd:
                /* Add a signal named "vdd" to the signal list. */
                  PtChainSig  = addchain(PtChainSig, (void*)fpgen_Glob.SigVdd);
                *aPtSigCount += 1;

                FgnMBK_TracePrintF( "%20s => vdd\n", PtCon->NAME );
                break;
            case C_GfpUT_TypeConVss:
                /* Add a signal named "vss" to the signal list. */
                  PtChainSig  = addchain(PtChainSig, (void*)fpgen_Glob.SigVss);
                *aPtSigCount += 1;

                FgnMBK_TracePrintF( "%20s => vss\n", PtCon->NAME );
                break;
            case C_GfpUT_TypeConImplicit:
                /* Name of the implicit signal and connector. */
                TmpName = concatname(aInstcName,PtCon->NAME);

                FgnMBK_TracePrintF( "%20s => %s\n", PtCon->NAME, TmpName );

                /* Creation of SIGNAL "TmpName". */
                PtChainSig = FgnMBK_NameToSig(PtChainSig,
                                                 TmpName,
                                     C_FgnMBK_TypeSigCTRL);
                *aPtSigCount += 1;

                /* Creation of CONNECTOR "TmpName". */
                (void)GfpUT_AddLoConEnd(WORK_LOFIG                     ,
                                        TmpName                        ,
                                        (losig_list*)(PtChainSig->DATA),
                                        PtCon->DIRECTION               );
                

                break;
            case C_GfpUT_TypeConNormal:
            default :
                SigType = C_FgnMBK_TypeSigDATA;

                for(PtChainCon  =  aChainConName    ;
                    PtChainCon !=  (chain_list*)NULL;
                    PtChainCon  = PtChainCon->NEXT)
                {
                    TmpName = (char*)PtChainCon->DATA;

                    if (TmpName == PtCon->NAME)
                    {
                        SigType = C_FgnMBK_TypeSigCTRL;

                        if (fpgen_Glob.ConMode == C_FgnMGN_ConModeIMPLICIT)
                        {
                            SigName = GfpUT_DecodMBKSigName(TmpName     ,
                                                           &SigNameIndex);
                            TmpName = GfpUT_MakeImplicitConName(aInstcName  ,
                                                                SigName     ,
                                                                SigNameIndex);
                        }
                        break;
                    }
                }

                /* Get the name of the signal :            */
                /* - DATA connector.                       */
                /* - CTRL connector when in EXPLICIT mode. */
                if (   (           SigType != C_FgnMBK_TypeSigCTRL    )
                    || (fpgen_Glob.ConMode != C_FgnMGN_ConModeIMPLICIT))
                {
                    /* Check the current signal. */
                    if (PtChainSigName == (chain_list*)NULL)
                    {
                        do
                        {
                          FgnMBK_TracePrintF( "%20s => ", PtCon->NAME );
                          FgnMBK_TracePrintF( "Unmatched\n" );

                          PtCon = PtCon->NEXT;
                        }
                        while( PtCon != (locon_list*)NULL );

                        FgnMBK_TraceFlush();
                        GfpUT_PrintError(fpgen_Glob.ModuleId,
                               E_FgnMBK_MatchInterfNoMoreSig,
                                         fpgen_Glob.CurrFunc);
                    }

                           TmpName = (char*)(PtChainSigName->DATA);
                    PtChainSigName = PtChainSigName->NEXT;

                    *aPtSigCount  += 1;
                }

                FgnMBK_TracePrintF( "%20s => %s\n", PtCon->NAME, TmpName );

                /* Creation of SIGNAL "TmpName". */
                PtChainSig = FgnMBK_NameToSig(PtChainSig,
                                                 TmpName,
                                                 SigType);

                if (   (           SigType == C_FgnMBK_TypeSigCTRL    )
                    && (fpgen_Glob.ConMode == C_FgnMGN_ConModeIMPLICIT))
                    /* Creation of CONNECTOR "TmpName". */
                    (void)GfpUT_AddLoConEnd(WORK_LOFIG                     ,
                                            TmpName                        ,
                                            (losig_list*)(PtChainSig->DATA),
                                            PtCon->DIRECTION               );
    
                break;
        }

    }

    /* Check if all the signal have been matched with the conns. */
    if (PtChainSigName != (chain_list*)NULL)
    {
        do
        {
          FgnMBK_TracePrintF( "%20s => ", "Unmatched" );
          FgnMBK_TracePrintF( "%s\n", (char*)(PtChainSigName->DATA) );

          PtChainSigName = PtChainSigName->NEXT;
        }
        while( PtChainSigName != (chain_list*)NULL );

        FgnMBK_TraceFlush();
        GfpUT_PrintError(fpgen_Glob.ModuleId,
               E_FgnMBK_MatchInterfNoMoreCon,
                         fpgen_Glob.CurrFunc);
    }

    /* Return the new head of signal chain list. */
    return(PtChainSig);
}


/*  Fonction  :  'AddPowerSig'.                                             */
/* ------------------------------------------------------------------------ */
/*  Detect power supplies terminals of the model and insert a signal in the
 *  instance signal list.
 */

static chain_list *AddPowerSig( aLoFig, aChainConName, aChainSigName )
  lofig_list *aLoFig;         /* Model associated to the instance.        */
  chain_list *aChainConName;  /* List of CTRL terminals names.            */
  chain_list *aChainSigName;  /* List of signals names (whithout powers). */
{
  locon_list *LoCon;
  chain_list *ChainSig, *ChainConName;
        char *TmpName, *MBKName;
        char *SigName, *ConName;
        char *SigRoot, *ConRoot;
        long  SigVect,  ConVect;
        long  SigOrig,  SigExtr, SigIncr;
        long  ConOrig,  ConExtr, ConIncr;
        long  SigType;


  SigIncr  = 0; SigRoot = (char*)NULL;
  ConIncr  = 0; ConRoot = (char*)NULL;
  ChainSig = (chain_list*)NULL;

  for( LoCon  = aLoFig->LOCON     ;
       LoCon != (locon_list*)NULL ;
       LoCon  =       LoCon->NEXT )
  {
    if (isvdd( LoCon->NAME ))
    {
      /* Add a VDD signal in the signal list. */
      ChainSig  = addchain( ChainSig, (void*)fpgen_Glob.SigVdd );

      continue;
    }


    if (isvss( LoCon->NAME ))
    {
      /* Add a VSS signal in the signal list. */
      ChainSig  = addchain( ChainSig, (void*)fpgen_Glob.SigVss );

      continue;
    }


    if (aChainSigName == (chain_list*)NULL)
    {
      /* There is more terminals in the figure
       * than signals in the instance.
       */

      /* Flush previously vectorized terminal/signal. */
      if (SigRoot != (char*)NULL)
      {
        if (SigOrig == SigExtr)
        {
          /* This is a one bit vector. */
          TmpName = UtBas_PrintF( "%s[%ld]", ConRoot, ConOrig );

          FgnMBK_TracePrintF( "%20s => %s[%ld]\n", TmpName, SigRoot, SigOrig );
        }
        else
        {
          /* This a real vector. */
          TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                                  ConRoot,
                                  ConOrig,
                                  ConExtr );

          FgnMBK_TracePrintF( "%20s => %s[%ld:%ld]\n",
                              TmpName,
                              SigRoot,
                              SigOrig,
                              SigExtr );
        }

        SigIncr = 0; SigRoot = (char*)NULL;
        ConIncr = 0; ConRoot = (char*)NULL;
      }

      for( ; LoCon != (locon_list*)NULL ; LoCon = LoCon->NEXT )
      {
        ConName = GfpUT_DecodMBKSigName( LoCon->NAME, &ConVect );
        ConName = UtBas_PrintF( "%s", ConName );

        if (ConVect == C_GfpUT_ImplicitConNoIndex)
        {
          /* Print previously stored vectorized terminal. */
          if (ConRoot != (char*)NULL)
          {
            if (ConOrig == ConExtr)
            {
              /* This is a one bit vector. */
              TmpName = UtBas_PrintF( "%s[%ld]", ConRoot, ConOrig );
            }
            else
            {
              /* This a real vector. */
              TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                                      ConRoot,
                                      ConOrig,
                                      ConExtr );
            }
          }
          FgnMBK_TracePrintF( "%20s => Unmatched\n", TmpName );

          if (ConVect != C_GfpUT_ImplicitConNoIndex)
          {
            ConName = UtBas_PrintF( "%s[%ld]", ConName, ConVect );
          }
          FgnMBK_TracePrintF( "%20s => Unmatched\n", ConName );

          ConIncr = 0; ConRoot = (char*)NULL;
        }
        else
        {
          if (ConRoot == (char*)NULL)
          {
            /* This is the first bit. */
            ConRoot = ConName; ConOrig = ConExtr = ConVect; ConIncr = 0;
          }
          else
          {
            if (ConIncr == 0)
            {
              ConIncr = ConVect - ConOrig;

              if (   ((ConIncr != -1) && (ConIncr != 1))
                  || (strcmp( ConRoot, ConName )) )
              {
                /* Bits are not contiguous. */
                TmpName = UtBas_PrintF( "%s[%ld]", ConRoot, ConOrig );

                FgnMBK_TracePrintF( "%20s => Unmatched\n", TmpName );

                ConRoot = ConName; ConOrig = ConExtr = ConVect; ConIncr = 0;
              }
              else
              {
                ConExtr = ConVect;
              }
            }
            else
            {
              /* Attempt to add a new bit to the current vector. */
              if (   (ConExtr + ConIncr != ConVect)
                  || (strcmp( ConRoot, ConName )) )
              {
                /* Bits are not contiguous. */

                TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                                        ConRoot,
                                        ConOrig,
                                        ConExtr );

                FgnMBK_TracePrintF( "%20s => Unmatched\n", TmpName );

                /* The new pair becomes the start bit. */
                ConRoot = ConName; ConOrig = ConExtr = ConVect; ConIncr = 0;
              }
              else
              {
                ConExtr = ConVect;
              }
            }
          }
        }
      }


      /* Flush the last vectorized terminal/Connal. */
      if (ConRoot != (char*)NULL)
      {
        if (ConOrig == ConExtr)
        {
          /* This is a one bit vector. */
          TmpName = UtBas_PrintF( "%s[%ld]", ConRoot, ConOrig );
        }
        else
        {
          /* This a real vector. */
          TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                                  ConRoot,
                                  ConOrig,
                                  ConExtr );
        }

        FgnMBK_TracePrintF( "%20s => Unmatched\n", TmpName );
      }


      GfpUT_PrintError( fpgen_Glob.ModuleId,
                        E_FgnMBK_MatchInterfNoMoreSig,
                        fpgen_Glob.CurrFunc );
    }


    MBKName = (char*)(aChainSigName->DATA);


    /* Attempt to recognize a CTRL terminal. */
    SigType = C_FgnMBK_TypeSigDATA;

    for( ChainConName  =  aChainConName     ;
         ChainConName !=  (chain_list*)NULL ;
         ChainConName  = ChainConName->NEXT )
    {
      if ((char*)(ChainConName->DATA) == LoCon->NAME)
      {
        SigType = C_FgnMBK_TypeSigCTRL; break;
      }
    }


    SigName = GfpUT_DecodMBKSigName( MBKName, &SigVect );
    SigName = UtBas_PrintF( "%s", SigName );

    ConName = GfpUT_DecodMBKSigName( LoCon->NAME, &ConVect );
    ConName = UtBas_PrintF( "%s", ConName );


    if (   (SigVect == C_GfpUT_ImplicitConNoIndex)
        || (ConVect == C_GfpUT_ImplicitConNoIndex))
    {
      /* At least one of the terminal is not vectorized. */

      /* Print previously stored vectorized terminal. */
      if (SigRoot != (char*)NULL)
      {
        if (SigOrig == SigExtr)
        {
          /* This is a one bit vector. */
          TmpName = UtBas_PrintF( "%s[%ld]", ConRoot, ConOrig );

          FgnMBK_TracePrintF( "%20s => %s[%ld]\n", TmpName, SigRoot, SigOrig );
        }
        else
        {
          /* This a real vector. */
          TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                                  ConRoot,
                                  ConOrig,
                                  ConExtr );

          FgnMBK_TracePrintF( "%20s => %s[%ld:%ld]\n",
                              TmpName,
                              SigRoot,
                              SigOrig,
                              SigExtr );
        }
      }

      /* Print the current pair terminal/signal. */
      if (SigVect != C_GfpUT_ImplicitConNoIndex)
      {
        SigName = UtBas_PrintF( "%s[%ld]", SigName, SigVect );
      }

      if (ConVect != C_GfpUT_ImplicitConNoIndex)
      {
        ConName = UtBas_PrintF( "%s[%ld]", ConName, ConVect );
      }

      FgnMBK_TracePrintF( "%20s => %s\n", ConName, SigName );

      SigIncr = 0; SigRoot = (char*)NULL;
      ConIncr = 0; ConRoot = (char*)NULL;
    }
    else
    {
      /* This is a pair of vectorized terminal/signal. */

      if (SigRoot == (char*)NULL)
      {
        /* This is the first bit. */
        SigRoot = SigName; SigOrig = SigExtr = SigVect; SigIncr = 0;
        ConRoot = ConName; ConOrig = ConExtr = ConVect; ConIncr = 0;
      }
      else
      {
        if (SigIncr == 0)
        {
          /* This is the second bit. */
          SigIncr = SigVect - SigOrig;
          ConIncr = ConVect - ConOrig;

          if (   ((SigIncr != -1) && (SigIncr != 1))
              || ((ConIncr != -1) && (ConIncr != 1))
              || (strcmp( SigRoot, SigName ))
              || (strcmp( ConRoot, ConName )) )
          {
            /* Bits are not contiguous. */
            TmpName = UtBas_PrintF( "%s[%ld]", ConRoot, ConOrig );

            FgnMBK_TracePrintF( "%20s => %s[%ld]\n",
                                TmpName,
                                SigRoot,
                                SigOrig );

            /* The new pair becomes the start bit. */
            SigRoot = SigName; SigOrig = SigExtr = SigVect; SigIncr = 0;
            ConRoot = ConName; ConOrig = ConExtr = ConVect; ConIncr = 0;
          }
          else
          {
            SigExtr = SigVect;
            ConExtr = ConVect;
          }
        }
        else
        {
          /* Attempt to add a new bit to the current vector. */
          if (   (SigExtr + SigIncr != SigVect)
              || (ConExtr + ConIncr != ConVect)
              || (strcmp( SigRoot, SigName ))
              || (strcmp( ConRoot, ConName )) )
          {
            /* Bits are not contiguous. */

            TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                                    ConRoot,
                                    ConOrig,
                                    ConExtr );

            FgnMBK_TracePrintF( "%20s => %s[%ld:%ld]\n",
                                TmpName,
                                SigRoot,
                                SigOrig,
                                SigExtr );

            /* The new pair becomes the start bit. */
            SigRoot = SigName; SigOrig = SigExtr = SigVect; SigIncr = 0;
            ConRoot = ConName; ConOrig = ConExtr = ConVect; ConIncr = 0;
          }
          else
          {
            SigExtr = SigVect;
            ConExtr = ConVect;
          }
        }
      }
    }

    /* Creation of SIGNAL "MBKName". */
    ChainSig = FgnMBK_NameToSig( ChainSig, MBKName, SigType );

    aChainSigName = aChainSigName->NEXT;
  }

  /* Print previously stored vectorized terminal. */
  if (SigRoot != (char*)NULL)
  {
    if (SigOrig == SigExtr)
    {
      /* This is a one bit vector. */
      TmpName = UtBas_PrintF( "%s[%ld]", ConRoot, ConOrig );

      FgnMBK_TracePrintF( "%20s => %s[%ld]\n", TmpName, SigRoot, SigOrig );
    }
    else
    {
      /* This a real vector. */
      TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                              ConRoot,
                              ConOrig,
                              ConExtr );

      FgnMBK_TracePrintF( "%20s => %s[%ld:%ld]\n",
                          TmpName,
                          SigRoot,
                          SigOrig,
                          SigExtr );
    }

    SigIncr = 0; SigRoot = (char*)NULL;
    ConIncr = 0; ConRoot = (char*)NULL;
  }

  if (aChainSigName != (chain_list*)NULL)
  {
    /* There is less terminals in the figure
     * than signals in the instance.
     */
    for( ; aChainSigName !=   (chain_list*)NULL ;
           aChainSigName  = aChainSigName->NEXT )
    {
      SigName = GfpUT_DecodMBKSigName(
                  (char*)(aChainSigName->DATA),
                  &SigVect );
      SigName = UtBas_PrintF( "%s", SigName );

      if (SigVect == C_GfpUT_ImplicitConNoIndex)
      {
        /* Print previously stored vectorized terminal. */
        if (SigRoot != (char*)NULL)
        {
          if (SigOrig == SigExtr)
          {
            /* This is a one bit vector. */
            TmpName = UtBas_PrintF( "%s[%ld]", SigRoot, SigOrig );
          }
          else
          {
            /* This a real vector. */
            TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                                    SigRoot,
                                    SigOrig,
                                    SigExtr );
          }
        }
        FgnMBK_TracePrintF( "%20s => %s\n", "Unmatched", TmpName );

        if (SigVect != C_GfpUT_ImplicitConNoIndex)
        {
          SigName = UtBas_PrintF( "%s[%ld]", SigName, SigVect );
        }
        FgnMBK_TracePrintF( "%20s => %s\n", "Unmatched", SigName );

        SigIncr = 0; SigRoot = (char*)NULL;
      }
      else
      {
        if (SigRoot == (char*)NULL)
        {
          /* This is the first bit. */
          SigRoot = SigName; SigOrig = SigExtr = SigVect; SigIncr = 0;
        }
        else
        {
          if (SigIncr == 0)
          {
            SigIncr = SigVect - SigOrig;

            if (   ((SigIncr != -1) && (SigIncr != 1))
                || (strcmp( SigRoot, SigName )) )
            {
              /* Bits are not Sigtiguous. */
              TmpName = UtBas_PrintF( "%s[%ld]", SigRoot, SigOrig );

              FgnMBK_TracePrintF( "%20s => %s\n", "Unmatched", TmpName );

              SigRoot = SigName; SigOrig = SigExtr = SigVect; SigIncr = 0;
            }
            else
            {
              SigExtr = SigVect;
            }
          }
          else
          {
            /* Attempt to add a new bit to the current vector. */
            if (   (SigExtr + SigIncr != SigVect)
                || (strcmp( SigRoot, SigName )) )
            {
              /* Bits are not Sigtiguous. */

              TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                                      SigRoot,
                                      SigOrig,
                                      SigExtr );

              FgnMBK_TracePrintF( "%20s => %s\n", "Unmatched", TmpName );

              /* The new pair becomes the start bit. */
              SigRoot = SigName; SigOrig = SigExtr = SigVect; SigIncr = 0;
            }
            else
            {
              SigExtr = SigVect;
            }
          }
        }
      }
    }


    /* Flush the last vectorized terminal. */
    if (SigRoot != (char*)NULL)
    {
      if (SigOrig == SigExtr)
      {
        /* This is a one bit vector. */
        TmpName = UtBas_PrintF( "%s[%ld]", SigRoot, SigOrig );
      }
      else
      {
        /* This a real vector. */
        TmpName = UtBas_PrintF( "%s[%ld:%ld]",
                                SigRoot,
                                SigOrig,
                                SigExtr );
      }
  
      FgnMBK_TracePrintF( "%20s => %s[%ld]\n",
                          "Unmatched",
                          SigRoot,
                          SigOrig );
    }
  

    GfpUT_PrintError( fpgen_Glob.ModuleId,
                      E_FgnMBK_MatchInterfNoMoreCon,
                      fpgen_Glob.CurrFunc );
  }

  return( ChainSig );
}


/*  Fonction  :  'FgnMBK_xvLoIns'.                                          */
/* ------------------------------------------------------------------------ */
/*  This function is derived from the genlib LOINS.                         */

extern void  FgnMBK_xvLoIns( aModelName    ,
                             aInstcName    ,
                             aChainCTRLConn,
                             aStdArgs      )
        char *aModelName;     /* Name of the model to be instanciate.   */
        char *aInstcName;     /* Name of the new instance.              */
  chain_list *aChainCTRLConn; /* List of the implicit connectors names. */
     va_list  aStdArgs;       /* List of the explicit signal names.     */
{
  lofig_list *PtLoFig;
  chain_list *ChainConName = (chain_list*)NULL;
  chain_list *ChainSigName = (chain_list*)NULL;
  chain_list *ChainSig     = (chain_list*)NULL;

  chain_list *PtChain;
        char  StringBuffer[C_SUPP_MaxLenString];
        char *TmpName;
        long  TmpCount;


  /* Load the logical model. */
  PtLoFig = getlofig( aModelName, 'P' );


  /* Build list of model unvectorized, implicit connectors names. */
  for( PtChain  =  aChainCTRLConn    ;
       PtChain !=  (chain_list*)NULL ;
       PtChain  = PtChain->NEXT      )
  {
    TmpName = (char*)(PtChain->DATA);

    if (! is_bus( TmpName ))
    {
      /* Case of sample signals.
       * Check name consistancy.
       */
      TmpName = GfpUT_CheckName( TmpName );

      /* Add signal name in front of the list. */
      ChainConName = addchain( ChainConName, (void*)namealloc( TmpName ));
    }
    else

      /* Bus case. */
      ChainConName = FgnMBK_BusToChain( ChainConName, TmpName , &TmpCount );
  }

  /* Build list of unvectorized signal names.
   * (check for name syntax and unvectorize buses)
   */
  while(TRUE)
  {
    if ((TmpName = (char*)va_arg(aStdArgs, char*)) == (char*)NULL) break;

    TmpName = strtok(strcpy(StringBuffer, TmpName), ",");

    do
    {
      FgnMBK_TracePrintF( "%20s : ", "(char *)Signal" );
      FgnMBK_TracePrintF( "(0x%08lX) ", TmpName );
      FgnMBK_TracePrintF(  "\"%s\".\n", TmpName );

      if (! is_bus( TmpName ))
      {
        /* Case of sample signals.
         * Check name consistancy.
         */
        TmpName = GfpUT_CheckName(TmpName);

        /* Add signal name in front of the list. */
        ChainSigName = addchain( ChainSigName, (void*)namealloc( TmpName ));
      }
      else
        /* Bus case. */
        ChainSigName = FgnMBK_BusToChain( ChainSigName , TmpName , &TmpCount );
    }
    while( ( TmpName = strtok( (char *)NULL, "," ) ) != (char *)NULL );
  }

  FgnMBK_TracePrintF( "%20s : ", "(char *)Signal" );
  FgnMBK_TracePrintF( "(0x%08lX) ", (char *)NULL );
  FgnMBK_TracePrintF(  "--> EOL.\n" );


  FgnMBK_TracePrintF( "*** FpGen xvLoins ***:MBK terminals assigments " );
  FgnMBK_TracePrintF( "(reverse order):\n" );


  /* Add implicit signals and connectors and convert the list of
   * name in a list of MBK signal objects.
   */
  /*  ChainSig = FgnMBK_MatchImplicitConn( aInstcName,
   *                                       ChainConName,
   *                                       ChainSigName,
   *                                       PtLoFig,
   *                                       &TmpCount );
   */
  ChainSig = AddPowerSig( PtLoFig, ChainConName, ChainSigName );

  /* Free two "chain_list" of name (implicit connectors and signals). */
  freechain( ChainSigName );
  freechain( ChainConName );

  /* Reverse order of MBK signals list. */
  ChainSig = reverse( ChainSig );

  /* Instanciation of the model. */
  (void)addloins( WORK_LOFIG, aInstcName, PtLoFig, ChainSig );

  /* Free the MBK signals list. */
  freechain( ChainSig );
}


/*  Function  :  'SaveMBKEnv'                                               */
/* ======================================================================== */

static void  SaveMBKEnv()
{
  fpgen_Glob.OLD_LOFIG = WORK_LOFIG;
            WORK_LOFIG = (lofig_list*)NULL;

  putenv( fpgen_Glob.FPGEN_LIB );
  WORK_LIB = fpgen_Glob.FPGEN_LIB + 13;
}


/*  Function  :  'RestoreMBKEnv'                                            */
/* ======================================================================== */

static void  RestoreMBKEnv()
{
  WORK_LOFIG = fpgen_Glob.OLD_LOFIG;

  putenv( fpgen_Glob.OLD_WORKL );
  WORK_LIB = fpgen_Glob.OLD_WORKL + 13;
}


/*  Function  :  'FgnMBK_ErrorHeader'                                       */
/* ------------------------------------------------------------------------ */

extern void  FgnMBK_ErrorHeader( aFuncName )
  char *aFuncName;
{
  fflush( stdout );
  fflush( stderr );

  fprintf( stderr, "\n*** FpGen Error ***:%s:", aFuncName );
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "FgnMBK_ErrorPrintF()"                               | |%
 * %| \--------------------------------------------------------------------/ |%
 */

extern void  FgnMBK_ErrorPrintF __FPV(( char *aFormat, ... ))
{
            va_list  aVaList;
    __KR_C(    char *aFormat; )

  __ANSI_C( va_start( aVaList, aFormat ); )
    __KR_C( va_start( aVaList ); )
    __KR_C( aFormat = va_arg( aVaList, char* ); )
  vfprintf( stderr, aFormat, aVaList );
  va_end( aVaList );
}


/*  Function  :  'FgnMBK_ErrorTrap'                                         */
/* ------------------------------------------------------------------------ */

extern void  FgnMBK_ErrorTrap( aSig )
  int  aSig;
{
  switch( aSig )
  {
    case SIGSEGV:

      FgnMBK_ErrorHeader( "Program receive signal SIGSEGV" );
      FgnMBK_ErrorPrintF( "\n  Please check carefully the parameters of " );
      FgnMBK_ErrorPrintF( "the following function call.\n\n\n" );

      break;

    case SIGTERM:

      FgnMBK_ErrorHeader( "Program receive signal SIGTERM" );
      FgnMBK_ErrorPrintF( "\n\n\n" );
      break;

    default:

      FgnMBK_ErrorHeader( "Program receive signal" );
      FgnMBK_ErrorPrintF( " \'%ld\'.\n", (long)aSig );
      FgnMBK_ErrorPrintF( "  This signal is not trapped !?!.\n" );
      break;
  }

  fpgen_Glob.TraceMode = C_FgnMGN_TraceModeOn;
  FgnMBK_TraceFlush();
  FgnMBK_ErrorPrintF( "\n\n\n" );

  exit( 1 );
}


/*  Function  :  'FgnMBK_TraceHeader'                                       */
/* ------------------------------------------------------------------------ */

extern void  FgnMBK_TraceHeader()
{
  fpgen_Glob.TraceMark = UtBas_PrintFMark();

  if (fpgen_Glob.TraceMode == C_FgnMGN_TraceModeOn)
  {
    fpgen_Glob.TraceStr = UtBas_PrintF( "*** FpGen Trace ***:\n" );
  }
  else
  {
    fpgen_Glob.TraceStr = UtBas_PrintF( "*** FpGen Context ***:\n" );
  }
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "FgnMBK_TracePrintF()"                               | |%
 * %| \--------------------------------------------------------------------/ |%
 */

extern void  FgnMBK_TracePrintF __FPV(( char* aFormat, ... ))
{
          va_list  ArgList;
  __KR_C(    char *aFormat; )

  __ANSI_C( va_start( ArgList, aFormat ); )
    __KR_C( va_start( ArgList ); )
    __KR_C( aFormat = va_arg( ArgList, char * ); )
  fpgen_Glob.TraceStr = UtBas_vPrintFAdd(
                            fpgen_Glob.TraceStr, aFormat, ArgList );
  va_end( ArgList );
}


/*  Function  :  'FgnMBK_TraceFlush'                                        */
/* ------------------------------------------------------------------------ */

extern void  FgnMBK_TraceFlush()
{
  if (fpgen_Glob.TraceMode == C_FgnMGN_TraceModeOn)
  {
    fprintf( stderr, fpgen_Glob.TraceStr );
  }

  UtBas_PrintFFree( fpgen_Glob.TraceMark );
}
