/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

 
/*
 * %% \textrm{This File Looks Nicer with\textsf{ LGrind} and \LaTeXe.}       %%
 * %|                                                                        |%
 * %| $Id: FpGenMGN.c,v 1.3 1999/10/21 13:50:11 jpc Exp $       |%
 * %|                                                                        |%
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |          Alliance CAD System 3.2b                                   | |%
 * %| |     F p G e n   P a c k a g e   1.10                               | |%
 * %| |                                                                    | |%
 * %| |  Copyright(c) 95-97, MASI, CAO-VLSI Team                           | |%
 * %| |                                                                    | |%
 * %| |  Author    :            Jean-Paul CHAPUT                           | |%
 * %| |  E-mail    :        cao-vlsi@masi.ibp.fr                           | |%
 * %| | ================================================================== | |%
 * %| |  Module    :        "FpGenMGN.c"                                   | |%
 * %| | ****************************************************************** | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 */


# include  <stdio.h>
# include  <string.h>

# include  "GenericGN.h"
# include  "FpGenMBK.h"
# include  "FpGenMGN.h"
# include  <genlib.h>


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "__DP_OP()"                                          | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %|  Generic macro for all the operator function calls.                    |%
 */

#define  __DP_OP( fname, fid )                                   \
  extern void  fname __FPV(( char *InstName, ... ))              \
  {                                                              \
              va_list  ArgList;                                  \
      __KR_C(    char *InstName; )                               \
                                                                 \
    /* Get the list of variable arguments. */                    \
    __ANSI_C( va_start( ArgList, InstName ); )                   \
      __KR_C( va_start( ArgList ); )                             \
      __KR_C( InstName = va_arg( ArgList, char * ); )            \
    /* Call the generic boolean operators generator. */          \
    FgnMBK_GenericMF( fid, InstName, ArgList );                  \
    /* Clean the list of variable arguments. */                  \
    va_end(  ArgList);                                           \
  }


/*  Function  :  'FgnMGN_Trace'                                             */
/* ------------------------------------------------------------------------ */
/*  Activate or disable the trace mode.                                     */

extern void  FgnMGN_Trace( aFile, aLine, aMode )
  char *aFile;
  long  aLine;
  long  aMode;
{
  char *ModeName[2];


  ModeName[0] = "OFF";
  ModeName[1] = "ON";

  fflush( stdout );
  fflush( stderr );

  switch( aMode )
  {
    case C_FgnMGN_TraceModeOn:
    case C_FgnMGN_TraceModeOff:

      fprintf( stderr, "FpGen:DP_TRACE:" );
      fprintf( stderr, "Trace mode %s at line %ld of file \"%s\".\n",
                        ModeName[aMode], aLine, aFile );

      fpgen_Glob.TraceMode = aMode;
      break;

    default:

      fprintf( stderr, "*** FpGen Warning ***:DP_TRACE" );
      fprintf( stderr, "Invalid trace mode %ld at line %ld of file \"%s\".\n",
                       aMode, aLine, aFile );
      fprintf( stderr, "                      " );
      fprintf( stderr, "Trace mode desabled.\n" );

      fpgen_Glob.TraceMode = C_FgnMGN_TraceModeOff;
      break;
  }
}


/*  Function  :  'DP_DEFLOFIG'.                                             */
/* ======================================================================== */
/*  Open a new Data-Path logical figure.                                    */

extern void  DP_DEFLOFIG(aFigName, aMaxWidth, aMsb0)
  char *aFigName;  /* Name of the newly created logical figure. */
  long  aMaxWidth; /* Maximum width of the data-path buses.     */
  long  aMsb0;     /* Index zero for the MSB.                   */
{
  static char  FuncName[] = "DP_DEFLOFIG";

  /* Set name of current function. */
  fpgen_Glob.CurrFunc = FuncName;

  FgnMBK_TraceHeader();
  FgnMBK_TracePrintF( "%20s : DP_DEFLOFIG.\n", "Interface-Function" );


  /* Actualize the number of calls to "DP_DEFLOFIG". */
  fpgen_Glob.CallDefLo += 1;

  /* First call to "DP_DEFLOFIG". */
  if ( fpgen_Glob.CallDefLo == 1L )
  {
    /* Initialisation of FpGen and Ut modules. */
    UtBas_Init();
    UtEnv_Init();
    UtTec_Init( UtEnv_Environ.RDS_TECHNO_NAME );
    UtLib_Init( UtEnv_Environ.FPGEN_TARGET );
    UtFlt_Init();
    UtXlg_Init();
    UtMLo_Init();
    UtMPh_Init();
    UtHir_Init();
    UtUsr_Init();
    FgnMBK_Init();
  }

  FgnMBK_TracePrintF( "%20s : ", "(char *)aFigName" );
  FgnMBK_TracePrintF( "(0x%08lX) ", aFigName );
  FgnMBK_TracePrintF(  "\"%s\".\n", aFigName );

  FgnMBK_TracePrintF( "%20s : ",        "(long)aMaxWidth" );
  FgnMBK_TracePrintF( "(0x--------)  %ld.\n", aMaxWidth  );

  FgnMBK_TracePrintF( "%20s : ", "(long)aMSB0" );
  FgnMBK_TracePrintF( "(0x--------)  %ld.\n", aMsb0 );


  /* Check the number of call to the function. */
  if ( fpgen_Glob.CallDefLo != fpgen_Glob.CallSavLo + 1 )
    GfpUT_PrintError( fpgen_Glob.ModuleId,
                      E_FgnMBK_DEFLOCall,
                      fpgen_Glob.CurrFunc,
                      aFigName );
  
  /* Check of argument "MaxWidth". */
  if (   ( aMaxWidth < 1                        )
      || ( aMaxWidth > C_GfpGN_LimitsMaxWidthDP ) )
    GfpUT_PrintError( fpgen_Glob.ModuleId,
                      E_FgnMBK_DEFLOWidthMax,
                      fpgen_Glob.CurrFunc,
                      aFigName ,
                      aMaxWidth,
                      C_GfpGN_LimitsMaxWidthDP );

  /* Check the parity of "MaxWidth". */
  if ( aMaxWidth % 2 )
    GfpUT_PrintError( fpgen_Glob.ModuleId,
                      E_FgnMBK_DEFLOWidthEven,
                      fpgen_Glob.CurrFunc,
                      aFigName,
                      aMaxWidth );

  /* Set global variable "WidthDP". */
  fpgen_Glob.WidthDP = aMaxWidth;

  /* Set global variable "Msb0". */
  switch( aMsb0 )
  {
    case LSB_INDEX_ZERO:
    case MSB_INDEX_ZERO:

      fpgen_Glob.Msb0 = aMsb0;
      break;

    default:

      GfpUT_PrintError( fpgen_Glob.ModuleId,
                        E_FgnMBK_DEFLOLsbPos,
                        fpgen_Glob.CurrFunc,
                        aFigName,
                        aMsb0 );
  }

  /* Set mode of CTRL terminals management.
   * (default mode : EXPLICIT)
   */
  fpgen_Glob.ConMode = C_FgnMGN_ConModeEXPLICIT;

  /* Set the current "lofig".
   * (copy of the GENLIB's WORK_LOFIG for checks)
   */
  DEF_LOFIG( aFigName );

  /* Perform additionnal MBK operations of DP_DEFLOFIG.  */
  FgnMBK_DefLoFig();

  FgnMBK_TraceFlush();
}


/*  Function  :  'DP_SAVLOFIG'.                                             */
/* ======================================================================== */
/*  Close and save on disk the previously generated Data-Path.              */

extern void  DP_SAVLOFIG()
{
  static char  FuncName[] = "DP_SAVLOFIG";

  /* Set name of current function. */
  fpgen_Glob.CurrFunc = FuncName;


  FgnMBK_TraceHeader();
  FgnMBK_TracePrintF( "%20s : DP_SAVELOFIG.\n", "Interface-Function" );

  /* Actualize the number of calls to "DP_SAVLOFIG". */
  fpgen_Glob.CallSavLo += 1;

  /* If the function has been call before "DP_DEFLOFIG"
   *    - Load 'FPGEN' specific error table.
   *    - Process error : "No previous call to "DP_DEFLOFIG".
   */
  if ( fpgen_Glob.CallDefLo == 0 )
  {
    FgnMBK_Init();
    GfpUT_PrintError( fpgen_Glob.ModuleId,
                      E_FgnMBK_SharedMissDefLo,
                      fpgen_Glob.CurrFunc );
  }

  /* Check the number of call to the function. */
  if ( fpgen_Glob.CallDefLo != fpgen_Glob.CallSavLo )
    GfpUT_PrintError( fpgen_Glob.ModuleId,
                      E_FgnMBK_SAVLOCall,
                       fpgen_Glob.CurrFunc );

  /* Perform MBK additionnal operations of DP_SAVLOFIG. */
  FgnMBK_SavLoFig();

  DP_FLATTEN();
  SAVE_LOFIG();

  FgnMBK_TraceFlush();
}


/*  Fonction  :  'DP_CONMODE'.                                    */
/* -------------------------------------------------------------- */
/*  Reset the management mode of CTRL terminals after the first
 *  call to "DP_DEFLOFIG".
 */

extern void  DP_CONMODE(aConMode)
	long  aConMode;
{
	static char  FuncName[] = "DP_CONMODE";

	/* Set the current function name. */
	fpgen_Glob.CurrFunc = FuncName;

	/* If the function has been call before "DP_DEFLOFIG"       */
	/*    - Load 'fpgen' specific error table.                  */
	/*    - Process error : "No previous call to "DP_DEFLOFIG". */
	if (fpgen_Glob.CallDefLo == 0)
	{
		FgnMBK_Init();
		GfpUT_PrintError(fpgen_Glob.ModuleId,
					E_FgnMBK_SharedMissDefLo,
						 fpgen_Glob.CurrFunc);
	}

	switch( aConMode )
	{
		case C_FgnMGN_ConModeDEFAULT  :
	 	case C_FgnMGN_ConModeEXPLICIT :
		case C_FgnMGN_ConModeIMPLICIT :
			fpgen_Glob.ConMode = aConMode;
			break;
		default :
			GfpUT_PrintError(fpgen_Glob.ModuleId,
						  E_FgnMBK_CONMODEUnknow,
							 fpgen_Glob.CurrFunc, aConMode);
	}
}


/*  Function  :  'DP_LOCON'.                                      */
/* -------------------------------------------------------------- */
/*  Add a new logical connector to the current figure.            */

extern void  DP_LOCON( aConName, aConDir, aConSig )
	char *aConName;  /* Name of the connector (vectors allowed). */
	char  aConDir;   /* Direction of the connector.              */
	char *aConSig;   /* Signal associated to the connector.      */
{
	static char  FuncName[] = "DP_LOCON";


  FgnMBK_TraceHeader();
  FgnMBK_TracePrintF( "%20s : DP_LOCON.\n", "Interface-Function" );


	/* Set the current function name. */
	fpgen_Glob.CurrFunc = FuncName;

    /* If the function have been call before DP_DEFLO :      */
	/*     - Initialisation of error data base.              */
	/*     - Process error : "no previous call to DP_DEFLO". */
	if (fpgen_Glob.CallDefLo == 0)
	{
		FgnMBK_Init();
		GfpUT_PrintError(fpgen_Glob.ModuleId,
					E_FgnMBK_SharedMissDefLo,
						 fpgen_Glob.CurrFunc);
	}

    FgnMBK_TracePrintF( "%20s : ", "(char *)aCon" );
    FgnMBK_TracePrintF( "(0x%08lX) ", aConName );
    FgnMBK_TracePrintF(  "\"%s\".\n", aConName );

    FgnMBK_TracePrintF( "%20s : ", "(long)aDir" );
    FgnMBK_TracePrintF( "(0x--------) \'%c\'.\n", aConDir  );

    FgnMBK_TracePrintF( "%20s : ", "(char *)aSig" );
    FgnMBK_TracePrintF( "(0x%08lX) ", aConSig );
    FgnMBK_TracePrintF(  "\"%s\".\n", aConSig );

	/* Detect power supplies. */
	if (!FgnMBK_LoConPower(aConName))
	{
	    /* Encapsulated GENLIB LOCON function call. */
	    LOCON(aConName, aConDir, aConSig);
	}

  FgnMBK_TraceFlush();
}


/*  Fonction  :  'DP_IMPORT'.                                     */
/* -------------------------------------------------------------- */
/*  Instanciate a data-path or a column in the current data-path.
 *    It must have his data terminals build with virtual 
 *  connectors. Implicits terminals are automatically detect and
 *  added to the interface.
 */

extern void  DP_IMPORT __FPV(( char *ModelName, char *InstName, ... ))
{
	static       char  FuncName[] = "DP_IMPORT";

			  va_list  ArgList;
      __KR_C(    char *ModelName; )
      __KR_C(    char *InstName;  )

	/* Set name of current function. */
	fpgen_Glob.CurrFunc = FuncName;

    /* If the function have been call before DP_DEFLO :      */
	/*     - Initialisation of error data base.              */
	/*     - Process error : "no previous call to DP_DEFLO". */
	if (fpgen_Glob.CallDefLo == 0)
	{
		FgnMBK_Init();
		GfpUT_PrintError(fpgen_Glob.ModuleId,
					E_FgnMBK_SharedMissDefLo,
						 fpgen_Glob.CurrFunc);
	}

	/* Get the list of variable arguments. */
	__ANSI_C( va_start( ArgList, InstName ); )
	  __KR_C( va_start( ArgList ); )

    __KR_C( ModelName = va_arg( ArgList, char * ); )
    __KR_C(  InstName = va_arg( ArgList, char * ); )

	/* Call to the importation function. */
	FgnMBK_Import( ModelName, InstName, ArgList );

	/* Clean the list of variable arguments. */
	va_end( ArgList);
}


/*  Function  :  'DP_FLATTEN'.                                    */
/* -------------------------------------------------------------- */
/*  Flatten the current logical figure to a normalized 2 levels
 *  netlist. It is just an encapsulation of the Data-Path UTili-
 *  ties function 'UtHir_HierarchFlatten'.
 */

extern void  DP_FLATTEN()
{ 
	static char          FuncName[] = "DP_FLATTEN";

	/* Set the current function name. */
	fpgen_Glob.CurrFunc = FuncName;

    /* If the function have been call before DP_DEFLO :      */
	/*     - Initialisation of error data base.              */
	/*     - Process error : "no previous call to DP_DEFLO". */
	if (fpgen_Glob.CallDefLo == 0)
	{
		FgnMBK_Init();
		GfpUT_PrintError(fpgen_Glob.ModuleId,
					E_FgnMBK_SharedMissDefLo,
						 fpgen_Glob.CurrFunc);
	}

	/* Call to flatten function. */
	FgnMBK_Flatten();
}


/*
 * /-------------------------------------------------------------------------\
 * |  Standart Macro-Function Calls.                                         |
 * \-------------------------------------------------------------------------/
 */

__DP_OP( DP_INV       , C_FgnMBK_FunctionInv       )
__DP_OP( DP_BUFF      , C_FgnMBK_FunctionBuff      )
__DP_OP( DP_AND2      , C_FgnMBK_FunctionAnd2      )
__DP_OP( DP_AND3      , C_FgnMBK_FunctionAnd3      )
__DP_OP( DP_NAND2     , C_FgnMBK_FunctionNand2     )
__DP_OP( DP_NAND3     , C_FgnMBK_FunctionNand3     )
__DP_OP( DP_OR2       , C_FgnMBK_FunctionOr2       )
__DP_OP( DP_OR3       , C_FgnMBK_FunctionOr3       )
__DP_OP( DP_NOR2      , C_FgnMBK_FunctionNor2      )
__DP_OP( DP_NOR3      , C_FgnMBK_FunctionNor3      )
__DP_OP( DP_XOR2      , C_FgnMBK_FunctionXor2      )
__DP_OP( DP_XNOR2     , C_FgnMBK_FunctionXnor2     )
__DP_OP( DP_NAND2MASK , C_FgnMBK_FunctionNand2mask )
__DP_OP( DP_NOR2MASK  , C_FgnMBK_FunctionNor2mask  )
__DP_OP( DP_XNOR2MASK , C_FgnMBK_FunctionXnor2mask )
__DP_OP( DP_NMUX2CS   , C_FgnMBK_FunctionNmux2cs   )
__DP_OP( DP_MUX2CS    , C_FgnMBK_FunctionMux2cs    )
__DP_OP( DP_MUX3CD    , C_FgnMBK_FunctionMux3cd    )
__DP_OP( DP_MUX3CS    , C_FgnMBK_FunctionMux3cs    )
__DP_OP( DP_MUX4CS    , C_FgnMBK_FunctionMux4cs    )
__DP_OP( DP_MUX5CS    , C_FgnMBK_FunctionMux5cs    )
__DP_OP( DP_MUX6CS    , C_FgnMBK_FunctionMux6cs    )
__DP_OP( DP_BUSE      , C_FgnMBK_FunctionBuse      )
__DP_OP( DP_NBUSE     , C_FgnMBK_FunctionNbuse     )
__DP_OP( DP_ADD2S     , C_FgnMBK_FunctionAdd2s     )
__DP_OP( DP_ADD2F     , C_FgnMBK_FunctionAdd2f     )
__DP_OP( DP_ADSB2F    , C_FgnMBK_FunctionAdsb2f    )
__DP_OP( DP_SHIFT     , C_FgnMBK_FunctionShift     )
__DP_OP( DP_CONST     , C_FgnMBK_FunctionConst     )
__DP_OP( DP_NUL       , C_FgnMBK_FunctionNul       )
__DP_OP( DP_PLATCH1   , C_FgnMBK_FunctionPl1       )
__DP_OP( DP_NPLATCH1  , C_FgnMBK_FunctionNpl1      )
__DP_OP( DP_PMSNW     , C_FgnMBK_FunctionPmsnw     )
__DP_OP( DP_PMSNWQ    , C_FgnMBK_FunctionPmsnwq    )
__DP_OP( DP_PMS       , C_FgnMBK_FunctionPms       )
__DP_OP( DP_PMSQ      , C_FgnMBK_FunctionPmsq      )
__DP_OP( DP_PMST      , C_FgnMBK_FunctionPmst      )
__DP_OP( DP_PMSTQ     , C_FgnMBK_FunctionPmstq     )
__DP_OP( DP_PDFFNW    , C_FgnMBK_FunctionPdffnw    )
__DP_OP( DP_PDFFNWQ   , C_FgnMBK_FunctionPdffnwq   )
__DP_OP( DP_PDFF      , C_FgnMBK_FunctionPdff      )
__DP_OP( DP_PDFFR     , C_FgnMBK_FunctionPdffr     )
__DP_OP( DP_PDFFT     , C_FgnMBK_FunctionPdfft     )
__DP_OP( DP_PDFFUT    , C_FgnMBK_FunctionPdffut    )
__DP_OP( DP_PDFFRT    , C_FgnMBK_FunctionPdffrt    )
__DP_OP( DP_PDFFTQ    , C_FgnMBK_FunctionPdfftq    )
__DP_OP( DP_RFG1C0    , C_FgnMBK_FunctionRfg1c0    )
__DP_OP( DP_RFG1CC    , C_FgnMBK_FunctionRfg1cc    )
__DP_OP( DP_RFG2C0    , C_FgnMBK_FunctionRfg2c0    )
__DP_OP( DP_RFG2CC    , C_FgnMBK_FunctionRfg2cc    )
__DP_OP( DP_RFGU1C0   , C_FgnMBK_FunctionRfgu1c0   )
__DP_OP( DP_RFGU1CC   , C_FgnMBK_FunctionRfgu1cc   )
__DP_OP( DP_RFGU2C0   , C_FgnMBK_FunctionRfgu2c0   )
__DP_OP( DP_RFGU2CC   , C_FgnMBK_FunctionRfgu2cc   )
__DP_OP( DP_MULT      , C_FgnMBK_FunctionMult      )
__DP_OP( DP_MULTP     , C_FgnMBK_FunctionMultp     )
