/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/*
 * %% \textrm{This File Looks Nicer with\textsf{ LGrind} and \LaTeXe.}       %%
 * %|                                                                        |%
 * %| $Id: SharedGN.h,v 1.1.1.1 1998/10/02 15:26:38 alliance Exp $       |%
 * %|                                                                        |%
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |          Alliance CAD System 3.2b                                   | |%
 * %| |     F p G e n   P a c k a g e   1.10                               | |%
 * %| |                                                                    | |%
 * %| |  Copyright(c) 95-97, MASI, CAO-VLSI Team                           | |%
 * %| |                                                                    | |%
 * %| |  Author    :            Jean-Paul CHAPUT                           | |%
 * %| |  E-mail    :        cao-vlsi@masi.ibp.fr                           | |%
 * %| | ================================================================== | |%
 * %| |  Module    :        "SharedGN.h"                                   | |%
 * %| | ****************************************************************** | |%
 * %| |  Declarations of all the stuff shared by `SharedGN.c', `FpLibGN.c' | |%
 * %| |  and `ScLibGN.c'. Theses files are all parts of the previous       | |%
 * %| |  `GenericGN.c' of version 1.x of FpGen.                            | |%
 * %| |    The declarations of constants and functions needed for using    | |%
 * %| |  the library are located in `GenericGN.h'.                         | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 */


# ifndef  __SharedGN__
#   define  __SharedGN__

/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |                     Constants Definitions                      | |%
 *   %| \----------------------------------------------------------------/ |%
 */



/* Constants for error messages. */
#   define       IE_GfpGN_WIDTH                2
#   define       IE_GfpGN_SLICE                3
#   define       IE_GfpGN_DRIVE                4
#   define       IE_GfpGN_NOVIEW               5
#   define       IE_GfpGN_OPENVBE              6
#   define       IE_GfpGN_FUNCTION             7
#   define       IE_GfpGN_WidthDP              8
#   define       IE_GfpGN_WidthOp              9
#   define       IE_GfpGN_SliceEven           10
/* Constants for warning messages. */
#   define       IW_GfpGN_PowerMerge           2
/* Flags for coding which view to generate. */
#   define       IF_GfpGN_ViewVHDL             1
#   define       IF_GfpGN_ViewNetlist          2
#   define       IF_GfpGN_ViewLayout           4
#   define       IF_GfpGN_ViewIcon             8
/* Flags for the functions specific arguments. */
#   define       IF_GfpGN_WidthOp              1
#   define       IF_GfpGN_Slice                2
#   define       IF_GfpGN_Drive                4
#   define       IF_GfpGN_Mask                 8
/* Flags for the various memory components (VHDL view). */
#   define       IF_GfpGN_MemMaster            1
#   define       IF_GfpGN_MemSlave             2
#   define       IF_GfpGN_MemScan              4
#   define       IF_GfpGN_MemReset             8
/* Some constants. */
#   define       IC_GfpGN_ModelNameMaxLen    256
#   define       IC_GfpGN_BusWideMaxLen      256
#   define       IC_GfpGN_PinNameMaxLen      256
#   define       IC_GfpGN_NameMaxLen         256
#   define       IC_GfpGN_PowerStringLen      16



/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |                      Types and Structures                      | |%
 *   %| \----------------------------------------------------------------/ |%
 */

typedef struct {
  long  ModuleId;      /* Module identifier.                    */
  long  PrintFMark;    /* Tag for self-allocated strings.       */
                       /* ------------------------------------- */
  long  aFunction;     /* Operator to generate.                 */
  char *aModelName;    /* Operator model name.                  */
  long  aWidthDP;      /* Data-path maximal width.              */
  long  aWidthOp;      /* Operator width.                       */
  long  aSlice;        /* LSB starting slice.                   */
  long  aDrive;        /* Output power (fan-out).               */
  long *aXlMask;       /* Logical Mask (constant).              */
  long  aMSB0;         /* MSB as zero index.                    */
  long  aPower;        /* Merge the power supplies.             */
  long  aViews;        /* The views to generate.                */
                       /* ------------------------------------- */
  FILE *PtFile;        /* VHDL file descriptor.                 */
  long  LeafCellIndex; /* Index of the leaf cells list inside   */
                       /* the "TablLeafCell[]".                 */
                       /* ------------------------------------- */
  void (*pFunctionArgs     ) __FP(( va_list ));
  long (*pFunctionCells    ) __FP(( void ));
  long (*pHaveBuff         ) __FP(( void ));
  void (*pCreateNetlist    ) __FP(( void ));
  void (*pCreateNetlistComp) __FP(( void ));
  void (*pCreateLayout     ) __FP(( void ));
  void (*pCreateLayoutComp ) __FP(( void ));
} GV_t;

extern GV_t GV;


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |                   Functions Declarations                       | |%
 *   %| \----------------------------------------------------------------/ |%
 */

/* Functions from module `SharedGN.c'. */
extern long  ShGN_llog2             __FP(( long value ));
extern long  ShGN_FunctionDrive     __FP(( void ));
extern char *ShGN_FunctionStr       __FP(( void ));
/* Functions from module `FpLibGN.c'. */
extern void  FpGN_FunctionArgs      __FP(( va_list ArgList ));
extern long  FpGN_FunctionCells     __FP(( void ));
extern long  FpGN_HaveBuff          __FP(( void ));
extern void  FpGN_CreateNetlist     __FP(( void ));
extern void  FpGN_CreateNetlistComp __FP(( void ));
extern void  FpGN_CreateLayout      __FP(( void ));
extern void  FpGN_CreateLayoutComp  __FP(( void ));
/* Functions from module `FpLibGN.c'. */
extern void  ScGN_FunctionArgs      __FP(( va_list ArgList ));
extern long  ScGN_FunctionCells     __FP(( void ));
extern long  ScGN_HaveBuff          __FP(( void ));
extern void  ScGN_CreateNetlist     __FP(( void ));
extern void  ScGN_CreateNetlistComp __FP(( void ));
extern void  ScGN_CreateLayout      __FP(( void ));
extern void  ScGN_CreateLayoutComp  __FP(( void ));

# endif
