/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/*
 * %% \textrm{This File Looks Nicer with\textsf{ LGrind} and \LaTeXe.}       %%
 * %|                                                                        |%
 * %| $Id: UtLib.c,v 1.1.1.1 1998/10/02 15:26:39 alliance Exp $          |%
 * %|                                                                        |%
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |          Alliance CAD System 3.2b                                   | |%
 * %| |     F p G e n   P a c k a g e   1.10                               | |%
 * %| |                                                                    | |%
 * %| |  Copyright(c) 95-97, MASI, CAO-VLSI Team                           | |%
 * %| |                                                                    | |%
 * %| |  Author    :            Jean-Paul CHAPUT                           | |%
 * %| |  E-mail    :        cao-vlsi@masi.ibp.fr                           | |%
 * %| | ================================================================== | |%
 * %| |  Module    :        "UtLib.c"                                      | |%
 * %| | ****************************************************************** | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 */


# include  <stdlib.h>

# include  "GenericGN.h"


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |             External Structures and Variables                  | |%
 *   %| \----------------------------------------------------------------/ |%
 *   %|                                                                    |%
 *   %| All those variables must be declared `extern' in "GenericGN.h".    |%
 */

       UtLib_V_t  UtLib_V;


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |             Internal Structures and Variables                  | |%
 *   %| \----------------------------------------------------------------/ |%
 */

#   define        IW_UtLib_TargetUnknown      2

static long  LV_mId;  /* "UtLib.c" module identifier. */


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                     Functions Definitions                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "UtLib_Init()"                                       | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| This function has to be called after `UtTec_Init' that read the envi-  |%
 * %| ronment variables needed to perform the inititialisations.             |%
 */

extern void  UtLib_Init( asLib )
  char *asLib;
{
  static long  FirstCall = TRUE;

  if (FirstCall == FALSE) return;
    FirstCall = FALSE;
  /* Attribution of a module identifier. */
  LV_mId = GfpUT_GetModuleId( "UtLib", F_GfpUT_ModuleAdd );
  /* Load the errors/warning DB. */
  GfpUT_AddWarng(
    LV_mId, IW_UtLib_TargetUnknown, "[%s]:Unknow target library \"%s\".");
  /* Set the default target cell library : FpLib. */
  UtLib_V.Type = C_UtLib_TargetFpLib;
  if (asLib != (char*)NULL) {
    /* Attempt to recognize a valid library name. */
    UtLib_V.Type = C_UtLib_TargetUnknown;
    if (!strncasecmp( asLib, "fplib", 5 )) UtLib_V.Type = C_UtLib_TargetFpLib;
    if (!strncasecmp( asLib, "sclib", 5 )) UtLib_V.Type = C_UtLib_TargetScLib;
    if (UtLib_V.Type == C_UtLib_TargetUnknown) {
      GfpUT_PrintWarng( LV_mId, IW_UtLib_TargetUnknown, "Init", asLib );
      UtLib_V.Type = C_UtLib_TargetFpLib;
    }
  }
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "UtLib_Target()"                                     | |%
 * %| \--------------------------------------------------------------------/ |%
 */

extern long  UtLib_Target()
{ return( UtLib_V.Type ); }
