/*
 * $Author: ac $
 * $Id: bflname.c,v 1.1 1999/10/21 12:51:52 ac Exp $
 * $Date: 1999/10/21 12:51:52 $
 */

/* Name allocation:
 *    On utilise un bflmap_t pour stoker les names. dans le champ key,
 *  on met le nom proprement dit, dans le champ data, on met le compte
 *  de references.
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include  BFL_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/* namealloc hash table size: the 1230th prime */
#define HASHVAL     10007

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

typedef struct chain {
    struct chain *NEXT;
    void *DATA;
} chain_list;

typedef struct name {
	struct name	*next;
	char		 data[1];
} name_t;

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

static chain_list *NAME_HASHTABLE[HASHVAL];
static char buffer[BUFSIZ];	/* buffer for namealloc strcpy */

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

#define bfltoupper(c)	(((c) >= 'a' && (c) <= 'z') * ('A' - 'a') + (c))
#define bfltolower(c)	(((c) >= 'A' && (c) <= 'Z') * ('a' - 'A') + (c))

/* Random hash function due to Don. E. Knuth, The Stanford Graph Base.
 *  Truly better than the previous one from my own experimentations.
 */
#define HASH_MULT 314159
#define HASH_PRIME 516595003

#define HASH_FUNC(inputname, name, code)                                     \
   do {                                                                      \
      while (*inputname) {                                                   \
         *name = bfltolower(*inputname);  	                             \
         code += (code ^ (code >> 1)) + HASH_MULT * (unsigned char) *name++; \
         while (code >= HASH_PRIME)                                          \
            code -= HASH_PRIME;                                              \
            inputname++;                                                     \
      }                                                                      \
      *name = '\0';                                                          \
      code %= HASHVAL;                                                       \
   } while (0)

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

static chain_list *
addchain(chain_list * next, void *data)
{
    chain_list *chain;

    chain = (chain_list *) bflmemalloc(sizeof(chain_list));
    chain->NEXT = next;
    chain->DATA = data;
    return chain;
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

char *
bflnamealloc(char *inputname)
{
    chain_list *pt;
    char *name = buffer;	/* ensure no modification of parameter string */
    int code = 0;

    if (inputname == NULL)
	return NULL;

    /* Beware, that's a define, ... */
    HASH_FUNC(inputname, name, code);

    for (pt = NAME_HASHTABLE[code]; pt; pt = pt->NEXT)
	if (!strcmp(buffer, (char *) pt->DATA))
	    return (char *) pt->DATA;

    name = (char *) bflalloc(strlen(buffer) + 1);
    (void) strcpy(name, buffer);
    NAME_HASHTABLE[code] = addchain(NAME_HASHTABLE[code], (void *) name);
    return (char *) (NAME_HASHTABLE[code]->DATA);

}

void
bflnamefree(char *name)
{
}

char *
bflnamefind(char *inputname)
{
    chain_list *pt;
    char *name = buffer;	/* ensure no modification of parameter string */
    register int code = 0;

    if (inputname == NULL)
	return NULL;

    HASH_FUNC(inputname, name, code);

    for (pt = NAME_HASHTABLE[code]; pt; pt = pt->NEXT)
	if (!strcmp(buffer, (char *) pt->DATA))
	    return (char *) pt->DATA;

    return NULL;
}
