/*
 * $Author: ac $
 * $Id: bfltree.c,v 1.2 1999/10/21 12:51:53 ac Exp $
 * $Date: 1999/10/21 12:51:53 $
 */
/*
 * $Log: bfltree.c,v $
 * Revision 1.2  1999/10/21 12:51:53  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:38  alliance
 * fpmap import
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include BFL_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

#define NIL		&__red__black__sentinel

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

static bfltreenode_t __red__black__sentinel =
{NIL, NIL, 0, bflblack, 0};

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

static void
bfltreerotateleft __P((bfltree_t * tree, bfltreenode_t * x));

static void
bfltreerotateright __P((bfltree_t * tree, bfltreenode_t * x));

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

static void
bfltreerotateleft(bfltree_t * tree, bfltreenode_t * x)
{
    bfltreenode_t *parent = x->right;

/* establish x->right link */
    x->right = parent->left;
    if (parent->left != NIL)
	parent->left->parent = x;

/* establish parent->parent link */
    parent->parent = x->parent;
    if (x == tree->header->parent)
	tree->header->parent = parent;
    else if (x == x->parent->left)
	x->parent->left = parent;
    else
	x->parent->right = parent;

/* link x and parent */
    parent->left = x;
    x->parent = parent;
}

static void
bfltreerotateright(bfltree_t * tree, bfltreenode_t * x)
{
    bfltreenode_t *parent = x->left;

/* establish x->left link */
    x->left = parent->right;
    if (parent->right != NIL)
	parent->right->parent = x;

/* establish parent->parent link */
    parent->parent = x->parent;
    if (x == tree->header->parent)
	tree->header->parent = parent;
    else if (x == x->parent->right)
	x->parent->right = parent;
    else
	x->parent->left = parent;

/* link x and parent */
    parent->right = x;
    x->parent = parent;
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

bfltree_t *
bfltreenew(bflcompar_fn compar, int dup)
{
    bfltree_t *tree;

    tree = (bfltree_t *) bflmemalloc(sizeof(bfltree_t));
    tree->duplicate = dup;
    tree->compar = compar;
    tree->size = 0;
    tree->header = (bfltreenode_t *) bflmemalloc(sizeof(bfltreenode_t));
    tree->header->color = bflred;
    tree->header->parent = NIL;
    tree->header->left = tree->header;
    tree->header->right = tree->header;
    return tree;
}

void
bfltreefree(bfltree_t * tree)
{
    bfltreeclean(tree);
    bflmemfree(tree->header, sizeof(bfltreenode_t));
    bflmemfree(tree, sizeof(bfltree_t));
}

bfltree_t *
bfltreecopy(bfltree_t * tree)
{
    bfltree_t *copy = bfltreenew(tree->compar, tree->duplicate);
    bfltreenode_t *first = bfltreebegin(tree);
    bfltreenode_t *last = bfltreeend(tree);

    while (first != last) {
	bfltreeinsert(copy, bfltreedata(first));
	first = bfltreenext(first);
    }
    return copy;
}

void
bfltreeclean(bfltree_t * tree)
{
    bfltreenode_t *first = bfltreebegin(tree);
    bfltreenode_t *last = bfltreeend(tree);

    while (first != last) {
	bfltreenode_t *next = bfltreenext(first);

	bfltreeerase(tree, first);
	first = next;
    }
}

bfltreenode_t *
bfltreeinsert(bfltree_t * tree, bflptr_t data)
{
    bfltreenode_t *current;
    bfltreenode_t *parent;
    bfltreenode_t *x;

/* find where node belongs */
    parent = tree->header;
    current = bfltreeroot(tree);
    while (current != NIL) {
	int comp = tree->compar(bfltreedata(current), data);

	if (!(tree->duplicate) && comp == 0) {
	    current->data = data;
	    return current;
	}
	parent = current;
	current = (comp < 0) ? current->left : current->right;
    }

    tree->size++;
    x = (bfltreenode_t *) bflmemalloc(sizeof(bfltreenode_t));
    x->data = data;
    x->left = NIL;
    x->right = NIL;
    x->parent = 0;
    x->color = bflred;

    if (parent == tree->header || tree->compar(parent->data, data) < 0) {
	parent->left = x;
	if (parent == tree->header) {
	    tree->header->parent = x;
	    tree->header->right = x;
	} else if (parent == tree->header->left)	/* leftmost */
	    tree->header->left = x;
    } else {
	parent->right = x;
	if (parent == tree->header->right)	/* rightmost */
	    tree->header->right = x;
    }
    x->parent = parent;
    current = x;
/* check bflred-bflblack properties */
    while (current != bfltreeroot(tree) && current->parent->color == bflred) {
    /* we have a violation */
	if (current->parent == current->parent->parent->left) {
	    parent = current->parent->parent->right;
	    if (parent->color == bflred) {
	    /* uncle is bflred */
		current->parent->color = bflblack;
		parent->color = bflblack;
		current->parent->parent->color = bflred;
		current = current->parent->parent;
	    } else {
	    /* uncle is bflblack */
		if (current == current->parent->right) {
		/* make current a left child */
		    current = current->parent;
		    bfltreerotateleft(tree, current);
		}
	    /* recolor and rotate */
		current->parent->color = bflblack;
		current->parent->parent->color = bflred;
		bfltreerotateright(tree, current->parent->parent);
	    }
	} else {
	/* mirror image of above code */
	    parent = current->parent->parent->left;
	    if (parent->color == bflred) {
	    /* uncle is bflred */
		current->parent->color = bflblack;
		parent->color = bflblack;
		current->parent->parent->color = bflred;
		current = current->parent->parent;
	    } else {
	    /* uncle is bflblack */
		if (current == current->parent->left) {
		    current = current->parent;
		    bfltreerotateright(tree, current);
		}
		current->parent->color = bflblack;
		current->parent->parent->color = bflred;
		bfltreerotateleft(tree, current->parent->parent);
	    }
	}
    }
    tree->header->parent->color = bflblack;	/* root->color */
    return x;
}

void
bfltreeerase(bfltree_t * tree, bfltreenode_t * node)
{
    bfltreenode_t *y = node;
    bfltreenode_t *x;

    if (y->left == NIL)
	x = y->right;
    else if (y->right == NIL)
	x = y->left;
    else {
	y = y->right;
	while (y->left != NIL)
	    y = y->left;
	x = y->right;
    }
    if (y != node) {		/* relink y in place of node */
	node->left->parent = y;
	y->left = node->left;
	if (y != node->right) {
	    x->parent = y->parent;	/* possibly x == NIL */
	    y->parent->left = x;	/* y must be a left child */
	    y->right = node->right;
	    node->right->parent = y;
	} else
	    x->parent = y;	/* needed in case x == NIL */
	if (tree->header->parent == node)
	    tree->header->parent = y;
	else if (node->parent->left == node)
	    node->parent->left = y;
	else
	    node->parent->right = y;
	y->parent = node->parent;
	bflswap(y->color, node->color);
	bflswap(y, node);
    /* 
     *y points to node to be actually deleted,
     *node points to old node's former successor
     */
    } else {			/* y == node */
	x->parent = y->parent;	/* possibly x == NIL */
	if (tree->header->parent == node)
	    tree->header->parent = x;
	else if (node->parent->left == node)
	    node->parent->left = x;
	else
	    node->parent->right = x;
	if (tree->header->left == node) {	/* leftmost */
	    if (node->right == NIL)
	    /* node->left must be NIL also */
		tree->header->left = node->parent;
    /* makes root == header if node == root */
	    else
		tree->header->left = bfltreemin(x);
	}
	if (tree->header->right == node) {	/* rightmost */
	    if (node->left == NIL)
	    /* node->right must be NIL also */
		tree->header->right = node->parent;
    /* makes root == header if node == root */
	    else		/* x == node->left */
		tree->header->right = bfltreemax(x);
	}
    }
    if (y->color != bflred) {
	while (x != tree->header->parent && x->color == bflblack)
	    if (x == x->parent->left) {
		bfltreenode_t *w = x->parent->right;

		if (w->color == bflred) {
		    w->color = bflblack;
		    x->parent->color = bflred;
		    bfltreerotateleft(tree, x->parent);
		    w = x->parent->right;
		}
		if (w->left->color == bflblack && w->right->color == bflblack) {
		    w->color = bflred;
		    x = x->parent;
		} else {
		    if (w->right->color == bflblack) {
			w->left->color = bflblack;
			w->color = bflred;
			bfltreerotateright(tree, w);
			w = x->parent->right;
		    }
		    w->color = x->parent->color;
		    x->parent->color = bflblack;
		    w->right->color = bflblack;
		    bfltreerotateleft(tree, x->parent);
		    break;
		}
	    } else {
	    /* 
	     * same as then clause with "right" and
	     * "left" exchanged
	     */
		bfltreenode_t *w = x->parent->left;

		if (w->color == bflred) {
		    w->color = bflblack;
		    x->parent->color = bflred;
		    bfltreerotateright(tree, x->parent);
		    w = x->parent->left;
		}
		if (w->right->color == bflblack && w->left->color == bflblack) {
		    w->color = bflred;
		    x = x->parent;
		} else {
		    if (w->left->color == bflblack) {
			w->right->color = bflblack;
			w->color = bflred;
			bfltreerotateright(tree, w);
			w = x->parent->left;
		    }
		    w->color = x->parent->color;
		    x->parent->color = bflblack;
		    w->left->color = bflblack;
		    bfltreerotateright(tree, x->parent);
		    break;
		}
	    }
	x->color = bflblack;
    }
    bflmemfree(y, sizeof(bfltreenode_t));
    tree->size--;
}

bfltreenode_t *
bfltreefind(bfltree_t * tree, bflptr_t data)
{
    bfltreenode_t *current;
    int comp;

    current = tree->header->parent;
    while (current != NIL) {
	comp = tree->compar(current->data, data);
	if (comp == 0)
	    return current;
	else
	    current = (comp < 0) ? current->left : current->right;
    }
    return (bfltreenode_t *) 0;
}

bfltreenode_t *
bfltreemin(bfltreenode_t * x)
{
    while (bfltreeleft(x) != NIL)
	x = bfltreeleft(x);
    return x;
}

bfltreenode_t *
bfltreemax(bfltreenode_t * x)
{
    while (bfltreeright(x) != NIL)
	x = bfltreeright(x);
    return x;
}

bfltreenode_t *
bfltreenext(bfltreenode_t * node)
{
    if (node->right != NIL) {
	node = node->right;
	while (node->left != NIL)
	    node = node->left;
    } else {
	bfltreenode_t *parent = node->parent;

	while (node == parent->right) {
	    node = parent;
	    parent = parent->parent;
	}
	if (node->right != parent)	/* necessary because of rightmost */
	    node = parent;
    }
    return node;
}

bfltreenode_t *
bfltreeprev(bfltreenode_t * node)
{
    if (node->color == bflred && node->parent->parent == node)
    /* check for header */
	node = node->right;	/* return rightmost */
    else if (node->left != NIL) {
	bfltreenode_t *parent = node->left;

	while (parent->right != NIL)
	    parent = parent->right;
	node = parent;
    } else {
	bfltreenode_t *parent = node->parent;

	while (node == parent->left) {
	    node = parent;
	    parent = parent->parent;
	}
	node = parent;
    }
    return node;
}
