/*
 * $Author: ac $
 * $Id: testqueue.c,v 1.1 1999/10/21 12:51:55 ac Exp $
 * $Date: 1999/10/21 12:51:55 $
 */

/* This program builds a Huffman encoding tree using a priority queue.
 */
#include <stdio.h>
#include <ctype.h>
#include  BFL_H

typedef struct node {
	int		 weight;
	unsigned char	 value;
	struct node	*child0;
	struct node	*child1;
} node_t;

static node_t *
node_new_leaf(unsigned char c, int w)
{
	node_t *node;

	node = (node_t *)bflmemalloc(sizeof(node_t));
	node->value = c;
	node->weight = w;
	node->child0 = NULL;
	node->child1 = NULL;
	return node;
}

static node_t *
node_new_internal(node_t *node0, node_t *node1)
{
	node_t *node;

	node = (node_t *)bflmemalloc(sizeof(node_t));
	node->value = '?';
	node->weight = node0->weight + node1->weight;
	node->child0 = node0;
	node->child1 = node1;
	return node;
}

static int
node_comp(const bflptr_t x, const bflptr_t y)
{
	return ((node_t *)y)->weight - ((node_t *)x)->weight;
}

static void
node_traverse(node_t *node, char *code)
{
	if (node->child0 != NULL) {
		char *code0 = bflstrcat(code, "0");
		char *code1 = bflstrcat(code, "1");
		node_traverse(node->child0, code0);
		node_traverse(node->child1, code1);
		free(code0);
		free(code1);
	} else {
		printf(" %c     (%3d)   %s\n", node->value, node->weight, code);
	}
}

int
main(int argc, char **argv)
{
	int count[256];
	int i, c;
	bflpqueue_t *pq;
	node_t *node0, *node1;

	for (i = 0; i < 256; ++i)
		count[i] = 0;

	while ((c = getchar()) != EOF)
		if (isprint(c)) ++count[c];

	pq = bflpqueuenew(node_comp);
	
	/* push the leaf nodes into the queue */
	for (i = 0; i < 256; ++i)
		if (count[i] != 0)
			bflpqueuepush(pq, node_new_leaf(i, count[i]));
	
	/* This removes the two smallest nodes from pq, and push a new
	 * internal node
	 */

	while (bflpqueuesize(pq) > 1) {
		node0 = (node_t *)bflpqueuepop(pq);
#if 1
{
	bflptr_t *first = bflvecbegin(pq);
	bflptr_t *last  = bflvecend(pq);
	while (first != last) {
		printf("%d ", ((node_t *)*first)->weight);
		first = bflvecnext(first);
	}
	printf("\n");
}
#endif
		node1 = (node_t *)bflpqueuepop(pq);
#if 1
{
	bflptr_t *first = bflvecbegin(pq);
	bflptr_t *last  = bflvecend(pq);
	while (first != last) {
		printf("%d ", ((node_t *)*first)->weight);
		first = bflvecnext(first);
	}
	printf("\n");
}
#endif
		bflpqueuepush(pq, node_new_internal(node0, node1));
#if 1
{
	bflptr_t *first = bflvecbegin(pq);
	bflptr_t *last  = bflvecend(pq);
	while (first != last) {
		printf("%d ", ((node_t *)*first)->weight);
		first = bflvecnext(first);
	}
	printf("\n");
}
#endif
	}

	/* Dump the result */
	printf(" Char  Weight  Code\n");
	node_traverse((node_t *)bflpqueuetop(pq), "");
	
	return 0;
}
