/*
 * $Author: ac $
 * $Id: testvec.c,v 1.1 1999/10/21 12:51:56 ac Exp $
 * $Date: 1999/10/21 12:51:56 $
 */

#include <assert.h>
#include <stdio.h>
#include BFL_H

void
printvec(vec)
    bflvec_t *vec;
{
    bflptr_t *first = bflvecbegin(vec);
    bflptr_t *last = bflvecend(vec);

    printf("#vec( %p ) : [ ", vec);
    while (first != last) {
	printf("%d ", (int) bflvecdata(first));
	first = bflvecnext(first);
    }
    puts("]");
}

void
test_sort()
{
    bflvec_t *vec = bflvecnew();

    printf("\ntest_sort\n");

    bflvecpushback(vec, (bflptr_t) 9);
    bflvecpushback(vec, (bflptr_t) 0);
    bflvecpushback(vec, (bflptr_t) 8);
    bflvecpushback(vec, (bflptr_t) 1);
    bflvecpushback(vec, (bflptr_t) 7);
    bflvecpushback(vec, (bflptr_t) 2);
    bflvecpushback(vec, (bflptr_t) 6);
    bflvecpushback(vec, (bflptr_t) 3);
    bflvecpushback(vec, (bflptr_t) 5);
    bflvecpushback(vec, (bflptr_t) 4);

    printvec(vec);
    bflvecsort(vec, bflnumcmp);
    printvec(vec);
    bflvecfree(vec);
}

int
main()
{
    int i;
    bflptr_t *first;
    bflptr_t *last;
    bflvec_t *vec = bflvecnew();

    assert(bflvecempty(vec) == 1);
    assert(bflvecsize(vec) == 0);

    /*
    for (i = 0; i < 10; i++)
	bflvecpushback(vec, (void *) i);
     */

    for (i = 9; i >= 0; i--)
	bflvecpushfront(vec, (void *) i);
    
    printvec(vec);
    assert(bflvecempty(vec) == 0);
    assert(bflvecsize(vec) == 10);

    i = 0;
    first = bflvecbegin(vec);
    last = bflvecend(vec);
    while (first != last) {
	assert(i == bflvecindex(vec, bflvecdata(first)));
	first = bflvecnext(first);
	i += 1;
    }

    i = 0;
    first = bflvecbegin(vec);
    last = bflvecend(vec);
    while (first != last) {
	assert(i == (int) bflvecdata(first));
	first = bflvecnext(first);
	i += 1;
    }

    i = 9;
    first = bflvecrbegin(vec);
    last = bflvecrend(vec);
    while (first != last) {
	assert(i == (int) bflvecdata(first));
	first = bflvecprev(first);
	i -= 1;
    }

    bflvecpopfront(vec);
    printvec(vec);

    assert(bflvecempty(vec) == 0);
    assert(bflvecsize(vec) == 9);

    i = 1;
    first = bflvecbegin(vec);
    last = bflvecend(vec);
    while (first != last) {
	assert(i == (int) bflvecdata(first));
	first = bflvecnext(first);
	i += 1;
    }

    bflvecpopback(vec);
    printvec(vec);

    assert(bflvecempty(vec) == 0);
    assert(bflvecsize(vec) == 8);

    i = 1;
    first = bflvecbegin(vec);
    last = bflvecend(vec);
    while (first != last) {
	assert(i == (int) bflvecdata(first));
	first = bflvecnext(first);
	i += 1;
    }

    assert((int) bflvecnth(vec, 2) == 3);

    assert(bflvecbegin(vec) + 2 == bflvecfind(vec, bflptrcmp, (void *) 3));

    bflvecremove(vec, bflptrcmp, (void *) 1);
    printvec(vec);

    assert(bflvecempty(vec) == 0);
    assert(bflvecsize(vec) == 7);

    i = 2;
    first = bflvecbegin(vec);
    last = bflvecend(vec);
    while (first != last) {
	assert(i == (int) bflvecdata(first));
	first = bflvecnext(first);
	i += 1;
    }

    bflvecclean(vec);
    printvec(vec);

    assert(bflvecempty(vec) == 1);
    assert(bflvecsize(vec) == 0);

    bflvecfree(vec);
    test_sort();
    puts("functional test good for 'tvec'");
    return 0;
}
