/*
 * $Author: ac $
 * $Id: fpm_network.h,v 1.1 1999/10/21 12:52:06 ac Exp $
 * $Date: 1999/10/21 12:52:06 $
 */
/*
 * $Log: fpm_network.h,v $
 * Revision 1.1  1999/10/21 12:52:06  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 */

#ifndef __fpm_network_h
#define __fpm_network_h

/*----------------------------------------------------------------------*/
/*                           Nested includes                            */
/*----------------------------------------------------------------------*/

#include <stdio.h>

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

#define FPN_VERBOSE             (1 << 0)
#define FPN_ABL_ORDER		(1 << 1)
#define FPN_FLATTEN_BEH		(1 << 2)

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

#ifndef __P
# ifdef __STDC__
#  define __P(args)	args
# else
#  define __P(args)	()
# endif
#endif

#define fpm_type_alloc(type)	 	(type *)autallocheap(sizeof(type ))
#define fpm_type_free(type, ptr)	autfreeheap(( char *)ptr, sizeof(type ))

#define fpm_atom_alloc()		fpm_type_alloc(fpm_atom_t)
#define fpm_expr_alloc()		fpm_type_alloc(fpm_expr_t)
#define fpm_network_alloc()		fpm_type_alloc(fpm_network_t)

#define fpm_atom_free(Atom)		fpm_type_free(fpm_atom_t,   Atom)
#define fpm_expr_free(Expr)		fpm_type_free(fpm_expr_t,  Expr)
#define fpm_network_free(Network)	fpm_type_free(fpm_network_t, Network)


/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

typedef struct fpm_atom {

	char		*NAME;		/* this atom's name		*/
	bddnode		*BDD;		/* this atom's bddnode		*/
	chain_list	*BDD_EXPR;	/* this atom's dependance	*/
	ptype_list	*USER;		/* application specific		*/
	
} fpm_atom_t;

typedef struct fpm_expr {

	char		*NAME;		/* this expr's name		*/
	bddnode		*BDD;		/* this expr's bddnode	*/
	chain_list	*BDD_ATOM;	/* this expr's support	*/
	long		 COST;		/* for optimize			*/
	ptype_list	*USER;		/* application specific		*/
	
} fpm_expr_t;

typedef struct fpm_network {
	
	char		*NAME;

	bflmap_t	*MAP_ATOM;	/* atom->NAME  - atom		*/
	bflmap_t	*MAP_EXPR;	/* expr->NAME - expr	*/

	bddnamein	*NAME_IN;
	long		 NAME_IN_SIZE;

	bddindex	*INDEX_IN;
	long		 INDEX_IN_SIZE;

	bddsystem	*BDD_SYSTEM;

} fpm_network_t;

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                         Function prototypes                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                         Fpn Network Functions                        */
/*----------------------------------------------------------------------*/

extern fpm_network_t *
fpm_network_new __P(( char *Name, bddsystem *System ));

extern void
fpm_network_delete __P(( fpm_network_t *Network ));

extern void
fpm_network_depend __P(( fpm_network_t *Network ));

extern void
fpm_network_view __P(( fpm_network_t *Network ));

/*----------------------------------------------------------------------*/
/*                          Fpm Atom Functions                          */
/*----------------------------------------------------------------------*/

extern fpm_atom_t *
fpm_atom_new __P(( fpm_network_t *Network, char *Name, int Index ));

extern void
fpm_atom_delete __P(( fpm_network_t *Network, fpm_atom_t *Atom ));

extern fpm_atom_t *
fpm_atom_find __P(( fpm_network_t *Network, char *Name ));

extern char *
fpm_atom_name __P(( fpm_network_t *Network, bddindex Index ));

extern bddvar
fpm_atom_var __P(( fpm_network_t *Network, fpm_atom_t *Atom ));

/*----------------------------------------------------------------------*/
/*                          Fpm Expr Functions                          */
/*----------------------------------------------------------------------*/

extern fpm_expr_t *
fpm_expr_new __P(( fpm_network_t *Network, char *Name, bddnode *Node ));

extern void
fpm_expr_delete __P(( fpm_network_t *Network, fpm_expr_t *Expr ));

extern fpm_expr_t *
fpm_expr_find __P(( fpm_network_t *Network, char *Name ));

extern bddnode *
fpm_atom_find_node __P(( fpm_network_t *Network, char *Name ));
	
extern void
fpm_expr_depend __P(( fpm_network_t *Network, fpm_expr_t *Expr ));
	
extern long
fpm_expr_num __P(( fpm_network_t *Network ));

#endif /* __fpm_network_h */
