/*
 * $Author: ac $
 * $Id: xnf.c,v 1.4 1999/10/21 12:52:15 ac Exp $
 * $Date: 1999/10/21 12:52:15 $
 */
/*
 * $Log: xnf.c,v $
 * Revision 1.4  1999/10/21 12:52:15  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.3  1999/08/27 18:46:29  czo
 * Plein de modif pour compiler sur RedHat6
 *
 * Revision 1.2  1999/05/18 15:39:11  ac
 * Bug fix in the xnf driver.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:42  alliance
 * fpmap import
 *
 */

/* ### ------------------------------------------------------- ### */
/* # Includes                                                    # */
/* ### ------------------------------------------------------- ### */

# include <stdio.h>
# include <string.h>
# include <time.h>
# include  MUT_H
# include  BFL_H
# include  AUT_H
# include  ABL_H
# include  ABE_H
# include  XNF_H

/* ### ------------------------------------------------------- ### */
/* # Constant declarations                                       # */
/* ### ------------------------------------------------------- ### */

/* ### ------------------------------------------------------- ### */
/* # Type declarations                                           # */
/* ### ------------------------------------------------------- ### */

/* ### ------------------------------------------------------- ### */
/* # Stucture declarations                                       # */
/* ### ------------------------------------------------------- ### */

/* ### ------------------------------------------------------- ### */
/* # Variable declarations                                       # */
/* ### ------------------------------------------------------- ### */

static authtable	*signal_table;

/* ### ------------------------------------------------------- ### */
/* # Macro declarations                                          # */
/* ### ------------------------------------------------------- ### */

/* ### ------------------------------------------------------- ### */
/* # Static function prototypes                                  # */
/* ### ------------------------------------------------------- ### */

#if defined(__STDC__)
#define ARG(args) args
#else
#define ARG(args) ()
#endif
 
static	char	*xnf_name ARG((char *name));
static	char	*xnf_rename ARG((char *name));
static	void	 xnf_support ARG((chain_list *expr, ptype_list **res));
static	int	 xnf_assoc ARG((char *var, ptype_list *assoc));
static	void	 xnf_infix ARG((FILE *fp, chain_list *equate, ptype_list *assoc));
static	void	 xnf_eqn_int ARG((FILE *fp, chain_list *equate, ptype_list *assoc, char *output));

#undef ARG

/* ### ------------------------------------------------------- ### */
/* # Definition of exported functions                            # */
/* ### ------------------------------------------------------- ### */

void xnf_prolog(fp, comment)
    FILE *fp;
    char *comment;
{
    time_t clock;
    
    signal_table = createauthtable(100);
    time (&clock);
    fprintf(fp, "LCANET, 5\n");
    fprintf(fp, "PROG, %s. date: %s", comment, ctime(&clock));
}

void xnf_epilog(fp)
    FILE *fp;
{
    fprintf(fp, "EOF\n");
    destroyauthtable(signal_table);
}

/*
 * xnf_ext(fp, name, direction)
 *	Declares an external signal. `name' is the name of the signal,
 * `direction' is one of 'I' (for input) or 'O' (for output). 
 */
void xnf_ext(fp, name, direction)
    FILE *fp;
    char *name;
    char direction;
{
    fprintf(fp, "EXT, %s, %c,\n", xnf_name(name), direction);
}

/*
 * xnf_ibuf(fp, pad, blknm)
 *	Declares an ibuf. input is the name of the pad (it should be declared
 * extern) and output is the internal name of the signal.
 */
void xnf_ibuf(fp, pad, blknm)
    FILE *fp;
    char *pad;
    char *blknm;
{
    static int coIbuf = 1;

    if (blknm)
	fprintf(fp, "SYM, IBUF_%d, IBUF, BLKNM=%s\n", coIbuf++, blknm);
    else
	fprintf(fp, "SYM, IBUF_%d, IBUF,\n", coIbuf++);
    fprintf(fp, "PIN, I, I, %s,\n", xnf_name(pad));
    fprintf(fp, "PIN, O, O, %s,\n", xnf_rename(pad));
    fprintf(fp, "END\n");
}

void xnf_bufg(fp, pad, input, blknm)
    FILE *fp;
    char *pad;
    char *input;
    char *blknm;
{
    static int coBufg = 1;

    if (blknm)
	fprintf(fp, "SYM, BUFG_%d, BUFG, BLKNM=%s\n", coBufg++, blknm);
    else
	fprintf(fp, "SYM, BUFG_%d, BUFG,\n", coBufg++);
    fprintf(fp, "PIN, O, O, %s,\n", xnf_name(input));
    fprintf(fp, "PIN, I, I, %s,\n", xnf_name(pad));
    fprintf(fp, "END\n");
}

void xnf_obuf(fp, pad, output, blknm)
    FILE *fp;
    char *pad;
    char *output;
    char *blknm;
{
    static int coObuf = 1;

    if (blknm)
	fprintf(fp, "SYM, OBUF_%d, OBUF, BLKNM=%s\n", coObuf++, blknm);
    else
	fprintf(fp, "SYM, OBUF_%d, OBUF,\n", coObuf++);
    fprintf(fp, "PIN, O, O, %s,\n", xnf_name(pad));
    fprintf(fp, "PIN, I, I, %s,\n", xnf_name(output));
    fprintf(fp, "END\n");
}

void xnf_obuft(fp, pad, output, tri, blknm)
    FILE *fp;
    char *pad;
    char *output;
    char *tri;
    char *blknm;
{
    static int coObuft = 1;

    if (blknm)
	fprintf(fp, "SYM, OBUFT_%d, OBUFT, BLKNM=%s\n", coObuft++, blknm);
    else
	fprintf(fp, "SYM, OBUFT_%d, OBUFT,\n", coObuft++);
    fprintf(fp, "PIN, O, O, %s,\n", xnf_name(pad));
    fprintf(fp, "PIN, I, I, %s,\n", xnf_name(output));
    fprintf(fp, "PIN, T, I, %s,, INV\n", xnf_name(tri));
    fprintf(fp, "END\n");
}

void
xnf_dff(fp, val, cnd, clk, reg, blknm)
    FILE *fp;
    char *val;
    char *cnd;
    char *clk;
    char *reg;
    char *blknm;
{
    static int coDff = 1;

    if (blknm)
	fprintf(fp, "SYM, DFF_%d, DFF, BLKNM=%s\n", coDff++, blknm);
    else
	fprintf(fp, "SYM, DFF_%d, DFF,\n", coDff++);

    fprintf(fp, "PIN, D, I, %s,\n", xnf_name(val));
    if (cnd) fprintf(fp, "PIN, CE, I, %s,\n", xnf_name(cnd));
    fprintf(fp, "PIN, C, I, %s,\n", xnf_name(clk));
    fprintf(fp, "PIN, Q, O, %s,\n", xnf_name(reg));
    fprintf(fp, "END\n");
}

void xnf_eqn(fp, equate, name)
    FILE	*fp;
    chain_list 	*equate;
    char	*name;
{
    ptype_list	*assoc	= 0;

    xnf_support(equate, &assoc);
    xnf_eqn_int(fp, equate, assoc, name);
    freeptype(assoc);
}

void
xnf_fmap(fp, equate, name, blknm)
    FILE	*fp;
    chain_list	*equate;
    char	*name;
    char	*blknm;
{
    static int	 coFmap = 1;
    ptype_list	*assoc = 0;
    ptype_list	*ptr;

    xnf_support(equate, &assoc);
    xnf_eqn_int(fp, equate, assoc, name);
    if (blknm)
	fprintf(fp, "SYM, FMAP_%d, FMAP, BLKNM=%s\n", coFmap++, blknm);
    else
	fprintf(fp, "SYM, FMAP_%d, FMAP,\n", coFmap++);

    for (ptr = assoc; ptr != NULL; ptr = ptr->NEXT)
	fprintf(fp, "PIN, I%ld, I, %s\n", ptr->TYPE, xnf_name((char *) ptr->DATA));
    fprintf(fp, "PIN, O, I, %s\n", xnf_name(name));
    fprintf(fp, "END\n");
    freeptype(assoc);
}

void
xnf_hmap(fp, equate, name, blknm)
    FILE	*fp;
    chain_list	*equate;
    char	*name;
    char	*blknm;
{
    static int coHmap = 1;
    ptype_list	*assoc = 0;
    ptype_list *ptr;

    xnf_support(equate, &assoc);
    xnf_eqn_int(fp, equate, assoc, name);
    if (blknm)
	fprintf(fp, "SYM, HMAP_%d, HMAP, BLKNM=%s\n", coHmap++, blknm);
    else
	fprintf(fp, "SYM, HMAP_%d, HMAP,\n", coHmap++);

    for (ptr = assoc; ptr != NULL; ptr = ptr->NEXT)
	fprintf(fp, "PIN, I%ld, I, %s\n", ptr->TYPE, xnf_name((char *) ptr->DATA));
    fprintf(fp, "PIN, O, I, %s\n", xnf_name(name));
    fprintf(fp, "END\n");
    freeptype(assoc);
}

void
xnf_tbuf(fp, val, cnd, name)
    FILE *fp;
    char *val;
    char *cnd;
    char *name;
{
    static int co_tbuf = 1;
    
    fprintf(fp, "SYM, TBUF_%d, TBUF,\n", co_tbuf++);
    fprintf(fp, "PIN, O, O, %s,\n", xnf_name(name));
    fprintf(fp, "PIN, I, I, %s,\n", xnf_name(val));
    fprintf(fp, "PIN, T, I, %s,, INV\n", xnf_name(cnd));
    fprintf(fp, "END\n");
}

char *
xnf_signal_name()
{
	static int count = 1;
	char buffer[512];

	do {
		sprintf( buffer, "xnf_signal%d", count++ );
	} while ( namefind( buffer ) != NULL );
	return namealloc( buffer );
}

/* ### ------------------------------------------------------- ### */
/* # Definition of internal functions                            # */
/* ### ------------------------------------------------------- ### */

/* ### ------------------------------------------------------- ### */
/* # Definition of static functions                              # */
/* ### ------------------------------------------------------- ### */

static char *
xnf_rename(char *name)
{
	char *new_name = xnf_signal_name();
	addauthelem(signal_table, name, (long)new_name);
	return new_name;
}

/* 
 * xnf_name(name)
 *	Make an identifier xnf compatible. If the name is a vector, translate
 * "toto 1" in "toto<1>".
 * 	If the name has been renamed (xnf_rename), returns the substitution.
 */
static char *xnf_name(name)
    char *name;
{
    if (strchr(name, ' ') == NULL)
	return name;
    else {
	
        char p[BUFSIZ];
        strcpy(p, name);
        for (name = p; *name; name++)
	    if (*name == ' ')
	        *name = '<';
	*name++ = '>';
	*name   = '\0';
        return namealloc(p);
    }
}

static void xnf_support(expr, res)
    chain_list *expr;
    ptype_list **res;
{
    if (expr->NEXT == NULL) {
	ptype_list *ptr;
	char *atom = (char *)expr->DATA;

	if (strcmp(atom, "'d'") && strcmp(atom, "'1'") && strcmp(atom, "'0'")) {
	    int i;
	    
	    for (ptr = *res, i = 1; ptr != NULL; ptr = ptr->NEXT, i++)
		if ((char*) ptr->DATA == atom)
		    return;
	    *res = addptype(*res, i, expr->DATA);
	}
    } else
	while ((expr = expr->NEXT))
	    xnf_support((chain_list*)expr->DATA, res);
}

/*
 * xnf_assoc(var, assoc)
 *	Returns the index of var in assoc.
 */
static int xnf_assoc(var, assoc)
    char *var;
    ptype_list *assoc;
{
    while (assoc) {
	if (!strcmp(var, (char *) assoc->DATA))
	    return assoc->TYPE;
	assoc = assoc->NEXT;
    }
    return (-1);
}

/*
 * xnf_infix(fp, equate, assoc)
 *	Prints an equation in the .xnf style. 
 */
static void xnf_infix(fp, equate, assoc)
    FILE *fp;
    chain_list *equate;
    ptype_list *assoc;
{
    short oper;

    if (ABL_ATOM(equate)) {
	if (ABL_ATOM_VALUE(equate) == ABL_ATOM_NAME_ZERO)
	    fprintf(fp, "0");
	else if  (ABL_ATOM_VALUE(equate) == ABL_ATOM_NAME_ONE)
	    fprintf(fp, "1");
	else
	    fprintf(fp, "I%d", (xnf_assoc((char *) equate->DATA, assoc) - 1));
    } else {
	oper = ABL_OPER(equate);
	if (oper == ABL_NOT) {
	    fprintf(fp, "~");
	    xnf_infix(fp, ABL_CADR(equate), assoc);
	} else {
	    if (getabloperpolar(oper) == ABL_POLAR_NEGATIVE) {
		fprintf(fp, "~");
		oper = getablopernot(oper);
	    }
	    fprintf(fp, "(");
	    while ((equate = ABL_CDR(equate))) {
		xnf_infix(fp, ABL_CAR(equate), assoc);
		if (ABL_CDR(equate)) {
		    switch (oper) {
		    case ABL_AND:
			fprintf(fp, "*");
			break;
		    case ABL_OR:
			fprintf(fp, "+");
			break;
		    case ABL_XOR:
			fprintf(fp, "@");
			break;
		    default:
			bflerror(1, "xnf_infix: bad xnf expression");
		    }
		}
	    }
	    fprintf(fp, ")");
	}
    }
}

static void xnf_eqn_int(fp, equate, assoc, name)
    FILE	*fp;
    chain_list 	*equate;
    ptype_list	*assoc;
    char	*name;
{
    static int coEqn = 1;
    ptype_list *ptr;

    fprintf(fp, "SYM, EQN_%d, EQN, EQN=", coEqn++);

    xnf_infix(fp, equate, assoc);
    fprintf(fp, "\n");
    for (ptr = assoc; ptr != NULL; ptr = ptr->NEXT)
        fprintf(fp, "PIN, I%ld, I, %s\n", (ptr->TYPE - 1), xnf_name((char *) ptr->DATA));
    fprintf(fp, "PIN, O, O, %s\n", xnf_name(name));
    fprintf(fp, "END\n");
}
