/*------------------------------------------------------------\
|                                                             |
| Tool    :                   GRAAL                           |
|                                                             |
| File    :                   Undo.c                          |
|                                                             |
| Author  :                Jacomme Ludovic                    |
|                                                             |
| Date    :                  28.03.95                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include MUT_H
# include MPH_H
# include RDS_H
# include RPR_H
# include RFM_H
# include GRM_H
# include GSB_H
# include "GRM_window.h"
# include "GRM_undo.h"

/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  graalundo *GraalHeadUndo = (graalundo *)NULL;
  graalundo *GraalHeadRedo = (graalundo *)NULL;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                       Alloc Functions                       |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        GraalAllocUndo                       |
|                                                             |
\------------------------------------------------------------*/

graalundo *GraalAllocUndo()
{
  return((graalundo *)rdsalloc(sizeof(graalundo), 1));
}

/*------------------------------------------------------------\
|                                                             |
|                      GraalAllocUndoRec                      |
|                                                             |
\------------------------------------------------------------*/

graalundorec *GraalAllocUndoRec()
{
  return((graalundorec *)rdsalloc(sizeof(graalundorec), 1));
}

/*------------------------------------------------------------\
|                                                             |
|                        Free Functions                       |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         GraalFreeUndo                       |
|                                                             |
\------------------------------------------------------------*/

void GraalFreeUndo( FreeUndo )

   graalundo *FreeUndo;
{
  rdsfree((char *)FreeUndo, sizeof(graalundo));
}

/*------------------------------------------------------------\
|                                                             |
|                       GraalFreeUndoRec                      |
|                                                             |
\------------------------------------------------------------*/

void GraalFreeUndoRec( FreeUndoRec )

   graalundorec *FreeUndoRec;
{
  rdsfree((char *)FreeUndoRec, sizeof(graalundorec));
}

/*------------------------------------------------------------\
|                                                             |
|                       GraalAddUndoRec                       |
|                                                             |
\------------------------------------------------------------*/

void GraalAddUndoRec( Rectangle )

   rdsrec_list *Rectangle;
{
  graalundorec *UndoRec;

  rdsbegin();

  UndoRec             = GraalAllocUndoRec();
  UndoRec->RECTANGLE  = Rectangle;
  UndoRec->NEXT       = GraalHeadUndo->UNDO;
  GraalHeadUndo->UNDO = UndoRec;

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                       GraalAddUndo                          |
|                                                             |
\------------------------------------------------------------*/

void GraalAddUndo()
 
{
  graalundo *NewUndo;

  rdsbegin();

  GraalDelUndo( &GraalHeadRedo );

  NewUndo       = GraalAllocUndo();
  NewUndo->NEXT = GraalHeadUndo;
  GraalHeadUndo = NewUndo;

  GraalRecomputeBound = GRAAL_TRUE;

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                       GraalEraseUndo                        |
|                                                             |
\------------------------------------------------------------*/

void GraalEraseUndo( HeadUndo )

  graalundo **HeadUndo;
{
  graalundo    *ScanUndo;
  graalundo    *DelUndo;
  graalundorec *ScanUndoRec;
  graalundorec *DelUndoRec;

  rdsbegin();

  ScanUndo  = *HeadUndo;
  *HeadUndo = (graalundo *)NULL;

  while ( ScanUndo != (graalundo *)NULL )
  {
    ScanUndoRec = ScanUndo->UNDO;

    while( ScanUndoRec != (graalundorec *)NULL )
    {
      DelUndoRec  = ScanUndoRec;
      ScanUndoRec = ScanUndoRec->NEXT;
    
      GraalFreeUndoRec( DelUndoRec );
    }

    DelUndo  = ScanUndo;
    ScanUndo = ScanUndo->NEXT;

    GraalFreeUndo( DelUndo );
  }

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                     GraalInitializeUndo                     |
|                                                             |
\------------------------------------------------------------*/

void GraalInitializeUndo()

{
  rdsbegin();

  GraalEraseUndo( &GraalHeadUndo );
  GraalEraseUndo( &GraalHeadRedo );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                        GraalDelUndo                         |
|                                                             |
\------------------------------------------------------------*/

void GraalDelUndo( HeadUndo )

  graalundo **HeadUndo;
{
  graalundo     *ScanUndo;
  graalundo     *DelUndo;
  graalundorec  *ScanUndoRec;
  graalundorec  *DelUndoRec;
  rdsrec_list   *ScanRec;
  graalundorec  *HeadDelete;
  graalundorec  *Delete;

  rdsbegin();
 
  HeadDelete = (graalundorec *)NULL;
  ScanUndo   = *HeadUndo;
  *HeadUndo  = (graalundo *)NULL;

  while ( ScanUndo != (graalundo *)NULL )
  {
    ScanUndoRec = ScanUndo->UNDO;
 
    while( ScanUndoRec != (graalundorec *)NULL )
    {
      ScanRec = ScanUndoRec->RECTANGLE; 
 
      if ( IsGraalDeleted( ScanRec ) )
      {
        GraalUndeleteRectangle( ScanRec );
 
        Delete       = ScanUndoRec;
        ScanUndoRec  = ScanUndoRec->NEXT;
        Delete->NEXT = HeadDelete;
        HeadDelete   = Delete;
      }
      else
      {
        DelUndoRec   = ScanUndoRec;
        ScanUndoRec  = ScanUndoRec->NEXT;

        GraalFreeUndoRec( DelUndoRec );
      }
    }

    DelUndo  = ScanUndo;
    ScanUndo = ScanUndo->NEXT;

    GraalFreeUndo( DelUndo );
  }

  while( HeadDelete != (graalundorec *)NULL )
  {
    Delete     = HeadDelete;
    HeadDelete = HeadDelete->NEXT;
 
    if ( IsRdsInstance( Delete->RECTANGLE ) )
    {
      GraalDelInstance( GRAAL_PREVIOUS( Delete->RECTANGLE ) );
    }
    else
    {
      GraalDelRectangle( Delete->RECTANGLE );
    }
 
    GraalFreeUndoRec( Delete );
  }
 
  rdsend();
}
