/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - CMOS Cells Compiler
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */




#ident  "$Id: c4_stretch.c,v 1.1.1.1 1998/10/02 15:25:08 alliance Exp $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include MPH_H
#include MLO_H
#include BEH_H
#include LOG_H

#include ICN_H
#include IAC_H
#include ICU_H

#include "c4_local.h"

/******************************************************************************
 ***                            c.c.c.c tools                               ***
 ***                    Complex Cmos Cells Compiler                         ***
 ******************************************************************************/


/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*
  =-=-=-=- low level func =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-[]=-
  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  SYNTAX : c4_genPCells(short, char *, chain_list *, char *, char *)
  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  EXEMPLE : c4_genPCells(NT1, "nt1_y", lc_io, "s", "io")
  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  FONCTION :
  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  ERROR : none
 *=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

char *
c4_genPCells (cellType, name, lc_io, out, in1, in2, in3, in4, in5)
     short cellType;
     char *name;
     chain_list *lc_io;
     char *out;
     char *in1, *in2, *in3, *in4, *in5;
{
  lofig_list *lFig;
  struct IconGate *not1icon ();
  struct IconGate *not1Picon ();
  struct IconGate *buficon ();
  struct IconGate *high1icon ();
  struct IconGate *low1icon ();
  struct IconGate *bus1icon ();
  struct IconGate *tieicon ();
  struct IconGate *latch1icon ();
  struct IconGate *latch1nicon ();
  struct IconGate *latch2icon ();
  struct IconGate *latch2nicon ();
  struct IconGate *bd1icon ();
  struct IconGate *bd1nicon ();

  (void) dellofig (name);
  switch (cellType)
    {
    case C4_N1:
      n1_p (name, out, in1);
      lofigNot (name, lc_io, out, in1);
      delicongate ((char *) ICONGATE_MODELNAME (not1icon (name, 1)));
      break;

      /* inverseur puissant */
    case C4_N1P:
      n1p_p (name, out, in1);
      lofigNotP (name, lc_io, out, in1);
      delicongate ((char *) ICONGATE_MODELNAME (not1Picon (name, 1)));
      break;
      /* buffer */
    case C4_BUF:
      b1_p (name, out, in1);
      lofigBuf (name, lc_io, out, in1);
      delicongate ((char *) ICONGATE_MODELNAME (buficon (name, 1)));
      break;

    case C4_TRI:
      bus_p (name, out, in1, in2);
      lofigBus (name, lc_io, out, in1, in2);
      delicongate ((char *) ICONGATE_MODELNAME (bus1icon (name, 1)));
      break;

    case C4_L1:
      l1_p (name, out, in1, in2);
      lofigLatch1 (name, lc_io, out, in1, in2, in3);
      delicongate ((char *) ICONGATE_MODELNAME (latch1icon (name, 1)));
      break;

    case C4_L1N:
      l1n_p (name, out, in1, in2);
      lofigLatch1N (name, lc_io, out, in1, in2, in3);
      delicongate ((char *) ICONGATE_MODELNAME (latch1nicon (name, 1)));
      break;

    case C4_L2:
      l2_p (name, out, in1, in2, in3, in4);
      lofigLatch2 (name, lc_io, out, in1, in2, in3, in4, in5);
      delicongate ((char *) ICONGATE_MODELNAME (latch2icon (name, 1)));
      break;

    case C4_L2N:
      l2n_p (name, out, in1, in2, in3, in4);
      lofigLatch2N (name, lc_io, out, in1, in2, in3, in4, in5);
      delicongate ((char *) ICONGATE_MODELNAME (latch2nicon (name, 1)));
      break;

    case C4_BD1:
      bd1_p (name, out, in1, in2);
      lofigBd1 (name, lc_io, out, in1, in2, in3, in4);
      delicongate ((char *) ICONGATE_MODELNAME (bd1icon (name, 1)));
      break;

    case C4_BD1N:
      bd1n_p (name, out, in1, in2);
      lofigBd1n (name, lc_io, out, in1, in2, in3, in4);
      delicongate ((char *) ICONGATE_MODELNAME (bd1nicon (name, 1)));
      break;

    case C4_TIE:
      tie_p (name);
      lofigTie (name, lc_io);
      delicongate ((char *) ICONGATE_MODELNAME (tieicon (name, 1)));
      break;

    case C4_HIGH:
      high_p (name, out);
      lofigHigh (name, lc_io, out);
      delicongate ((char *) ICONGATE_MODELNAME (high1icon (name, 1)));
      break;

    case C4_LOW:
      low_p (name, out);
      lofigLow (name, lc_io, out);
      delicongate ((char *) ICONGATE_MODELNAME (low1icon (name, 1)));
      break;
    }

  (void) dellofig (name);
  lFig = getlofig (name, 'A');
  savelofig (lFig);

  return name;
}
