/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Standard Cell Mapping
 * Date    : 1991,92
 * Author  : Pascale Allegre
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: mapping_display.c,v 1.1.1.1 1998/10/02 15:25:09 alliance Exp $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include LOG_H
#include BEH_H
#include MLO_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"


/******************************************************************************/
/*********** displayChain_liste : affiche une chain_list  *********************/
/******************************************************************************/
void 
displayChain_liste (liste)
     chain_list *liste;
{
  while (liste)
    {
      printf ("%s ", liste->DATA);
      liste = CDR (liste);
    }
  printf ("\n");
}

/******************************************************************************/
/******* displayArite : affiche l'expression et son arite *********************/
/******************************************************************************/
void
displayArite (exp)
     chain_list *exp;
{
  if (!exp)
    printf ("Null expression");
  if (ATOM (exp))
    printf (" %s", VALUE_ATOM (exp));
  else
    {
      printf (" (%s", searchCharOper (OPER (exp)));
      if (ARITE (exp))
	printf (" %d", ARITE (exp));
      mapExpr (displayArite, exp);
      printf (")");
    }
}

/****************************************************************************/
/* --------------------------------------------------------------------
   displayBehMap : affiche les signaux d'une befig et leur ABL

   nb = 1 : befig comportementale
   nb = 0 : befig structurelle

   -------------------------------------------------------------------- */
void
displayBehMap (befig, nb)
     befig_list *befig;
     short nb;
{
  beout_list *o;
  bereg_list *r;
  bebus_list *b;
  beaux_list *a;

  fprintf (stderr, "----- display Befig (%s) -----\n", nb ? "comportementale" : "structurelle");
  printf (" OUT :\n");
  for (o = befig->BEOUT; o; o = o->NEXT)
    {
      printf ("   => %s =", o->NAME);
      displayArite (o->ABL);
      printf ("\n");
    }
  printf (" AUX :\n");
  for (a = befig->BEAUX; a; a = a->NEXT)
    {
      printf ("   => %s =", a->NAME);
      displayArite (a->ABL);
      printf ("\n");
    }
  printf (" BUS :\n");
  for (b = befig->BEBUS; b; b = b->NEXT)
    {
      biabl_list *bi;

      printf ("   => %s\n", b->NAME);
      for (bi = b->BIABL; bi; bi = bi->NEXT)
	{
	  printf ("         = VAL :");
	  displayArite (bi->VALABL);
	  printf ("\n");
	  if (nb)
	    {
	      printf ("         = CND :");
	      displayArite (bi->CNDABL);
	      printf ("\n");
	    }
	}
    }
  printf (" REG :\n");
  for (r = befig->BEREG; r; r = r->NEXT)
    {
      biabl_list *bi;

      if (!nb)
	{
	  printf ("   => %s =", r->NAME);
	  displayArite (r->BIABL->VALABL);
	  printf ("\n");
	  displayArite (r->BIABL->CNDABL);
	  printf ("\n");
	}
      else
	{
	  printf ("   => %s\n", r->NAME);
	  for (bi = r->BIABL; bi; bi = bi->NEXT)
	    {
	      printf ("\tVAL : ");
	      displayArite (bi->VALABL);
	      printf ("\n");
	      printf ("\tCOM : %s", (bi->USER) ? "(STABLE)" : "");
	      displayArite (bi->CNDABL);
	      printf ("\n");
	    }
	}
    }
  fprintf (stderr, "----- Fin displayBehMap -----\n");
}

void
nbreRegles ()
{
  declenList *d;
  regle_list *r;
  short i = 0;

  for (d = mappSys->ptdecl; d; d = d->NEXT)
    for (r = d->regles; r; r = r->NEXT)
      i++;
  printf ("%d rules generated\n", i);
}



void
displayDeclencheurs ()
{
  declenList *p;

  printf ("\n***************************************************************");
  printf ("\n		Lecture de mappSys->ptdecl\n");
  printf ("***************************************************************\n");
  for (p = mappSys->ptdecl; p; p = p->NEXT)
    {
      printf ("===============================\n");
      printf ("===> nom du declencheur : %s\n", p->name);
      printf ("===============================\n");
      if (p->regles)
	{
	  regle_list *r;

	  for (r = p->regles; r; r = r->NEXT)
	    {
	      printf ("---> ");
	      displayArite (r->premisse);
	      printf (" -->");
	      displayArite (r->conclusion);
	      printf ("\n");
/*
   displayAlist(r->var);
   printf("\n");
   printf(" profondeur premisse = %d\n",r->prof_premi);
 */
	    }
	}
    }
  printf ("\n");
}

/******************************************************************************/
/***** displayCell : affichage de chaque cellule logique et de ces champs *****/
/******************************************************************************/
void
displayCell (c)
     cellList *c;

{
  ptype_list *pt;
  begen_list *gen;
  Alist *in;
  tabNumList *dt, *ct;
  ptype_list *st;
  chain_list *out;
  num_list *d, *dtVal, *ctVal;
  biabl_list *i;

  printf ("\n*****************************************\n");
  printf ("Nom de la cellule logique : %s\n", c->nomlog);
  printf ("Nom de la cellule par defaut : %s\n", c->nom);
  printf ("*****************************************\n");

  printf ("   Type de la cellule : %c\n", c->type);
  if ((c->type == 'l') || (c->type == 't') || (c->type == 'c'))
    {
      printf ("   Fonction logique de la cellule : ");
      displayArite (c->val_abl->VALABL);
      printf ("\n");
    }
  else
    for (i = c->val_abl; i; i = i->NEXT)
      {
	printf ("   Fonction logique & commande : ");
	displayArite (i->VALABL);
	printf (" & ");
	displayArite (i->CNDABL);
	printf ("\n");
      }

  printf ("   Interface de la cellule : ");
  for (in = c->interface; in; in = in->NEXT)
    printf ("%s ", in->name);

  printf ("\n   Entrees de la cellule : ");
  for (in = c->entrees; in; in = in->NEXT)
    printf ("%s ", in->name);

  printf ("\n   Delais  de la cellule : ");
  for (d = c->delais; d; d = d->NEXT)
    printf ("%ld ", d->DATA);

  printf ("\n   Sorties de la cellule : ");
  for (out = c->sorties; out; out = out->NEXT)
    printf ("%s ", (char *)out->DATA);

  printf ("\n   Surface de la cellule : %ld\n", c->surface);

  printf ("   Enveloppe de la cellule : ");
  if (c->envel)
    displayArite (c->envel);
  printf ("\n");

  printf ("     Delais_tech :");
  for (dt = c->delais_tech; dt; dt = dt->NEXT)
    {
      printf ("\n     %s : ", dt->name);
      for (dtVal = dt->ptNum; dtVal; dtVal = dtVal->NEXT)
	printf (" %ld ", dtVal->DATA);
    }

  printf ("\n   Capa_tech :");
  for (ct = c->capa_tech; ct; ct = ct->NEXT)
    {
      printf ("\n     %s : ", ct->name);
      for (ctVal = ct->ptNum; ctVal; ctVal = ctVal->NEXT)
	printf (" %ld ", ctVal->DATA);
    }

  printf ("\n   Surface_tech :");
  for (st = c->surface_tech; st; st = st->NEXT)
    printf ("\n     %s = %ld", (char *)st->DATA, st->TYPE);

  printf ("\n   Resistance_tech :");
  for (st = c->resistance_tech; st; st = st->NEXT)
    printf ("\n     %s = %ld", (char *)st->DATA, st->TYPE);

  printf ("\n   Generic :");
  for (pt = c->generic; pt; pt = pt->NEXT)
    {
      printf ("\n     Porte %s\n", (char *) pt->DATA);
      for (gen = (begen_list *) pt->TYPE; gen; gen = gen->NEXT)
	printf ("%s=%d ", gen->NAME, *(int *) gen->VALUE);
    }
  printf ("\n");
}


void 
displayTHSignaux ()
{
  int i;
  pElemTH pEl = mappSys->sig_hash->pElem;

  printf ("=============================================\n");
  printf ("        Lecture de mappSys->sig_hash\n");
  printf ("=============================================\n");

  printf ("length = %d\t		count = %d\n", mappSys->sig_hash->length, mappSys->sig_hash->count);
  printf ("=============================================\n");
  for (i = 0; i < mappSys->sig_hash->length; i++)
    {
      if ((pEl->value != VIDETH) && (pEl->value != DELETETH))
	{
	  chain_list *abl;

	  printf ("NOM du signal : %s\n", pEl->key);
	  printf ("\ttype\t\t= %d\n", ((sig *) pEl->value)->type);
	  printf ("\tdelais\t\t= %ld\n", ((sig *) pEl->value)->delais);
	  printf ("\tcapa\t\t= %ld\n", ((sig *) pEl->value)->capa);
	  printf ("\tcout\t\t= %d\n", ((sig *) pEl->value)->cout);
	  for (abl = ((sig *) pEl->value)->ablList; abl; abl = abl->NEXT)
	    {
	      printf ("\tABL = ");
	      displayArite (CAR (abl));
	      printf ("\n");
	    }
	  printf ("\n");
	}
      pEl++;
    }
}

void 
displayVirtuel (file,line)
     char * file;
     int line;
{
  ptype_list *i;
  chain_list *v;


  printf ("------------------------ display de mappSys->ref ----------------- %s:%d\n", file, line);

  for (i = mappSys->ref; i; i = CDR (i))
    {
      printf ("==> NIVEAU = %ld\n", i->TYPE);
      for (v = (chain_list *) (i->DATA); v; v = CDR (v))
	printVirtuel (v->DATA);
    }
  printf ("-------------------------------END---------------------------------\n");
}

void 
displayTHVirtuel  (file,line)
     char * file;
     int line;
{
  int i;
  static int czo_counter=0;
  pElemTH pEl = mappSys->virt_hash->pElem;

  czo_counter++;
  printf ("=============================================\n");
  printf ("        Lecture de mappSys->virt_hash   - %s:%d\n", file, line);

  printf ("length = %d\t		count = %d\n", mappSys->virt_hash->length, mappSys->virt_hash->count);
  printf ("=============================================\n");
  fprintf (stderr, "%d cellules generees\n", mappSys->virt_hash->count);
  for (i = 0; i < mappSys->virt_hash->length; i++)
    {
      if ((pEl->value != VIDETH) && (pEl->value != DELETETH))
	{
		printVirtuel (pEl->key); 
	  /*	printf(" -> printVirtuel : %d = %s\n",czo_counter, pEl->key); */

	}

      pEl++;
    }
}

void 
printVirtuel (nameVirt)
     char *nameVirt;
{

  virtuelList *virt = (virtuelList *) testObj (nameVirt, 'v');
  chain_list *sol = virt->solutions;

  printf (" -> printVirtuel : NOM = %s\n", nameVirt);
  printf ("\tdelais\t\t= %ld\n", virt->delais);
  printf ("\tsurf_cone\t= %ld\n", virt->surface_cone);
  printf ("\tcapa\t\t= %ld\n", virt->capa);
  printf ("\tdel_rc\t\t= %ld\n", virt->del_rc);
  printf ("\tsolutions\t= ");
  displayArite (CAR (virt->solutions)); 
  for (sol = CDR (virt->solutions); sol; sol = CDR (sol))
    {
      printf ("\n\t\t\t  ");
      displayArite (CAR (sol));
    }
  printf ("\n\tfanout\t\t= ");
  displayChain_liste (virt->fanout);
  printf ("\tcellule\t\t= %s\n", virt->cellule);
  if (virt->op_log)
  printf ("\top_log\t\t= %s\n", (virt->op_log)->nomlog);

}
