/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Standard Cell Mapping
 * Date    : 1991,92
 * Author  : Pascale Allegre
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



/* $Id: mapping_local.h,v 1.1.1.1 1998/10/02 15:25:09 alliance Exp $ */

#ifndef MAPPING_LOCAL_H
#define MAPPING_LOCAL_H

/* polarite */
#define NEG 0
#define POS 1
#define AUT 2


#define ARITE(exp) (int)CDR(CAR(exp))

#define createATOM(var) addchain((chain_list *)NULL,(void *)var)


typedef struct tabNum
  {
    struct tabNum *NEXT;
    char *name;
    num_list *ptNum;
  }
tabNumList;

typedef struct cellule
  {
    struct cellule *NEXT;
    char type;
    Alist *interface, *entrees;
    chain_list *sorties;
    num_list *delais;
    long surface;
    char *nomlog, *nom;
    tabNumList *delais_tech, *capa_tech;
    ptype_list *surface_tech, *resistance_tech;
    ptype_list *generic;
    struct biabl *val_abl;
    chain_list *envel;
    short util;
    pCircuit circuit;
    pNode BDD;
  }
cellList;

typedef struct regle
  {
    struct regle *NEXT;
    char *nom;
    chain_list *premisse, *conclusion;
    short condition, prof_premi;
    int util;
  }
regle_list;

typedef struct declencheur
  {
    struct declencheur *NEXT;
    char *name;
    regle_list *regles;
  }
declenList;

typedef struct s_sig
  {
    short type;
    char *name;
    long delais_in, delais, capa;
    int cout;
    chain_list *ablList, *util;
  }
sig;

typedef struct virtuel
  {
    chain_list *solutions;
    long delais, surface_cone, capa, del_rc;
    chain_list *fanout;
    cellList *op_log;
    char *cellule;
  }
virtuelList;

typedef struct mapping
  {
    cellList *ptcell;
    declenList *ptdecl;
    ptype_list *ref;
    short delta, coef, coutSig;
    pTH sig_hash, cell_hash, virt_hash, abl_hash;
  }
mapping;




extern mapping *mappSys;

extern int cptVirtuel;

extern char *nameUn, *nameZero, *nameHigh, *nameLow;
extern char *nameLatch, *nameBD, *nameBux, *nameBus, *nameNot1, *nameRest;



#ifndef __P
# if defined(__STDC__) ||  defined(__GNUC__)
#  define __P(x) x
# else
#  define __P(x) ()
# endif
#endif


/* Prototypes from mapping_arbres.c */

  extern    chain_list * genArbo __P((chain_list *expr, sig *signal, short niv, short init));
  extern    chain_list * remp __P((char *inst, chain_list *abl, short indSol));
  extern    chain_list * delFanout __P((chain_list *fanout, char *nom));
  extern          char * cherchePorte __P((chain_list *abl, int niv));
  extern           void  geleVirtuel __P((sig *signal, chain_list *abl));
  extern           void  genVar __P((virtuelList *inst, char *name, short niv, short init));
  extern           void  supp __P(());
  extern           void  suppVirtuel __P((char *nameInst, chain_list *liste));
  extern           void  delVirt __P((char *name));


/* Prototypes from mapping_compile.c */

  extern      cellList * initCellList __P((cellList *cellSuite, char *name, char type, befig_list *phys));
  extern       mapping * initMapping __P(());
  extern      cellList * endMapping __P(());
  extern          char * renameSig __P((char *name));
  extern            int  compileLibrary __P(());
  extern    chain_list * ablContraint __P((cellList *cell));


/* Prototypes from mapping_cout.c */

  extern           long  delais __P((chain_list *expr));
  extern           long  delRC __P((virtuelList *virt, chain_list *expr));
  extern           long  entrance __P((virtuelList *virtFan, char *name));
  extern           void  capa __P((char *name, virtuelList *virt));
  extern           long  surface __P((chain_list *expr));
  extern           long  surface_cone __P((virtuelList *cell));
  extern           long  Mapcout __P((chain_list *expr));
  extern           void  genDelta __P((sig *signal));


/* Prototypes from mapping_decomp.c */

  extern    chain_list * balance __P((chain_list *abl));
  extern    chain_list * decomp __P((chain_list *expr));
  extern    chain_list * searchObj __P((char *name, pTH pTab));
  extern          short  searchMax __P((chain_list *liste, short n));
  extern          char * eltNeutre __P((short operateur));
  extern    chain_list * decomposition __P((chain_list *abl, short n, short max, short plus, short inverse));
  extern    chain_list * balanceMax __P((chain_list *listeOp, chain_list *abl, int max, int *maxretour, short inverse));
  extern    chain_list * echantil __P((chain_list *exp, int delta));
  extern           void  permutVar __P((chain_list *exp, chain_list *envel));
  extern    chain_list * triVar __P((chain_list *exp));


/* Prototypes from mapping_display.c */

  extern           void  displayChain_liste __P((chain_list *liste));
  extern           void  displayArite __P((chain_list *exp));
  extern           void  displayBehMap __P((befig_list *befig, short nb));
  extern           void  nbreRegles __P(());
  extern           void  displayDeclencheurs __P(());
  extern           void  displayCell __P((cellList *c));
  extern           void  displayTHSignaux __P(());
  extern           void  displayVirtuel __P((char *file, int line));
  extern           void  displayTHVirtuel __P((char *file, int line));
  extern           void  printVirtuel __P((char *nameVirt));


/* Prototypes from mapping_generic.c */

  extern           long  searchGeneric __P((cellList *cell, char *cellPhy, char *name));
  extern            int  loadPolarity __P((cellList *cell));
  extern           long  resistIn __P((cellList *cell, char *nomPhy, char *nameIn));
  extern           long  delaiIn __P((cellList *cell, char *nomPhy, char *nameIn));
  extern      num_list * findDelaisGen __P((cellList *cell, char *nomPhy, int resist));
  extern      num_list * findCapaGen __P((cellList *cell, char *nomPhy));
  extern            int  resistance __P((cellList *cell, char *nomPhy));
  extern          char * findResistanceGen __P((cellList *cell));
  extern            int  recupGeneric __P((befig_list *beh, cellList *cell, int hauteur));
  extern           long  getDelai __P((cellList *cell, chain_list *expr));
  extern           void  triExpr __P((chain_list *expr, cellList *cell));
  extern    chain_list * memberChain_list __P((void *data, chain_list *chaine));
  extern          short  member __P((char *in, bereg_list *regList));
  extern         Alist * initPort __P((bepor_list *portList));
  extern         Alist * initIn __P((berin_list *inList, bereg_list *regList));
  extern    chain_list * initOut __P((beout_list *outList));
  extern            int  commut __P((chain_list *exp));
  extern           void  calculEnvel __P((chain_list *expr));
  extern    biabl_list * reCopyBiabl __P((biabl_list *biabl, short n));
  extern         Alist * triOrdreInterface __P((Alist *entrees, Alist *interface));
  extern           void  mirrorExpr __P((chain_list *expr));
  extern    chain_list * normMappingExpr __P((chain_list *expr));


/* Prototypes from mapping_mapper.c */

  extern           void  changeAblBeh __P(());
  extern           long  memberDelaypi __P((char *name, ptype_list *listeDel));
  extern           void  genBd __P(());
  extern           void  genDecomp __P((ptype_list *ref));
  extern           void  genInfer __P((ptype_list *ref));
  extern    chain_list * signauxFils __P((chain_list *expr));
  extern           void  traiteExpLog __P((sig *inst, short coutSig));
  extern    chain_list * traiteBus __P((biabl_list *biabl));
  extern    chain_list * traiteReg __P((biabl_list *biabl));
  extern           void  parcours __P((sig *signal, chain_list *abl));
  extern    ptype_list * evalue __P((chain_list *expr1));
  extern           void  libere __P((ptype_list *sol, ptype_list *head));
  extern    chain_list * polarite __P((chain_list *expr));
  extern           void  mapper __P(());


/* Prototypes from mapping_mbk.c */

  extern           void  maxDepthExpr __P((befig_list *beh, chain_list *expr, pTH nCellTH, pTH nConeTH, int *numberCell, int *numberCone));
  extern           void  maxDepthBst __P((befig_list *beh, int *numberCell, int *numberCone));
  extern            int  Surface __P((char *name));
  extern          char * nameInst __P((char *name));
  extern    lofig_list * genMBKBeh __P((befig_list *beh, short capa, short sc, char *nomFig));
  extern    chain_list * recupLosig __P((pTH losigHash, cellList *cell, char *name, chain_list *abl));


/* Prototypes from mapping_moteur.c */

  extern           void  ssExprArite __P((chain_list *prem, Alist *varRegle));
  extern    chain_list * decArite __P((regle_list *regle, chain_list *expr, Alist *var));
  extern    chain_list * backTrack __P((chain_list *expr));
  extern          short  filtre __P((chain_list *prem, chain_list *expr, Alist *tabVar));
  extern    chain_list * moteur __P((regle_list *regle, chain_list *expr));
  extern            int  permutOp __P((chain_list *expr1, chain_list *expr2, short tri, short profPrem, short prof));
  extern    chain_list * remplace __P((chain_list *conc, Alist *tabVar));
  extern          short  verifExpr __P((chain_list *prem, chain_list *exp));


/* Prototypes from mapping_oper.c */

  extern           void  createTabOper __P(());
  extern           void  addOper __P((char *name));
  extern            int  searchNumOper __P((char *name));
  extern          char * searchCharOper __P((int num));
  extern      cellList * operToCell __P((chain_list *expr));
  extern    chain_list * invExpr __P((chain_list *abl));


/* Prototypes from mapping_regles.c */

  extern          short  simple __P((chain_list *exp));
  extern          short  simpleInverse __P((chain_list *exp));
  extern          char * cellOpose __P((char *nom));
  extern          short  cellXor __P((cellList *cell));
  extern          short  invOper __P((short oper));
  extern          short  morgOper __P((short oper));
  extern    chain_list * moinsNot __P((chain_list *exp));
  extern            int  numberOcVarExpr __P((chain_list *exp, char *name));
  extern    chain_list * simplifNot __P((chain_list *exp));
  extern    chain_list * appendChain_list __P((chain_list *l1, chain_list *l2));
  extern    chain_list * combin __P((chain_list *list));
  extern    chain_list * solutions __P((chain_list *exp, short signe, short op));
  extern    chain_list * fct __P((chain_list *exp));
  extern          short  findEqual __P((chain_list *abl, chain_list *liste));
  extern    chain_list * triCommut __P((chain_list *liste));
  extern    chain_list * consExpr __P((int oper, Alist *e));
  extern    chain_list * regOpposee __P((cellList *cell));
  extern    chain_list * opposees __P((chain_list *liste));
  extern    chain_list * conclu __P((chain_list *prems, chain_list *conc));
  extern           void  acons __P((Alist *aListe, char *in, char *var));
  extern          char * cassqVal __P((Alist *aListe, char *val));
  extern           void  alistEntrees __P((chain_list *premisse, Alist *entrees));
  extern    chain_list * alistPrem __P((chain_list *premisse, Alist *entrees));
  extern    chain_list * alistConc __P((chain_list *conc, Alist *entrees));
  extern           void  reportNot __P((chain_list *liste, cellList *cell));
  extern    chain_list * reglesXor __P((cellList *cell));
  extern    chain_list * reglesCellLogique __P((cellList *cell));
  extern            int  fonc __P((cellList *cell));
  extern           void  regleOperVHDL __P((cellList *cell));
  extern          short  testOperVHDL __P((cellList *cell));
  extern          char * varF __P((short n, char *var));
  extern           void  parcoursConclu __P((chain_list *conclusion, Alist *entrees));
  extern           void  filtrePremisse __P((chain_list *prem, cellList *cell, ptype_list *pTList));
  extern    chain_list * filtreConclusion __P((chain_list *conc, cellList *cell));
  extern           void  argsSsexpr __P((chain_list *args, cellList *cell, ptype_list *pTList));
  extern           void  filtrage __P((chain_list *Regle, cellList *cell));
  extern    chain_list * reglesCellRegistre __P((cellList *cell, char type));
  extern    chain_list * reglesCellBus __P((cellList *cell));
  extern    chain_list * remplaceNot __P((chain_list *abl, chain_list *conc));
  extern    chain_list * findBestNot __P((regle_list *regles));
  extern           void  reglesCellules __P(());
  extern    regle_list * genereObj __P((chain_list *Regle));
  extern           void  addDeclen __P((regle_list *regle, char *nameDeclen));


/* Prototypes from mapping_util.c */

  extern    ptype_list * minPtype __P((ptype_list *liste));
  extern           void  displayAlist __P((Alist *ptTL));
  extern           void  initAlist __P((chain_list *expr, Alist *tab));
  extern          char * cassq __P((Alist *ptTL, char *name));
  extern           void  resetAlist __P((Alist *ptTL, short type));
  extern           void  freeAlist __P((Alist *ptTL));
  extern          char * exprToCharRed __P((chain_list *expr));
  extern           void  createRef __P((long niv));
  extern           void  addRef __P((char *cell, long niv));
  extern           void  delRef __P((char *cell));
  extern    ptype_list * searchRefNiv __P((long niv));
  extern    chain_list * searchRefName __P((char *name, long niv));
  extern    ptype_list * searchRef __P((char *name));
  extern    chain_list * expanse __P((chain_list *abl));
  extern           void  statistiques __P(());
  extern           void  createAux __P((char *nom, virtuelList *ptVirt, char type));
  extern           void  newBefig __P(());
  extern           void  sort2Expr __P((chain_list *expr, int (*func1)(), int (*func2)(), int dir1, int dir2));
  extern            int  funcNorm2 __P((chain_list *expr));

#endif
