/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Netlist Optimisation
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: netop_enveloppe.c,v 1.1.1.1 1998/10/02 15:25:10 alliance Exp $"

#include <stdlib.h>

#include MUT_H
#include BEH_H
#include LOG_H

#include MLO_H
#include MPH_H
#include "lax_param.h"
#include "sl_local.h"

#include "mapping_local.h"
#include "netop_local.h"

/*-----------------------------------------------------------------------------
inListe  : renvoie si data fait partie de la chain_list (si oui = 1)
-------------------------------------------------------------------------------
retour   : un short
-----------------------------------------------------------------------------*/
short 
inListe (liste, data)
     chain_list *liste;
     char *data;
{
  chain_list *auxiliaire = liste;

  while (auxiliaire)
    {
      if ((char *) auxiliaire->DATA == data)
	return 1;
      auxiliaire = auxiliaire->NEXT;
    };
  return 0;
}

/*-----------------------------------------------------------------------------
inListeAtom : renvoie si data fait partie de la chain_list d'atomes (si oui = 1)
-------------------------------------------------------------------------------
retour      : un short
-----------------------------------------------------------------------------*/
short 
inListeAtom (liste, data)
     chain_list *liste;
     char *data;
{
  chain_list *auxiliaire = liste;

  while (auxiliaire)
    {
      if (inListe ((chain_list *) auxiliaire->DATA, data))
	return 1;
      auxiliaire = auxiliaire->NEXT;
    };
  return 0;
}

/*-----------------------------------------------------------------------------
completeListe : ajoute a l'enveloppe les entrees n'apparaissant pas encore
-------------------------------------------------------------------------------
retour        : chain_list de chain_list
-----------------------------------------------------------------------------*/
chain_list *
completeListe (liste, entrees)
     chain_list *liste;
     Alist *entrees;
{
  Alist *in;

  for (in = entrees; in; in = in->NEXT)
    {
      if (!inListeAtom (liste, in->name))
	liste = addchain (liste, addchain ((chain_list *) NULL, (void *) in->name));
    }
  return liste;
}

/*---------------------------------------------------------------------------
longueur : renvoie la longueur d'une chain_list = nbre d'elements
-----------------------------------------------------------------------------
retour   : short
---------------------------------------------------------------------------*/
short 
longueur (liste)
     chain_list *liste;
{
  short length = 0;

  while (liste)
    {
      length++;
      liste = liste->NEXT;
    }
  return length;
}

/*---------------------------------------------------------------------------
egalListe : renvoie si tous les elts de liste1 sont dans liste2
-----------------------------------------------------------------------------
retour    : si oui = 1
---------------------------------------------------------------------------*/
short 
egalListe (liste1, liste2)
     chain_list *liste1, *liste2;
{
  chain_list *aux;

  for (aux = liste1; aux; aux = aux->NEXT)
    if (!inListe (liste2, (char *) aux->DATA))
      return 0;
  return 1;
}

/*---------------------------------------------------------------------------
intersection : renvoie si l'intersection de 2 chain_list est vide
-----------------------------------------------------------------------------
retour       : si vide = 0
---------------------------------------------------------------------------*/
short 
intersection (liste1, liste2)
     chain_list *liste1, *liste2;
{
  chain_list *aux;

  for (aux = liste1; aux; aux = aux->NEXT)
    if (inListe (liste2, (char *) aux->DATA))
      return 1;
  return 0;
}

/*---------------------------------------------------------------------------
testElimineListe : retourne si la liste doit etre elimine de l'enveloppe
-----------------------------------------------------------------------------
retour           : 1 = elimine ou 0
---------------------------------------------------------------------------*/
short 
testElimineListe (liste1, liste2)
     chain_list *liste1, *liste2;
{
  short elimine = 0;

  if (intersection (liste1, liste2) != 0)	/* intersection non vide */
    if (longueur (liste1) > longueur (liste2))
      {
	if (egalListe (liste1, liste2))
	  elimine = 1;
      }
    else
      {
	if (egalListe (liste2, liste1))
	  elimine = 1;
      }
  return elimine;
}

/*---------------------------------------------------------------------------
elimineListe : elimine un element d'une chain_list
-----------------------------------------------------------------------------
retour       : une chain_list de chain_list ...
---------------------------------------------------------------------------*/
chain_list *
elimineListe (elt, liste)
     chain_list *elt, *liste;
{
  if (elt == liste)
    return liste->NEXT;
  else
    {
      chain_list *p1 = liste;
      chain_list *p2 = liste->NEXT;

      while (p2 && (p2 != elt))
	{
	  p2 = p2->NEXT;
	  p1 = p1->NEXT;
	}
      p1->NEXT = p2->NEXT;
      p2->NEXT = (chain_list *) NULL;
      freechain (p2);
      return liste;
    }
}

/*---------------------------------------------------------------------------
filtreListeAtom : elimine les listes redondantes
                  dans le calcul de l'enveloppe d'une porte
-----------------------------------------------------------------------------
retour          : une chain_list de chain_list ...
                  chaque ss-chain_list est une liste
                  de connecteurs interchangeables
---------------------------------------------------------------------------*/
chain_list *
filtreListeAtom (liste)
     chain_list *liste;
{
  chain_list *aux = liste, *queue;

  while (aux)
    {
      short elimine = 0;
      chain_list *next = aux->NEXT;

      for (queue = aux->NEXT; queue; queue = queue->NEXT)
	if (testElimineListe ((chain_list *) aux->DATA, (chain_list *) queue->DATA))
	  elimine = 1;
      if (elimine)
	liste = elimineListe (aux, liste);
      aux = next;
    }
  return liste;
}

/*---------------------------------------------------------------------------
listeAtom : construit une liste de sous-listes d'atomes interchangeables
            d'apres l'ABL de la fonction logique de la porte
-----------------------------------------------------------------------------
retour    : une chain_list de chain_list ...
---------------------------------------------------------------------------*/
chain_list *
listeAtom (expr)
     chain_list *expr;
{
  chain_list *retour = (chain_list *) NULL;
  chain_list *inter1 = (chain_list *) NULL;
  chain_list *inter2 = (chain_list *) NULL;

  if (!ATOM (expr))
    {
      while ((expr = CDR (expr)))
	if (ATOM (CAR (expr)))
	  /* liste de atomes se trouvant au meme niveau de l'expression */
	  inter1 = addchain (inter1, (void *) VALUE_ATOM (CAR (expr)));
	else
	  {
	    /* appel recursif sur les ss-expressions */
	    inter2 = listeAtom (CAR (expr));
	    while (inter2)
	      {
		/* mise a plat de l'appel recursif */
		retour = addchain (retour, (void *) inter2->DATA);
		inter2 = inter2->NEXT;
	      }
	  }
      if (inter1 && inter1->NEXT)
	retour = addchain (retour, (void *) inter1);
    }
  return retour;
}

/*---------------------------------------------------------------------------
displayEnvel :
-----------------------------------------------------------------------------
retour    : affichage
---------------------------------------------------------------------------*/
void 
displayEnvel (liste)
     chain_list *liste;
{
  chain_list *lst, *l;

  for (lst = liste; lst; lst = lst->NEXT)
    {
      printf ("(");
      for (l = (chain_list *) lst->DATA; l; l = l->NEXT)
	printf ("%s ", (char *) l->DATA);
      printf (")");
    }
  printf ("\n");
}
