/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Netlist Optimisation
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: netop_load_global.c,v 1.1.1.1 1998/10/02 15:25:10 alliance Exp $"

#include <stdio.h>

#include MUT_H
#include BEH_H
#include MLO_H
#include MPH_H
#include LOG_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"
#include "netop_local.h"


int cptBuffer;
int cptPower;
int nbGates;

long nbChgt;

pTH RAZTH;
pTH assignTH;





/* *********************** Assignation de Fanin ***************************** */

/*---------------------------------------------------------------------------
displayCombi :
-----------------------------------------------------------------------------
retour     :
---------------------------------------------------------------------------*/
void 
displayCombi (liste)
     ptype_list *liste;
{
  ptype_list *l, *data;

  printf ("displayCombi :\n");
  for (l = liste; l; l = l->NEXT)
    {
      printf ("   %ld =", l->TYPE);
      for (data = (ptype_list *) l->DATA; data; data = data->NEXT)
	printf (" %s", (char *) ((locon_list *) data->DATA)->NAME);
      printf ("\n");
    }
}

/*---------------------------------------------------------------------------
moins :
-----------------------------------------------------------------------------
retour     :
---------------------------------------------------------------------------*/
ptype_list *
moins (liste, data)
     ptype_list *liste;
     void *data;
{
  ptype_list *l, *retour = (ptype_list *) NULL;

  for (l = liste; l; l = l->NEXT)
    if (l->DATA != data)
      retour = addptype (retour, (long) 0, l->DATA);

  return retour;
}

/*---------------------------------------------------------------------------
addAll :
-----------------------------------------------------------------------------
retour     :
---------------------------------------------------------------------------*/
ptype_list *
addAll (liste, data)
     ptype_list *liste;
     void *data;
{
  ptype_list *l;

  for (l = liste; l; l = l->NEXT)
    l->DATA = (void *) addptype ((ptype_list *) l->DATA, (long) 0, data);

  return liste;
}

/*---------------------------------------------------------------------------
combinaison :
-----------------------------------------------------------------------------
retour     :
---------------------------------------------------------------------------*/
ptype_list *
combinaison (liste)
     ptype_list *liste;
{
  if (liste->NEXT)
    {
      ptype_list *l, *retour = (ptype_list *) NULL, *solution = (ptype_list *) NULL;
      ptype_list *sol;

      for (l = liste; l; l = l->NEXT)
	{
	  ptype_list *listemoins = moins (liste, l->DATA);

	  solution = addAll (combinaison (listemoins), l->DATA);
/*
   freeptype(listemoins);
 */
	  for (sol = solution; sol; sol = sol->NEXT)
	    retour = addptype (retour, (long) 0, (void *) sol->DATA);
	}
      return retour;
    }
  else
    return addptype ((ptype_list *) NULL, (long) 0, (void *) liste);
}

/*---------------------------------------------------------------------------
assignFaninLofig : assignation de fanin
-----------------------------------------------------------------------------
retour      :
---------------------------------------------------------------------------*/
void 
assignFaninLofig (lofig)
     lofig_list *lofig;
{
  loins_list *inst;

  printf ("ERROR : les delais ne sont pas remis a jour ...\n");

  nbChgt = 0;
  assignTH = createTH (1000);

  printf ("\nPin assignment\n");
  for (inst = lofig->LOINS; inst; inst = inst->NEXT)
    if (searchTH (assignTH, inst->INSNAME) == EMPTYTH)
      assignFaninEx (inst, 1);	/* 1 = meilleur */

  destroyTH (assignTH);

  printf ("  ==> %ld new assignments\n", nbChgt);
}

/*---------------------------------------------------------------------------
assignFaninEx : assignation de fanin sur une instance (methode exhaustive)
-----------------------------------------------------------------------------
retour      : les signaux sont eventuellemnt echanges sur les connecteurs
              de l'instance a traiter
---------------------------------------------------------------------------*/
void 
assignFaninEx (inst, meilleur)
     loins_list *inst;
     short meilleur;
{
  locon_list *con;
  int polarite = cellLoadPolarity (inst->FIGNAME);
  long maxRetUp = 0, maxRetDw = 0;
  losig_list *sigOut;
  long sCin;

/*
   printf("assignFaninEx of %s [%s]\n", inst->INSNAME, inst->FIGNAME);
 */

  /* Appel recursif sur les entrees */
  for (con = inst->LOCON; con; con = con->NEXT)
    if (con->DIRECTION == IN &&
	!isvdd (con->NAME) && !isvss (con->NAME))
      {
	int user = searchTH (genUpTH, getsigname (con->SIG));

	if (user != EMPTYTH)
	  {
	    loins_list *inInst = (loins_list *) user;

	    if (searchTH (assignTH, inInst->INSNAME) == EMPTYTH)
	      assignFaninEx (inInst, meilleur);
	  }
      }

/*
   printf("  assignFaninEx of %s [%s]\n", inst->INSNAME, inst->FIGNAME);
 */

  if (cellLoadType (inst->FIGNAME) == REG || cellLoadType (inst->FIGNAME) == BUS)
    {
      /* bus ou registre */
      return;
    }

  /* table de hachage des instances deja traitees */
  addTH (assignTH, inst->INSNAME, 1);

  sigOut = ((locon_list *) searchTH (loconOutTH, inst->INSNAME))->SIG;
  sCin = (long) searchTH (capaTH, getsigname (sigOut));

  if (polarite == AUT)
    {
      /*  remise a jour des retards sur l'instance */
      for (con = inst->LOCON; con; con = con->NEXT)
	{
	  if (con->DIRECTION == IN &&
	      !isvdd (con->NAME) && !isvss (con->NAME))
	    {
	      long retardGenUp = (long) getptype (con->SIG->USER, (long) RETARDUP)->DATA;
	      long retardGenDw = (long) getptype (con->SIG->USER, (long) RETARDDW)->DATA;

	      long retardInstUp = MAX (retardGenUp + getDel (inst, con, sigOut, "hh", 0),
			 retardGenDw + getDel (inst, con, sigOut, "lh", 0));
	      long retardInstDw = MAX (retardGenUp + getDel (inst, con, sigOut, "hl", 0),
			 retardGenDw + getDel (inst, con, sigOut, "ll", 0));

	      if (retardInstDw > maxRetDw)
		maxRetDw = retardInstDw;
	      if (retardInstUp > maxRetUp)
		maxRetUp = retardInstUp;
	    }
	}			/* fin parcours connecteurs */

      if (sigOut->USER && getptype (sigOut->USER, (long) RETARDUP))
	{
	  /* mise a jour des retards */
	  sigOut->USER = delptype (sigOut->USER, (long) RETARDUP);
	  sigOut->USER = delptype (sigOut->USER, (long) RETARDDW);
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDUP, (void *) maxRetUp);
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDDW, (void *) maxRetDw);
	}
      else
	{
	  printf ("ERROR : No delay for instance %s on output signal %s\n",
		  inst->INSNAME, (char *)sigOut->NAMECHAIN->DATA);
	  exit (1);
	}
    }
  else
    {				/* polarite POS ou NEG */
      int changement = 0;
      long oldUp = (long) getptype (sigOut->USER, (long) RETARDUP)->DATA;
      long oldDw = (long) getptype (sigOut->USER, (long) RETARDDW)->DATA;
      chain_list *enveloppe = cellLoadEnvel (inst->FIGNAME), *env;

/*
   displayEnvel(enveloppe);
 */

      if (NO_TRACE >= 2)
	{
	  printf (" Traitement of %s [%s], input capacitances sum = %ld\n",
		  inst->INSNAME, inst->FIGNAME, sCin);
	  printf ("       delay of %s [%s] - UP = %ld - DOWN = %ld - Cin Sum = %ld\n",
		  inst->INSNAME, inst->FIGNAME, oldUp, oldDw, sCin);
	}

      /* Parcours des connecteurs de l'enveloppe de l'instance a traiter */
      for (env = enveloppe; env; env = env->NEXT)
/*
   if (((chain_list *)env->DATA)->NEXT)
 */
	{
	  /* on a plus d'une entree dans la sous-liste d'enveloppe */
	  ptype_list *chainSig = NULL, *lstSig;
	  ptype_list *chainCon = NULL, *lstCon;
	  ptype_list *allCombi = NULL, *ssLstCon;
	  ptype_list *sol;
	  long max = 0, min = 0;
	  short nbSol = 0;
	  ptype_list *liste = getptype (sigOut->USER, (long) LOFIGCHAIN);
	  chain_list *lstConnect;
	  long minSlack = 999999;
	  long nbPolOptimUp = 0, nbPolOptimDw = 0;
	  int trans;

	  /* on recherche les connecteurs IN faisant partie de l'enveloppe */
	  for (con = inst->LOCON; con; con = con->NEXT)
	    {
	      if (con->DIRECTION == IN &&
		  !isvdd (con->NAME) && !isvss (con->NAME) &&
		  inListe ((chain_list *) env->DATA, con->NAME))
		{
		  long delaiConUp, delaiConDw;
		  long delUp, delDw;

		  /* calcul du Tp+RC Up et Dw pour le connecteur */
		  switch (polarite)
		    {
		    case POS:
		      delaiConUp = cellLoadTp (inst->FIGNAME, con->NAME, HH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 sCin) / 1000;
		      delaiConDw = cellLoadTp (inst->FIGNAME, con->NAME, LL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 sCin) / 1000;
		      break;
		    case NEG:
		      delaiConUp = cellLoadTp (inst->FIGNAME, con->NAME, LH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 sCin) / 1000;
		      delaiConDw = cellLoadTp (inst->FIGNAME, con->NAME, HL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 sCin) / 1000;
		      break;
		    }

		  if (NO_TRACE >= 3)
		    printf ("    on connector %s : intrinsic delay UP = %ld - intrinsic delay DOWN = %ld\n",
			    con->NAME, delaiConUp, delaiConDw);

		  /* sauvegarde des Tp+RC par connecteurs dans USER */
		  con->USER = addptype (con->USER, (long) LOCONUP, (void *) delaiConUp);
		  con->USER = addptype (con->USER, (long) LOCONDW, (void *) delaiConDw);

		  /* Liste des signaux et des connecteurs de l'enveloppe */
		  chainCon = addptype (chainCon, (long) 0, (void *) con);
		  chainSig = addptype (chainSig, (long) 0, (void *) con->SIG);

		  /* on recupere les retards sur les signaux */
		  delUp = (long) getptype (con->SIG->USER, (long) RETARDUP)->DATA;
		  delDw = (long) getptype (con->SIG->USER, (long) RETARDDW)->DATA;

		  if (NO_TRACE >= 3)
		    printf ("    on signal %s : delay UP = %ld - delay DOWN = %ld\n",
			  (char *) con->SIG->NAMECHAIN->DATA, delUp, delDw);

		}		/* if connecteur IN et dans enveloppe */
	    }			/* for sur les connecteurs de l'instance */

	  allCombi = combinaison (chainCon);

	  /* parcours des permutations des connecteurs de l'enveloppe */
	  for (lstCon = allCombi; lstCon; lstCon = lstCon->NEXT)
	    {
	      long maxUp = 0, maxDw = 0;

/*
 */
	      if (NO_TRACE >= 2)
		printf ("     Solution number %d\n", ++nbSol);

	      for (lstSig = chainSig, ssLstCon = (ptype_list *) lstCon->DATA;
		   lstSig;
		   lstSig = lstSig->NEXT, ssLstCon = ssLstCon->NEXT)
		{
		  long delaiInstUp, delaiInstDw;

		  /* on recupere les retards sur les signaux */
		  long delaiGenUp =
		  (long) getptype (((losig_list *) (lstSig->DATA))->USER,
				   (long) RETARDUP)->DATA;
		  long delaiGenDw =
		  (long) getptype (((losig_list *) (lstSig->DATA))->USER,
				   (long) RETARDDW)->DATA;

		  /* on recupere les Tp+RC Up et Dw des connecteurs */
		  long delaiConUp =
		  (long) getptype (((locon_list *) (ssLstCon->DATA))->USER,
				   (long) LOCONUP)->DATA;
		  long delaiConDw =
		  (long) getptype (((locon_list *) (ssLstCon->DATA))->USER,
				   (long) LOCONDW)->DATA;

		  /* pour le connecteur et le signal calcul de :
		     MAX(delaiGenUp + Tp+RC Up, delaiGenDw + Tp+RC Dw) */
		  switch (polarite)
		    {
		    case POS:
		      delaiInstUp = delaiGenUp + delaiConUp;
		      delaiInstDw = delaiGenDw + delaiConDw;
		      break;
		    case NEG:
		      delaiInstUp = delaiGenDw + delaiConUp;
		      delaiInstDw = delaiGenUp + delaiConDw;
		      break;
		    }
		  if (NO_TRACE >= 3)
		    printf ("       signal %s on connector %s : UP = %ld - DOWN = %ld\n",
		    (char *) ((losig_list *) lstSig->DATA)->NAMECHAIN->DATA,
			    ((locon_list *) ssLstCon->DATA)->NAME,
			    delaiInstUp, delaiInstDw);
		  if (maxUp <= delaiInstUp)
		    maxUp = delaiInstUp;
		  if (maxDw <= delaiInstDw)
		    maxDw = delaiInstDw;
		}
	      lstCon->TYPE = (long) addptype ((ptype_list *) NULL, maxUp, maxDw);
	    }			/* fin parcours des permutations */

/*
   displayCombi(allCombi);
 */

	  if (!liste)
	    {
	      printf ("ASSIGN FANIN : No type LOFIGCHAIN for %s\n", inst->INSNAME);
	      exit (1);
	    }

	  /* recherche du fanout avec le + petit slack */
	  if (0)
	    for (lstConnect = (chain_list *) liste->DATA; lstConnect; lstConnect = lstConnect->NEXT)
	      {
		locon_list *conAux = (locon_list *) lstConnect->DATA;

		if (conAux->ROOT != (void *) inst)
		  {
		    if ((conAux->TYPE == INTERNAL) && (conAux->DIRECTION == IN))
		      {
			loins_list *cellOut = (loins_list *) conAux->ROOT;
			long slackUp = getInfoInst (cellOut, SLACKUP);
			long slackDw = getInfoInst (cellOut, SLACKDW);

			if (slackUp == minSlack)
			  {
			    if (cellLoadPolarity (cellOut->FIGNAME) == POS)
			      nbPolOptimUp++;
			    else if (cellLoadPolarity (cellOut->FIGNAME) == NEG)
			      nbPolOptimDw++;
			  }
			else if (slackUp < minSlack)
			  {
			    minSlack = slackUp;
			    if (cellLoadPolarity (cellOut->FIGNAME) == POS)
			      {
				nbPolOptimUp = 1;
				nbPolOptimDw = 0;
			      }
			    else if (cellLoadPolarity (cellOut->FIGNAME) == NEG)
			      {
				nbPolOptimUp = 0;
				nbPolOptimDw = 1;
			      }
			    else
			      {
				nbPolOptimUp = 0;
				nbPolOptimDw = 0;
			      }
			  }
			if (slackDw == minSlack)
			  {
			    if (cellLoadPolarity (cellOut->FIGNAME) == POS)
			      nbPolOptimDw++;
			    else if (cellLoadPolarity (cellOut->FIGNAME) == NEG)
			      nbPolOptimUp++;
			  }
			else if (slackDw < minSlack)
			  {
			    minSlack = slackDw;
			    if (cellLoadPolarity (cellOut->FIGNAME) == POS)
			      {
				nbPolOptimUp = 0;
				nbPolOptimDw = 1;
			      }
			    else if (cellLoadPolarity (cellOut->FIGNAME) == NEG)
			      {
				nbPolOptimUp = 1;
				nbPolOptimDw = 0;
			      }
			    else
			      {
				nbPolOptimUp = 0;
				nbPolOptimDw = 0;
			      }
			  }
		      }
		    else
		      {
			if (conAux->TYPE != EXTERNAL)
			  printf ("ASSIGN FANIN : Unknown connector '%s' of %s[%s]\n",
				  conAux->NAME, ((loins_list *) conAux->ROOT)->INSNAME,
				  ((loins_list *) conAux->ROOT)->FIGNAME);
		      }
		  }
	      }

	  if (nbPolOptimUp == nbPolOptimDw)
	    trans = AUT;
	  else if (nbPolOptimUp < nbPolOptimDw)
	    trans = POS;
	  else
	    trans = NEG;

/*
   if (trans == POS)
   printf("Pour %s on cherche a optimiser retard UP\n", inst->INSNAME);
   else if (trans == NEG)
   printf("Pour %s on cherche a optimiser retard DOWN\n", inst->INSNAME); 
   else 
   printf("Pour %s on cherche a optimiser retard AUTRE\n", inst->INSNAME); 
 */

	  if (meilleur)
	    {
	      max = 999999;
	      min = 999999;
	    }
	  else
	    {
	      max = 0;
	      min = 0;
	    }

	  for (lstCon = allCombi; lstCon; lstCon = lstCon->NEXT)
	    if (meilleur)
	      {
		if (trans == POS)
		  {
		    if (((ptype_list *) lstCon->TYPE)->TYPE < max)
		      {
			max = ((ptype_list *) lstCon->TYPE)->TYPE;
			min = (long) ((ptype_list *) lstCon->TYPE)->DATA;
			sol = lstCon;
		      }
		  }
		else if (trans == NEG)
		  {
		    if ((long) ((ptype_list *) lstCon->TYPE)->DATA < max)
		      {
			max = (long) ((ptype_list *) lstCon->TYPE)->DATA;
			min = (long) ((ptype_list *) lstCon->TYPE)->TYPE;
			sol = lstCon;
		      }
		  }
		else
		  {
		    /* trans == AUT */
		    ptype_list *ret = (ptype_list *) lstCon->TYPE;
		    long retUp = ret->TYPE;
		    long retDw = (long) ret->DATA;

		    if ((MAX (retUp, retDw)) < max)
		      {
			max = (MAX (retUp, retDw));
			min = (MIN (retUp, retDw));
			sol = lstCon;
		      }
		    else if ((max == (MAX (retUp, retDw))) &&
			     ((MIN (retUp, retDw)) < min))
		      {
			min = (MIN (retUp, retDw));
			sol = lstCon;
		      }
		  }
	      }
	    else
	      /* a faire : pire cas */
	    if (lstCon->TYPE > max)
	      {
		max = lstCon->TYPE;
		sol = lstCon;
	      }
	  freeptype (chainCon);
	  chainCon = (ptype_list *) sol->DATA;

	  /* Mise en oeuvre de la solution choisie */
	  for (lstSig = chainSig, lstCon = chainCon;
	       lstSig; lstSig = lstSig->NEXT, lstCon = lstCon->NEXT)
	    {
	      long delaiGenUp, delaiGenDw;
	      long delaiInstUp, delaiInstDw;

	      if (((locon_list *) lstCon->DATA)->SIG != (losig_list *) lstSig->DATA)
		{
		  ((locon_list *) lstCon->DATA)->SIG = (losig_list *) lstSig->DATA;
		  if (changement == 0)
		    {
		      if (NO_TRACE >= 1)
			printf ("    new assignment on %s[%s] :\n",
				inst->INSNAME, inst->FIGNAME);
		      changement = 1;
		    }
		}
	      delaiGenUp = (long) (getptype (((losig_list *) lstSig->DATA)->USER, (long) RETARDUP)->DATA);
	      delaiGenDw = (long) (getptype (((losig_list *) lstSig->DATA)->USER, (long) RETARDDW)->DATA);
	      delaiInstUp = (long) (getptype (((locon_list *) lstCon->DATA)->USER, (long) LOCONUP)->DATA);
	      delaiInstDw = (long) (getptype (((locon_list *) lstCon->DATA)->USER, (long) LOCONDW)->DATA);

	      switch (polarite)
		{
		case POS:
		  delaiInstUp = delaiGenUp + delaiInstUp;
		  delaiInstDw = delaiGenDw + delaiInstDw;
		  break;
		case NEG:
		  delaiInstUp = delaiGenDw + delaiInstUp;
		  delaiInstDw = delaiGenUp + delaiInstDw;
		  break;
		}
	      if (NO_TRACE >= 3)
		printf ("     -> signal %s on connector %s\n",
			(char *)((locon_list *) lstCon->DATA)->SIG->NAMECHAIN->DATA,
			((locon_list *) lstCon->DATA)->NAME);

	      if (delaiInstDw > maxRetDw)
		maxRetDw = delaiInstDw;
	      if (delaiInstUp > maxRetUp)
		maxRetUp = delaiInstUp;
	    }
/*
   for(pListe = allCombi; pListe; pListe = pListe->NEXT)
   {
   freeptype((ptype_list *)pListe->DATA);
   pListe->DATA = NULL;
   }
 */
	  freeptype (allCombi);
	  freeptype (chainSig);
	}			/* fin du parcours de l'enveloppe */

      /* sauvegarde du nouveau retard de l'instance */
      if (sigOut->USER && getptype (sigOut->USER, (long) RETARDUP))
	{
	  sigOut->USER = delptype (sigOut->USER, (long) RETARDUP);
	  sigOut->USER = delptype (sigOut->USER, (long) RETARDDW);
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDUP, (void *) maxRetUp);
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDDW, (void *) maxRetDw);
	}
      else
	{
	  printf ("ERROR : No delay for instance %s on output signal %s\n",
		  inst->INSNAME, (char *)sigOut->NAMECHAIN->DATA);
	  exit (1);
	}
      if (NO_TRACE >= 2)
	printf ("   new delay of %s [%s] - UP = %ld - DOWN = %ld - Cin Sum = %ld\n",
		inst->INSNAME, inst->FIGNAME, maxRetUp, maxRetDw, sCin);

      nbChgt = nbChgt + changement;
    }				/* instance de polarite POS ou NEG */
}				/* traitement de l'instance */
/* ******************** Fin de l'assignation de Fanin *********************** */

/*---------------------------------------------------------------------------
triFanout  : tri le fanout de inst en slack croissant (sauf la porte except)
-----------------------------------------------------------------------------
retour : ptype_list triee en ordre croissant du slack minimum
---------------------------------------------------------------------------*/
ptype_list *
triFanout (inst, conIn, sens, except)
     loins_list *inst;
     locon_list *conIn;
     int sens;
     loins_list *except;
{
  int userOut;
  locon_list *conOut;
  losig_list *sigOut;
  ptype_list *liste;
  chain_list *lstCon;
  ptype_list *lstFanout = (ptype_list *) NULL;

  if (inst)
    {
      userOut = searchTH (loconOutTH, inst->INSNAME);

      if (userOut != EMPTYTH)
	conOut = (locon_list *) userOut;
    }
  else if (conIn)
    conOut = conIn;
  else
    {
      if (inst)
	printf ("ERROR triFanout : No type LOCONOUT for %s\n", inst->INSNAME);
      exit (1);
    }

  sigOut = conOut->SIG;
  liste = getptype (sigOut->USER, (long) LOFIGCHAIN);

  for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
    {
      locon_list *conAux = (locon_list *) lstCon->DATA;

      if ((inst && (conAux->ROOT != (void *) inst)) || (conIn && (conAux != conIn)))
	{
	  if (conAux->TYPE == INTERNAL)
	    /* connecteur interne */
	    if (conAux->DIRECTION == IN)
	      {
		loins_list *cellOut = (loins_list *) conAux->ROOT;

		if (cellOut != except)
		  {
		    /* portes attaquees par l'instance a traiter */
		    long slack = MIN (getInfoInst (cellOut, SLACKUP),
				      getInfoInst (cellOut, SLACKDW));
/*
   printf("Fanout de %s = %s (slack MIN = %d)\n",
   inst->INSNAME, cellOut->INSNAME, slack);
 */
		    lstFanout = tripType (lstFanout, slack, cellOut, sens);
		  }
	      }
	}
    }

  return lstFanout;
}

/*---------------------------------------------------------------------------
RAZTiming  : 
-----------------------------------------------------------------------------
retour : 
---------------------------------------------------------------------------*/
void 
RAZTiming (lofig, inst)
     lofig_list *lofig;
     loins_list *inst;
{
  if (inst && (searchTH (RAZTH, inst->INSNAME) == EMPTYTH))
    {
      int userOut = searchTH (loconOutTH, inst->INSNAME);
      losig_list *sigOut;
      ptype_list *liste;
      chain_list *lstCon;

/*
   printf("RAZ %s\n", inst->INSNAME);
 */
      addTH (RAZTH, inst->INSNAME, 1);

      if (userOut != EMPTYTH)
	{
	  sigOut = ((locon_list *) userOut)->SIG;
	  liste = getptype (sigOut->USER, (long) LOFIGCHAIN);
	}
      else
	{
	  printf ("RAZTIMING : No type LOCONOUT for %s\n", inst->INSNAME);
	  exit (1);
	}

      if (!liste)
	{
	  printf ("ERROR : RAZTiming : No type LOFIGCHAIN for %s\n", inst->INSNAME);
	  exit (1);
	}

      if (sigOut->USER && getptype (sigOut->USER, (long) RETARDUP))
	sigOut->USER = delptype (sigOut->USER, (long) RETARDUP);
      if (sigOut->USER && getptype (sigOut->USER, (long) RETARDDW))
	sigOut->USER = delptype (sigOut->USER, (long) RETARDDW);
      if (sigOut->USER && getptype (sigOut->USER, (long) PROF))
	sigOut->USER = delptype (sigOut->USER, (long) PROF);
      if (sigOut->USER && getptype (sigOut->USER, (long) LOINSUP))
	sigOut->USER = delptype (sigOut->USER, (long) LOINSUP);
      if (sigOut->USER && getptype (sigOut->USER, (long) LOINSDW))
	sigOut->USER = delptype (sigOut->USER, (long) LOINSDW);
      if (sigOut->USER && getptype (sigOut->USER, (long) LOCONUP))
	sigOut->USER = delptype (sigOut->USER, (long) LOCONUP);
      if (sigOut->USER && getptype (sigOut->USER, (long) LOCONDW))
	sigOut->USER = delptype (sigOut->USER, (long) LOCONDW);
      if (!sigOut->USER)
	printf ("WARNING : RAZTiming = No USER for %s\n", inst->INSNAME);

      /* on  parcourt les connecteurs attaches au connecteur de sortie de inst */
      for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
	{
	  locon_list *conAux = (locon_list *) lstCon->DATA;

	  if (conAux->ROOT != (void *) inst && conAux->TYPE == INTERNAL)
	    {
	      loins_list *fanout = (loins_list *) conAux->ROOT;

	      /* on doit pouvoir memoriser les cellules sur lesquelles
	         le menage est deja fait pour eviter les appels redondants ... TH */

	      RAZTiming ((lofig_list *) NULL, fanout);
	    }
	}
    }

  if (lofig)
    {
      loins_list *cell;

      for (cell = lofig->LOINS; cell; cell = cell->NEXT)
	{
	  int userOut = searchTH (loconOutTH, cell->INSNAME);
	  losig_list *sigOut;

	  if (userOut)
	    sigOut = ((locon_list *) userOut)->SIG;
	  else
	    {
	      printf ("RAZTIMING : No type LOCONOUT for %s\n", inst->INSNAME);
	      exit (1);
	    }

	  if (sigOut->USER && getptype (sigOut->USER, (long) REQUIUP))
	    {
	      sigOut->USER = delptype (sigOut->USER, (long) REQUIUP);
	      sigOut->USER = delptype (sigOut->USER, (long) REQUIDW);
	      sigOut->USER = delptype (sigOut->USER, (long) SLACKUP);
	      sigOut->USER = delptype (sigOut->USER, (long) SLACKDW);
	      sigOut->USER = delptype (sigOut->USER, (long) CRITIC);
	    }
	}
    }
}

/*---------------------------------------------------------------------------
puissanceLofig  : adaptation en puissance des portes
-----------------------------------------------------------------------------
retour : rien. La lofig est modifiee (changement des noms de modeles)
---------------------------------------------------------------------------*/
void 
puissanceLofig (limite)
     int limite;
{
  ptype_list *pinst;
  int nbTraite = 0, nbIter = 0;
  short OK = 1;
  pTH puissanceTH = createTH (500);

  printf ("\nPower Gate Optimization\n");

  cptPower = 0;

  while (OK == 1)
    {
      OK = 0;
/*
   printf("--- ITERATION %d ---\n", nbIter);
 */

      nbTraite = 0;
      for (pinst = listeCritic; pinst && (nbTraite == 0); pinst = pinst->NEXT)
	{
	  loins_list *inst = (loins_list *) pinst->DATA;
	  cellList *model = (cellList *) searchTH (NO_CELLTH, inst->FIGNAME);
	  locon_list *conOut = (locon_list *) searchTH (loconOutTH, inst->INSNAME);
	  long sCin = (long) searchTH (capaTH, getsigname (conOut->SIG));
	  int polarity = cellLoadPolarity (inst->FIGNAME);

	  if ((searchTH (puissanceTH, inst->INSNAME) == EMPTYTH) &&
	      (model->surface_tech->NEXT))
	    {
	      /* il existe plusieurs versions de cette instance */
	      ptype_list *modelPuissant;
	      long delaiUp = 0, delaiDw = 0, delaiConUp = 0, delaiConDw = 0;
	      long delaiPuissantUp, delaiPuissantDw;
	      locon_list *con;
	      char *modelMin = inst->FIGNAME;

	      /* calcul du delai intrinseque de l'instance */
	      for (con = inst->LOCON; con; con = con->NEXT)
		if ((con->DIRECTION == IN) &&
		    !isvdd (con->NAME) && !isvss (con->NAME))
		  {
		    long rcUp = (cellLoadR (inst->FIGNAME, con->NAME, NUP) * sCin) / 1000;
		    long rcDw = (cellLoadR (inst->FIGNAME, con->NAME, NDOWN) * sCin) / 1000;

		    switch (polarity)
		      {
		      case POS:
			delaiConUp = cellLoadTp (inst->FIGNAME, con->NAME, HH) + rcUp;
			delaiConDw = cellLoadTp (inst->FIGNAME, con->NAME, LL) + rcDw;
			break;
		      case NEG:
			delaiConUp = cellLoadTp (inst->FIGNAME, con->NAME, LH) + rcUp;
			delaiConDw = cellLoadTp (inst->FIGNAME, con->NAME, HL) + rcDw;
			break;
		      case AUT:
			delaiConUp = MAX (
			 (cellLoadTp (inst->FIGNAME, con->NAME, LH) + rcUp),
			(cellLoadTp (inst->FIGNAME, con->NAME, HH) + rcUp));
			delaiConDw = MAX (
			 (cellLoadTp (inst->FIGNAME, con->NAME, HL) + rcDw),
			(cellLoadTp (inst->FIGNAME, con->NAME, LL) + rcDw));
			break;
		      }
		    if (delaiConUp > delaiUp)
		      delaiUp = delaiConUp;
		    if (delaiConDw > delaiDw)
		      delaiDw = delaiConDw;
		  }

	      if (NO_TRACE >= 1)
		printf ("* Power for %s [%s] (critic = %ld) sum Cin = %ld delay = %ld %ld\n",
			inst->INSNAME, inst->FIGNAME, pinst->TYPE, sCin, delaiUp, delaiDw);

	      delaiPuissantUp = delaiUp;
	      delaiPuissantDw = delaiDw;

	      for (modelPuissant = model->surface_tech; modelPuissant;
		   modelPuissant = modelPuissant->NEXT)
		{
		  long delaiIntrinsequeUp = 0;
		  long delaiIntrinsequeDw = 0;
		  char *nameModel = (char *) modelPuissant->DATA;
		  int polMod = cellLoadPolarity (nameModel);

		  if (nameModel != inst->FIGNAME)
		    {
		      for (con = inst->LOCON; con; con = con->NEXT)
			if ((con->DIRECTION == IN) &&
			    !isvdd (con->NAME) && !isvss (con->NAME))
			  {
			    long delaiIntrinsequeConUp = 0;
			    long delaiIntrinsequeConDw = 0;
			    long rcUpPerte = 0;
			    long rcDwPerte = 0;
			    long rcUp = (cellLoadR (nameModel, con->NAME, NUP) * sCin) / 1000;
			    long rcDw = (cellLoadR (nameModel, con->NAME, NDOWN) * sCin) / 1000;
			    int gen = searchTH (genUpTH, getsigname (con->SIG));
			    loins_list *instGener;

			    if (gen != EMPTYTH)
			      {
				long rUp;
				long rDw;
				long diffCin = cellLoadCin (nameModel, con->NAME) -
				cellLoadCin (inst->FIGNAME, con->NAME);
				instGener = (loins_list *) gen;
				rUp = cellLoadR (instGener->FIGNAME, (char *) NULL, NUP);
				rDw = cellLoadR (instGener->FIGNAME, (char *) NULL, NDOWN);
				rcUpPerte = (rUp * diffCin) / 1000;
				rcDwPerte = (rDw * diffCin) / 1000;
/*
   if (NO_TRACE >= 2)
   printf("      porte genere = %s (diff = %d), perte UP = %d DW = %d\n", instGener->FIGNAME, diffCin, rcUpPerte, rcDwPerte);
 */
			      }
			    else
			      {
				long rUp = resistanceNot (NUP);
				long rDw = resistanceNot (NDOWN);
				long diffCin = cellLoadCin (nameModel, con->NAME) - cellLoadCin (inst->FIGNAME, con->NAME);
				rcUpPerte = (rUp * diffCin) / 1000;
				rcDwPerte = (rDw * diffCin) / 1000;
/*
   if (NO_TRACE >= 2)
   printf("      connecteur E (diff = %d), perte UP = %d DW = %d\n", diffCin, rcUpPerte, rcDwPerte);
 */
			      }
			    switch (polMod)
			      {
			      case POS:
				delaiIntrinsequeConUp =
				  rcUpPerte + cellLoadTp (nameModel, con->NAME, HH) + rcUp;
				delaiIntrinsequeConDw =
				  rcDwPerte + cellLoadTp (nameModel, con->NAME, LL) + rcDw;
				break;
			      case NEG:
				delaiIntrinsequeConUp =
				  rcDwPerte + cellLoadTp (nameModel, con->NAME, LH) + rcUp;
				delaiIntrinsequeConDw =
				  rcUpPerte + cellLoadTp (nameModel, con->NAME, HL) + rcDw;
				break;
			      case AUT:
				delaiIntrinsequeConUp = MAX (
							      rcDwPerte + cellLoadTp (nameModel, con->NAME, LH) + rcUp,
							      rcUpPerte + cellLoadTp (nameModel, con->NAME, HH) + rcUp);
				delaiIntrinsequeConDw = MAX (
							      rcUpPerte + cellLoadTp (nameModel, con->NAME, HL) + rcDw,
							      rcDwPerte + cellLoadTp (nameModel, con->NAME, LL) + rcDw);
				break;
			      }
			    if (delaiIntrinsequeConUp > delaiIntrinsequeUp)
			      delaiIntrinsequeUp = delaiIntrinsequeConUp;
			    if (delaiIntrinsequeConDw > delaiIntrinsequeDw)
			      delaiIntrinsequeDw = delaiIntrinsequeConDw;
			  }
/*
 */
		      if (NO_TRACE >= 2)
			printf ("    delai intrinseque [%s] = %ld %ld\n",
			 nameModel, delaiIntrinsequeUp, delaiIntrinsequeDw);
		      if (((delaiIntrinsequeUp < delaiPuissantUp) &&
			   (delaiIntrinsequeDw < delaiPuissantDw)) ||
			  ((delaiPuissantUp > delaiPuissantDw) &&
			   (delaiIntrinsequeUp < delaiPuissantUp)) ||
			  ((delaiPuissantUp < delaiPuissantDw) &&
			   (delaiIntrinsequeDw < delaiPuissantDw)))
			{
			  delaiPuissantUp = delaiIntrinsequeUp;
			  delaiPuissantDw = delaiIntrinsequeDw;
			  modelMin = nameModel;
			}
		    }
		}		/* fin parcours des differents modeles */
	      if (inst->FIGNAME != modelMin)
		{
		  long gainUp = delaiUp - delaiPuissantUp;
		  long gainDw = delaiDw - delaiPuissantDw;
		  long gain = MIN (gainUp, gainDw);

		  if (((pinst->TYPE == 0) && (gain >= limite)) ||
		      ((pinst->TYPE != 0) && (gain >= limite * nbIter)))
		    {
		      long oldDelaiUp = delaiCriticUp;
		      long oldDelaiDw = delaiCriticDw;
		      char *oldFigname = inst->FIGNAME;
		      locon_list *conIn;

		      if (NO_TRACE >= 3)
			displayTimingInst (inst, 0, 1);

		      printf ("    ==> Model to use for %s [%s] : %s\n",
			      inst->INSNAME, inst->FIGNAME, modelMin);

		      /* changement effectif de modele */
		      inst->FIGNAME = modelMin;

		      /* remise a jour des capa & delai sur les signaux en entree de inst */
		      for (conIn = inst->LOCON; conIn; conIn = conIn->NEXT)
			if ((conIn->DIRECTION == IN) &&
			    !isvdd (conIn->NAME) && !isvss (conIn->NAME))
			  {
			    long diffCin = cellLoadCin (inst->FIGNAME, conIn->NAME) -
			    cellLoadCin (oldFigname, conIn->NAME);
			    long oldCin = (long) searchTH (capaTH, getsigname (conIn->SIG));
			    long newCin = oldCin + diffCin;
			    int gen, genCon;

			    addTH (capaTH, getsigname (conIn->SIG), (int) newCin);
			    if ((gen = searchTH (genUpTH, getsigname (conIn->SIG))) !=
				EMPTYTH)
			      majDelaiInst ((loins_list *) gen, newCin);
			    else if ((genCon = searchTH (genConTH, getsigname (conIn->SIG))) !=
				     EMPTYTH)
			      majDelaiCon ((locon_list *) genCon, newCin);
			    else
			      {
				printf ("ERROR repower : no type GENERUP ou GENCON for %s\n",
					inst->INSNAME);
				exit (-1);
			      }
			  }
		      majDelaiInst (inst, sCin);

		      cptPower++;
		      addTH (puissanceTH, inst->INSNAME, 1);
		      nbTraite++;
		      /* mise a jour des retards et temps requis */
		      calcul_retard (NO_LOG, 0);
		      calcul_requi (NO_LOG, delaiCriticUp, delaiCriticDw, 2);
		      if ((delaiCriticUp > oldDelaiUp) ||
			  (delaiCriticDw > oldDelaiDw))
			OK = 0;
		      else
			OK = 1;
		      if (NO_TRACE == 2)
			displayTimingInst (inst, 0, 1);
		    }
		}
/*
   else
   printf("    ==> Modele a utiliser : %s\n", modelMin);
 */
	    }
	  else if (NO_TRACE >= 3)
	    printf ("No power model for %s [%s]\n",
		    inst->INSNAME, inst->FIGNAME);
	}			/* parcours des instances de la lofig */
      nbIter++;
    }				/* while OK */
  printf ("\n\t %d repowered gates\n\n", cptPower);
  destroyTH (puissanceTH);
}

/*---------------------------------------------------------------------------
checheBuffer: 
-----------------------------------------------------------------------------
retour : 
---------------------------------------------------------------------------*/
void 
chercheBuffer (porteCritic, type)
     loins_list *porteCritic;
     long type;
{
  loins_list *porte = porteCritic;
  loins_list *porteSuiv = porteCritic;
  ptype_list *chemin;
  short continu = 1;

/*
   printf("Cherche BUFFER %s : ", (type == NUP)?"NUP":"NDW");
   printf("<-%s[%s]", porte->INSNAME, porte->FIGNAME);
 */
  while (continu && (chemin = getptype (porte->USER, (long) ((type == NUP) ? LOINSUP : LOINSDW))))
    if (chemin)
      {
	if ((type == NUP) && (cellLoadPolarity (porte->FIGNAME) == NEG))
	  type = NDOWN;
	else if ((type == NDOWN) && (cellLoadPolarity (porte->FIGNAME) == NEG))
	  type = NUP;
	porteSuiv = porte;
	porte = (loins_list *) chemin->DATA;
/*
   printf("<-%s[%s]", porte->INSNAME, porte->FIGNAME);
 */
	if (cellLoadType (porte->FIGNAME) == BUFF)
	  {
	    losig_list *buffSigOut = ((locon_list *) searchTH (loconOutTH, porte->INSNAME))->SIG;
	    locon_list *conInSuiv, *con;
	    locon_list *conOut = (locon_list *) searchTH (loconOutTH, porte->INSNAME);
	    ptype_list *critic = getptype (porte->USER, (long) LOINSUP);
	    locon_list *conOutCritic;
	    ptype_list *listeCritic;
	    long cinBuff = (long) searchTH (capaTH, getsigname (conOut->SIG));
	    long retardBuffUp = cellLoadTp (porte->FIGNAME, cellLoadIn (porte->FIGNAME)->name, HH) + ((cellLoadR (porte->FIGNAME, cellLoadIn (porte->FIGNAME)->name, NUP) * cinBuff) / 1000);
	    long retardBuffDw = cellLoadTp (porte->FIGNAME, cellLoadIn (porte->FIGNAME)->name, LL) + ((cellLoadR (porte->FIGNAME, cellLoadIn (porte->FIGNAME)->name, NDOWN) * cinBuff) / 1000);
	    long retardBuff = MAX (retardBuffUp, retardBuffDw);
	    ptype_list *triFan, *triFanBuff, *tri;

	    printf ("  BUFFER(Suiv %s)\n  ", porteSuiv->INSNAME);
	    displayTimingInst (porte, 0, 0);

	    if (!critic)
	      {
		/* connecteur d'entree attaque le buffer */
		conOutCritic = (locon_list *) getptype (porte->USER, (long) LOCONUP)->DATA;
		triFan = triFanout ((loins_list *) NULL, conOutCritic,
				    NDOWN, (loins_list *) NULL);
	      }
	    else
	      {
		conOutCritic = (locon_list *) searchTH (loconOutTH,
				    ((loins_list *) critic->DATA)->INSNAME);
		triFan = triFanout ((loins_list *) critic->DATA,
			   (locon_list *) NULL, NDOWN, (loins_list *) NULL);
	      }

	    listeCritic = getptype (conOutCritic->SIG->USER, (long) LOFIGCHAIN);

	    if (critic)
	      printf ("Fanout of %s\n", ((loins_list *) critic->DATA)->INSNAME);
	    else
	      printf ("Fanout of input %s\n", conOutCritic->NAME);
/*
   for(lstCon = (chain_list *)listeCritic->DATA; lstCon; lstCon = lstCon->NEXT)
   {
   locon_list *conAux = (locon_list *)lstCon->DATA;

   if (conAux->ROOT != (void *)porte)
   if (conAux->TYPE == INTERNAL)
   if (conAux->DIRECTION == IN)
   {
   loins_list *cellOut = (loins_list *)conAux->ROOT;
   locon_list *conOut = (locon_list *)searchTH(loconOutTH,
   cellOut->INSNAME);
   ptype_list *slackUp = getptype(conOut->SIG->USER, (long)SLACKUP);
   ptype_list *slackDw = getptype(conOut->SIG->USER, (long)SLACKDW);

   printf("Slack of %s : Up = %d - Dw = %d\n", cellOut->INSNAME,
   (long)slackUp->DATA, (long)slackDw->DATA);
   }
   }
 */

	    tri = triFan;
	    while (tri && tri->TYPE > retardBuff)
	      {
		loins_list *cellFan = (loins_list *) tri->DATA;

		printf ("retardBuff = %ld (cin = %ld) - slack of %s = %ld\n", retardBuff, cinBuff,
			cellFan->INSNAME, tri->TYPE);
		if (critic)
		  {
		    for (con = cellFan->LOCON; con; con = con->NEXT)
		      if (con->SIG == conOutCritic->SIG)
			cinBuff += cellLoadCin (cellFan->FIGNAME, con->NAME);
		  }
		else
		  cinBuff += cellLoadCin ((char *) NULL, conOutCritic->NAME);
		retardBuffUp = cellLoadTp (porte->FIGNAME, cellLoadIn (porte->FIGNAME)->name, HH) + ((cellLoadR (porte->FIGNAME, cellLoadIn (porte->FIGNAME)->name, NUP) * cinBuff) / 1000);
		retardBuffDw = cellLoadTp (porte->FIGNAME, cellLoadIn (porte->FIGNAME)->name, LL) + ((cellLoadR (porte->FIGNAME, cellLoadIn (porte->FIGNAME)->name, NDOWN) * cinBuff) / 1000);
		retardBuff = MAX (retardBuffUp, retardBuffDw);
		tri = tri->NEXT;
	      }
	    if (tri && tri->NEXT)
	      {
		freeptype (tri->NEXT);
		tri->NEXT = NULL;
	      }

	    printf ("Fanout of %s\n", porte->INSNAME);
	    triFanBuff = triFanout (porte, (locon_list *) NULL,
				    NUP, (loins_list *) NULL);

/*
   for(lstCon = (chain_list *)liste->DATA; lstCon; lstCon = lstCon->NEXT)
   {
   locon_list *conAux = (locon_list *)lstCon->DATA;

   if (conAux->ROOT != (void *)porte)
   if (conAux->TYPE == INTERNAL)
   if (conAux->DIRECTION == IN)
   {
   loins_list *cellOut = (loins_list *)conAux->ROOT;
   locon_list *conOut = (locon_list *)searchTH(loconOutTH,
   cellOut->INSNAME);
   ptype_list *slackUp = getptype(conOut->SIG->USER, (long)SLACKUP);
   ptype_list *slackDw = getptype(conOut->SIG->USER, (long)SLACKDW);

   printf("Slack of %s : Up = %d - Dw = %d\n", cellOut->INSNAME, (long)slackUp->DATA, (long)slackDw->DATA);
   }
   }
 */

	    tri = triFanBuff;
	    while (tri && (tri->TYPE < retardBuff / 4))
	      {
		printf ("porte %s slack = %ld\n", ((loins_list *) tri->DATA)->INSNAME, tri->TYPE);
		/* on deconnecte le chemin critique du buffer */
		for (con = ((loins_list *) tri->DATA)->LOCON; con; con = con->NEXT)
		  if ((con->DIRECTION == IN) && (con->SIG == buffSigOut))
		    conInSuiv = con;
		for (con = porte->LOCON; con; con = con->NEXT)
		  if ((con->DIRECTION == IN) && !isvdd (con->NAME) && !isvss (con->NAME))
		    conInSuiv->SIG = con->SIG;
		tri = tri->NEXT;
	      }
	    freeptype (triFanBuff);

	    /* on connecte triFan au buffer */
	    for (tri = triFan; tri; tri = tri->NEXT)
	      {
		loins_list *cellFan = (loins_list *) tri->DATA;

		for (con = cellFan->LOCON; con; con = con->NEXT)
		  if (con->SIG == conOutCritic->SIG)
		    con->SIG = conOut->SIG;

	      }

	    calcul_retard (NO_LOG, 2);
	    calcul_requi (NO_LOG, delaiCriticUp, delaiCriticDw, 1);
	    continu = 0;
	  }
      }
  printf ("\n");
}

/*---------------------------------------------------------------------------
insereBuffer : 
	          |--
	  -----   |-- 		  	  -----   |--
	--|   |   |--			--|   |   |   1'
	--| g |---|-- 1		==>	--| g |---|--
	--|   |   |--  			--|   |   |              |--
	  -----   |--          		  -----   |  ----------  |
	          |--          		          |--||buffer||--|-- 1"
	                                    	     ----------  |
					                         |--
-----------------------------------------------------------------------------
retour : un buffer est insere dans la netlist
---------------------------------------------------------------------------*/
void 
insereBuffer (instTraite, conIn, pOk, pEncore, nameBuff)
     loins_list *instTraite;
     locon_list *conIn;
     short *pOk, *pEncore;
     char *nameBuff;
{
  ptype_list *triFan;
  ptype_list *tri, *t;
  long chargeBuff = 0;
  long retardBuffDw;
  long retardBuffUp;
  long retardBuff;
  int nbChargeBuff = 0;
  locon_list *con;
  short continu = 1;
  short nbDirect = 1;
  losig_list *sigOut;
  long sCin;
  long ROutUp;
  long ROutDw;
  char *inBuff = cellLoadIn (nameBuff)->name;
  long cinBuff = cellLoadCin (nameBuff, inBuff);
  loins_list *instDirect;
  ptype_list *fanoutInst;

  if (instTraite)
    triFan = triFanout (instTraite, (locon_list *) NULL, NUP, (loins_list *) NULL);
  else
    triFan = triFanout ((loins_list *) NULL, conIn, NUP, (loins_list *) NULL);


  if (!triFan)
    {
      /* le fanout est vide */
      *pOk = 0;
      *pEncore = 1;

      return;
    }

  instDirect = (loins_list *) triFan->DATA;
  fanoutInst = addptype ((ptype_list *) NULL, triFan->TYPE, triFan->DATA);

  if (instTraite)
    {
      sigOut = ((locon_list *) searchTH (loconOutTH, instTraite->INSNAME))->SIG;
      ROutUp = cellLoadR (instTraite->FIGNAME, (char *) NULL, NUP);
      ROutDw = cellLoadR (instTraite->FIGNAME, (char *) NULL, NDOWN);
    }
  else
    {
      sigOut = conIn->SIG;
      ROutUp = resistanceNot (NUP);
      ROutDw = resistanceNot (NDOWN);
    }
  sCin = (long) searchTH (capaTH, getsigname (sigOut));

  for (con = ((loins_list *) triFan->DATA)->LOCON; con; con = con->NEXT)
    if (con->SIG == sigOut)
      chargeBuff = sCin - cellLoadCin (instDirect->FIGNAME, con->NAME);

/*
   if (instTraite)
   {
   printf(" insereBuffer sur %s[%s] charge %d, nb fanout %d :\n",
   instTraite->INSNAME, instTraite->FIGNAME, sCin, nbFan);
   printf("  Gate direct %s (slack %d) chargeBuff %d\n",
   instDirect->INSNAME, triFan->TYPE, chargeBuff);
   }
   else
   {
   printf(" insereBuffer sur entree %s charge %d, nb fanout %d :\n",
   conIn->NAME, sCin, nbFan);
   printf("  Gate direct %s (slack %d) chargeBuff %d\n",
   instDirect->INSNAME, triFan->TYPE, chargeBuff);
   }
 */


  /* ATTENTION : partie de test ... A ENLEVER !!! */
/*
   tri = triFan->NEXT;
   while (tri)
   {
   long cas1, cas2;
   instDirect =  (loins_list *)tri->DATA;

   printf("Porte %s : chargeBuff %d RBuff %d TpBuff %d cinBuff %d Rg %d\n",
   instDirect->INSNAME, chargeBuff, cellLoadR(nameBuff, inBuff, NUP),
   cellLoadTp(nameBuff, inBuff, HH), cinBuff, ROutUp);

   cas1 = (ROutUp * (cinBuff-chargeBuff)) / 1000;
   cas2 = (ROutUp * (cinBuff-chargeBuff)) / 1000 +
   cellLoadTp(nameBuff, inBuff, HH) +
   (cellLoadR(nameBuff, inBuff, NDOWN) * chargeBuff) / 1000;

   printf("    porte %s : cas1 %ld, cas2 %ld = %ld - slack %ld\n", instDirect->INSNAME,
   cas1, cas2, cas2 - cas1, tri->TYPE);

   for(con = instDirect->LOCON; con; con = con->NEXT)
   if (con->SIG == sigOut)
   {
   cin = cellLoadCin(instDirect->FIGNAME, con->NAME);
   chargeBuff -= cin;
   }

   tri = tri->NEXT;
   }
 */

  tri = triFan->NEXT;
  while (tri && continu)
    {
      instDirect = (loins_list *) tri->DATA;

/* BUG : Y aurait-il un bug ? 27/10/1994 
   retardBuffUp = ((cellLoadR(nameBuff, inBuff, NUP) * chargeBuff) -
   (ROutUp * cin)) / 1000 +
   cellLoadTp(nameBuff, inBuff, HH);
   retardBuffDw = ((cellLoadR(nameBuff, inBuff, NDOWN) * chargeBuff) -
   (ROutDw * cin)) / 1000 +
   cellLoadTp(nameBuff, inBuff, LL);
   retardBuff = (retardBuffUp > retardBuffDw)? retardBuffUp : retardBuffDw;
 */

/*
   printf("    UP : chargeBuff %d RBuff %d TpBuff %d cinBuff %d Rg %d\n",
   chargeBuff, cellLoadR(nameBuff, inBuff, NUP),
   cellLoadTp(nameBuff, inBuff, HH), cinBuff, ROutUp);
 */

      retardBuffUp = ((cellLoadR (nameBuff, inBuff, NUP) * chargeBuff) -
		      (ROutUp * (chargeBuff - cinBuff))) / 1000 +
	cellLoadTp (nameBuff, inBuff, HH);
      retardBuffDw = ((cellLoadR (nameBuff, inBuff, NDOWN) * chargeBuff) -
		      (ROutDw * (chargeBuff - cinBuff))) / 1000 +
	cellLoadTp (nameBuff, inBuff, LL);

      retardBuff = (retardBuffUp > retardBuffDw) ? retardBuffUp : retardBuffDw;
      retardBuff += 500 + (nbGates / 6);
/*
   retardBuff += 500;
   retardBuff += 700;
   printf("    porte %s : retard %ld - slack %ld\n", instDirect->INSNAME,
   retardBuff, tri->TYPE);
 */

      if (tri->TYPE < retardBuff)
	{
	  /* slack < retard cause par le buffer */
	  fanoutInst = addptype (fanoutInst, tri->TYPE, tri->DATA);
	  for (con = instDirect->LOCON; con; con = con->NEXT)
	    if (con->SIG == sigOut)
	      chargeBuff -= cellLoadCin (instDirect->FIGNAME, con->NAME);

	  nbDirect++;
/*
   printf("  Gate direct %s (slack %d) retardBuff = %d chargeBuff = %d\n",
   ((loins_list *)tri->DATA)->INSNAME, tri->TYPE, retardBuff, chargeBuff);
 */

	  tri = tri->NEXT;
	}
      else
	continu = 0;
    }
  if (tri)
    {
      t = tri;
/*
   printf("  first Gate on buffer %s (slack %d) retardBuff %d\n",
   ((loins_list *)t->DATA)->INSNAME, t->TYPE, retardBuff);
 */

      while (t)
	{
	  nbChargeBuff++;
	  t = t->NEXT;
	}
/*
   if (nbChargeBuff > 1)
   {
   printf("    %d fanouts directs\n", nbDirect);
   printf(" ..... nb fanout of buffer = %d\n", nbChargeBuff);
   }
 */

    }
  freeptype (triFan);

  if (nbChargeBuff > 1)
    {
      /* ajout d'un buffer conditionnel */
      if ((nbChargeBuff < 4) && (NO_TRACE))
	printf ("WARNING : Buffer added after %s for %d gates\n",
	    (instTraite) ? instTraite->INSNAME : conIn->NAME, nbChargeBuff);
      if (instTraite)
	addBufferCond (NO_LOG, instTraite, (locon_list *) NULL, nameBuff, fanoutInst);
      else
	addBufferCond (NO_LOG, (loins_list *) NULL, conIn, nameBuff, fanoutInst);

      freeptype (fanoutInst);
      *pOk = 0;
      *pEncore = 1;

      /* remise a jour des retards et temps requis */
      calcul_retard (NO_LOG, 0);
      calcul_requi (NO_LOG, delaiCriticUp, delaiCriticDw, 1);
/*
   if (delaiCriticUp > delaiCriticDw)
   chercheBuffer(porteCriticUp, NUP);
   else
   chercheBuffer(porteCriticDw, NDOWN);
 */
    }
/*
   else
   {
   printf("Buffer drives only %d gates\n", nbChargeBuff);
   }
 */
}

/*---------------------------------------------------------------------------
remonter	:
-----------------------------------------------------------------------------
retour		:
---------------------------------------------------------------------------*/
void
remonter (instTraite, retardMoyen, capaPlus)
     loins_list *instTraite;
     long retardMoyen, *capaPlus;
{
  long addCapa = 0;
  long charge, nbFan, buff;
  locon_list *con;
  ptype_list *fi;
  ptype_list *fan = computFanout (instTraite, (locon_list *) NULL,
				  &charge, &nbFan, &buff);
  losig_list *sigIn, *sigOut;
  long oldCin;

  printf ("   REMONTER %s[%s] - retard = %ld - capa+ = %ld\n",
	  instTraite->INSNAME, instTraite->FIGNAME, retardMoyen, *capaPlus);

  /* recherche signal a l'entree du buffer en cours de traitement */
  for (con = instTraite->LOCON; con; con = con->NEXT)
    if ((con->DIRECTION == IN) && !isvdd (con->NAME) && !isvss (con->NAME))
      sigIn = con->SIG;
    else if (con->DIRECTION == OUT)
      sigOut = con->SIG;

  if (buff > 0)
    {
      for (fi = fan; fi && cellLoadType (((loins_list *) fi->DATA)->FIGNAME) != BUFF;
	   fi = fi->NEXT)
	;
      if (fi)
	{
	  loins_list *cellOut = (loins_list *) fi->DATA;
	  remonter (cellOut, retardMoyen, capaPlus);
	}
    }

  printf ("    Remonter %s[%s] - retard = %ld - capa+ = %ld\n",
	  instTraite->INSNAME, instTraite->FIGNAME, retardMoyen, *capaPlus);

  for (fi = fan; fi; fi = fi->NEXT)
    {
      loins_list *cellOut = (loins_list *) fi->DATA;

      if (cellLoadType (cellOut->FIGNAME) != BUFF)
	{
	  long slackUp = getInfoInst (cellOut, SLACKUP);
	  long slackDw = getInfoInst (cellOut, SLACKDW);
	  long slack = (slackUp > slackDw) ? slackDw : slackUp;
	  long ROutUp = cellLoadR (instTraite->FIGNAME, (char *) NULL, NUP);
	  long ROutDw = cellLoadR (instTraite->FIGNAME, (char *) NULL, NDOWN);
	  long RMoyen = (ROutUp > ROutDw) ? ROutUp : ROutDw;

	  if (slack < (1000 + retardMoyen + RMoyen * (*capaPlus)))
	    {
	      addCapa += fi->TYPE;

	      /* l'instance du fanin est remontee d'un buffer */

	      for (con = cellOut->LOCON; con; con = con->NEXT)
		if ((con->DIRECTION == IN) && (con->SIG == sigOut))
		  {
		    con->SIG = sigIn;
		    printf ("\t %s deconnectee (slack %ld)\n",
			    cellOut->INSNAME, slack);
		  }
	    }
	}
    }
  if ((addCapa != 0) || (*capaPlus != 0))
    {
      oldCin = (long) searchTH (capaTH, getsigname (sigOut));
      majDelaiInst (instTraite, (oldCin - addCapa + *capaPlus));
      *capaPlus = addCapa;
    }
}

/*---------------------------------------------------------------------------
arbreBuffer	: insertion d'un buffer + re-equilibrage de l'arbre de buffers
-----------------------------------------------------------------------------
retour		: un buffer est insere dans la netlist
---------------------------------------------------------------------------*/
void 
arbreBuffer (instTraite, conIn, lstFanout)
     loins_list *instTraite;
     locon_list *conIn;
     ptype_list *lstFanout;
{
  ptype_list *pt;
  long capaSup = 0;
  locon_list *conOutInst;
  long chargeInst;

  printf ("Arbre %s\n", (instTraite) ? instTraite->INSNAME : conIn->NAME);

  if (instTraite)
    {
      conOutInst = (locon_list *) searchTH (loconOutTH, instTraite->INSNAME);
      chargeInst = searchTH (capaTH, getsigname (conOutInst->SIG));
    }
  else
    chargeInst = searchTH (capaTH, getsigname (conIn->SIG));

  for (pt = lstFanout; pt; pt = pt->NEXT)
    {
      loins_list *cellOut = (loins_list *) pt->DATA;

      if (cellLoadType (cellOut->FIGNAME) == BUFF)
	{
	  long RBufUp = cellLoadR (cellOut->FIGNAME, (char *) NULL, NUP);
	  long RBufDw = cellLoadR (cellOut->FIGNAME, (char *) NULL, NDOWN);
	  long RBuf = (RBufUp > RBufDw) ? RBufUp : RBufDw;
	  long ROutUp, ROutDw, ROut, retardMoyen;

	  if (instTraite)
	    {
	      ROutUp = cellLoadR (instTraite->FIGNAME, (char *) NULL, NUP);
	      ROutDw = cellLoadR (instTraite->FIGNAME, (char *) NULL, NDOWN);
	    }
	  else
	    {
	      ROutUp = resistanceNot (NUP);
	      ROutDw = resistanceNot (NDOWN);
	    }

	  ROut = (ROutUp > ROutDw) ? ROutUp : ROutDw;

	  /* ATTENTION : rechercher capa et TP buffer */
	  retardMoyen = (chargeInst * (RBuf - ROut)) / 2000 +
	    (ROut * 54) / 1000 + 930;

	  remonter (cellOut, retardMoyen, &capaSup);
	  lofigchain (NO_LOG);

	  /* remise a jour des retards et temps requis */
	  calcul_retard (NO_LOG, 0);
	  /* on ne recalcule pas la liste critique */
	  calcul_requi (NO_LOG, delaiCriticUp, delaiCriticDw, 0);
	}
    }
}

/*---------------------------------------------------------------------------
bufferLofig	: insertion de buffers dans la lofig
-----------------------------------------------------------------------------
retour	     	: lofig modifiee
-----------------------------------------------------------------------------
remarque	: trier les portes selon RC et/ou prof ...
---------------------------------------------------------------------------*/
void 
bufferLofig (minFanIn, minFanInst, limite, foisLimite)
     int minFanIn, minFanInst, limite, foisLimite;
{
  cellList *cells = NO_CELLS;
  int maxBuffIn = 3;

  /* recherche des buffers */
  while (cells && cells->type != 't')
    cells = cells->NEXT;

  if (cells)
    {
      /* il existe bien des buffers dans la bibliotheque */
      ptype_list *cellsPhy;
      long minR = 9999999;
      long maxR = 0;
      char *nameBuff = (char *) NULL;
      char *nameBuff2 = (char *) NULL;
      short fois = 0;
      short OK = 1, encore = 1;
      cptBuffer = 0;

      printf ("\nBuffer Optimization\n");

      /* on choisit le buffer le moins resistif */
      for (cellsPhy = cells->resistance_tech; cellsPhy;
	   cellsPhy = cellsPhy->NEXT)
	{
	  char *nameCellsPhy = (char *) cellsPhy->DATA;
	  long ROutUp = cellLoadR (nameCellsPhy, (char *) NULL, NUP);

	  if (ROutUp > maxR)
	    {
	      maxR = ROutUp;
	      nameBuff2 = nameCellsPhy;
	    }

	  if (ROutUp < minR)
	    {
	      minR = ROutUp;
	      nameBuff = nameCellsPhy;
	    }
	}

      assignTH = createTH (50);

      while (encore)
	{
	  ptype_list *p;
	  int val;
	  long oldDelaiUp = delaiCriticUp;
	  long oldDelaiDw = delaiCriticDw;

	  OK = 1;
	  encore = 0;

	  /* on teste si l'entree est + chargee que la porte la + chargee */
	  if (listeCritic->TYPE > listeCritic->NEXT->TYPE)
	    {
	      /* traitement de l'entree */
	      if (listeCritic->DATA)
		{
		  locon_list *conTraite = (locon_list *) listeCritic->DATA;
		  ptype_list *pBuf;
		  int optimIn = -1;

		  for (pBuf = NO_BUFFPI; pBuf; pBuf = pBuf->NEXT)
		    if ((char *) pBuf->DATA == conTraite->NAME)
		      optimIn = pBuf->TYPE;

		  if (optimIn != 0)
		    if ((val = searchTH (assignTH, conTraite->NAME)) < maxBuffIn)
		      {
/*
   ptype_list *fanout = computFanout(NULL, conTraite,
   &charge, &nbFan, &buff);
 */
			if (val == EMPTYTH)
			  val = 0;

/*
   printf("FANOUT du connecteur %s - charge = %ld - nbFan = %ld - buff=%ld\n",
   conTraite->NAME, charge, nbFan, buff);
   for(f = fanout; f; f = f->NEXT)
   printf(" fanout %s[%s] cin=%ld\n",
   ((loins_list *)f->DATA)->INSNAME, ((loins_list *)f->DATA)->FIGNAME,
   f->TYPE);

   if (nbFan > minFanIn)
   {
   if (buff > 0)
   arbreBuffer((loins_list *)NULL, conTraite, fanout);
   insereBuffer((loins_list *)NULL, conTraite, &OK, &encore,
   nameBuff);
   if (OK == 1)
   addTH(assignTH, conTraite->NAME, val+1);
   }
 */
			insereBuffer ((loins_list *) NULL, conTraite, &OK, &encore,
				      nameBuff);
/*
   if (OK == 0)
   insereBuffer((loins_list *)NULL, conTraite, &OK, &encore,
   nameBuff2);
 */

		      }
		}
	    }

	  /* traitement des instances  : suite de la liste */
	  for (p = listeCritic->NEXT; p && OK; p = p->NEXT)
	    {
	      loins_list *instTraite = (loins_list *) p->DATA;
/*
   ptype_list *fanout = computFanout(instTraite, NULL, &charge, &nbFan, &buff);
   printf("FANOUT de %s[%s] - charge = %ld - nbFan = %ld - buff=%ld\n",
   instTraite->INSNAME, instTraite->FIGNAME, charge, nbFan, buff);
   for(f = fanout; f; f = f->NEXT)
   printf(" fanout %s[%s] cin=%ld\n",
   ((loins_list *)f->DATA)->INSNAME, ((loins_list *)f->DATA)->FIGNAME,
   f->TYPE);
   if (nbFan > minFanInst)
   {
   if (buff > 0)
   arbreBuffer(instTraite, (locon_list *)NULL, fanout);
   insereBuffer(instTraite, (locon_list *)NULL, &OK, &encore, nameBuff);
   }
 */
	      insereBuffer (instTraite, (locon_list *) NULL, &OK, &encore, nameBuff);
/*
   if (OK == 1)
   insereBuffer(instTraite, (locon_list *)NULL, &OK, &encore, nameBuff2);
 */
	    }
/*
   displayLofig(NO_LOG, 4);
 */
	  if (((oldDelaiUp - delaiCriticUp) + (oldDelaiDw - delaiCriticDw)) < limite)
	    fois++;
	  else
	    fois = 0;
	  if (fois >= foisLimite)
	    encore = 0;
/*
   printf("fois = %d\n", fois);
 */
	}
      destroyTH (assignTH);
      printf ("\n\t %d inserted buffers\n\n", cptBuffer);
    }
}

/*---------------------------------------------------------------------------
  insereInverseur : insere un buffer inverseur apres inv1 pour decharger g.
  ---------------------------------------------------------------------------
	         |--
	  ----   |-- 1		  	  ----   |--
	--|  |   |--			--|  |   |   1'
	--|g |---|--		==>	--|g |---|--
	--|  |   |  			--|  |   |
	  ----   |  |\   |--		  ----   |  ||\   |--
	         |--| o--|   2		         |--|| o--|   2
	            |/   |--		            ||/   |--
					                  |  ||\   |--
	  	   inv1			                  |--|| o--|   1"
					                     ||/   |--

-----------------------------------------------------------------------------
retour : un inverseur puissant est insere dans la netlist.
-----------------------------------------------------------------------------*/
void 
insereInverseur (instTraite, conIn, pOk, pEncore, minFan, nameBuff, inv1)
     loins_list *instTraite;	/* g ou NULL si connecteur */
     locon_list *conIn;		/* ou NULL si instance   */
     short *pOk, *pEncore;
     int minFan;
     char *nameBuff;
     loins_list *inv1;		/* pointeur sur instance inv1 */
{
  ptype_list *pt;
  int nbFan = 0;
  ptype_list *triFan;
  /* charge de l'inverseur (2) */
  locon_list *conOutInv = (locon_list *) searchTH (loconOutTH, inv1->INSNAME);
  long somInv = (long) searchTH (capaTH, getsigname (conOutInv->SIG));

  if (instTraite)
    triFan = triFanout (instTraite, (locon_list *) NULL, NUP, inv1);
  else
    triFan = triFanout ((loins_list *) NULL, conIn, NUP, inv1);

  for (pt = triFan; pt && (nbFan < minFan); pt = pt->NEXT)
    nbFan++;

  if (nbFan >= minFan)
    {
      ptype_list *tri, *t;
      loins_list *instDirect = (loins_list *) triFan->DATA;
      ptype_list *fanoutInst = addptype ((ptype_list *) NULL,
					 triFan->TYPE, triFan->DATA);
      long chargeBuff = 0;
      long retardBuffDw;
      long retardBuffUp;
      long retardBuff;
      int nbChargeBuff = 0;
      locon_list *con;
      short continu = 1;
      short nbDirect = 1;
      losig_list *sigOut;
      long sCin;
      long ROutUp;
      long ROutDw;
      char *inBuff = cellLoadIn (nameBuff)->name;

      /* l'inverseur inv1 est mis dans le fanout direct */
      fanoutInst = addptype (fanoutInst, 0, (void *) inv1);

      if (instTraite)
	{
	  sigOut = ((locon_list *) searchTH (loconOutTH, instTraite->INSNAME))->SIG;
	  ROutUp = cellLoadR (instTraite->FIGNAME, (char *) NULL, NUP);
	  ROutDw = cellLoadR (instTraite->FIGNAME, (char *) NULL, NDOWN);
	}
      else
	{
	  sigOut = conIn->SIG;
	  ROutUp = resistanceNot (NUP);
	  ROutDw = resistanceNot (NDOWN);
	}
      sCin = (long) searchTH (capaTH, getsigname (sigOut));

      /* init de chargeBuff : charge totale de inst - les 2 inst directes */
      for (con = instDirect->LOCON; con; con = con->NEXT)
	if (con->SIG == sigOut)
	  chargeBuff = sCin - cellLoadCin (instDirect->FIGNAME, con->NAME)
	    - cellLoadCin (nameBuff, inBuff);

/*
 */
      if (instTraite)
	printf ("  sur %s[%s] capa %ld : %s[%s] slack %ld-...",
		instTraite->INSNAME, instTraite->FIGNAME, sCin,
		((loins_list *) triFan->DATA)->INSNAME,
		((loins_list *) triFan->DATA)->FIGNAME,
		triFan->TYPE);
      else
	printf ("  sur %s capa %ld : %s[%s] slack %ld-...",
		conIn->NAME, sCin,
		((loins_list *) triFan->DATA)->INSNAME,
		((loins_list *) triFan->DATA)->FIGNAME,
		triFan->TYPE);

      tri = triFan->NEXT;
      while (tri && continu)
	{
	  long cin = 0;
	  instDirect = (loins_list *) tri->DATA;

	  for (con = instDirect->LOCON; con; con = con->NEXT)
	    if (con->SIG == sigOut)
	      cin = cellLoadCin (instDirect->FIGNAME, con->NAME);

	  retardBuffUp = ((((cellLoadR (nameBuff, inBuff, NUP) - ROutUp) *
			    chargeBuff) +
			   (cellLoadR (nameBuff, inBuff, NDOWN) *
		       (somInv + cellLoadCin (nameBuff, inBuff)))) / 1000) +
	    cellLoadTp (nameBuff, inBuff, LH) +
	    cellLoadTp (nameBuff, inBuff, HL);
	  retardBuffDw = ((((cellLoadR (nameBuff, inBuff, NDOWN) - ROutUp) *
			    chargeBuff) +
			   (cellLoadR (nameBuff, inBuff, NUP) *
		       (somInv + cellLoadCin (nameBuff, inBuff)))) / 1000) +
	    cellLoadTp (nameBuff, inBuff, HL) +
	    cellLoadTp (nameBuff, inBuff, LH);

	  retardBuff = (retardBuffUp > retardBuffDw) ? retardBuffUp : retardBuffDw;

	  if (NO_TRACE >= 3)
	    printf (" Gate %s (slack %ld) delay = %ld\n",
		((loins_list *) tri->DATA)->INSNAME, tri->TYPE, retardBuff);
	  if (tri->TYPE < retardBuff)
	    {
/*
   printf(" %s %d-", ((loins_list *)tri->DATA)->INSNAME, tri->TYPE);
 */
	      fanoutInst = addptype (fanoutInst, tri->TYPE, tri->DATA);
	      chargeBuff -= cin;
	      nbDirect++;
	      tri = tri->NEXT;
	    }
	  else
	    continu = 0;
	}

/*
   printf(" %d fanouts\n", nbDirect);
   printf("\n     sur %s :", nameBuff);
 */
      if (tri)
	{
	  t = tri;
/*
   printf(" %s %d-", ((loins_list *)t->DATA)->INSNAME, t->TYPE);
 */
	  while (t)
	    {
	      nbChargeBuff++;
	      t = t->NEXT;
	    }
/*
   if (nbChargeBuff > 1)
   printf(" ..... buff fanout = %d\n", nbChargeBuff);
   else
   printf("\n");
 */
	}
      freeptype (triFan);

      /* ajout d'un buffer inverseur conditionnel */
      if (nbChargeBuff > 1)
	{
	  if (instTraite)
	    addInversCond (NO_LOG, instTraite, inv1, (locon_list *) NULL,
			   nameBuff, fanoutInst);
	  else
	    addInversCond (NO_LOG, (loins_list *) NULL, inv1, conIn,
			   nameBuff, fanoutInst);

	  freeptype (fanoutInst);
	  *pOk = 0;
	  *pEncore = 1;

	  /* remise a jour des retards et temps requis */
	  calcul_retard (NO_LOG, 0);
	  calcul_requi (NO_LOG, delaiCriticUp, delaiCriticDw, 1);
/*
   if (delaiCriticUp > delaiCriticDw)
   chercheBuffer(porteCriticUp, NUP);
   else
   chercheBuffer(porteCriticDw, NDOWN);
 */
	}
    }
/*
   else
   {
   if (instTraite)
   printf("Cell %s [%s] drives only %d gates\n", instTraite->INSNAME,
   instTraite->FIGNAME, nbFan);
   else
   printf("Con %s drives only %d gates\n", conIn->NAME, nbFan);
   }
 */
}

/*---------------------------------------------------------------------------
inverseurLofig	: bufferisation avec des inverseurs puissants.
-----------------------------------------------------------------------------
retour		: des inverseurs sont inseres dans la lofig.
remarque : trier les portes selon RC et/ou prof ...
---------------------------------------------------------------------------*/
void 
inverseurLofig (minFanIn, minFanInst, limite, foisLimite)
     int minFanIn, minFanInst, limite, foisLimite;
{
  cellList *cells = NO_CELLS;
  int maxBuffIn = 2;

  while (cells && cells->nomlog != NOT1)
    cells = cells->NEXT;
  if (cells)
    {
      ptype_list *cellsPhy;
      long minR = 9999999;
      char *nameBuff = (char *) NULL;
      short fois = 0;
      short OK = 1, encore = 1;
      cptBuffer = 0;

      printf ("\nInverted Buffer Optimization\n");

      /* on choisit le buffer inverseur le moins resistif */
      for (cellsPhy = cells->resistance_tech; cellsPhy;
	   cellsPhy = cellsPhy->NEXT)
	{
	  char *nameCellsPhy = (char *) cellsPhy->DATA;
	  long ROutUp = cellLoadR (nameCellsPhy, (char *) NULL, NUP);
	  if (ROutUp < minR)
	    {
	      minR = ROutUp;
	      nameBuff = nameCellsPhy;
	    }
	}
/*
   printf("\tBuffer inverseur = %s-\n", nameBuff);
 */

      assignTH = createTH (50);

      while (encore)
	{
	  ptype_list *p;
	  int val;
	  long oldDelaiUp = delaiCriticUp;
	  long oldDelaiDw = delaiCriticDw;

	  OK = 1;
	  encore = 0;

	  /* traitement de l'entree */
	  if (listeCritic->DATA)
	    if ((val = searchTH (assignTH,
				 ((locon_list *) listeCritic->DATA)->NAME))
		< maxBuffIn)
	      {
		locon_list *conTraite = (locon_list *) listeCritic->DATA;
		ptype_list *liste = getptype (conTraite->SIG->USER, (long) LOFIGCHAIN);
		chain_list *lstCon;
		int cptInv = 0;
		loins_list *instInv = (loins_list *) NULL;

		if (val == EMPTYTH)
		  val = 0;
		addTH (assignTH, conTraite->NAME, val + 1);

		/* recherche si inverseur existe dans le fanout de conTraite */
		for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
		  {
		    locon_list *conAux = (locon_list *) lstCon->DATA;

		    if (conAux->ROOT != (void *) NO_LOG)
		      {
			loins_list *fanout = (loins_list *) conAux->ROOT;
			int val = searchTH (NO_CELLTH, fanout->FIGNAME);

			if ((val == EMPTYTH) || (val == DELETETH))
			  {
			    printf ("ERROR : %s not found\n", fanout->INSNAME);
			    exit (-1);
			  }
			if (((cellList *) val)->nomlog == NOT1)
			  {
			    cptInv++;
			    instInv = fanout;
			  }
		      }
		  }

/*
   if (cptInv > 0)
   printf("insere inv apres con %s, nb inv == %d, nom = %s[%s]\n",
   conTraite->NAME, cptInv, instInv->INSNAME, instInv->FIGNAME);
 */

		if (cptInv > 0)
		  {
		    if (instInv->FIGNAME != nameBuff)
		      addPuissante (NO_LOG, instInv, nameBuff);
		    insereInverseur ((loins_list *) NULL, conTraite, &OK, &encore,
				     minFanIn, nameBuff, instInv);
		  }
		else
		  insereBuffer ((loins_list *) NULL, conTraite, &OK, &encore, nameBuff);
	      }

	  /* traitement des instances  : suite de la liste */
	  for (p = listeCritic->NEXT; p && OK; p = p->NEXT)
	    {
	      loins_list *instTraite = (loins_list *) p->DATA;
	      locon_list *conOut = (locon_list *) searchTH (loconOutTH,
						       instTraite->INSNAME);
	      ptype_list *liste = getptype (conOut->SIG->USER, (long) LOFIGCHAIN);
	      chain_list *lstCon;
	      int cptInv = 0;
	      loins_list *instInv = (loins_list *) NULL;

/*
   displayTimingInst(instTraite, 0, 0);
 */
	      /* recherche si inverseur existe dans le fanout de instTraite */
	      for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
		{
		  locon_list *conAux = (locon_list *) lstCon->DATA;

		  if ((conAux->ROOT != (void *) instTraite) &&
		      (conAux->TYPE == INTERNAL))
		    {
		      loins_list *fanout = (loins_list *) conAux->ROOT;
		      int val = searchTH (NO_CELLTH, fanout->FIGNAME);

		      if ((val == EMPTYTH) || (val == DELETETH))
			{
			  printf ("ERROR : %s not found\n", fanout->INSNAME);
			  exit (-1);
			}
		      if (((cellList *) val)->nomlog == NOT1)
			{
			  cptInv++;
			  instInv = fanout;
			}
		    }
		}

/*
   if (cptInv > 0)
   printf("insere inv apres inst %s, nb inv == %d, nom = %s[%s]\n",
   instTraite->INSNAME, cptInv, instInv->INSNAME, instInv->FIGNAME);
   else
   printf("insere inv apres inst %s, nb inv == %d\n",
   instTraite->INSNAME, cptInv);
 */

	      if (cptInv > 0)
		insereInverseur (instTraite, (locon_list *) NULL, &OK, &encore,
				 minFanInst, nameBuff, instInv);
	      else
		insereBuffer (instTraite, (locon_list *) NULL, &OK, &encore, nameBuff);
	    }
/*
   displayLofig(NO_LOG, 4);
 */
	  if (((oldDelaiUp - delaiCriticUp) + (oldDelaiDw - delaiCriticDw)) < limite)
	    fois++;
	  else
	    fois = 0;
	  if (fois >= foisLimite)
	    encore = 0;
/*
   printf("fois = %d\n", fois);
 */
	}
      destroyTH (assignTH);
/*
   printf("\n\t %d inserted buffers\n\n", cptBuffer);
 */
    }
  else
    printf ("Pas de cellules inverseuses\n");
}
