/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Netlist Optimisation
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: netop_load_incdelai.c,v 1.1.1.1 1998/10/02 15:25:10 alliance Exp $"

/****************************************************************************/
/*    Produit :  NetOptim (calcul incremental des retards)                  */
/****************************************************************************/

#include <stdio.h>
#include <math.h>
#include MUT_H
#include BEH_H
#include MLO_H
#include MPH_H
#include LOG_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"
#include "netop_local.h"



long delaiCriticUp;
long delaiCriticDw;
long delaiCriticMinUp;
long delaiCriticMinDw;
loins_list *porteCriticUp;
loins_list *porteCriticDw;
loins_list *porteCriticMinUp;
loins_list *porteCriticMinDw;

long profondeurCircuit;
loins_list *instProf;

ptype_list *listeCritic;

/* Tables de hasch */

/* key   : nom d'instance, value = 1 si extremite de chemin */
pTH pathTH;

/* key   : nom de signal, value : somme capas sur le signal */
pTH capaTH;

/* key   : index de signal_nom de connecteur_transition */
pTH delaiTH;

/* value : TP+RC en fct du connecteur et de la transition  */
pTH delaiminTH;

/* key   : nom d'instance, value : pointeur sur connecteur de sortie */
pTH loconOutTH;

/* key   : nom de signal, value : pointeur sur le connecteur qui drive le signal */
pTH genConTH;

/* key   : nom de signal, value : pointeur sur l'instance qui drive le signal */
pTH genUpTH;

/* key   : nom de signal, value : pointeur sur l'instance qui drive le signal */
pTH genDwTH;



/*---------------------------------------------------------------------------
RAZsig 		: fait des delptype sur les retards, temps requis ...
		  attaches a un signal
-----------------------------------------------------------------------------
retour      	: rien
---------------------------------------------------------------------------*/
void 
RAZsig (signal)
losig_list *signal;
{
  if (signal->USER && getptype (signal->USER, (long) RETARDUP))
    signal->USER = delptype (signal->USER, (long) RETARDUP);
  if (signal->USER && getptype (signal->USER, (long) RETARDDW))
    signal->USER = delptype (signal->USER, (long) RETARDDW);
  if (signal->USER && getptype (signal->USER, (long) RETARDMINUP))
    signal->USER = delptype (signal->USER, (long) RETARDMINUP);
  if (signal->USER && getptype (signal->USER, (long) RETARDMINDW))
    signal->USER = delptype (signal->USER, (long) RETARDMINDW);
  if (signal->USER && getptype (signal->USER, (long) REQUIUP))
    signal->USER = delptype (signal->USER, (long) REQUIUP);
  if (signal->USER && getptype (signal->USER, (long) REQUIDW))
    signal->USER = delptype (signal->USER, (long) REQUIDW);
  if (signal->USER && getptype (signal->USER, (long) SLACKUP))
    signal->USER = delptype (signal->USER, (long) SLACKUP);
  if (signal->USER && getptype (signal->USER, (long) SLACKDW))
    signal->USER = delptype (signal->USER, (long) SLACKDW);
  if (signal->USER && getptype (signal->USER, (long) PROF))
    signal->USER = delptype (signal->USER, (long) PROF);
  if (signal->USER && getptype (signal->USER, (long) CRITIC))
    signal->USER = delptype (signal->USER, (long) CRITIC);
  if (signal->USER && getptype (signal->USER, (long) LOINSUP))
    signal->USER = delptype (signal->USER, (long) LOINSUP);
  if (signal->USER && getptype (signal->USER, (long) LOINSDW))
    signal->USER = delptype (signal->USER, (long) LOINSDW);
  if (signal->USER && getptype (signal->USER, (long) LOINSMINUP))
    signal->USER = delptype (signal->USER, (long) LOINSMINUP);
  if (signal->USER && getptype (signal->USER, (long) LOINSMINDW))
    signal->USER = delptype (signal->USER, (long) LOINSMINDW);
  if (signal->USER && getptype (signal->USER, (long) LOCONUP))
    signal->USER = delptype (signal->USER, (long) LOCONUP);
  if (signal->USER && getptype (signal->USER, (long) LOCONDW))
    signal->USER = delptype (signal->USER, (long) LOCONDW);
  if (!signal->USER)
    printf ("WARNING : RAZsig = No field USER for signal %s\n",
	    getsigname (signal));
}

/*---------------------------------------------------------------------------
RAZlofig 	: fait des delptype sur les retards, temps requis ...
		  de toute la lofig.
-----------------------------------------------------------------------------
retour      	: rien
---------------------------------------------------------------------------*/
void 
RAZlofig (lofig)
     lofig_list *lofig;
{
  loins_list *inst;
  locon_list *con;

  /* Parcours des connecteurs externes d'entree */
  for (con = lofig->LOCON; con; con = con->NEXT)
    if (con->DIRECTION == IN && !isvdd (con->NAME) && !isvss (con->NAME))
      RAZsig (con->SIG);

  /* parcours des instances */
  for (inst = lofig->LOINS; inst; inst = inst->NEXT)
    if (cellLoadType (inst->FIGNAME) != TIE)
      {
	int val = searchTH (loconOutTH, inst->INSNAME);

	if (val != EMPTYTH)
	  {
	    losig_list *sigOut = ((locon_list *) val)->SIG;

	    RAZsig (sigOut);
	  }
	else
	  {
	    printf ("ERROR RAZlofig : no type LOCONOUT for %s[%s]\n",
		    inst->INSNAME, inst->FIGNAME);
	    exit (-1);
	  }
      }
}


/*---------------------------------------------------------------------------
majDelaiCon : calcul des delais d'un connecteur d'entree : Tp + RC et
	       memorisation dans une table de hasch.
	       On tient compte des resistances & retards sur les entrees :
               NO_RESISTPI et NO_DELAYPI
-----------------------------------------------------------------------------
retour       : void
---------------------------------------------------------------------------*/
void 
majDelaiCon (con, sCin)
     locon_list *con;
     long sCin;
{
  long delIn = getDelayIn (NO_DELAYPI, con->NAME);
  long rUp = getResistIn (NO_RESISTPI, con->NAME);
  long rDw = rUp;
  long rNotUp = resistanceNot (NUP);
  long rNotDw = resistanceNot (NDOWN);
  long capaInv = capaNot ();
  char *inNameHH = (char *) mbkalloc (strlen (con->NAME) + 10);
  char *inNameLL = (char *) mbkalloc (strlen (con->NAME) + 10);
  long delaiHH, delaiLL, delaiMinHH, delaiMinLL;

  if (rUp == 0)
    rUp = rNotUp;
  if (rDw == 0)
    rDw = rNotDw;

  delaiHH = delIn + (rUp * sCin) / 1000;
  delaiMinHH = delIn + (rUp * capaInv) / 1000;
  delaiLL = delIn + (rDw * sCin) / 1000;
  delaiMinLL = delIn + (rDw * capaInv) / 1000;

  sprintf (inNameHH, "%ld_%s_hh", con->SIG->INDEX, con->NAME);
  addTH (delaiTH, namealloc (inNameHH), delaiHH);
  addTH (delaiminTH, namealloc (inNameHH), delaiMinHH);
  mbkfree ((void *) inNameHH);

  sprintf (inNameLL, "%ld_%s_ll", con->SIG->INDEX, con->NAME);
  addTH (delaiTH, namealloc (inNameLL), delaiLL);
  addTH (delaiminTH, namealloc (inNameLL), delaiMinLL);
  mbkfree ((void *) inNameLL);

  if (NO_TRACE >= 2)
    printf ("  delay of %s [input] - UP = %ld - DOWN = %ld - Cin Sum = %ld\n",
     con->NAME, delIn + rUp * sCin / 1000, delIn + rDw * sCin / 1000, sCin);
}

/*---------------------------------------------------------------------------
majDelaiInst : calcul des delais d'une instance : Tp + RC et
	       memorisation dans une table de hasch.
-----------------------------------------------------------------------------
retour       : void
---------------------------------------------------------------------------*/
void 
majDelaiInst (inst, sCin)
     loins_list *inst;
     long sCin;
{
  long capaInv = capaNot ();
  char *inNameHH;
  char *inNameLL;
  char *inNameHL;
  char *inNameLH;
  long delaiInstHH = 0, delaiInstLL = 0, delaiInstLH = 0, delaiInstHL = 0;
  long delaiInstMinHH = 0, delaiInstMinLL = 0;
  long delaiInstMinHL = 0, delaiInstMinLH = 0;
  losig_list *sigOut;
  int val = searchTH (loconOutTH, inst->INSNAME);

  if (val != EMPTYTH)
    sigOut = ((locon_list *) val)->SIG;
  else
    {
      printf ("ERROR majDelaiInst : no type LOCONOUT for %s[%s]\n",
	      inst->INSNAME, inst->FIGNAME);
      exit (-1);
    }

  if (cellLoadType (inst->FIGNAME) == REG)
    {
      /* cas d'un registre */
      long rUp = cellLoadR (inst->FIGNAME, NULL, NUP);
      long rDw = cellLoadR (inst->FIGNAME, NULL, NDOWN);

      /* memorisation du delai du registre  sur le signal de sortie */
      delaiInstHH = (rUp * sCin) / 1000;
      delaiInstMinHH = (rUp * capaInv) / 1000;
      delaiInstLL = (rDw * sCin) / 1000;
      delaiInstMinLL = (rDw * capaInv) / 1000;

      inNameHH = (char *) mbkalloc (strlen (inst->INSNAME) + 10);
      sprintf (inNameHH, "%ld_%s_hh", sigOut->INDEX, inst->INSNAME);
      addTH (delaiTH, namealloc (inNameHH), delaiInstHH);
      addTH (delaiminTH, namealloc (inNameHH), delaiInstMinHH);
      mbkfree ((void *) inNameHH);

      inNameLL = (char *) mbkalloc (strlen (inst->INSNAME) + 10);
      sprintf (inNameLL, "%ld_%s_ll", sigOut->INDEX, inst->INSNAME);
      addTH (delaiTH, namealloc (inNameLL), delaiInstLL);
      addTH (delaiminTH, namealloc (inNameLL), delaiInstMinLL);
      mbkfree ((void *) inNameLL);

    }				/* fin du traitement des registres */
  else
    {
      /* cas d'une cellule combinatoire */
      int polarityInst = cellLoadPolarity (inst->FIGNAME);

      if (NO_TRACE >= 3)
	printf (" Computing delay on %s [%s], sum Cin = %ld\n",
		inst->INSNAME, inst->FIGNAME, sCin);

      if (cellLoadType (inst->FIGNAME) == CONST)
	{
	  /* Cas des cellules zero_y et one_y : pas d'entrees */
	  delaiInstHH = (cellLoadR (inst->FIGNAME, (char *) NULL, NUP) * sCin) / 1000;
	  delaiInstLL = (cellLoadR (inst->FIGNAME, (char *) NULL, NDOWN) * sCin) / 1000;
	  delaiInstMinHH = (cellLoadR (inst->FIGNAME, (char *) NULL, NUP) * capaInv) / 1000;
	  delaiInstMinLL = (cellLoadR (inst->FIGNAME, (char *) NULL, NDOWN) * capaInv) / 1000;

	  inNameHH = (char *) mbkalloc (strlen (inst->INSNAME) + 10);
	  sprintf (inNameHH, "%ld_%s_hh", sigOut->INDEX, inst->INSNAME);
	  addTH (delaiTH, namealloc (inNameHH), delaiInstHH);
	  addTH (delaiminTH, namealloc (inNameHH), delaiInstMinHH);
	  mbkfree ((void *) inNameHH);

	  inNameLL = (char *) mbkalloc (strlen (inst->INSNAME) + 10);
	  sprintf (inNameLL, "%ld_%s_ll", sigOut->INDEX, inst->INSNAME);
	  addTH (delaiTH, namealloc (inNameLL), delaiInstLL);
	  addTH (delaiminTH, namealloc (inNameLL), delaiInstMinLL);
	  mbkfree ((void *) inNameLL);
	}
      else if (cellLoadType (inst->FIGNAME) == BUS)
	{
	  locon_list *con;
	  chain_list *lstCon;
	  ptype_list *liste = getptype (sigOut->USER, (long) LOFIGCHAIN);


	  /* parcours des connecteurs sur le signal bus */
	  for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
	    {
	      locon_list *conAux = (locon_list *) lstCon->DATA;

	      /* recherche de toutes les portes tristate sur le signal bus */
	      if ((conAux->TYPE == INTERNAL) &&
		  ((conAux->DIRECTION == TRISTATE) ||
		   (conAux->DIRECTION == TRANSCV)))
		{
		  loins_list *porteBus = (loins_list *) conAux->ROOT;
		  int val = searchTH (NO_CELLTH, porteBus->FIGNAME);
		  cellList *cell;

		  printf ("-Bus =  %s-\n", porteBus->INSNAME);

		  if (val == EMPTYTH)
		    {
		      printf ("\nTIMING ANALYSIS : le modele %s n'est pas charge en memoire (cellLoad !)\n", porteBus->FIGNAME);
		      exit (1);
		    }

		  cell = (cellList *) val;
		  if (cellLoadPolarity (porteBus->FIGNAME) == AUT)
		    {
		      printf ("Other polarity for gate %s [%s]\n",
			      cell->nom, porteBus->FIGNAME);
		      exit (1);
		    }

		  /* calcul du delai sur la porte  bus */
		  for (con = porteBus->LOCON; con; con = con->NEXT)
		    {
		      if (con->DIRECTION == IN &&
			  !isvdd (con->NAME) && !isvss (con->NAME))
			{

			  if (searchExpr (cell->val_abl->VALABL, con->NAME))
			    {
			      /* entree dans le VALABL */
			      switch (cellLoadPolarity (porteBus->FIGNAME))
				{
				case POS:
				  delaiInstHH =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     sCin) / 1000;
				  delaiInstLL =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     sCin) / 1000;
				  delaiInstMinHH =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     capaInv) / 1000;
				  delaiInstMinLL =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     capaInv) / 1000;
				  break;
				case NEG:
				  delaiInstLH =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     sCin) / 1000;
				  delaiInstHL =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     sCin) / 1000;
				  delaiInstMinLH =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     capaInv) / 1000;
				  delaiInstMinHL =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     capaInv) / 1000;
				  break;
				}
			    }
			  else if (searchExpr (cell->val_abl->CNDABL, con->NAME))
			    {
			      if (ATOM (cell->val_abl->CNDABL))
				{
				  delaiInstHH =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     sCin) / 1000;
				  delaiInstMinHH =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     capaInv) / 1000;
				  delaiInstLL =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     sCin) / 1000;
				  delaiInstMinLL =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     capaInv) / 1000;
				}
			      else
				{
				  delaiInstHH =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     sCin) / 1000;
				  delaiInstMinHH =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     capaInv) / 1000;
				  delaiInstLL =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     sCin) / 1000;
				  delaiInstMinLL =
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     capaInv) / 1000;
				}
			    }
			  else
			    {
			      printf ("On tristate %s, the input %s neither in the command nor in the data\n", porteBus->INSNAME, con->NAME);
			      exit (1);
			    }

/*
   if (NO_TRACE >= 3)
 */
			  printf ("INST %s Con %s Delai %ld %ld\n",
				  porteBus->INSNAME, con->NAME,
				  delaiInstHH, delaiInstLL);

			}	/* if connecteur d'entree */
		    }		/* parcours connecteurs du bus (for) */
		}		/* traitement de chaque bus */
	    }			/* parcours des connecteurs sur le signal bus */

	  /* memorisation des delais sur le signal tristate */
	  inNameHH = (char *) mbkalloc (strlen (inst->INSNAME) + 10);
	  sprintf (inNameHH, "%ld_%s_hh", sigOut->INDEX, inst->INSNAME);
	  addTH (delaiTH, namealloc (inNameHH), delaiInstMinHH);
	  addTH (delaiminTH, namealloc (inNameHH), delaiInstHH);
	  mbkfree ((void *) inNameHH);

	  inNameLL = (char *) mbkalloc (strlen (inst->INSNAME) + 10);
	  sprintf (inNameLL, "%ld_%s_ll", sigOut->INDEX, inst->INSNAME);
	  addTH (delaiTH, namealloc (inNameLL), delaiInstMinLL);
	  addTH (delaiminTH, namealloc (inNameLL), delaiInstMinLL);
	  mbkfree ((void *) inNameLL);
	}			/* fin BUS */
      else
	{
	  locon_list *con;

/****** cellule combi ******/
	  for (con = inst->LOCON; con; con = con->NEXT)
	    {
	      if (con->DIRECTION == IN &&
		  !isvdd (con->NAME) && !isvss (con->NAME))
		{
		  /* Calcul du delai du connecteur de l'instance a traiter
		     en fonction de sa polarite */
		  switch (polarityInst)
		    {
		    case POS:
		      delaiInstHH = cellLoadTp (inst->FIGNAME, con->NAME, HH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 sCin) / 1000;
		      delaiInstLL = cellLoadTp (inst->FIGNAME, con->NAME, LL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 sCin) / 1000;
		      delaiInstMinHH = cellLoadTp (inst->FIGNAME, con->NAME, HH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 capaInv) / 1000;
		      delaiInstMinLL = cellLoadTp (inst->FIGNAME, con->NAME, LL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 capaInv) / 1000;

		      inNameHH = (char *) mbkalloc (strlen (con->NAME) + 10);
		      sprintf (inNameHH, "%ld_%s_hh", sigOut->INDEX, con->NAME);
		      addTH (delaiTH, namealloc (inNameHH), delaiInstHH);
		      addTH (delaiminTH, namealloc (inNameHH), delaiInstMinHH);
		      mbkfree ((void *) inNameHH);

		      inNameLL = (char *) mbkalloc (strlen (con->NAME) + 10);
		      sprintf (inNameLL, "%ld_%s_ll", sigOut->INDEX, con->NAME);
		      addTH (delaiTH, namealloc (inNameLL), delaiInstLL);
		      addTH (delaiminTH, namealloc (inNameLL), delaiInstMinLL);
		      mbkfree ((void *) inNameLL);
		      break;
		    case NEG:
		      delaiInstLH = cellLoadTp (inst->FIGNAME, con->NAME, LH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 sCin) / 1000;
		      delaiInstHL = cellLoadTp (inst->FIGNAME, con->NAME, HL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 sCin) / 1000;
		      delaiInstMinLH = cellLoadTp (inst->FIGNAME, con->NAME, LH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 capaInv) / 1000;
		      delaiInstMinHL = cellLoadTp (inst->FIGNAME, con->NAME, HL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 capaInv) / 1000;

		      inNameLH = (char *) mbkalloc (strlen (con->NAME) + 10);
		      sprintf (inNameLH, "%ld_%s_lh", sigOut->INDEX, con->NAME);
		      addTH (delaiTH, namealloc (inNameLH), delaiInstLH);
		      addTH (delaiminTH, namealloc (inNameLH), delaiInstMinLH);
		      mbkfree ((void *) inNameLH);

		      inNameHL = (char *) mbkalloc (strlen (con->NAME) + 10);
		      sprintf (inNameHL, "%ld_%s_hl", sigOut->INDEX, con->NAME);
		      addTH (delaiTH, namealloc (inNameHL), delaiInstHL);
		      addTH (delaiminTH, namealloc (inNameHL), delaiInstMinHL);
		      mbkfree ((void *) inNameHL);
		      break;
		    case AUT:
		      delaiInstHH =
			cellLoadTp (inst->FIGNAME, con->NAME, HH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 sCin) / 1000;
		      delaiInstLH =
			cellLoadTp (inst->FIGNAME, con->NAME, LH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 sCin) / 1000;
		      delaiInstHL =
			cellLoadTp (inst->FIGNAME, con->NAME, HL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 sCin) / 1000;
		      delaiInstLL =
			cellLoadTp (inst->FIGNAME, con->NAME, LL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 sCin) / 1000;
		      delaiInstMinHH =
			cellLoadTp (inst->FIGNAME, con->NAME, HH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 capaInv) / 1000;
		      delaiInstMinLH =
			cellLoadTp (inst->FIGNAME, con->NAME, LH) +
			(cellLoadR (inst->FIGNAME, con->NAME, NUP) *
			 capaInv) / 1000;
		      delaiInstMinHL =
			cellLoadTp (inst->FIGNAME, con->NAME, HL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 capaInv) / 1000;
		      delaiInstMinLL =
			cellLoadTp (inst->FIGNAME, con->NAME, LL) +
			(cellLoadR (inst->FIGNAME, con->NAME, NDOWN) *
			 capaInv) / 1000;

		      inNameHH = (char *) mbkalloc (strlen (con->NAME) + 10);
		      sprintf (inNameHH, "%ld_%s_hh", sigOut->INDEX, con->NAME);
		      addTH (delaiTH, namealloc (inNameHH), delaiInstHH);
		      addTH (delaiminTH, namealloc (inNameHH), delaiInstMinHH);
		      mbkfree ((void *) inNameHH);

		      inNameLL = (char *) mbkalloc (strlen (con->NAME) + 10);
		      sprintf (inNameLL, "%ld_%s_ll", sigOut->INDEX, con->NAME);
		      addTH (delaiTH, namealloc (inNameLL), delaiInstLL);
		      addTH (delaiminTH, namealloc (inNameLL), delaiInstMinLL);
		      mbkfree ((void *) inNameLL);

		      inNameLH = (char *) mbkalloc (strlen (con->NAME) + 10);
		      sprintf (inNameLH, "%ld_%s_lh", sigOut->INDEX, con->NAME);
		      addTH (delaiTH, namealloc (inNameLH), delaiInstLH);
		      addTH (delaiminTH, namealloc (inNameLH), delaiInstMinLH);
		      mbkfree ((void *) inNameLH);

		      inNameHL = (char *) mbkalloc (strlen (con->NAME) + 10);
		      sprintf (inNameHL, "%ld_%s_hl", sigOut->INDEX, con->NAME);
		      addTH (delaiTH, namealloc (inNameHL), delaiInstHL);
		      addTH (delaiminTH, namealloc (inNameHL), delaiInstMinHL);
		      mbkfree ((void *) inNameHL);
		      break;
		    }		/* switch sur la polarite */

		  if (NO_TRACE >= 3)
		    printf ("INST %s Con %s - delai HH %ld LL %ld LH %ld HL %ld\n",
			 inst->INSNAME, con->NAME, delaiInstHH, delaiInstLL,
			    delaiInstLH, delaiInstHL);
		}		/* if connecteur d'entree */
	      else if (con->DIRECTION != OUT && !isvdd (con->NAME) && !isvss (con->NAME))
		printf ("TIMING ANALYSIS : Unknown direction for connector %s of %s\n", con->NAME, inst->INSNAME);
	    }			/* parcours des connecteurs de l'instance (for) */
	}			/* fin cellule combi */
    }
}



/*---------------------------------------------------------------------------
capaMoyenne	:  renvoie la capa d'entree moyenne de SCLIB
-----------------------------------------------------------------------------
retour          : int
---------------------------------------------------------------------------*/
long 
capaMoyenne ()
{
  long cinSomme = 0;
  cellList *cells;
  long nbCells = 0;

  for (cells = NO_CELLS; cells; cells = cells->NEXT)
    {
      ptype_list *c;

      for (c = cells->surface_tech; c; c = c->NEXT)
	{
	  long sumCinCell = 0;
	  long nbIn = 0;
	  Alist *in;

	  for (in = cells->entrees; in; in = in->NEXT)
	    {
	      char *CIN = (char *) mbkalloc (5 + strlen (in->name));

	      sprintf (CIN, "cin_%s", in->name);
	      sumCinCell += (long) searchGeneric (cells, c->DATA, CIN);
	      nbIn++;
	    }
	  if (nbIn != 0)
	    {
/*
   printf("Cell %s --> capa %ld\n", cells->nomlog, sumCinCell/nbIn);
 */
	      cinSomme += sumCinCell / nbIn;
	      nbCells++;
	    }
	}
    }
  if (NO_TRACE > 0)
    printf ("nbCells = %ld capa moyenne = %ld\n", nbCells, cinSomme / nbCells);

  return cinSomme / nbCells;
}

/*---------------------------------------------------------------------------

	       memorisation dans une table de hasch.
	       On tient compte des resistances & retards sur les entrees :
               NO_RESISTPI et NO_DELAYPI
-----------------------------------------------------------------------------
retour       : void
---------------------------------------------------------------------------*/
void 
prepareLofig (lofig)
     lofig_list *lofig;
{
  loins_list *inst;
  locon_list *con;

  /* pour une netlist extraite par lynx -f */
/*
   pTH routTH = createTH(2*nbGates);
   lofig_list *lofigTrans;
   losig_list *sig;
   char *nameRout = mbkalloc(strlen(NO_NAME) + 2);

   IN_LO[0] = 'f';
   IN_LO[1] = 'n';
   IN_LO[2] = 'e';
   sprintf(nameRout, "%sr", NO_NAME);
   lofigTrans = getlofig(nameRout, 'A');
   if (lofigTrans->LOINS)
   printf("LOINS ds %s.%s\n", lofigTrans->NAME, IN_LO);

   for(sig = lofigTrans->LOSIG; sig; sig = sig->NEXT)
   {
   addTH(routTH, getsigname(sig), (int)(1000.0*sig->CAPA));
   }
 */

  /* FIN : pour une netlist extraite par lynx -f */

  /* chainage signal - liste de connecteurs */
  lofigchain (lofig);

  /* initialisation des tables de hasch */
  pathTH = createTH (100);
  capaTH = createTH (2 * nbGates);
  delaiTH = createTH (2 * nbGates);
  delaiminTH = createTH (2 * nbGates);
  loconOutTH = createTH (nbGates);
  genConTH = createTH (nbGates);
  genUpTH = createTH (nbGates);
  genDwTH = createTH (nbGates);

  /* Parcours des connecteurs externes d'entree */
  for (con = lofig->LOCON; con; con = con->NEXT)
    {
      if (con->DIRECTION == IN && !isvdd (con->NAME) && !isvss (con->NAME))
	{
	  chain_list *lstCon;
	  long sCin = 0;
	  long nbFanout = 0;
	  ptype_list *liste = getptype (con->SIG->USER, (long) LOFIGCHAIN);

	  /* parcours du fanout et calcul de la somme des capas */
	  for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
	    {
	      locon_list *conAux = (locon_list *) lstCon->DATA;

	      if (conAux->ROOT != (void *) lofig)
		{
		  if (conAux->TYPE == EXTERNAL)
		    /* connecteur externe */
		    sCin += cellLoadCin ((char *) NULL, conAux->NAME);
		  else
		    /* connecteur interne */
		  if (conAux->DIRECTION == IN)
		    {
		      loins_list *cellOut = (loins_list *) conAux->ROOT;

		      sCin += cellLoadCin (cellOut->FIGNAME, conAux->NAME);
		      nbFanout++;
		    }
		  else
		    printf ("DELAY on PRIMARY INPUT : Unknown connector '%s'\n",
			    conAux->NAME);
		}
	    }

/*
   printf("Connecteur ext. %s : somme capa = %d, nb fanout = %d %d, routage = %ld\n",
   con->NAME, sCin, nbFanout, (long)sqrt((double)nbFanout), (long)((con->SIG->CAPA)*1000.0));
   printf("Fanout IN %s = %ld - Capa in = %ld, capa TAS = %-6.2f\n",
   con->NAME, nbFanout, sCin, (con->SIG->CAPA)*1000.00);
 */

/*
   printf("Connecteur %s : somme capa = %d, nb fanout = %d, routage = %d\n",
   con->NAME, sCin, nbFanout, searchTH(routTH, getsigname(con->SIG)));
 */

/*
   sCin = (long)((con->SIG->CAPA)*1000.0);
 */

	  addTH (capaTH, getsigname (con->SIG), sCin);
	  addTH (genConTH, getsigname (con->SIG), (int) con);

	  majDelaiCon (con, sCin);
	}
    }				/* fin parcours des connecteurs */

  /* parcours des instances */
  for (inst = lofig->LOINS; inst; inst = inst->NEXT)
    if (cellLoadType (inst->FIGNAME) != TIE)
      {
	int nbOut = 0;
	locon_list *conOut;

	for (con = inst->LOCON; con; con = con->NEXT)
	  {
	    if ((con->DIRECTION == OUT) ||
		(con->DIRECTION == INOUT) ||
		(con->DIRECTION == TRISTATE) ||
		(con->DIRECTION == TRANSCV))
	      {
		/* connecteur de sortie */
		nbOut++;
		conOut = con;

		addTH (loconOutTH, inst->INSNAME, (int) conOut);
		addTH (genUpTH, getsigname (con->SIG), (int) inst);
		addTH (genDwTH, getsigname (con->SIG), (int) inst);
	      }
	    else
	      {
		if (con->DIRECTION != IN)
		  printf ("WARNING : Unknown direction for connector %s of instance %s [%s]\n",
			  con->NAME, inst->INSNAME, inst->FIGNAME);
	      }
	  }

	if (nbOut != 1)
	  printf ("WARNING : Number of output connectors = %d for %s [%s]\n",
		  nbOut, inst->INSNAME, inst->FIGNAME);

	if (cellLoadType (inst->FIGNAME) == REG)
	  {
	    /* cas d'un registre */
	    losig_list *sigOut = conOut->SIG;
	    ptype_list *liste = getptype (sigOut->USER, (long) LOFIGCHAIN);
	    chain_list *lstCon;
	    long sCin = 0, nbFanout = 0;

	    if (!liste)
	      {
		printf ("TIMING ANALYSIS : No type LOFIGCHAIN for %s[%s] on %s\n",
			inst->INSNAME, inst->FIGNAME, conOut->NAME);
		exit (1);
	      }

	    /* pour le connecteur de sortie, calculer son fanout (somme CIN) */
	    for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
	      {
		locon_list *conAux = (locon_list *) lstCon->DATA;

		if (conAux->ROOT != (void *) inst)
		  {
		    if (conAux->TYPE == EXTERNAL)
		      {
			/* connecteur externe */
			sCin += cellLoadCin ((char *) NULL, conAux->NAME);
			nbFanout++;
		      }
		    else
		      /* connecteur interne */
		    if (conAux->DIRECTION == IN)
		      {
			loins_list *cellOut = (loins_list *) conAux->ROOT;

			/* portes attaquees par l'instance a traiter */
			sCin += cellLoadCin (cellOut->FIGNAME, conAux->NAME);
			nbFanout++;
		      }
		    else
		      {
			printf ("TIMING ANALYSIS : Unknown connector '%s' of %s[%s]\n",
			conAux->NAME, ((loins_list *) conAux->ROOT)->INSNAME,
				((loins_list *) conAux->ROOT)->FIGNAME);
		      }
		  }
	      }			/* fin calcul de la somme des capas */



	    /* memorisation de SCIN sur le signal de sortie de inst */
	    addTH (capaTH, getsigname (sigOut), sCin);

	    majDelaiInst (inst, sCin);

	  }			/* fin du traitement des registres */
	else
	  {
	    /* cas d'une cellule combinatoire */
	    losig_list *sigOut = conOut->SIG;
	    ptype_list *liste = getptype (sigOut->USER, (long) LOFIGCHAIN);
	    short conExt = 0, conReg = 0;
	    long sCin = 0, nbFanout = 0;
	    chain_list *lstCon;

	    if (!liste)
	      {
		printf ("TIMING ANALYSIS : No type LOFIGCHAIN for %s\n", inst->INSNAME);
		exit (1);
	      }

	    /* pour le connecteur de sortie, calculer son fanout (somme CIN) */
	    for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
	      {
		locon_list *conAux = (locon_list *) lstCon->DATA;

		if (conAux->ROOT != (void *) inst)
		  {
		    if (conAux->TYPE == EXTERNAL)
		      {
			/* connecteur externe */
			sCin += cellLoadCin ((char *) NULL, conAux->NAME);
			nbFanout++;
			conExt += 1;
		      }
		    else
		      /* connecteur interne */
		    if (conAux->DIRECTION == IN)
		      {
			loins_list *cellOut = (loins_list *) conAux->ROOT;

			/* portes attaquees par l'instance a traiter */
			sCin += cellLoadCin (cellOut->FIGNAME, conAux->NAME);
			nbFanout++;
			if (cellLoadType (cellOut->FIGNAME) == REG)
			  conReg += 1;
		      }
		    else
		      {
			if ((conAux->DIRECTION != TRANSCV) &&
			    (conAux->DIRECTION != TRISTATE))
			  printf ("TIMING ANALYSIS : Unknown connector '%s' of %s[%s]\n",
				  conAux->NAME, ((loins_list *) conAux->ROOT)->INSNAME,
				  ((loins_list *) conAux->ROOT)->FIGNAME);
		      }
		  }
	      }			/* fin du calcul de la somme CIN */

/*
   printf("Instance %s[%s]\t: somme capa = %d, nb fanout = %d, routage = %d\n",
   inst->INSNAME, inst->FIGNAME, sCin, nbFanout, searchTH(routTH, getsigname(sigOut)));
 */

/*
   printf("Fanout %s[%s]\t = %ld - Capa in = %ld, capa TAS = %-6.2f\n",
   inst->INSNAME, inst->FIGNAME, nbFanout,
   sCin, (sigOut->CAPA)*1000.00);
 */


	    /* somme capa = 0 et une seul connecteur sur le signal */
	    if (sCin == 0 && !((chain_list *) liste->DATA)->NEXT)
	      {
		/* detection des portes avec la sortie en l'air */
		printf ("\n\n\tNetlist %s has gate %s [%s] without fanouts.\n\n",
			NO_LOG->NAME, inst->INSNAME, inst->FIGNAME);
		exit (1);
	      }

	    /* memorisation de la somme des capa */
	    addTH (capaTH, getsigname (sigOut), sCin);

	    /* l'instance attaque au - un connecteur externe ou un registre */
	    /* c'est donc une fin de chemin */
	    if ((conExt > 0) || (conReg > 0))
	      addTH (pathTH, inst->INSNAME, 1);

	    if (NO_TRACE >= 3)
	      printf (" Computing delay on %s [%s], sum Cin = %ld\n",
		      inst->INSNAME, inst->FIGNAME, sCin);

	    majDelaiInst (inst, sCin);

	  }
      }				/* parcours des instances != TIE */

  /* pour une netlist extraite par lynx -f */
/*
   destroyTH(routTH);
 */
}

/*---------------------------------------------------------------------------
calcul_retard : calcul des retards pour les instances d'une LOFIG
-----------------------------------------------------------------------------
retour       : les retards sont memorises dans le champ USER des signaux
---------------------------------------------------------------------------*/
void 
calcul_retard (lofig, affich)
     lofig_list *lofig;
     short affich;
{
  loins_list *inst;

  RAZlofig (lofig);

  delaiCriticUp = delaiCriticDw = 0;
  porteCriticUp = porteCriticDw = NULL;
  delaiCriticMinUp = delaiCriticMinDw = 0;
  profondeurCircuit = 0;
  instProf = NULL;

  /* parcours des instances de la figure */
  for (inst = lofig->LOINS; inst; inst = inst->NEXT)
    if (cellLoadType (inst->FIGNAME) != TIE)
      {
	int val = searchTH (loconOutTH, inst->INSNAME);

	if (val != EMPTYTH)
	  {
	    losig_list *sigOut = ((locon_list *) val)->SIG;

	    if (!getptype (sigOut->USER, (long) RETARDUP))
	      retard (inst);
	  }
	else
	  {
	    printf ("ERROR calcul_retard : no type LOCONOUT for %s[%s]\n",
		    inst->INSNAME, inst->FIGNAME);
	    exit (-1);
	  }
      }

  if (affich)
    {
      cheminCritique (porteCriticUp, NUP, delaiCriticUp, 0);
      cheminCritique (porteCriticDw, NDOWN, delaiCriticDw, 0);
      if (affich >= 4)
	{
	  printf ("\n Maximal depth = %ld sur %s\n\n",
		  profondeurCircuit, instProf->INSNAME);
	  cheminCritique (porteCriticMinUp, NUP, delaiCriticMinUp, 1);
	  cheminCritique (porteCriticMinDw, NDOWN, delaiCriticMinDw, 1);
	}
    }
  else
    printf ("  Timing Analysis : Delay UP %ld (%s) - Delay DW %ld (%s)\n",
	    delaiCriticUp, porteCriticUp->INSNAME,
	    delaiCriticDw, porteCriticDw->INSNAME);


/*
   printf("End of Timing Analysis on %s\n\n", lofig->NAME);
 */
}

/*---------------------------------------------------------------------------
getDel  : recupere le delai (Tp + RC) dans la table de hash delaiTH
			   --------
			--o|      |   |--	o = con
			---| inst |---|-- sig
			---|      |   |--
			   --------
			   <------><---->
			      Tp   + RC
-----------------------------------------------------------------------------
retour : un long
---------------------------------------------------------------------------*/
long 
getDel (inst, con, signal, trans, min)
     loins_list *inst;		/* NULL pour une entree primaire */
     locon_list *con;		/* NULL pour un registre, un bus et une constante */
     losig_list *signal;		/* signal attache a con ou a la sortie de inst */
     char *trans;
     short min;
{
  int val;

  if (con)
    {
      char *inName = (char *) mbkalloc (strlen (con->NAME) + 10);

      sprintf (inName, "%ld_%s_%s", signal->INDEX, con->NAME, trans);
      val = searchTH ((min == 0) ? delaiTH : delaiminTH, namealloc (inName));
      mbkfree ((void *) inName);
    }
  else
    {
      char *inName = (char *) mbkalloc (strlen (inst->INSNAME) + 10);

      sprintf (inName, "%ld_%s_%s", signal->INDEX, inst->INSNAME, trans);
      val = searchTH ((min == 0) ? delaiTH : delaiminTH, namealloc (inName));
      mbkfree ((void *) inName);
    }
  if ((val == EMPTYTH) || (val == DELETETH))
    {
      if (con)
	printf ("ERROR timing analysis : no delay %s for %s[%s/%s] on signal %ld\n",
		trans, inst->INSNAME, inst->FIGNAME, con->NAME, signal->INDEX);
      else
	printf ("ERROR timing analysis : no delay %s for %s[%s]\n",
		trans, inst->INSNAME, inst->FIGNAME);
      exit (1);
    }
  else
    return (long) val;
}

/*---------------------------------------------------------------------------
retard  : calcul du retard et de la profondeur d'une instance 
         (appel recursif sur les entrees)
-----------------------------------------------------------------------------
retour : RIEN. Memorisation dans le champ USER du signal de sortie
         Type RETARDUP, RETARDDW et PROF
---------------------------------------------------------------------------*/
void 
retard (inst)
     loins_list *inst;
{
  if (cellLoadType (inst->FIGNAME) == REG)
    {
      int val = searchTH (loconOutTH, inst->INSNAME);
      losig_list *sigOut;

      if (val != EMPTYTH)
	sigOut = ((locon_list *) val)->SIG;
      else
	{
	  printf ("TIMING ANALYSIS : No type LOCONOUT for %s[%s]\n",
		  inst->INSNAME, inst->FIGNAME);
	  exit (1);
	}

      /* recuperer les delais calcules par prepareLofig */
      sigOut->USER = addptype (sigOut->USER, (long) RETARDUP,
			     (void *) getDel (inst, NULL, sigOut, "hh", 0));
      sigOut->USER = addptype (sigOut->USER, (long) RETARDDW,
			     (void *) getDel (inst, NULL, sigOut, "ll", 0));
      sigOut->USER = addptype (sigOut->USER, (long) RETARDMINUP,
			     (void *) getDel (inst, NULL, sigOut, "hh", 1));
      sigOut->USER = addptype (sigOut->USER, (long) RETARDMINDW,
			     (void *) getDel (inst, NULL, sigOut, "ll", 1));

      /* initialisation de la profondeur du registre a 1 */
      sigOut->USER = addptype (sigOut->USER, (long) PROF, (void *) 1);

      return;
    }
  else
    {
      /* cas d'une cellule combinatoire */
      int polarityInst = cellLoadPolarity (inst->FIGNAME);
      int val = searchTH (loconOutTH, inst->INSNAME);
      locon_list *conOut;
      losig_list *sigOut;
      ptype_list *liste;
      long sCin = 0;
      chain_list *lstCon;
      long maxDelUp = 0, maxDelDw = 0;
      long maxDelMinUp = 0, maxDelMinDw = 0;
      void *criticUp = NULL, *criticDw = NULL;
      void *criticMinUp = NULL, *criticMinDw = NULL;
      long maxProfond = 0;
      long genreUp, genreDw;
      long genreMinUp, genreMinDw;
      locon_list *con;

      if (val != EMPTYTH)
	{
	  conOut = (locon_list *) val;
	  sigOut = conOut->SIG;
	  liste = getptype (sigOut->USER, (long) LOFIGCHAIN);
	}
      else
	{
	  printf ("TIMING ANALYSIS : No type LOCONOUT for %s\n", inst->INSNAME);
	  exit (1);
	}

      if (!liste)
	{
	  printf ("TIMING ANALYSIS : No type LOFIGCHAIN for %s\n", inst->INSNAME);
	  exit (1);
	}

      /* pour le connecteur de sortie, recuperer son fanout (somme CIN) */
      sCin = (long) searchTH (capaTH, getsigname (sigOut));

      if (NO_TRACE >= 3)
	printf (" Computing delay on %s [%s], sum Cin = %ld\n",
		inst->INSNAME, inst->FIGNAME, sCin);

      /* Parcours des connecteurs de l'instance a traiter pour
         trouver les connecteurs d'entree et recuperer les retards
         des signaux sur ces connecteurs */

      if (cellLoadType (inst->FIGNAME) == CONST)
	{
	  /* Cas des cellules zero_y et one_y : pas d'entrees */
	  int val = searchTH (loconOutTH, inst->INSNAME);
	  losig_list *sigOut;

	  if (val != EMPTYTH)
	    sigOut = ((locon_list *) val)->SIG;
	  else
	    {
	      printf ("TIMING ANALYSIS : No type LOCONOUT for %s[%s]\n",
		      inst->INSNAME, inst->FIGNAME);
	      exit (1);
	    }

	  /* recuperer les delais calcules par prepareLofig */
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDUP,
			     (void *) getDel (inst, NULL, sigOut, "hh", 0));
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDDW,
			     (void *) getDel (inst, NULL, sigOut, "ll", 0));
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDMINUP,
			     (void *) getDel (inst, NULL, sigOut, "hh", 1));
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDMINDW,
			     (void *) getDel (inst, NULL, sigOut, "ll", 1));

	  /* initialisation de la profondeur de la constante a 1 */
	  sigOut->USER = addptype (sigOut->USER, (long) PROF, (void *) 1);
	}
      else if (cellLoadType (inst->FIGNAME) == BUS)
	{
	  long delaiInstUp = 0, delaiInstDw = 0;
	  long delaiInstMinUp = 0, delaiInstMinDw = 0;
	  loins_list *porteBusCriticUp, *porteBusCriticDw;

	  /* parcours des connecteurs sur le signal bus */
	  for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
	    {
	      locon_list *conAux = (locon_list *) lstCon->DATA;

	      /* recherche de toutes les portes tristate sur le signal bus */
	      if ((conAux->TYPE == INTERNAL) &&
		  ((conAux->DIRECTION == TRISTATE) ||
		   (conAux->DIRECTION == TRANSCV)))
		{
		  loins_list *porteBus = (loins_list *) conAux->ROOT;
		  int val = searchTH (NO_CELLTH, porteBus->FIGNAME);
		  cellList *cell;
		  long maxDelBusUp = 0, maxDelBusDw = 0;
		  long maxDelMinBusUp = 0, maxDelMinBusDw = 0;
		  long genreBusDw, genreBusUp;
		  long genreMinBusDw, genreMinBusUp;
		  void *criticBusDw = NULL, *criticBusUp = NULL;
		  void *criticMinBusDw = NULL, *criticMinBusUp = NULL;

		  printf ("-Bus =  %s-\n", porteBus->INSNAME);

		  if (val == EMPTYTH)
		    {
		      printf ("\nTIMING ANALYSIS : le modele %s n'est pas charge en memoire (cellLoad !)\n", porteBus->FIGNAME);
		      exit (1);
		    }

		  cell = (cellList *) val;
		  if (cellLoadPolarity (porteBus->FIGNAME) == AUT)
		    {
		      printf ("Other polarity for gate %s [%s]\n",
			      cell->nom, porteBus->FIGNAME);
		      exit (1);
		    }

		  /* calcul du retard sur la porte  bus */
		  for (con = porteBus->LOCON; con; con = con->NEXT)
		    {
		      if (con->DIRECTION == IN &&
			  !isvdd (con->NAME) && !isvss (con->NAME))
			{
			  loins_list *cellGenereUp = (loins_list *) NULL;
			  loins_list *cellGenereDw = (loins_list *) NULL;
			  locon_list *conGenere = (locon_list *) NULL;
			  long delaiGenUp = 0, delaiGenDw = 0;
			  long delaiGenMinUp = 0, delaiGenMinDw = 0;
			  long profond = 0;
			  ptype_list *user;
			  int gen;

			  if ((gen = searchTH (genUpTH, getsigname (con->SIG))) != EMPTYTH)
			    {
			      /* C'est une porte qui genere le signal */
			      cellGenereUp = (loins_list *) gen;
			      cellGenereDw = (loins_list *) searchTH (genDwTH,
						     getsigname (con->SIG));
			    }
			  else
			    /* C'est un connecteur qui genere le signal */
			    conGenere = (locon_list *) searchTH (genConTH,
						     getsigname (con->SIG));
			  if (!(user = getptype (con->SIG->USER, (long) RETARDUP)))
			    {
			      /* retards pas encore calcules sur l'entree */
			      if (NO_TRACE >= 3)
				printf ("  Gate %s generates signal on connector %s of %s\n",
					cellGenereUp->INSNAME, con->NAME, porteBus->INSNAME);

			      /* appel recursif pour la porte generant le signal */
			      retard (cellGenereUp);
			    }
			  /* retards calcules maintenant... */
			  delaiGenUp = (long) (getptype (con->SIG->USER,
						    (long) RETARDUP))->DATA;
			  delaiGenDw = (long) (getptype (con->SIG->USER,
						    (long) RETARDDW))->DATA;
			  delaiGenMinUp = (long) (getptype (con->SIG->USER,
						 (long) RETARDMINUP))->DATA;
			  delaiGenMinDw = (long) (getptype (con->SIG->USER,
						 (long) RETARDMINDW))->DATA;
			  profond = (long) (getptype (con->SIG->USER,
						      (long) PROF)->DATA);
			  if (searchExpr (cell->val_abl->VALABL, con->NAME))
			    {
			      /* entree dans le VALABL */
			      switch (cellLoadPolarity (porteBus->FIGNAME))
				{
				case POS:
				  delaiInstUp = delaiGenUp +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     sCin) / 1000;
				  delaiInstDw = delaiGenDw +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     sCin) / 1000;
				  delaiInstMinUp = delaiGenMinUp +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     capaNot ()) / 1000;
				  delaiInstMinDw = delaiGenMinDw +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     capaNot ()) / 1000;
				  break;
				case NEG:
				  delaiInstUp = delaiGenDw +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     sCin) / 1000;
				  delaiInstDw = delaiGenUp +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     sCin) / 1000;
				  delaiInstMinUp = delaiGenMinDw +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     capaNot ()) / 1000;
				  delaiInstMinDw = delaiGenMinUp +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     capaNot ()) / 1000;
				  break;
				}
			    }
			  else if (searchExpr (cell->val_abl->CNDABL, con->NAME))
			    {
			      if (ATOM (cell->val_abl->CNDABL))
				{
				  delaiInstUp = delaiGenUp +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     sCin) / 1000;
				  delaiInstMinUp = delaiGenMinUp +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     capaNot ()) / 1000;
				  delaiInstDw = delaiGenUp +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     sCin) / 1000;
				  delaiInstMinDw = delaiGenMinUp +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, HL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     capaNot ()) / 1000;
				}
			      else
				{
				  delaiInstUp = delaiGenDw +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     sCin) / 1000;
				  delaiInstMinUp = delaiGenMinDw +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LH) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NUP) *
				     capaNot ()) / 1000;
				  delaiInstDw = delaiGenDw +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     sCin) / 1000;
				  delaiInstMinDw = delaiGenMinDw +
				    cellLoadTp (porteBus->FIGNAME, con->NAME, LL) +
				    (cellLoadR (porteBus->FIGNAME, con->NAME, NDOWN) *
				     capaNot ()) / 1000;
				}
			    }
			  else
			    {
			      printf ("Entree ni dans la commande ni dans la data : %s\n", con->NAME);
			      exit (1);
			    }
			  if (profond > maxProfond)
			    maxProfond = profond;

			  if (delaiInstDw > maxDelBusDw)
			    {
			      if (cellGenereUp)
				{
				  if (cellLoadPolarity (porteBus->FIGNAME) == POS)
				    criticBusDw = (void *) cellGenereDw;
				  else
				    criticBusDw = (void *) cellGenereUp;
				  genreBusDw = LOINSDW;
				}
			      else
				{
				  criticBusDw = (void *) conGenere;
				  genreBusDw = LOCONDW;
				}
			      maxDelBusDw = delaiInstDw;
			    }
			  if (delaiInstUp > maxDelBusUp)
			    {
			      if (cellGenereUp)
				{
				  if (cellLoadPolarity (porteBus->FIGNAME) == POS)
				    criticBusUp = (void *) cellGenereUp;
				  else
				    criticBusUp = (void *) cellGenereDw;
				  genreBusUp = LOINSUP;
				}
			      else
				{
				  criticBusUp = (void *) conGenere;
				  genreBusUp = LOCONUP;
				}
			      maxDelBusUp = delaiInstUp;
			    }
			  if (delaiInstMinDw > maxDelMinBusDw)
			    {
			      if (cellGenereUp)
				{
				  if (cellLoadPolarity (porteBus->FIGNAME) == POS)
				    criticMinBusDw = (void *) cellGenereDw;
				  else
				    criticMinBusDw = (void *) cellGenereUp;
				  genreMinBusDw = LOINSMINDW;
				}
			      else
				{
				  criticMinBusDw = (void *) conGenere;
				  genreMinBusDw = LOCONMINDW;
				}
			      maxDelMinBusDw = delaiInstMinDw;
			    }
			  if (delaiInstMinUp > maxDelMinBusUp)
			    {
			      if (cellGenereUp)
				{
				  if (cellLoadPolarity (porteBus->FIGNAME) == POS)
				    criticMinBusUp = (void *) cellGenereUp;
				  else
				    criticMinBusUp = (void *) cellGenereDw;
				  genreMinBusUp = LOINSMINUP;
				}
			      else
				{
				  criticMinBusUp = (void *) conGenere;
				  genreMinUp = LOCONMINUP;
				}
			      maxDelMinBusUp = delaiInstMinUp;
			    }
/*
   if (NO_TRACE >= 3)
 */
			  printf ("INST %s Con %s Gen [%s] %ld %ld Delai %ld %ld\n",
				  porteBus->INSNAME, con->NAME,
				  (cellGenereUp) ? cellGenereUp->INSNAME : conGenere->NAME,
			  delaiGenUp, delaiGenDw, delaiInstUp, delaiInstDw);
			}	/* if connecteur d'entree */
		    }		/* parcours connecteurs du bus */
		  printf ("Bus %s cUp = %s cDw = %s\n", porteBus->INSNAME,
			  ((loins_list *) criticBusUp)->INSNAME,
			  ((loins_list *) criticBusDw)->INSNAME);
		  porteBus->USER = addptype (porteBus->USER,
					  genreBusUp, (void *) criticBusUp);
		  porteBus->USER = addptype (porteBus->USER,
					  genreBusDw, (void *) criticBusDw);
		  porteBus->USER = addptype (porteBus->USER,
				    genreMinBusUp, (void *) criticMinBusUp);
		  porteBus->USER = addptype (porteBus->USER,
				    genreMinBusDw, (void *) criticMinBusDw);
		  if (maxDelBusDw > maxDelDw)
		    {
		      criticDw = criticBusDw;
		      genreDw = genreBusDw;
		      maxDelDw = maxDelBusDw;
		      porteBusCriticDw = porteBus;
		    }
		  if (maxDelBusUp > maxDelUp)
		    {
		      criticUp = criticBusUp;
		      genreUp = genreBusUp;
		      maxDelUp = maxDelBusUp;
		      porteBusCriticUp = porteBus;
		    }
		  if (maxDelMinBusDw > maxDelMinDw)
		    {
		      criticMinDw = criticMinBusDw;
		      genreMinDw = genreMinBusDw;
		      maxDelMinDw = maxDelMinBusDw;
		    }
		  if (maxDelMinBusUp > maxDelMinUp)
		    {
		      criticMinUp = criticMinBusDw;
		      genreMinUp = genreMinBusUp;
		      maxDelMinUp = maxDelMinBusUp;
		    }
		}		/* traitement de chaque bus */
	    }			/* parcours des instances sur le signal bus */

	  /* introduire porte critique pour generation sur signal bus */
	  printf ("--Bus critic Up = %s (%ld), Bus critic Dw = %s (%ld)\n",
		  porteBusCriticUp->INSNAME, maxDelUp,
		  porteBusCriticDw->INSNAME, maxDelDw);
	  addTH (genUpTH, getsigname (sigOut), (int) porteBusCriticUp);
	  addTH (genDwTH, getsigname (sigOut), (int) porteBusCriticDw);

	  /* memorisation des retards sur le signal tristate */
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDUP,
				   (void *) maxDelUp);
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDDW,
				   (void *) maxDelDw);
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDMINUP,
				   (void *) maxDelMinUp);
	  sigOut->USER = addptype (sigOut->USER, (long) RETARDMINDW,
				   (void *) maxDelMinDw);
	  sigOut->USER = addptype (sigOut->USER, (long) PROF,
				   (void *) maxProfond);

	  /* mise a jour des profondeurs et delais max pour le circuit */
	  if (maxProfond > profondeurCircuit)
	    {
	      profondeurCircuit = maxProfond;
	    }

	  if (searchTH (pathTH, inst->INSNAME) == 1)
	    {
	      if (maxDelUp > delaiCriticUp)
		{
		  porteCriticUp = porteBusCriticUp;
		  delaiCriticUp = maxDelUp;
		}
	      if (maxDelDw > delaiCriticDw)
		{
		  porteCriticDw = porteBusCriticDw;
		  delaiCriticDw = maxDelDw;
		}
	      if (maxDelMinUp > delaiCriticMinUp)
		{
		  porteCriticMinUp = porteBusCriticDw;
		  delaiCriticMinUp = maxDelMinUp;
		}
	      if (maxDelMinDw > delaiCriticMinDw)
		{
		  porteCriticMinDw = porteBusCriticDw;
		  delaiCriticMinDw = maxDelMinDw;
		}
	    }
	}			/* fin BUS */
      else
/****** cellule combi ******/
	for (con = inst->LOCON; con; con = con->NEXT)
	  {
	    if (con->DIRECTION == IN &&
		!isvdd (con->NAME) && !isvss (con->NAME))
	      {
		loins_list *cellGenereUp = (loins_list *) NULL;
		loins_list *cellGenereDw = (loins_list *) NULL;
		locon_list *conGenere = (locon_list *) NULL;
		long delaiInstUp = 0, delaiInstDw = 0;
		long delaiInstMinUp = 0, delaiInstMinDw = 0;
		long delaiGenUp = 0, delaiGenDw = 0;
		long delaiGenMinUp = 0, delaiGenMinDw = 0;
		long profond = 0;
		ptype_list *user;
		int gen;

		if ((gen = searchTH (genUpTH, getsigname (con->SIG))) != EMPTYTH)
		  {
		    /* C'est une porte qui genere le signal */
		    cellGenereUp = (loins_list *) gen;
		    cellGenereDw = (loins_list *) searchTH (genDwTH,
						     getsigname (con->SIG));
		  }
		else
		  {
		    int genCon = searchTH (genConTH, getsigname (con->SIG));

		    /* C'est un connecteur qui genere le signal */
		    if (genCon != EMPTYTH)
		      {
			conGenere = (locon_list *) genCon;
		      }
		    else if (isvdd (getsigname (con->SIG)) ||
			     isvss (getsigname (con->SIG)))
		      {
			/* alimentations conncectees aux portes ... */
			printf ("\nERROR : gate %s[%s] has signal %s on the input %s.\n\n",
				inst->INSNAME, inst->FIGNAME,
				getsigname (con->SIG), con->NAME);
			exit (-1);
		      }
		    else
		      {
			printf ("\nERROR : gate %s[%s] has input signal %s [%s] driven neither by a gate nor by a conncetor.\n\n",
				inst->INSNAME, inst->FIGNAME,
				getsigname (con->SIG), con->NAME);
			exit (-1);
		      }
		  }
		if (!(user = getptype (con->SIG->USER, (long) RETARDUP)))
		  {
		    /* retards pas encore calcules sur le connecteur d'entree */
		    if (cellGenereUp)
		      {
			/* retards de la porte fanin pas encore calcules */
			if (NO_TRACE >= 3)
			  printf ("  Gate %s generates signal on connector %s of %s\n",
			   cellGenereUp->INSNAME, con->NAME, inst->INSNAME);

			/* appel recursif pour la porte qui genere le signal */
			retard (cellGenereUp);

			/* retards calcules maintenant... */
			delaiGenUp = (long) (getptype (con->SIG->USER, (long) RETARDUP))->DATA;
			delaiGenDw = (long) (getptype (con->SIG->USER, (long) RETARDDW))->DATA;
			delaiGenMinUp = (long) (getptype (con->SIG->USER, (long) RETARDMINUP))->DATA;
			delaiGenMinDw = (long) (getptype (con->SIG->USER, (long) RETARDMINDW))->DATA;
			profond = (long) (getptype (con->SIG->USER, (long) PROF)->DATA);
		      }
		    else
		      {
			/* retards sur l'entree a recuperer */
			delaiGenUp = getDel (NULL, conGenere, conGenere->SIG,
					     "hh", 0);
			delaiGenDw = getDel (NULL, conGenere, conGenere->SIG,
					     "ll", 0);
			delaiGenMinUp = getDel (NULL, conGenere, conGenere->SIG,
						"hh", 1);
			delaiGenMinDw = getDel (NULL, conGenere, conGenere->SIG,
						"ll", 1);
			profond = (long) 0;
			conGenere->SIG->USER = addptype (conGenere->SIG->USER,
							 (long) RETARDUP,
						       (void *) delaiGenUp);
			conGenere->SIG->USER = addptype (conGenere->SIG->USER,
							 (long) RETARDMINUP,
						    (void *) delaiGenMinUp);
			conGenere->SIG->USER = addptype (conGenere->SIG->USER,
							 (long) RETARDDW,
						       (void *) delaiGenDw);
			conGenere->SIG->USER = addptype (conGenere->SIG->USER,
							 (long) RETARDMINDW,
						    (void *) delaiGenMinDw);
			conGenere->SIG->USER = addptype (conGenere->SIG->USER,
							 (long) PROF,
							 (void *) profond);

		      }
		  }
		else
		  {
		    delaiGenUp = (long) (getptype (con->SIG->USER, (long) RETARDUP))->DATA;
		    delaiGenDw = (long) (getptype (con->SIG->USER, (long) RETARDDW))->DATA;
		    delaiGenMinUp = (long) (getptype (con->SIG->USER, (long) RETARDMINUP))->DATA;
		    delaiGenMinDw = (long) (getptype (con->SIG->USER, (long) RETARDMINDW))->DATA;
		    profond = (long) (getptype (con->SIG->USER, (long) PROF)->DATA);
		  }

		/* Calcul du retard du connecteur de l'instance a traiter
		   en fonction de sa polarite */
		switch (polarityInst)
		  {
		  case POS:
		    delaiInstUp = delaiGenUp +
		      getDel (inst, con, sigOut, "hh", 0);
		    delaiInstDw = delaiGenDw +
		      getDel (inst, con, sigOut, "ll", 0);
		    delaiInstMinUp = delaiGenMinUp +
		      getDel (inst, con, sigOut, "hh", 1);
		    delaiInstMinDw = delaiGenMinDw +
		      getDel (inst, con, sigOut, "ll", 1);
		    break;
		  case NEG:
		    delaiInstUp = delaiGenDw +
		      getDel (inst, con, sigOut, "lh", 0);
		    delaiInstDw = delaiGenUp +
		      getDel (inst, con, sigOut, "hl", 0);
		    delaiInstMinUp = delaiGenMinDw +
		      getDel (inst, con, sigOut, "lh", 1);
		    delaiInstMinDw = delaiGenMinUp +
		      getDel (inst, con, sigOut, "hl", 1);
		    break;
		  case AUT:
		    delaiInstUp = MAX (
					delaiGenUp +
					getDel (inst, con, sigOut, "hh", 0),
					delaiGenDw +
					getDel (inst, con, sigOut, "lh", 0));
		    delaiInstDw = MAX (
					delaiGenUp +
					getDel (inst, con, sigOut, "hl", 0),
					delaiGenDw +
					getDel (inst, con, sigOut, "ll", 0));
		    delaiInstMinUp = MAX (
					   delaiGenMinUp +
					getDel (inst, con, sigOut, "hh", 1),
					   delaiGenMinDw +
				       getDel (inst, con, sigOut, "lh", 1));
		    delaiInstMinDw = MAX (
					   delaiGenMinUp +
					getDel (inst, con, sigOut, "hl", 1),
					   delaiGenMinDw +
				       getDel (inst, con, sigOut, "ll", 1));
		    break;
		  }		/* switch sur la polarite */

		if (profond > maxProfond)
		  maxProfond = profond;

		if (delaiInstDw > maxDelDw)
		  {
		    if (cellGenereUp)
		      {
			if (cellLoadPolarity (inst->FIGNAME) == POS)
			  criticDw = (void *) cellGenereDw;
			else
			  criticDw = (void *) cellGenereUp;
			genreDw = LOINSDW;
		      }
		    else
		      {
			criticDw = (void *) conGenere;
			genreDw = LOCONDW;
		      }
		    maxDelDw = delaiInstDw;
		  }
		if (delaiInstUp > maxDelUp)
		  {
		    if (cellGenereUp)
		      {
			if (cellLoadPolarity (inst->FIGNAME) == POS)
			  criticUp = (void *) cellGenereUp;
			else
			  criticUp = (void *) cellGenereDw;
			genreUp = LOINSUP;
		      }
		    else
		      {
			criticUp = (void *) conGenere;
			genreUp = LOCONUP;
		      }
		    maxDelUp = delaiInstUp;
		  }
		if (delaiInstMinDw > maxDelMinDw)
		  {
		    if (cellGenereUp)
		      {
			if (cellLoadPolarity (inst->FIGNAME) == POS)
			  criticMinDw = (void *) cellGenereDw;
			else
			  criticMinDw = (void *) cellGenereUp;
			genreMinDw = LOINSMINDW;
		      }
		    else
		      {
			criticMinDw = (void *) conGenere;
			genreMinDw = LOCONMINDW;
		      }
		    maxDelMinDw = delaiInstMinDw;
		  }
		if (delaiInstMinUp > maxDelMinUp)
		  {
		    if (cellGenereUp)
		      {
			if (cellLoadPolarity (inst->FIGNAME) == POS)
			  criticMinUp = (void *) cellGenereUp;
			else
			  criticMinUp = (void *) cellGenereDw;
			genreMinUp = LOINSMINUP;
		      }
		    else
		      {
			criticMinUp = (void *) conGenere;
			genreMinUp = LOCONMINUP;
		      }
		    maxDelMinUp = delaiInstMinUp;
		  }
		if (NO_TRACE >= 3)
		  printf ("INST %s Con %s Gen [%s] %ld %ld Retard %ld %ld\n",
			  inst->INSNAME, con->NAME,
		   (cellGenereUp) ? cellGenereUp->INSNAME : conGenere->NAME,
			  delaiGenUp, delaiGenDw, delaiInstUp, delaiInstDw);
	      }			/* if connecteur d'entree */
	    else if (con->DIRECTION != OUT && !isvdd (con->NAME) && !isvss (con->NAME))
	      printf ("TIMING ANALYSIS : Unknown direction for connector %s of %s\n", con->NAME, inst->INSNAME);
	  }			/* parcours des connecteurs de l'instance (for) */

      maxProfond++;

      if (NO_TRACE >= 2)
	printf ("  arrival time of %s [%s] - UP %ld - DW %ld - Cin Sum %ld - Depth = %ld\n",
		inst->INSNAME, inst->FIGNAME,
		maxDelUp, maxDelDw, sCin, maxProfond);
      if (NO_TRACE >= 3)
	printf ("  cell fanin critique : UP %s  DW %s\n",
		(genreUp == LOINSUP) ? ((loins_list *) criticUp)->INSNAME : ((locon_list *) criticUp)->NAME,
		(genreDw == LOINSDW) ? ((loins_list *) criticDw)->INSNAME : ((locon_list *) criticDw)->NAME);

      /* memorisation des retards critiques sur le signal de sortie de inst */
      sigOut->USER = addptype (sigOut->USER, (long) RETARDUP, (void *) maxDelUp);
      sigOut->USER = addptype (sigOut->USER, (long) RETARDDW, (void *) maxDelDw);
      sigOut->USER = addptype (sigOut->USER, (long) RETARDMINUP, (void *) maxDelMinUp);
      sigOut->USER = addptype (sigOut->USER, (long) RETARDMINDW, (void *) maxDelMinDw);
      sigOut->USER = addptype (sigOut->USER, (long) PROF, (void *) maxProfond);

      if (cellLoadType (inst->FIGNAME) != CONST)
	{
	  inst->USER = addptype (inst->USER, genreMinUp, (void *) criticMinUp);
	  inst->USER = addptype (inst->USER, genreMinDw, (void *) criticMinDw);
	  inst->USER = addptype (inst->USER, genreUp, (void *) criticUp);
	  inst->USER = addptype (inst->USER, genreDw, (void *) criticDw);
	}

      /* mise a jour des profondeurs et retards max pour le circuit */
      if (maxProfond > profondeurCircuit)
	{
	  profondeurCircuit = maxProfond;
	  instProf = inst;
	}

      if (searchTH (pathTH, inst->INSNAME) == 1)
	{
	  if (maxDelUp > delaiCriticUp)
	    {
	      porteCriticUp = inst;
	      delaiCriticUp = maxDelUp;
	    }
	  if (maxDelDw > delaiCriticDw)
	    {
	      porteCriticDw = inst;
	      delaiCriticDw = maxDelDw;
	    }
	  if (maxDelMinUp > delaiCriticMinUp)
	    {
	      porteCriticMinUp = inst;
	      delaiCriticMinUp = maxDelMinUp;
	    }
	  if (maxDelMinDw > delaiCriticMinDw)
	    {
	      porteCriticMinDw = inst;
	      delaiCriticMinDw = maxDelMinDw;
	    }
	}
    }
}

/*---------------------------------------------------------------------------
calcul_requi : calcul des temps requis pour les instances d'une LOFIG
-----------------------------------------------------------------------------
retour       : les temps requis sont memorises dans le champ USER des signaux
---------------------------------------------------------------------------*/
void 
calcul_requi (lofig, requiUp, requiDw, tout)
     lofig_list *lofig;
     long requiUp, requiDw;
     short tout;
{
  loins_list *inst;

/*
   printf("Required times computation on %s\n", lofig->NAME);
 */

  if (tout > 0)
    {
      if (listeCritic)
	freeptype (listeCritic);
      listeCritic = (ptype_list *) NULL;
    }

  /* parcours des instances de la figure */
  for (inst = lofig->LOINS; inst; inst = inst->NEXT)
    if (cellLoadType (inst->FIGNAME) != TIE)
      {
	int val = searchTH (loconOutTH, inst->INSNAME);
	losig_list *sigOut = ((locon_list *) val)->SIG;

	if (!getptype (sigOut->USER, (long) REQUIUP))
	  requi (inst, requiUp, requiDw, tout);
      }

  if (tout == 1)
    {
      /* construction de la liste critique (portes du chemin critique) */
      loins_list *porte = (loins_list *) NULL;
      int type;
      losig_list *porteOut;
      ptype_list *chemin;
      ptype_list *conCritic = (ptype_list *) NULL;

      if (delaiCriticUp > delaiCriticDw)
	{
	  porte = porteCriticUp;
	  type = NUP;
	}
      else
	{
	  porte = porteCriticDw;
	  type = NDOWN;
	}
      porteOut = ((locon_list *) searchTH (loconOutTH, porte->INSNAME))->SIG;
      listeCritic = tripType (listeCritic,
			    (long) searchTH (capaTH, getsigname (porteOut)),
			      (void *) porte, NDOWN);
      while ((chemin = getptype (porte->USER, (long) ((type == NUP) ? LOINSUP : LOINSDW))))
	{
	  if ((type == NUP) && (cellLoadPolarity (porte->FIGNAME) == NEG))
	    type = NDOWN;
	  else if ((type == NDOWN) && (cellLoadPolarity (porte->FIGNAME) == NEG))
	    type = NUP;
	  porte = (loins_list *) chemin->DATA;
	  porteOut = ((locon_list *) searchTH (loconOutTH, porte->INSNAME))->SIG;
	  listeCritic = tripType (listeCritic,
			    (long) searchTH (capaTH, getsigname (porteOut)),
				  (void *) porte, NDOWN);
	}
      conCritic = getptype (porte->USER, (long) ((type == NUP) ? LOCONUP : LOCONDW));
      if (conCritic)
	listeCritic = addptype (listeCritic,
				(long) searchTH (capaTH,
			getsigname (((locon_list *) conCritic->DATA)->SIG)),
				conCritic->DATA);
      else
	listeCritic = addptype (listeCritic, (long) 0, (void *) NULL);
    }

/*
   printf("End of required times computation on %s\n\n", lofig->NAME);

   if (tout == 1)
   {
   ptype_list *p;

   for(p = listeCritic->NEXT; p; p = p->NEXT)
   printf("%s = %d\n", ((loins_list *)p->DATA)->INSNAME, p->TYPE);
   }
 */
}

/*---------------------------------------------------------------------------
requi  : calcul du temps requis d'une instance (appel recursif sur les sorties)
-----------------------------------------------------------------------------
retour : RIEN. Memorisation des temps requis dans le champ USER des signaux
---------------------------------------------------------------------------*/
void 
requi (inst, requiUp, requiDw, tout)
     loins_list *inst;
     long requiUp, requiDw;
     short tout;		/* tout = 2 : toutes les portes sont mises ds listeCritic  */
		/* tout = 3 : portes mises ds listeCritic si critic < 100  */
		/* tout = 1 : portes mises ds listeCritic si critic = 0    */
		/* tout = 0 : listeCritic n'est pas recalculee             */
{
  int val = searchTH (loconOutTH, inst->INSNAME);
  locon_list *conOut;
  losig_list *sigOut;
  ptype_list *liste;

  chain_list *lstCon;
  long minRequiUp = 999999, minRequiDw = 999999;
  long slackUp, slackDw, aspetCritic, slack, prof;
  long capaOut, RcUpInst, RcDwInst, RcInst;

  if (cellLoadType (inst->FIGNAME) == BUS)
    printf ("BUS : %s [%s]\n", inst->INSNAME, inst->FIGNAME);

  if (val != EMPTYTH)
    {
      conOut = (locon_list *) val;
      sigOut = conOut->SIG;
      liste = getptype (sigOut->USER, (long) LOFIGCHAIN);
    }
  else
    {
      printf ("TIMING ANALYSIS : No type LOCONOUT for %s\n", inst->INSNAME);
      exit (1);
    }

  if (!liste)
    {
      printf ("TIMING ANALYSIS : No type LOFIGCHAIN for %s\n", inst->INSNAME);
      exit (1);
    }

  if (NO_TRACE >= 3)
    printf (" Computing required time on %s [%s]\n",
	    inst->INSNAME, inst->FIGNAME);

  /* on parcourt les connecteurs attaches au connecteur de sortie de inst */
  for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
    {
      locon_list *conAux = (locon_list *) lstCon->DATA;

      if (conAux->ROOT != (void *) inst)
	{
	  long requiInstUp = 0, requiInstDw = 0;

	  if ((conAux->TYPE == INTERNAL) && (conAux->DIRECTION == IN))
	    {
	      loins_list *fanout = (loins_list *) conAux->ROOT;

	      if (cellLoadType (fanout->FIGNAME) == REG)
		{
		  requiInstUp = requiUp;
		  requiInstDw = requiDw;
		}
	      else
		{
		  losig_list *fanoutSig;
		  locon_list *fanoutCon;
		  long requiOutUp = 0, requiOutDw = 0;
		  ptype_list *user;
		  long sCin = 0;

		  fanoutCon = (locon_list *) searchTH (loconOutTH, fanout->INSNAME);
		  fanoutSig = fanoutCon->SIG;
		  sCin = (long) searchTH (capaTH, getsigname (fanoutSig));

		  if ((user = getptype (fanoutSig->USER, (long) REQUIUP)))
		    {
		      /* required times deja calcules ... */
		      requiOutUp = (long) user->DATA;
		      requiOutDw = (long) (getptype (fanoutSig->USER, (long) REQUIDW))->DATA;
		    }
		  else
		    {
		      /* appel recursif pour le fanout */
		      requi (fanout, requiUp, requiDw, tout);
		      /* on recupere les required times tout juste calcules */
		      requiOutUp = (long) (getptype (fanoutSig->USER, (long) REQUIUP))->DATA;
		      requiOutDw = (long) (getptype (fanoutSig->USER, (long) REQUIDW))->DATA;
		    }		/* else : required times pas encore calcules */


		  /* Calcul du required time */
		  if (cellLoadType (fanout->FIGNAME) == BUS)
		    {
		      long RcUp, RcDw;
		      cellList *cell = (cellList *) searchTH (NO_CELLTH, fanout->FIGNAME);

		      RcUp = (cellLoadR (fanout->FIGNAME, conAux->NAME, NUP) * sCin) / 1000;
		      RcDw = (cellLoadR (fanout->FIGNAME, conAux->NAME, NDOWN) * sCin) / 1000;

		      if (cellLoadPolarity (fanout->FIGNAME) == AUT)
			{
			  printf ("Other polarity for gate %s [%s]\n",
				  cell->nom, fanout->FIGNAME);
			  exit (1);
			}

		      if (searchExpr (cell->val_abl->VALABL, conAux->NAME))
			{
			  /* entree dans le VALABL */
			  switch (cellLoadPolarity (fanout->FIGNAME))
			    {
			    case POS:
			      requiInstUp = requiOutUp -
				cellLoadTp (fanout->FIGNAME, conAux->NAME, HH) -
				RcUp;
			      requiInstDw = requiOutDw -
				cellLoadTp (fanout->FIGNAME, conAux->NAME, LL) -
				RcDw;
			      break;
			    case NEG:
			      requiInstUp = requiOutDw -
				cellLoadTp (fanout->FIGNAME, conAux->NAME, HL) -
				RcDw;
			      requiInstDw = requiOutUp -
				cellLoadTp (fanout->FIGNAME, conAux->NAME, LH) -
				RcUp;
			      break;
			    }
			}
		      else if (searchExpr (cell->val_abl->CNDABL, conAux->NAME))
			{
			  if (ATOM (cell->val_abl->CNDABL))
			    {
			      long requiInstUp1 = requiOutUp -
			      cellLoadTp (fanout->FIGNAME, conAux->NAME, HH) -
			      RcUp;
			      long requiInstUp2 = requiOutDw -
			      cellLoadTp (fanout->FIGNAME, conAux->NAME, HL) -
			      RcDw;
			      requiInstUp = MIN (requiInstUp1, requiInstUp2);
			      requiInstDw = requiDw;
			    }
			  else
			    {
			      long requiInstDw1 = requiOutUp -
			      cellLoadTp (fanout->FIGNAME, conAux->NAME, LH) -
			      RcUp;
			      long requiInstDw2 = requiOutDw -
			      cellLoadTp (fanout->FIGNAME, conAux->NAME, LL) -
			      RcDw;
			      requiInstDw = MIN (requiInstDw1, requiInstDw2);
			      requiInstUp = requiUp;
			    }
			}
		    }		/* fin BUS */
		  else
		    {
		      /* cellule combi */
		      switch (cellLoadPolarity (fanout->FIGNAME))
			{
			case POS:
			  requiInstUp = requiOutUp -
			    getDel (fanout, conAux, fanoutCon->SIG, "hh", 0);
			  requiInstDw = requiOutDw -
			    getDel (fanout, conAux, fanoutCon->SIG, "ll", 0);
			  break;
			case NEG:
			  requiInstUp = requiOutDw -
			    getDel (fanout, conAux, fanoutCon->SIG, "hl", 0);
			  requiInstDw = requiOutUp -
			    getDel (fanout, conAux, fanoutCon->SIG, "lh", 0);
			  break;
			case AUT:
			  requiInstUp = MIN (
					      (requiOutUp -
			  getDel (fanout, conAux, fanoutCon->SIG, "hh", 0)),
					      (requiOutDw -
			  getDel (fanout, conAux, fanoutCon->SIG, "hl", 0)));
			  requiInstDw = MIN (
					      (requiOutUp -
			  getDel (fanout, conAux, fanoutCon->SIG, "lh", 0)),
					      (requiOutDw -
			  getDel (fanout, conAux, fanoutCon->SIG, "ll", 0)));
			  break;
			}	/* switch sur la polarite */
		    }		/* cellule combi  != BUS */
		}		/* pas un registre */
	    }			/* connecteur interne */
	  else if (conAux->TYPE == EXTERNAL)
	    {
/*
   printf("Fanout Con = %s\n", conAux->NAME);
 */
	      requiInstUp = requiUp;
	      requiInstDw = requiDw;
	    }
	  if (((conAux->TYPE == INTERNAL) && (conAux->DIRECTION == IN)) ||
	      (conAux->TYPE == EXTERNAL))
	    {
	      if (requiInstDw < minRequiDw)
		minRequiDw = requiInstDw;
	      if (requiInstUp < minRequiUp)
		minRequiUp = requiInstUp;
	    }
	}
    }				/* parcours des connecteurs du fanout de l'instance (for) */

  if (NO_TRACE >= 2)
    printf ("   required time of %s [%s] -UP = %ld DW = %ld\n",
	    inst->INSNAME, inst->FIGNAME, minRequiUp, minRequiDw);

  slackUp = minRequiUp - (long) (getptype (sigOut->USER, (long) RETARDUP)->DATA);
  slackDw = minRequiDw - (long) (getptype (sigOut->USER, (long) RETARDDW)->DATA);
  slack = (slackUp < slackDw) ? slackUp : slackDw;
  prof = (long) getptype (sigOut->USER, (long) PROF)->DATA;

  capaOut = (long) searchTH (capaTH, getsigname (sigOut));
  RcUpInst = (cellLoadR (inst->FIGNAME, (char *) NULL, NUP) * capaOut) / 1000;
  RcDwInst = (cellLoadR (inst->FIGNAME, (char *) NULL, NDOWN) * capaOut) / 1000;
  RcInst = (RcUpInst > RcDwInst) ? RcUpInst : RcDwInst;

/*
   if ((RcUpInst <= 0) || (RcDwInst <= 0))
   printf("Res a 0 pour %s [%s] RcUp=%d RcDw=%d capaOut=%d\n", inst->INSNAME, inst->FIGNAME, RcUpInst, RcDwInst, capaOut);
 */
/* 10/06/1997 : pour nouvelle biblio : (RcInst + 1) ok */

  if ((slack != 0) && ((slack * 10 * prof) <= RcInst))
    aspetCritic = 1;
  else
      aspetCritic = (slack * 10 * prof) / (RcInst + 1) ;
/* 28/03/96 : pour comparaison avec [MAR90] ...
   if ((slack != 0) && ((slack * 10) <= (RcInst * prof)))
   aspetCritic = 1;
   else
   aspetCritic = (slack * 10) / (RcInst * prof);
 */

  /* memorisation des required times sur le signal de sortie de inst */
  sigOut->USER = addptype (sigOut->USER, (long) REQUIUP, (void *) minRequiUp);
  sigOut->USER = addptype (sigOut->USER, (long) REQUIDW, (void *) minRequiDw);
  sigOut->USER = addptype (sigOut->USER, (long) SLACKUP, (void *) slackUp);
  sigOut->USER = addptype (sigOut->USER, (long) SLACKDW, (void *) slackDw);
  sigOut->USER = addptype (sigOut->USER, (long) CRITIC, (void *) aspetCritic);

  if (RcUpInst < 0 || RcDwInst < 0)
    printf (" !!! %s [%s] RC = %ld / %ld\n",
	    inst->INSNAME, inst->FIGNAME, RcUpInst, RcDwInst);
  if (minRequiUp < 0 || minRequiDw < 0)
    printf (" !!! %s [%s] Requi = %ld / %ld\n", inst->INSNAME, inst->FIGNAME, minRequiUp, minRequiDw);
  if (slackUp < 0 || slackDw < 0)
    printf (" !!! %s [%s] Slack = %ld / %ld\n", inst->INSNAME, inst->FIGNAME, slackUp, slackDw);
  if (aspetCritic < 0)
    printf (" !!! %s critic = %ld\n", inst->INSNAME, aspetCritic);
/*
   if (aspetCritic == 0)
   displayTimingInst(inst, 0, 0);
 */

  if (tout == 2)
    listeCritic = tripType (listeCritic, aspetCritic, (void *) inst, NUP);
/*
   listeCritic = tripType(listeCritic, RcInst, (void *)inst, NDOWN);
 */
  else if ((tout == 3) && (aspetCritic < 100))
    listeCritic = tripType (listeCritic, aspetCritic, (void *) inst, NUP);
}
