/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */




#ident  "$Id: sl_facto_cliqueZ.c,v 1.1.1.1 1998/10/02 15:25:11 alliance Exp $"

#include <stdio.h>
#include <stdlib.h>



#include MUT_H
#include LOG_H
#include BEH_H

#include "lax_param.h"
#include "sl_local.h"

#define EXTRACT 15
#define SIMPLIFLOC 1

#define TAILLEGRAPH 1000

/*-------------------------------------------------------------------------
upVarAllDoubleBdd  : remonte tous les index de 2*(index-1) dans un graphe
---------------------------------------------------------------------------
retour          :
---------------------------------------------------------------------------*/
pNode
upVarAllDoubleBdd (pt)
     pNode pt;
{
  if (pt->index < 2)
    return (pt);

  return (initVertexBdd (2 * (pt->index - 1), upVarAllDoubleBdd (pt->high),
			 upVarAllDoubleBdd (pt->low)));
}

/*-------------------------------------------------------------------------
downVarAllDoubleBdd  : descend tous les index de (index)/2 + 1 dans un graphe
---------------------------------------------------------------------------
retour          :
---------------------------------------------------------------------------*/
pNode
downVarAllDoubleBdd (pt)
     pNode pt;
{
  if (pt->index < 2)
    return (pt);

  return (initVertexBdd ((pt->index) / 2 + 1, downVarAllDoubleBdd (pt->high),
			 downVarAllDoubleBdd (pt->low)));
}


/*-------------------------------------------------------------------------
suptautology    : calcule si il existe une couverture commune a deux
                  fonctions phi-Boolennes 
---------------------------------------------------------------------------
retour          : 0 ou 1 
---------------------------------------------------------------------------*/
int
suptautology (x, y)
     pNode x, y;
{
  if ((applyBinBdd (AND, x->low, y->high) == zero) &&
      (applyBinBdd (AND, x->high, y->low) == zero))
    return 1;
  else
    return 0;
}

/*-------------------------------------------------------------------------
simplifONOFF    : simplifie ON et OFF avec des regles :
                       1. assignations a tautologie
                       2. elimination du noeud pere de ON 
---------------------------------------------------------------------------
retour          : si modif 
---------------------------------------------------------------------------*/
int
simplifONOFF (pX, ppY)
     pNode pX;
     pNode *ppY;
{
  int modifONOFF = 0;

  *ppY = pX;

  /* index egaux */

  if ((pX->low)->index == (pX->high)->index && ((pX->low)->index > 1))
    {
      pNode pON, pOFF;

      pON = pX->high;
      pOFF = pX->low;

      if ((pON->high == zero) && (pOFF->high == zero))
	{
	  *ppY = initVertexBdd (pX->index, pON->low, pOFF->low);
	  modifONOFF = 1;
	}
      if ((pON->low == zero) && (pOFF->low == zero))
	{
	  *ppY = initVertexBdd (pX->index, pON->high, pOFF->high);
	  modifONOFF = 1;
	}
    }
  return modifONOFF;
}

/*-------------------------------------------------------------------------
extractClique    : extrait les cliques  
---------------------------------------------------------------------------
retour          : une liste chainee correspondant a une clique 
---------------------------------------------------------------------------*/
chain_list *
extractClique (tabCouple, count)
     struct couple *tabCouple;
     int count;
{
  int cm;
  int max;
  chain_list *choix;
  chain_list *clique;
  int i;

  /* calcul du card. max dans le tableau tabCouple */
  max = 0;
  cm = 0;
  for (i = 0; i < count; i++)
    {
      if (max < tabCouple[i].count)
	{
	  max = tabCouple[i].count;
	  cm = i;
	}
    }
  /* il n'existe plus de clique */

  if (max == 0)
    return NULL;

  choix = tabCouple[cm].lst;
  clique = addchain (NULL, cm);

  while (choix != NULL)
    {				/* calcul du card. max dans la liste choix */
      chain_list *lst = choix;

      max = 0;
      cm = 0;
      while (lst)
	{
	  if (max < tabCouple[(int) lst->DATA].count)
	    {
	      max = tabCouple[(int) lst->DATA].count;
	      cm = (int) lst->DATA;
	    }
	  lst = lst->NEXT;
	}

      choix = (chain_list *) interLC (choix, tabCouple[cm].lst);
      clique = addchain (clique, cm);
    }
  return clique;
}

/*-------------------------------------------------------------------------
calculClique    : extrait les couples ON - OFF a un niveau  
---------------------------------------------------------------------------
retour          : 1 si modif 
---------------------------------------------------------------------------*/
int
calculClique (tabCouple, count)
     struct couple *tabCouple;
     int count;
{
  int i, j;
  int modifONOFF = 0;
  chain_list *lst;

  /* valeur par defaut pour y */

  for (i = 0; i < count; i++)
    {
      tabCouple[i].y = tabCouple[i].x;
      tabCouple[i].lst = NULL;
      tabCouple[i].count = 0;
    }

  printf ("Algo de clique : nombre de noeuds = %d\n", count);

  /* creation du graphe */

  for (i = 0; i < count; i++)
    {
      for (j = i + 1; j < count; j++)
	{
	  if (suptautology (tabCouple[i].x, tabCouple[j].x))
	    {
	      printf ("*** Edge between %d and %d\n", i, j);

	      tabCouple[i].lst = addchain (tabCouple[i].lst, j);
	      tabCouple[i].count = tabCouple[i].count + 1;
	      tabCouple[j].lst = addchain (tabCouple[j].lst, i);
	      tabCouple[j].count = tabCouple[j].count + 1;
	    }
	}
    }

  /* calcul de couverture pour les couples isoles 
     de noeuds relies par une arete */

  for (i = 0; i < count; i++)
    {
      pNode pY;

      /* une arete unique entre deux noeuds */

      if ((tabCouple[i].count == 1) &&
	  (tabCouple[(int) (tabCouple[i].lst)->DATA].count == 1))
	{
	  int j;

	  j = (int) (tabCouple[i].lst)->DATA;

	  printf ("*** Reso : %d - %d\n", i, j);

/*
   printf("couple[i].x \n");
   displayBdd(tabCouple[i].x,1);
   printf("couple[i].y \n");
   displayBdd(tabCouple[i].y,1);
 */

	  pY = applyBinBdd (OR, tabCouple[i].x, tabCouple[j].y);

/*
   printf("U-x \n");
   displayBdd(pY,1);
 */

	  /* report de la couverture sur les couples de BDD concernes */

	  tabCouple[i].y = pY;
	  tabCouple[j].y = pY;

	  freechain (tabCouple[i].lst);
	  tabCouple[i].lst = NULL;
	  tabCouple[i].count = 0;

	  freechain (tabCouple[j].lst);
	  tabCouple[j].lst = NULL;
	  tabCouple[j].count = 0;

	  modifONOFF = 1;
	}
    }

  /* calcul des cliques */

  while ((lst = (chain_list *) extractClique (tabCouple, count)))
    {
      chain_list *lstp;
      pNode pY;

      modifONOFF = 1;

      printf ("clique = [");
      lstp = lst;
      while (lstp)
	{
	  printf (" %d", (int) lstp->DATA);
	  lstp = lstp->NEXT;
	}
      printf (" ]\n");

      /* elimination des noeuds de la clique et de leurs aretes */

      lstp = lst;
      while (lstp)
	{
	  freechain (tabCouple[(int) lstp->DATA].lst);
	  tabCouple[(int) lstp->DATA].lst = NULL;
	  tabCouple[(int) lstp->DATA].count = 0;
	  lstp = lstp->NEXT;
	}

      for (i = 0; i < count; i++)
	{
	  if (tabCouple[i].lst != NULL)
	    {
	      lstp = tabCouple[i].lst;
	      tabCouple[i].lst = (chain_list *) extractLC (lst, tabCouple[i].lst);
	      tabCouple[i].count = countChain_list (tabCouple[i].lst);
	      freechain (lstp);
	    }
	}

      /* calcul de la couverture simple */
      lstp = lst;
      pY = zero;
      while (lstp)
	{
	  pY = applyBinBdd (OR, tabCouple[(int) lstp->DATA].x, pY);
	  lstp = lstp->NEXT;
	}

      /* report de la couverture sur les couples de BDD concernes */
      lstp = lst;
      while (lstp)
	{
	  tabCouple[(int) lstp->DATA].y = pY;
	  lstp = lstp->NEXT;
	}
    }

  /* simplifications locales directes */

  if (SIMPLIFLOC)
    {
      for (i = 0; i < count; i++)
	{
	  pNode pY;

	  if (simplifONOFF (tabCouple[i].x, &pY))
	    {

/*
   printf("couple[i].x \n");
   displayBdd(tabCouple[i].x,1);
   printf("apres simplif :\n");
 */

	      tabCouple[i].y = pY;
	      modifONOFF = 1;

/*
   printf("couple[i].y \n");
   displayBdd(tabCouple[i].y,1);
 */
	      /* on verifie ... on ne sait jamais ! */

	      if (!suptautology (tabCouple[i].x, tabCouple[i].y))
		{
		  printf ("Error simplif loc\n");
		  exit (-1);
		}
	    }
	}
    }

  return (modifONOFF);
}

/*-------------------------------------------------------------------------
setONOFF        : positionne les couples ON - OFF a un niveau  
---------------------------------------------------------------------------
retour          : deux nouveaux pON et pOFF 
---------------------------------------------------------------------------*/
void
setONOFF (pZ, ppONp, ppOFFp, tabCouple, count, level)
     pNode pZ;
     pNode *ppONp, *ppOFFp;
     struct couple *tabCouple;
     int count;
     int level;
{
  int i;


  if (pZ->index == level)
    {
      for (i = 0; i < count; i++)
	{
	  if (pZ == tabCouple[i].x)
	    {
	      *ppONp = (tabCouple[i].y)->high;
	      *ppOFFp = (tabCouple[i].y)->low;
	      return;
	    }
	}

      printf ("Error : cannot find the node :\n");
      displayBdd (pZ, 1);
      exit (-1);
    }
  else
    {
      pNode pONpLow, pONpHigh, pOFFpLow, pOFFpHigh;

      if (pZ->index > level)
	{
	  setONOFF (pZ->high, &pONpHigh, &pOFFpHigh, tabCouple, count, level);
	  setONOFF (pZ->low, &pONpLow, &pOFFpLow, tabCouple, count, level);

	  *ppONp = initVertexBdd (pZ->index, pONpHigh, pONpLow);
	  *ppOFFp = initVertexBdd (pZ->index, pOFFpHigh, pOFFpLow);
	}
      else
	/* on = off  => dc = 1 et on = 0,off = 0 ... */
	{
	  *ppONp = zero;
	  *ppOFFp = zero;
	}
    }
}

/*-------------------------------------------------------------------------
extractONOFF    : extrait les couples ON - OFF a un niveau  
---------------------------------------------------------------------------
retour          : remplit un tableau et met a jour un compteur 
---------------------------------------------------------------------------*/
void
extractONOFF (pZ, tabCouple, pCount, level)
     pNode pZ;
     struct couple *tabCouple;
     int *pCount;
     int level;
{
  int i;

  /* couple deja traite */

/*====	il faut creer un cache memoire specifique ...

pBddLoc = searchTabLoc(sysBdd.pMC,pON,pOFF,EXTRACT);
if ((int) pBddLoc == level) return;

========================================================*/


  if (pZ->index == level)
    {

      for (i = 0; i < *pCount; i++)
	{
	  if (pZ == tabCouple[i].x)
	    {
	      return;
	    }
	}

      tabCouple[*pCount].x = pZ;
      *pCount = *pCount + 1;
      if (*pCount == TAILLEGRAPH)
	{
	  printf ("Error : Too many nodes for the cliques algorithm ...\n");
	  exit (-1);
	}
    }
  else
    {
      if (pZ->index > level)
	{
	  extractONOFF (pZ->high, tabCouple, pCount, level);
	  extractONOFF (pZ->low, tabCouple, pCount, level);
	}
    }

/* ===========================================

   addTabLoc(sysBdd.pMC,pON,pOFF,level,EXTRACT);

   =========================================== */

}

/*-------------------------------------------------------------------------
cliqueBdd       : minimise un Bdd avec ses DC 
---------------------------------------------------------------------------
retour          : un BDD simplifie 
---------------------------------------------------------------------------*/
pNode
cliqueBdd (pON, pDC)
     pNode pON, pDC;
{
  pNode pOFF;
  int *tabSupport;
  chain_list *support, *saveSupport;
  struct couple *pCouple;
  int i;
  pNode pZ;

  /* on remonte les index pour laisser de la place pour Z */

  pON = upVarAllDoubleBdd (pON);
  pDC = upVarAllDoubleBdd (pDC);

  /* on calcule le off-set */

  pOFF = notBdd (applyBinBdd (OR, pON, pDC));

  printf ("zero = %p, one = %p\n", zero, one);

  /* support de pON */

  support = (chain_list *) supportIndexBdd (pON, 1);
  tabSupport = (int *) mbkalloc ((pON->index + 2) * sizeof (int));
  saveSupport = support;

  for (i = 0; i <= pON->index; i++)
    tabSupport[i] = 0;

  while (support)
    {
      tabSupport[(int) support->DATA] = 1;
      support = support->NEXT;
    }
  freechain (saveSupport);

  /* tableau des couples ON-OFF */

  pCouple = (struct couple *) mbkalloc (TAILLEGRAPH * sizeof (struct couple));

  /* on descend 'Z' de l'index de la racine de pON  a 2 */

  for (i = pON->index; i >= 2; i--)
    {
      pNode pONp, pOFFp;
      pNode termZ;

      if (tabSupport[i] == 1)	/* la variable i existe dans pON */
	{
	  int numCouples = 0;
	  int modifONOFF = 0;

	  printf ("niveau = %d\n", i);

	  termZ = createNodeTermBdd (i + 1);

	  /* on amene la variable Z */

	  pZ = applyBinBdd (OR, applyBinBdd (AND, termZ, pON),
			    applyBinBdd (AND, notBdd (termZ), pOFF));

	  printf ("pZ = \n");
	  displayGraphicBdd (pZ);

	  extractONOFF (pZ, pCouple, &numCouples, i + 1);
	  modifONOFF = calculClique (pCouple, numCouples);

	  /* on reporte les modifications sur pOn et pOFF */

	  if (modifONOFF)
	    {
	      setONOFF (pZ, &pONp, &pOFFp, pCouple, numCouples, i + 1);
	      pON = pONp;
	      pOFF = pOFFp;
	    }
	}
    }

  pON = downVarAllDoubleBdd (pON);

  mbkfree (pCouple);
  return (pON);
}

/*-------------------------------------------------------------------------
cliqueMultiBdd  : minimise un multi-BDD avec ses DC 
---------------------------------------------------------------------------
retour          : un multi-BDD simplifie 
---------------------------------------------------------------------------*/
void
cliqueMultiBdd (pON, pDC, numBdd)
     pNode *pON, *pDC;
     int numBdd;
{
  pNode *pOFF;
  int i, j;
  int indexMax;
  struct couple *pCouple;
  pNode *pZ;

  /* on remonte les index pour laisser de la place pour Z */

  for (i = 0; i < numBdd; i++)
    {
      pON[i] = upVarAllDoubleBdd (pON[i]);
      pDC[i] = upVarAllDoubleBdd (pDC[i]);
    }

  pZ = (pNode *) mbkalloc (numBdd * sizeof (pNode));

  pOFF = (pNode *) mbkalloc (numBdd * sizeof (pNode));

  /* on calcule le off-set du multi-Bdd 
     et l'index maximun */

  indexMax = 0;
  for (i = 0; i < numBdd; i++)
    {
      pOFF[i] = notBdd (applyBinBdd (OR, pON[i], pDC[i]));
      if (indexMax < pON[i]->index)
	indexMax = pON[i]->index;
    }

  /* tableau des couples ON-OFF */

  pCouple = (struct couple *) mbkalloc (TAILLEGRAPH * sizeof (struct couple));

  /* on descend 'Z' de l'index indexMax  a 2 */

  printf ("zero = %p, one = %p\n", zero, one);

  for (i = indexMax; i >= 2; i--)
    {
      pNode pONp, pOFFp;
      int numCouples;
      int modifONOFF;
      pNode termZ;

      numCouples = 0;
      modifONOFF = 0;

      /* traitement si l'index est pair */
      if ((i & 1) == 0)
	{
	  printf ("niveau = %d\n", i);

	  termZ = createNodeTermBdd (i + 1);

	  /* calcul du multi-BDD avec Z */

	  for (j = 0; j < numBdd; j++)
	    pZ[j] = applyBinBdd (OR, applyBinBdd (AND, termZ, pON[j]),
				 applyBinBdd (AND, notBdd (termZ), pOFF[j]));

	  /* extraction de tous les couples distincts ON,OFF */

	  for (j = 0; j < numBdd; j++)
	    {
	      extractONOFF (pZ[j], pCouple, &numCouples, i + 1);
	    }

	  modifONOFF = calculClique (pCouple, numCouples);

	  /* on reporte les modifications sur pON et pOFF */

	  if (modifONOFF)
	    {
/*
   for (j=0; j < numCouples; j++)
   {
   printf("x[%d] = \n",j);
   displayBdd(pCouple[j].x,1);
   }
 */

	      for (j = 0; j < numBdd; j++)
		{
		  setONOFF (pZ[j], &pONp, &pOFFp, pCouple, numCouples, i + 1);


		  pON[j] = pONp;
		  pOFF[j] = pOFFp;

/*
   printf("apres simplif pON[%d] = \n",j);
   displayBdd(pON[j],1);
   printf("apres simplif pOFF[%d] = \n",j);
   displayBdd(pOFF[j],1);
 */

		}
	    }
	}
    }

  /* on descend les index */

  for (i = 0; i < numBdd; i++)
    {
      pON[i] = downVarAllDoubleBdd (pON[i]);
      pDC[i] = downVarAllDoubleBdd (pDC[i]);
    }

  mbkfree (pCouple);
  mbkfree (pOFF);
}
