/*------------------------------------------------------------\
|                                                             |
| Tool    :                     FVH                           |
|                                                             |
| File    :                   fvherror.c                      |
|                                                             |
| Author  :                 Jacomme Ludovic                   |
|                                                             |
| Date    :                   04.12.96                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include FSM_H

# include "fvherror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/

void fvh_error( Error, Text, Value, File, Line )

    char  Error;
    char *Text;
    long  Value;
    char *File;
    long  Line;
{
  char *Name;

  Name = mbkstrdup( File );
  Name[ strlen( File ) - 1 ] = '\0';

  fprintf( stderr, "%s%ld ", Name, Line );

  switch( Error )
  {
    case FVH_ERROR_OPEN_FILE :
      fprintf( stderr, "Unable to open file \"%s\"\n", Text );
    break;
    case FVH_ERROR_PARSE_FILE :
      fprintf( stderr, "Unable to parse file \"%s\"\n", Text );
    break;
    case FVH_ERROR_SIGNAL_NOT_FOUND :
      fprintf( stderr, "Signal \"%s\" not found line %ld\n", Text, Value );
    break;
    case FVH_ERROR_ILLEGAL_DEFAULT_ASSIGN :
      fprintf( stderr, "Illegal default output assignation\n" );
    break;
    case FVH_ERROR_NO_STATE :
      fprintf( stderr, "Number of state is zero in \"%s\"\n", Text );
    break;
    case FVH_ERROR_CONTROL_TYPE :
      fprintf( stderr, "Bad control type size in \"%s\" line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_ILLEGAL_CASE :
      fprintf( stderr, "Illegal case condition, type \"%s\" line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_ILLEGAL_STATE :
      fprintf( stderr, "Illegal state name \"%s\" in case condition line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_ILLEGAL_CONTROL :
      fprintf( stderr, "Illegal control name \"%s\" in case condition line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_ILLEGAL_INSTRUCTION :
      fprintf( stderr, "Illegal instruction type %ld\n", (long)Text );
    break;
    case FVH_ERROR_ILLEGAL_STATE_ASSIGN :
      fprintf( stderr, "Illegal state name \"%s\" in assignation line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_ILLEGAL_STACK_ASSIGN :
      fprintf( stderr, "Illegal stack name \"%s\" in assignation line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_ILLEGAL_CONTROL_ASSIGN :
      fprintf( stderr, "Illegal control name \"%s\" in assignation line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_ILLEGAL_ASSIGNATION :
      fprintf( stderr, "Illegal use of \"%s\" in main process line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_MAIN_PROCESS :
      fprintf( stderr, "Illegal figure \"%s\" missing main process\n", Text );
    break;
    case FVH_ERROR_CLOCK_PROCESS :
      fprintf( stderr, "Illegal figure \"%s\" missing clock process\n", Text );
    break;
    case FVH_ERROR_PROCESS_SENSITIVITY :
      fprintf( stderr, "Illegal sensitivity list in the process \"%s\"\n", Text );
    break;
    case FVH_ERROR_MISSING_SIGNAL :
      fprintf( stderr, "\"%s\" must be in the sensitivity list line %ld\n", Text, Value );
    break;
    case FVH_ERROR_IN_MAIN_PROCESS :
      fprintf( stderr, "Illegal use of \"%s\" in main process line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_IN_CLOCK_PROCESS :
      fprintf( stderr, "Illegal use of \"%s\" in clock process line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_ILLEGAL_OUT_ASSIGN :
      fprintf( stderr, "Illegal output \"%s\" assignation line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_MISSING_CLOCK_PORT :
      fprintf( stderr, "Missing clock declaration in figure \"%s\"\n",  Text );
    break;
    case FVH_ERROR_STACK_SIZE_ZERO :
      fprintf( stderr, "Size of stack is zero in \"%s\"\n",  Text );
    break;
    case FVH_ERROR_MISSING_STACK_CONTROL :
      fprintf( stderr, "Missing stack control in figure \"%s\" line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_MISSING_STATE_CONTROL :
      fprintf( stderr, "Missing state control in figure \"%s\"\n",  Text );
    break;
    case FVH_ERROR_NO_CONTROL_CONDITION :
      fprintf( stderr, "Missing control condition in figure \"%s\" line %ld\n",  Text, Value );
    break;
    case FVH_ERROR_BAD_STACK_CONTROL :
      fprintf( stderr, "Bad stack control in figure \"%s\"\n",  Text );
    break;
    case FVH_ERROR_ILLEGAL_IF_CONDITION :
      fprintf( stderr, "Illegal IF condition line %ld\n", Value );
    break;
    case FVH_ERROR_TWO_PROCESS :
      fprintf( stderr, "Should have two process in figure \"%s\"\n", Text );
    break;
    case FVH_ERROR_ILLEGAL_PROCESS_BODY :
      fprintf( stderr, "Illegal body structure for process \"%s\" line %ld\n", Text, Value );
    break;
    default :
      fprintf( stderr, "Internal error number %d\n", Error );
  }

  autexit( 1 );
}

void fvh_warning( Warning, Text, Value, File, Line )

    char  Warning;
    char *Text;
    long  Value;
    char *File;
    long  Line;
{
  char *Name;

  Name = mbkstrdup( File );
  Name[ strlen( File ) - 1 ] = '\0';

  fprintf( stderr, "%s%ld ", Name, Line );

  switch ( Warning )
  {
    default :
      fprintf( stderr, "Internal warning number %d\n", Warning );
  }
}
