/*------------------------------------------------------------\
|                                                             |
| Tool    :                     FVH                           |
|                                                             |
| File    :                   fvhparse.c                      |
|                                                             |
| Author  :                 Jacomme Ludovic                   |
|                                                             |
| Date    :                   04.12.96                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include FSM_H
# include FVH_H

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include "fvherror.h"
# include "fvhfbh2fsm.h"
# include "fvhparse.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  extern FILE *fbl_y_in;
  extern char  FBL_ERRFLG;

  bddcircuit *FblBddCircuit = (bddcircuit *)0;
  bddsystem  *FblBddSystem  = (bddsystem  *)0;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        vhdlloadfsmfig                       |
|                                                             |
\------------------------------------------------------------*/

static void loc_trap_signal()
{
  fprintf( stderr, "fvh: Internal error !\n" );
  autexit( 1 );
}

void vhdlloadfsmfig( FsmFigure, FigureName )

  fsmfig_list *FsmFigure;
  char        *FigureName;
{
  bddsystem   *SaveBddSystem;
  bddcircuit  *SaveBddCircuit;
  int          Error;
  void        (*PreviousHandler)();

  fbl_y_in = mbkfopen( FigureName, "fsm", "r" );

  if ( fbl_y_in == (FILE *)0 )
  {
    fvherror( FVH_ERROR_OPEN_FILE, FigureName, 0 );
  }

  SaveBddSystem  = BddLocalSystem;
  SaveBddCircuit = BddLocalCircuit;

  if ( FblBddSystem == (bddsystem *)0 )
  {
    FblBddSystem  = createbddsystem( 50, 1000, 200, 50000000 );
    FblBddCircuit = createbddcircuit( "-select-", 200, 0, FblBddSystem );
  }

  BddLocalSystem  = FblBddSystem;
  BddLocalCircuit = FblBddCircuit;

  PreviousHandler = signal( SIGSEGV, loc_trap_signal );
  Error = fbl_y_parse();
  signal( SIGSEGV, PreviousHandler );

  BddLocalSystem  = SaveBddSystem;
  BddLocalCircuit = SaveBddCircuit;

  if ( ( Error      ) ||
       ( FBL_ERRFLG ) )
  {
    fvherror( FVH_ERROR_PARSE_FILE, FigureName, 0 );
  }

  FvhFbh2Fsm( FBL_HEADFIG, FsmFigure );

  if ( ! IsFsmFigMixedRtl( FsmFigure ) )
  {
    fbh_delfbfig( FBL_HEADFIG, FBL_HEADFIG, 'Y' );
  }
}
