/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RTD                           |
|                                                             |
| File    :                   rtd_drive.c                     |
|                                                             |
| Author  :                 Jacomme Ludovic                   |
|                                                             |
| Date    :                   01.01.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include VEX_H
# include RTN_H

# include <stdio.h>
# include <stdlib.h>
# include <string.h>

# include "rtd_error.h"
# include "rtd_drive.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  static FILE *RtlFile;
  static long  RtlExprLength;
  static char  RtlBuffer[ 128 ];

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                     RtlWriteNameFile                        |
|                                                             |
\------------------------------------------------------------*/

static void RtlWriteNameFile( Name )

  char *Name;
{
  int Length;

  Length = strlen( Name );

  RtlExprLength += Length;

  if ( RtlExprLength > 80 )
  {
    fprintf( RtlFile, "\n" );
    RtlExprLength = Length;
  }

  fprintf( RtlFile, Name );
}

/*------------------------------------------------------------\
|                                                             |
|                     RtlWriteExprPrefix                      |
|                                                             |
\------------------------------------------------------------*/

static void RtlWriteExprPrefix( Expr )

  vexexpr *Expr;
{
  char       *Name;
  chain_list *ScanOper;

  if ( IsVexNodeAtom( Expr ) )
  {
    sprintf( RtlBuffer, "{%d,%d,%x}%s", 
             Expr->LEFT, Expr->RIGHT, Expr->TYPE, GetVexAtomValue( Expr ) );
    RtlWriteNameFile( RtlBuffer );
  }
  else
  {
    if ( IsVexNodeOper( Expr ) )
    {
      Name = getvexoperuppername( GetVexOperValue( Expr ) );
    }
    else
    {
      Name = GetVexFuncValue( Expr );
    }

    sprintf( RtlBuffer, "({%d,%d,%x}%s", 
             Expr->LEFT, Expr->RIGHT, Expr->TYPE, Name );
    RtlWriteNameFile( RtlBuffer );

    for ( ScanOper  = Expr->OPERAND;
          ScanOper != (chain_list *)0;
          ScanOper  = ScanOper->NEXT )
    {
      RtlWriteNameFile( "~" );
      RtlWriteExprPrefix( GetVexOperand( ScanOper ) );
    }

    RtlWriteNameFile( ")" );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                        RtlWriteExpr                         |
|                                                             |
\------------------------------------------------------------*/

static void RtlWriteExpr( Label, Expr )

  char    *Label;
  vexexpr *Expr;
{
  if ( Expr != (vexexpr *)0 )
  {
    RtlExprLength = 0;

    fprintf( RtlFile, "%s\t\n", Label );
    RtlWriteExprPrefix( Expr );
    fprintf( RtlFile, "\n" );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                      RtlWriteOneSymbol                      |
|                                                             |
\------------------------------------------------------------*/

static void RtlWriteOneSymbol( RtlSymbol )

  rtlsym *RtlSymbol;
{
  fprintf( RtlFile, "SYMBOL\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%s\n", 
           (long)RtlSymbol->INDEX,
           (long)RtlSymbol->INIT,
           (long)RtlSymbol->DRIVE,
           (long)RtlSymbol->EFFEC,
           (long)RtlSymbol->EVENT,
           (long)RtlSymbol->FLAGS,
           RtlSymbol->NAME );
}

/*------------------------------------------------------------\
|                                                             |
|                        RtlWriteOneDeclar                    |
|                                                             |
\------------------------------------------------------------*/

static void RtlWriteOneDeclar( RtlDeclar )

  rtldecl_list *RtlDeclar;
{
  int Width;
  int Scan;

  fprintf( RtlFile, "DECLAR\t%ld\t%ld\t%ld\t%ld\t%ld", 
           (long)RtlDeclar->DIR,
           (long)RtlDeclar->TYPE,
           (long)RtlDeclar->KIND,
           (long)RtlDeclar->BASE,
           (long)RtlDeclar->FLAGS );

  RtlWriteExpr( "", RtlDeclar->VEX_ATOM );
  RtlWriteExpr( "INIT", RtlDeclar->VEX_INIT );

  Width = RtlDeclar->VEX_ATOM->WIDTH;

  for ( Scan = 0; Scan < Width; Scan++ )
  {
    RtlWriteOneSymbol( &RtlDeclar->DECL_SYM[ Scan ] );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                        RtlWriteDeclar                       |
|                                                             |
\------------------------------------------------------------*/

static void RtlWriteDeclar( RtlFigure )

  rtlfig_list *RtlFigure;
{
  rtldecl_list *ScanDeclar;
  int           Type;

  for ( Type = 0; Type < RTL_MAX_DECLAR_TYPE; Type++ )
  {
    for ( ScanDeclar  = RtlFigure->DECLAR[ Type ];
          ScanDeclar != (rtldecl_list *)0;
          ScanDeclar  = ScanDeclar->NEXT )
    {
      RtlWriteOneDeclar( ScanDeclar);
    }
  }
}


/*------------------------------------------------------------\
|                                                             |
|                        RtlWriteAssign                       |
|                                                             |
\------------------------------------------------------------*/

static void RtlWriteAssign( RtlFigure )

  rtlfig_list *RtlFigure;
{
  rtlasg_list   *ScanAsg;
  rtlbivex_list *ScanBiVex;

  for ( ScanAsg  = RtlFigure->ASSIGN;
        ScanAsg != (rtlasg_list *)0;
        ScanAsg  = ScanAsg->NEXT )
  {
    fprintf( RtlFile, "ASG\t%ld\t%ld\t%ld\n", 
             (long)ScanAsg->TYPE,
             (long)ScanAsg->REG_TYPE,
             (long)ScanAsg->FLAGS );

    RtlWriteExpr( "ASG_ATOM", ScanAsg->VEX_ATOM );
    RtlWriteExpr( "ASG_DATA", ScanAsg->VEX_DATA );

    for ( ScanBiVex  = ScanAsg->BIVEX;
          ScanBiVex != (rtlbivex_list *)0;
          ScanBiVex  = ScanBiVex->NEXT )
    {
      fprintf( RtlFile, "ASG_BIVEX\t%ld\t%ld\n",
               (long)ScanBiVex->TYPE,
               (long)ScanBiVex->FLAGS );

      RtlWriteExpr( "BIVEX_COND", ScanBiVex->VEX_COND );
      RtlWriteExpr( "BIVEX_DATA", ScanBiVex->VEX_DATA );
    }
  }
}

/*------------------------------------------------------------\
|                                                             |
|                        RtlWriteFsm                          |
|                                                             |
\------------------------------------------------------------*/

static void RtlWriteFsm( RtlFigure )

  rtlfig_list *RtlFigure;
{
  /* TO BE DONE */
}

/*------------------------------------------------------------\
|                                                             |
|                        RtlWriteFigure                       |
|                                                             |
\------------------------------------------------------------*/

static void RtlWriteFigure( RtlFigure )

  rtlfig_list *RtlFigure;
{
  fprintf( RtlFile, "FIGURE\t%ld\t%s\n",
           (long)RtlFigure->FLAGS,
           RtlFigure->NAME );

  RtlWriteDeclar( RtlFigure );
  RtlWriteAssign( RtlFigure );
  RtlWriteFsm( RtlFigure );
}

/*------------------------------------------------------------\
|                                                             |
|                         savertlfig                          |
|                                                             |
\------------------------------------------------------------*/

void savertlfig( RtlFigure )

  rtlfig_list *RtlFigure;
{
  RtlFile = mbkfopen( RtlFigure->NAME, "rtl", "w" );

  if ( RtlFile == (FILE *)0 )
  {
    rtderror( RTD_ERROR_OPEN_FILE, RtlFigure->NAME, 0 );
  }

  RtlWriteFigure( RtlFigure );

  fclose( RtlFile );
}
