/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Rtn                           |
|                                                             |
| File    :                 Rtn Errors                        |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   04.07.96                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <signal.h>
# include MUT_H
# include AUT_H
# include VEX_H
# include RTN_H
# include "rtnerror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Rtl Error Function                  |
|                                                             |
\------------------------------------------------------------*/

void rtl_error( Error, Text1, Text2, File, Line )

  int   Error;
  char *Text1;
  char *Text2;
  char *File;
  long  Line;
{
  char *Name;

  Name = mbkstrdup( File );
  Name[ strlen( File ) - 1 ] = '\0';

  fprintf( stderr, "%s%ld ", Name, Line );

  switch( Error )
  {
    case RTL_PROC_EXIST_ERROR :
      fprintf( stderr, "process %s exists already\n", Text1 );
    break;
    case RTL_FUNC_EXIST_ERROR :
      fprintf( stderr, "function %s exists already\n", Text1 );
    break;
    case RTL_PLACE_EXIST_ERROR :
      fprintf( stderr, "place %s exists already\n", Text1 );
    break;
    case RTL_TRANS_EXIST_ERROR :
      fprintf( stderr, "transition %s exists already\n", Text1 );
    break;
    case RTL_PORT_EXIST_ERROR :
      fprintf( stderr, "port %s exists already\n", Text1 );
    break;
    case RTL_ARC_PLACE_EXIST_ERROR :
      fprintf( stderr, "arc place %s to %s exists already\n", Text1, Text2 );
    break;
    case RTL_ARC_TRANS_EXIST_ERROR :
      fprintf( stderr, "arc trans %s to %s exists already\n", Text1, Text2 );
    break;
    case RTL_UNKNOWN_DECLAR_TYPE :
      fprintf( stderr, "unknown declar type %ld for %s\n", (long)Text1, Text2 );
    break;
    case RTL_SIGNAL_EXIST_ERROR :
      fprintf( stderr, "signal %s exists already\n", Text1 );
    break;
    case RTL_VARIABLE_EXIST_ERROR :
      fprintf( stderr, "variable %s exists already\n", Text1 );
    break;
    case RTL_DEFINE_EXIST_ERROR :
      fprintf( stderr, "define %s exists already\n", Text1 );
    break;
    case RTL_ARGUMENT_EXIST_ERROR :
      fprintf( stderr, "argument %s exists already\n", Text1 );
    break;
    case RTL_CONSTANT_EXIST_ERROR :
      fprintf( stderr, "constant %s exists already\n", Text1 );
    break;
    case RTL_DEL_ARC_PLACE_ERROR :
      fprintf( stderr, "unable to delete arc from place %s\n", Text1 );
    break;
    case RTL_DEL_ARC_TRANS_ERROR :
      fprintf( stderr, "unable to delete arc from trans %s\n", Text1 );
    break;
    case RTL_TRANS_FATHER_UNKNOWN :
      fprintf( stderr, "unknown father type %ld for transition %s\n", (long)Text1, Text2 );
    break;
    case RTL_DECLAR_EXIST_ERROR   :
      fprintf( stderr, "symbol %s already declared\n", Text1 );
    break;
    case RTL_BAD_DECLAR_TYPE   :
      fprintf( stderr, "bad declaration type %ld\n", (long)Text1 );
    break;
    case RTL_BAD_ASSIGN_TYPE   :
      fprintf( stderr, "bad assign %s, type %ld\n", Text1, (long)Text2 );
    break;
    case RTL_UNDECLARED_SYMBOL   :
      fprintf( stderr, "undeclared symbol %s\n", Text1 );
    break;
    case RTL_ILLEGAL_POSITION_ERROR :
      fprintf( stderr, "illegal position for symbol %s %ld\n", Text1, (long)Text2 );
    break;
    case RTL_ILLEGAL_INDEX_ERROR :
      fprintf( stderr, "illegal index for symbol %s %ld\n", Text1, (long)Text2 );
    break;
    default :

      fprintf( stderr, "unknown internal error %d !\n", Error );
  }

  autexit( 1 );
}

/*------------------------------------------------------------\
|                                                             |
|                         Rtl Warning Function                |
|                                                             |
\------------------------------------------------------------*/

void rtl_warning( Warning, Text1, Text2, File, Line )

  int   Warning;
  char *Text1;
  char *Text2;
  char *File;
  long  Line;
{
  char *Name;

  Name = mbkstrdup( File );
  Name[ strlen( File ) - 1 ] = '\0';

  fprintf( stderr, "%s%ld ", Name, Line );

  switch ( Warning )
  {
    default :

      fprintf( stderr, "unknown internal warning %d !\n", Warning );
  }
}
