/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Sch                           |
|                                                             |
| File    :                  schview.c                        |
|                                                             |
| Date    :                   11.03.97                        |
|                                                             |
| Authors :               Picault  Stephane                   |
|                         Miramond   Benoit                   |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <stdlib.h>

# include MUT_H
# include AUT_H
# include SCL_H

# include "schview.h"
# include "scherror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                       Sch View Functions                    |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        Sch View Wire                        |
|                                                             |
\------------------------------------------------------------*/

void viewschwir( Wire )

  schwir_list *Wire;
{
  fprintf( stdout, "\n--> Wire" );
  fprintf( stdout, "\n\tX : %ld", Wire->X );
  fprintf( stdout, "\n\tY : %ld", Wire->Y );     
  fprintf( stdout, "\n\tDX : %ld", Wire->DX ); 
  fprintf( stdout, "\n\tDX : %ld", Wire->DY ); 
  fprintf( stdout, "\n\tFLAGS : %lx", Wire->FLAGS );
  fprintf( stdout, "\n\tUSER  : %lx", (long)Wire->USER );
  fprintf( stdout, "\n<-- Wire" );
}

/*------------------------------------------------------------\
|                                                             |
|                        Sch View Net                         |
|                                                             |
/------------------------------------------------------------*/

void viewschnet( Net )

  schnet_list *Net ;
{
  schwir_list *ScanWire;
  schcon_list *ScanCon;
  chain_list  *ScanChain;
  char        *RootName;

  fprintf( stdout, "\n\t--> Net" );
  fprintf( stdout, "\n\tNUMBER_IN : %ld", Net->NUMBER_IN );
  fprintf( stdout, "\n\tNUMBER_OUT : %ld", Net->NUMBER_OUT ); 
  fprintf( stdout, "\n\t\tFLAGS : %lx", Net->FLAGS );
  
  fprintf( stdout, "\n\t\tSOURCE_TYPE : %d", (int)Net->SOURCE_TYPE );
  
  fprintf( stdout, "\n\t\tUSER  : %lx", (long)Net->USER );

  fprintf( stdout, "\n\t\tCON_NET :" );

  for ( ScanChain  = Net->CON_NET;
        ScanChain != (chain_list *)0;
        ScanChain  = ScanChain->NEXT )
  {
    ScanCon = (schcon_list *)ScanChain->DATA;

    if ( ScanCon->ROOT_TYPE == SCH_ROOT_CON_FIG )
    {
      RootName = ((schfig_list *)ScanCon->ROOT)->NAME;
      fprintf( stdout, "\n\t\t # %s\tof fig %s", ScanCon->NAME, RootName );
    }
    else
    {
      RootName = ((schbox_list *)ScanCon->ROOT)->NAME;
      fprintf( stdout, "\n\t\t # %s\tof box %s", ScanCon->NAME, RootName );
    }
  }

  fprintf( stdout, "\n\tWIRE :" );

  for ( ScanWire  = Net->WIRE;
        ScanWire != (schwir_list *)0;
        ScanWire  = ScanWire->NEXT )
  {
    viewschwir( ScanWire );
  }

  fprintf( stdout, "\n<-- Net" );
}

/*------------------------------------------------------------\
|                                                             |
|                        Sch View Connector                   |
|                                                             |
\------------------------------------------------------------*/

void viewschcon( Con )

  schcon_list *Con ;
{
  fprintf( stdout, "\n\t--> Con" );
  fprintf( stdout, "\n\t\tNAME : %s", Con->NAME );
  fprintf( stdout, "\n\t\tFLAGS     : %lx", Con->FLAGS );
  fprintf( stdout, "\n\t\tSOURCE_TYPE    : %d", (int)Con->SOURCE_TYPE );
  fprintf( stdout, "\n\t\tUSER      : %lx", (long)Con->USER );
  fprintf( stdout, "\n\t\tDIR       : %d", (int)Con->DIR );
  fprintf( stdout, "\n\t\tTYPE      : Ox%x", (int)Con->TYPE );
  fprintf( stdout, "\n\t\tROOT_TYPE : %d", (int)Con->ROOT_TYPE );
  fprintf( stdout, "\n\t\tX_REL     : %lx", Con->X_REL );
  fprintf( stdout, "\n\t\tY_REL     : %lx", Con->Y_REL );
  fprintf( stdout, "\n\t<-- Con" );
}

/*------------------------------------------------------------\
|                                                             |
|                     Sch View Box                            |
|                                                             |
\------------------------------------------------------------*/

void viewschbox( Box )

  schbox_list *Box;
{
  schcon_list *ScanCon;

  fprintf( stdout, "\n--> Box" );
  fprintf( stdout, "\n\tNAME : %s", Box->NAME );
  fprintf( stdout, "\n\tFLAGS : %lx", Box->FLAGS );
  fprintf( stdout, "\n\tSOURCE_TYPE: %d", (int)Box->SOURCE_TYPE );
  fprintf( stdout, "\n\tUSER  : %lx", (long)Box->USER );
  fprintf( stdout, "\n\tNUMBER_IN : %ld", Box->NUMBER_IN );
  fprintf( stdout, "\n\tNUMBER_OUT : %ld", Box->NUMBER_OUT ); 
  fprintf( stdout, "\n\tX          : %ld", Box->X );
  fprintf( stdout, "\n\tY          : %ld", Box->Y );
  fprintf( stdout, "\n\tDX         : %ld", Box->DX );
  fprintf( stdout, "\n\tDY         : %ld", Box->DY );

  fprintf( stdout, "\n\tCON_IN : " );

  for ( ScanCon  = Box->CON_IN;
        ScanCon != (schcon_list *)0;
        ScanCon  = ScanCon->NEXT )
  {
    viewschcon( ScanCon );
  }

  fprintf( stdout, "\n\tCON_OUT : " );

  for ( ScanCon  = Box->CON_OUT;
        ScanCon != (schcon_list *)0;
        ScanCon  = ScanCon->NEXT )
  {
    viewschcon( ScanCon );
  }

  fprintf( stdout, "\n<-- Box" );
}

/*------------------------------------------------------------\
|                                                             |
|                        Sch View Figure                      |
|                                                             |
\------------------------------------------------------------*/

void viewschfig( Figure )

  schfig_list *Figure;
{
  schcon_list   *ScanCon;
  schbox_list   *ScanBox;
  schnet_list   *ScanNet;

  fprintf( stdout, "\n--> Figure" );
  fprintf( stdout, "\n\tNAME : %s", Figure->NAME  );

  fprintf( stdout, "\n\tFLAGS : %lx", Figure->FLAGS );
  fprintf( stdout, "\n\tSOURCE_TYPE: %d", (int)Figure->SOURCE_TYPE );
  fprintf( stdout, "\n\tUSER  : %lx", (long)Figure->USER );
  fprintf( stdout, "\n\tNUMBER_IN : %ld", Figure->NUMBER_IN );
  fprintf( stdout, "\n\tNUMBER_OUT : %ld", Figure->NUMBER_OUT );
  fprintf( stdout, "\n\tX          : %ld", Figure->X );
  fprintf( stdout, "\n\tY          : %ld", Figure->Y );

  fprintf( stdout, "\n\tCON_IN : " );

  for ( ScanCon  = Figure->CON_IN;
        ScanCon != (schcon_list *)0;
        ScanCon  = ScanCon->NEXT )
  {
    viewschcon( ScanCon );
  }

  fprintf( stdout, "\n\tCON_OUT : " );

  for ( ScanCon  = Figure->CON_OUT;
        ScanCon != (schcon_list *)0;
        ScanCon  = ScanCon->NEXT )
  {
    viewschcon( ScanCon );
  }

  fprintf( stdout, "\n\tBOX : " );

  for ( ScanBox  = Figure->BOX;
        ScanBox != (schbox_list *)0;
        ScanBox  = ScanBox->NEXT )
  {
    viewschbox( ScanBox );
  }

  fprintf( stdout, "\n\tNET : " );

  for ( ScanNet  = Figure->NET;
        ScanNet != (schnet_list *)0;
        ScanNet  = ScanNet->NEXT )
  {
    viewschnet( ScanNet );
  }

  fprintf( stdout, "\n<-- Figure\n" );
}
