/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Scp                           |
|                                                             |
| File    :                 Scp Build                         |
|                                                             |
| Authors :              Picault  Stephane                    |
|                        Miramond   Benoit                    |
|                                                             |
| Date    :                  04.03.98                         |
|                                                             |
\------------------------------------------------------------*/

/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include MUT_H
#include AUT_H
#include SCL_H

#include "scp.h"
#include "scpbuild.h"

#define BOX_SIZE_MIN 8
#define BOX_SIZE_MAX (BOX_SIZE_MIN*3)

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        Locals  Functions                    |
|                                                             |
\------------------------------------------------------------*/

static int SchNetCompare __P((schnet_list **, schnet_list **));
static int CheckNetIntersection __P((chain_list *, schnet_list *));

/*------------------------------------------------------------\
|                                                             |
|                Function Loc Destroy Transparence            |
|                                                             |
\------------------------------------------------------------*/
static void 
loc_destroytransparence(Figure, Transparence)
    schfig_list    *Figure;
    schbox_list    *Transparence;
{
    schnet_list    *SchNet;
    schwir_list    *SchWire;

    SchNet = Transparence->CON_IN->NET;
    SchWire = addschwir(Figure, SchNet);

    SchWire->X = Transparence->X + Transparence->CON_IN->X_REL;
    SchWire->Y = Transparence->Y + Transparence->CON_IN->Y_REL;
    SchWire->DX = Transparence->X + Transparence->CON_OUT->X_REL;
    SchWire->DY = Transparence->Y + Transparence->CON_OUT->Y_REL;

/*  return(0); */
}

/*------------------------------------------------------------\
|                                                             |
|                    Function Loc Destroy Cluster             |
|                                                             |
\------------------------------------------------------------*/
static void 
loc_destroycluster(Figure, Cluster)
    schfig_list    *Figure;
    schbox_list    *Cluster;
{

    schwir_list    *SchWire;
    schbox_list    *Box;
    schbox_list    *Box1;
    schbox_list    *Box2;
    schcon_list    *Con;
    schcon_list    *ConCast;
    schnet_list    *SchNet;
    schcon_list    *ClusterCon;
    chain_list     *ScanBox;
    chain_list     *ScanCon;

    ScanBox = (chain_list *) Cluster->SOURCE;
    Box = (schbox_list *) ScanBox->DATA;
    ScanBox = ScanBox->NEXT;
    Box2 = (schbox_list *) ScanBox->DATA;
    ScanBox = ScanBox->NEXT;
    Box1 = (schbox_list *) ScanBox->DATA;


    Box1->USER = (void *) Cluster->USER;
    Box1->X = Cluster->X;
    Box1->Y = Cluster->Y;
    Box1->DX = (Cluster->DX / 3);

    for (Con = Box1->CON_OUT;
         Con != (schcon_list *) 0;
         Con = Con->NEXT) {
        Con->X_REL = Box1->DX + 2;
    };

    for (Con = Box1->CON_IN;
         Con != (schcon_list *) 0;
         Con = Con->NEXT) {
        ClusterCon = (schcon_list *) Con->USER;

        SchNet = Con->NET;

        for (ScanCon = SchNet->CON_NET;
             ScanCon != (chain_list *) 0;
             ScanCon = ScanCon->NEXT) {
            ConCast = (schcon_list *) ScanCon->DATA;
            if (ConCast == ClusterCon) {
                ScanCon->DATA = (void *) Con;
            }
        }
    };

    Box2->USER = (void *) Cluster->USER;
    Box2->X = Cluster->X;
    Box2->Y = Cluster->Y + Box1->DY;
    Box2->DX = (Cluster->DX / 3);

    for (Con = Box2->CON_OUT;
         Con != (schcon_list *) 0;
         Con = Con->NEXT) {
        Con->X_REL = Box2->DX + 2;
    }

    for (Con = Box2->CON_IN;
         Con != (schcon_list *) 0;
         Con = Con->NEXT) {
        ClusterCon = (schcon_list *) Con->USER;

        SchNet = Con->NET;

        for (ScanCon = SchNet->CON_NET;
             ScanCon != (chain_list *) 0;
             ScanCon = ScanCon->NEXT) {
            ConCast = (schcon_list *) ScanCon->DATA;
            if (ConCast == ClusterCon) {
                ScanCon->DATA = (void *) Con;
            }
        }
    }

    Box->USER = (void *) Cluster->USER;
    Box->X = Cluster->X + ((Cluster->DX * 2) / 3);
    Box->Y = Cluster->Y + (Cluster->CON_OUT->Y_REL - 4);
    Box->DX = (Cluster->DX / 3);

    for (Con = Box->CON_OUT;
         Con != (schcon_list *) 0;
         Con = Con->NEXT) {
        Con->X_REL = Box->DX + 2;

        ClusterCon = (schcon_list *) Con->USER;

        SchNet = Con->NET;

        for (ScanCon = SchNet->CON_NET;
             ScanCon != (chain_list *) 0;
             ScanCon = ScanCon->NEXT) {
            ConCast = (schcon_list *) ScanCon->DATA;
            if (ConCast == ClusterCon) {
                ScanCon->DATA = (void *) Con;
            }
        }
    }

    SchWire = addschwir(Figure, Box1->CON_OUT->NET);

    SchWire->X = Box1->X + Box1->CON_OUT->X_REL;
    SchWire->Y = Box1->Y + Box1->CON_OUT->Y_REL;
    SchWire->DX = Box->X + Box->CON_IN->NEXT->X_REL;
    SchWire->DY = Box->Y + Box->CON_IN->NEXT->Y_REL;

    SchWire = addschwir(Figure, Box2->CON_OUT->NET);

    SchWire->X = Box2->X + Box2->CON_OUT->X_REL;
    SchWire->Y = Box2->Y + Box2->CON_OUT->Y_REL;
    SchWire->DX = Box->X + Box->CON_IN->X_REL;
    SchWire->DY = Box->Y + Box->CON_IN->Y_REL;

}


/*------------------------------------------------------------\
|                                                             |
|                   Function  Loc Build Cluster               |
|                                                             |
\------------------------------------------------------------*/
static void 
loc_buildcluster(Figure)
    schfig_list    *Figure;
{
    schcon_list    *Con1;
    schbox_list    *Box;
    schcon_list    *Con2;
    schcon_list    *SchCon;
    schcon_list    *ConCast;
    schcon_list    *ClusterCon;
    schbox_list    *Box1;
    schbox_list    *Box2;
    schbox_list    *Cluster;
    schnet_list    *SchNet;

    chain_list     *ScanCon;
    int             Go;


    for (Box = Figure->BOX;
         Box != (schbox_list *) 0;
         Box = Box->NEXT) {

        Go = 1;

        if ((!IsSchBoxCluster(Box)) &&
            (!IsSchBoxClusterised(Box)) &&
            (Box->NUMBER_IN == 2) &&
            (Box->NUMBER_OUT == 1)) {
            Con2 = (schcon_list *) Box->CON_IN;
            Con1 = (schcon_list *) Box->CON_IN->NEXT;

            if ((Con1->NET->NUMBER_OUT == 1) && (Con2->NET->NUMBER_OUT == 1)
            && (Con1->NET->NUMBER_IN == 1) && (Con2->NET->NUMBER_IN == 1)) {
                for (ScanCon = Con1->NET->CON_NET;
                     ScanCon != (chain_list *) 0;
                     ScanCon = ScanCon->NEXT) {
                    ConCast = (schcon_list *) ScanCon->DATA;

                    if (ConCast != Con1) {
                        if (ConCast->ROOT_TYPE == SCH_ROOT_CON_BOX) {
                            Box1 = ConCast->ROOT;

                            if (IsSchBoxClusterised(Box1)
                                || IsSchBoxCluster(Box1)
                                || (Box1->NUMBER_OUT != 1)) {
                                Go = 0;
                            }
                        }
                        else
                            Go = 0;
                    }
                }

                if (Go == 1) {
                    for (ScanCon = Con2->NET->CON_NET;
                         ScanCon != (chain_list *) 0;
                         ScanCon = ScanCon->NEXT) {
                        ConCast = (schcon_list *) ScanCon->DATA;

                        if (ConCast != Con2) {
                            if (ConCast->ROOT_TYPE == SCH_ROOT_CON_BOX) {
                                Box2 = ConCast->ROOT;

                                if (IsSchBoxClusterised(Box2)
                                    || IsSchBoxCluster(Box2)
                                    || (Box2->NUMBER_OUT != 1)) {
                                    Go = 0;
                                }
                            }
                            else
                                Go = 0;
                        }
                    }
                }
                if (Go == 1) {
                    Cluster = addschbox(Figure, "Cluster");

                    SetSchBoxCluster(Cluster);

                    SetSchBoxClusterised(Box1);
                    Cluster->SOURCE = (chain_list *) addchain((chain_list *) Cluster->SOURCE
                                                            ,(void *) Box1);

                    Box1->CON_IN = (schcon_list *) reverse((chain_list *) Box1->CON_IN);

                    for (SchCon = Box1->CON_IN;
                         SchCon != (schcon_list *) 0;
                         SchCon = SchCon->NEXT) {
                        ClusterCon = addschboxconin(Figure, Cluster, SchCon->NAME);

                        SchNet = SchCon->NET;

                        for (ScanCon = SchNet->CON_NET;
                             ScanCon != (chain_list *) 0;
                             ScanCon = ScanCon->NEXT) {
                            ConCast = (schcon_list *) ScanCon->DATA;
                            if (ConCast == SchCon) {
                                ScanCon->DATA = (void *) ClusterCon;
                            }
                        }

                        ClusterCon->NET = SchNet;

                        SchCon->USER = (void *) ClusterCon;
                    }

                    Box1->CON_IN = (schcon_list *) reverse((chain_list *) Box1->CON_IN);

                    SetSchBoxClusterised(Box2);
                    Cluster->SOURCE = (chain_list *) addchain((chain_list *) Cluster->SOURCE
                                                            ,(void *) Box2);

                    Box2->CON_IN = (schcon_list *) reverse((chain_list *) Box2->CON_IN);

                    for (SchCon = Box2->CON_IN;
                         SchCon != (schcon_list *) 0;
                         SchCon = SchCon->NEXT) {
                        ClusterCon = addschboxconin(Figure, Cluster, SchCon->NAME);

                        SchNet = SchCon->NET;

                        for (ScanCon = SchNet->CON_NET;
                             ScanCon != (chain_list *) 0;
                             ScanCon = ScanCon->NEXT) {
                            ConCast = (schcon_list *) ScanCon->DATA;
                            if (ConCast == SchCon) {
                                ScanCon->DATA = (void *) ClusterCon;
                            }
                        }

                        ClusterCon->NET = SchNet;

                        SchCon->USER = (void *) ClusterCon;

                    }

                    Box2->CON_IN = (schcon_list *) reverse((chain_list *) Box2->CON_IN);

                    SetSchBoxClusterised(Box);
                    Cluster->SOURCE = (chain_list *) addchain((chain_list *) Cluster->SOURCE
                                                              ,(void *) Box);

                    for (SchCon = Box->CON_OUT;
                         SchCon != (schcon_list *) 0;
                         SchCon = SchCon->NEXT) {
                        ClusterCon = addschboxconout(Figure, Cluster, SchCon->NAME);

                        SchNet = SchCon->NET;

                        for (ScanCon = SchNet->CON_NET;
                             ScanCon != (chain_list *) 0;
                             ScanCon = ScanCon->NEXT) {
                            ConCast = (schcon_list *) ScanCon->DATA;
                            if (ConCast == SchCon) {
                                ScanCon->DATA = (void *) ClusterCon;
                            }
                        }

                        ClusterCon->NET = SchNet;
                        SchCon->USER = (void *) ClusterCon;
                    }
                }
            }
        }
    }
}

/*------------------------------------------------------------\
|                                                             |
|                  Function  Loc Build Place First            |
|                                                             |
\------------------------------------------------------------*/
static scpcol_list *
loc_buildplacefirst(Figure)
    schfig_list    *Figure;
{
    scpcol_list    *ColMax;
    scpcel_list    *Cell;
    scpcol_list    *Column;
    schbox_list    *Box;
    schcon_list    *Con;

    long            MaxCumulY = 0;
    long            CumulY;


    for (Column = HEAD_SCPCOL;
         Column != (scpcol_list *) 0;
         Column = Column->NEXT) {
        CumulY = 0;

        for (Cell = Column->CELL;
             Cell != (scpcel_list *) 0;
             Cell = Cell->NEXT) {
            if (Cell->TYPE == SCP_CEL_BOX) {
                Box = (schbox_list *) Cell->ROOT;
                Box->Y = CumulY;
                CumulY += Box->DY + 5;
                if (Box->DX > Column->MAX_DX_CELL) {
                    Column->MAX_DX_CELL = Box->DX;
                }
            }
            else if (Cell->TYPE == SCP_CEL_CON) {
                Con = (schcon_list *) Cell->ROOT;
                Con->Y_REL = CumulY;
                CumulY += 3;
                if (2 > Column->MAX_DX_CELL) {
                    Column->MAX_DX_CELL = 2;
                }
            }
        }
        if (MaxCumulY < CumulY) {
            ColMax = Column;
            MaxCumulY = CumulY;
        }
    }
    return (ColMax);
}


/*------------------------------------------------------------\
|                                                             |
|                Function  Loc Build Place In                 |
|                                                             |
\------------------------------------------------------------*/
static schfig_list *
loc_buildplacein(Figure, ColumnFrom, ColumnTo)
    schfig_list    *Figure;
    scpcol_list    *ColumnFrom;
    scpcol_list    *ColumnTo;
{
    scpcel_list    *CellTo;
    schcon_list    *ConTo;
    schbox_list    *BoxTo;

    scpcel_list    *CellFrom;
    schbox_list    *BoxFrom;
    schcon_list    *ConFrom;

    chain_list     *Chain;
    char            Begin = 1;

    long            CumulY;

    for (CellTo = ColumnTo->LAST_CELL; CellTo; CellTo = CellTo->PREV) {

        if (CellTo->TYPE == SCP_CEL_BOX) {
            BoxTo = (schbox_list *) CellTo->ROOT;

            for (ConTo = BoxTo->CON_IN; ConTo; ConTo = ConTo->NEXT) {
                for (Chain = ConTo->NET->CON_NET; Chain; Chain = Chain->NEXT) {
                    ConFrom = (schcon_list *) Chain->DATA;

                    if (IsSchConInternal(ConFrom) && IsSchConOut(ConFrom)) {
                        BoxFrom = (schbox_list *) ConFrom->ROOT;

                        if (IsSchBoxTaged(BoxFrom) && !IsSchBoxPlaced(BoxFrom)) {
                            CellFrom = (scpcel_list *) BoxFrom->USER;
                            if (CellFrom->COL == ColumnFrom) {
                                if (Begin) {
                                    /* juste une idee de grandeur */
                                    CumulY = BoxTo->Y + 2 * (BoxFrom->DY) + (BoxTo->DY);
                                    Begin = 0;
                                }

                                if ((BoxTo->Y + ConTo->Y_REL - ConFrom->Y_REL + BoxFrom->DY) <= CumulY) {
                                    BoxFrom->Y = BoxTo->Y + ConTo->Y_REL - ConFrom->Y_REL;
                                }
                                else {
                                    BoxFrom->Y = CumulY - BoxFrom->DY;
                                }
                                CumulY = BoxFrom->Y - 1;
                                SetSchBoxPlaced(BoxFrom);
                            }
                        }
                    }
                    if (IsSchConExternal(ConFrom) && IsSchConIn(ConFrom)
                    && !IsSchConPlaced(ConFrom) && IsSchConTaged(ConFrom)) {
                        CellFrom = (scpcel_list *) ConFrom->USER;
                        if (CellFrom->COL == ColumnFrom) {
                            if (!IsSchConPlaced(ConFrom)) {
                                ConFrom->Y_REL = (BoxTo->Y + ConTo->Y_REL);
                                SetSchConPlaced(ConFrom);
                            }
                        }
                    }
                }
            }
        }
    }
    return (Figure);
}

/*------------------------------------------------------------\
|                                                             |
|                Function  Loc Build Place Out                |
|                                                             |
\------------------------------------------------------------*/
static schfig_list *
loc_buildplaceout(Figure, ColumnFrom, ColumnTo)
    schfig_list    *Figure;
    scpcol_list    *ColumnFrom;
    scpcol_list    *ColumnTo;
{
    chain_list     *Chain;
    char            Begin = 1;

    schbox_list    *BoxFrom;
    schcon_list    *ConFrom;
    scpcel_list    *CellFrom;

    schcon_list    *ConTo;
    scpcel_list    *CellTo;
    schbox_list    *BoxTo;

    long            CumulY;

    for (CellTo = ColumnTo->LAST_CELL; CellTo; CellTo = CellTo->PREV) {

        if (CellTo->TYPE == SCP_CEL_BOX) {
            BoxTo = (schbox_list *) CellTo->ROOT;

            for (ConTo = BoxTo->CON_IN; ConTo && !IsSchBoxPlaced(BoxTo); ConTo = ConTo->NEXT) {
                for (Chain = ConTo->NET->CON_NET; Chain && !IsSchBoxPlaced(BoxTo); Chain = Chain->NEXT) {
                    ConFrom = (schcon_list *) Chain->DATA;
                    if (ConFrom != ConTo) {
                        if (IsSchConInternal(ConFrom) && IsSchConOut(ConFrom)) {
                            BoxFrom = (schbox_list *) ConFrom->ROOT;
                            if (IsSchBoxTaged(BoxFrom)) {
                                CellFrom = (scpcel_list *) BoxFrom->USER;
                                if (CellFrom->COL == ColumnFrom) {
                                    if (Begin) {
                                        CumulY = BoxFrom->Y + 2 * (BoxTo->DY + BoxFrom->DY);
                                        Begin = 0;
                                    }
                                    if ((BoxFrom->Y + ConFrom->Y_REL - ConTo->Y_REL + BoxTo->DY) <= CumulY) {
                                        SetSchBoxPlaced(BoxTo);
                                        BoxTo->Y = BoxFrom->Y + ConFrom->Y_REL - ConTo->Y_REL;
                                        CumulY = BoxTo->Y - 1;
                                    }
                                    else {        /*
                                                   BoxTo->Y =  CumulY-BoxTo->DY;
                                                 */
                                    }
                                }
                            }
                        }
                    }
                }
            }
            if (!IsSchBoxPlaced(BoxTo)) {
                BoxTo->Y = CumulY - BoxTo->DY;
                SetSchBoxPlaced(BoxTo);
                CumulY = BoxTo->Y - 1;
            }
        }

        if (CellTo->TYPE == SCP_CEL_CON) {
            ConTo = (schcon_list *) CellTo->ROOT;

            for (Chain = ConTo->NET->CON_NET; Chain && !IsSchConPlaced(ConTo); Chain = Chain->NEXT) {
                ConFrom = (schcon_list *) Chain->DATA;
                if (ConFrom != ConTo) {
                    if (IsSchConInternal(ConFrom) && IsSchConOut(ConFrom)) {
                        BoxFrom = (schbox_list *) ConFrom->ROOT;
                        if (IsSchBoxTaged(BoxFrom)) {
                            CellFrom = (scpcel_list *) BoxFrom->USER;
                            if (CellFrom->COL == ColumnFrom) {
                                if (Begin) {
                                    CumulY = BoxFrom->Y + 2 * (BoxFrom->DY);
                                    Begin = 0;
                                }
                                if ((BoxFrom->Y + ConFrom->Y_REL + 2) <= CumulY) {
                                    SetSchConPlaced(ConTo);
                                    ConTo->Y_REL = BoxFrom->Y + ConFrom->Y_REL;
                                    CumulY = ConTo->Y_REL - 1;
                                }
                                else {        /*
                                           ConTo->Y_REL =  CumulY;
                                         */
                                }
                            }
                        }
                    }
                }
            }
            if (Begin && IsSchConPlaced(ConTo)) {
                ConTo->Y_REL = CumulY;
                SetSchConPlaced(ConTo);
                CumulY = ConTo->Y_REL - 1;
            }
        }
    }
    return (Figure);
}

/*------------------------------------------------------------\
|                                                             |
|                  Function  Loc Build Place                  |
|                                                             |
\------------------------------------------------------------*/
static schfig_list *
loc_buildplace(Figure)
    schfig_list    *Figure;
{

    scpcol_list    *Column;
    scpcol_list    *MaxColumn;

    MaxColumn = loc_buildplacefirst(Figure);

    for (Column = MaxColumn; Column->NEXT; Column = Column->NEXT) {
        loc_buildplaceout(Figure, Column, Column->NEXT);
    }

    HEAD_SCPCOL = (scpcol_list *) reverse((chain_list *) HEAD_SCPCOL);

    for (Column = MaxColumn; Column->NEXT; Column = Column->NEXT) {
        loc_buildplacein(Figure, Column->NEXT, Column);
    }

    HEAD_SCPCOL = (scpcol_list *) reverse((chain_list *) HEAD_SCPCOL);

    return (Figure);
}

/*------------------------------------------------------------\
|                                                             |
|                 Function  Loc Routage Canal                 |
|                                                             |
\------------------------------------------------------------*/
static long 
loc_routage_canal(Figure, ColumnFrom, ColumnTo, Xdep)
    schfig_list    *Figure;
    scpcol_list    *ColumnFrom;
    scpcol_list    *ColumnTo;
    long            Xdep;
{
    scpcel_list    *ScanCell;
    schbox_list    *SchBox;
    schnet_list    *SchNet;
    schnet_list   **SchNetTable;
    schwir_list    *SchWire;
    schcon_list    *ScanCon;
    schcon_list    *SchCon;
    chain_list     *SchNetList = NULL;
    chain_list     *CurChannelNets = NULL;
    chain_list     *ptchain;
    long            numschnet = 0;
    long            channel;
    long            i = 0;

    Xdep += ColumnFrom->MAX_DX_CELL;

    /* Traverse Canal Input and Output Connectors to */
    /* obtain list of SchNets in Canal and their     */
    /* vertical extremities with direction           */

    /* Channel Input Connectors */

    for (ScanCell = ColumnFrom->CELL; ScanCell; ScanCell = ScanCell->NEXT) {
        if (ScanCell->TYPE == SCP_CEL_BOX) {
            SchBox = (schbox_list *) ScanCell->ROOT;
            for (ScanCon = SchBox->CON_OUT; ScanCon; ScanCon = ScanCon->NEXT) {
                SchNet = ScanCon->NET;
                if (!IsSchNetCanal(SchNet)) {
                    numschnet++;
                    SetSchNetCanal(SchNet);
                    SchNetList = addchain(SchNetList, SchNet);
                    SchNet->YMAX = SchBox->Y + ScanCon->Y_REL;
                    SchNet->YMIN = SchBox->Y + ScanCon->Y_REL;
                }
                else if ((SchBox->Y + ScanCon->Y_REL) > (SchNet->YMAX)) {
                    SchNet->YMAX = SchBox->Y + ScanCon->Y_REL;
                }
                else if ((SchBox->Y + ScanCon->Y_REL) < (SchNet->YMIN)) {
                    SchNet->YMIN = SchBox->Y + ScanCon->Y_REL;
                }
            }
        }
        else if (ScanCell->TYPE == SCP_CEL_CON) {
            SchCon = (schcon_list *) ScanCell->ROOT;

            SchNet = SchCon->NET;
            if (!IsSchNetCanal(SchNet)) {
                numschnet++;
                SetSchNetCanal(SchNet);
                SchNetList = addchain(SchNetList, SchNet);
                SchNet->YMAX = SchCon->Y_REL;
                SchNet->YMIN = SchCon->Y_REL;
            }
            if ((SchCon->Y_REL) > (SchNet->YMAX)) {
                SchNet->YMAX = SchCon->Y_REL;
            }
            if ((SchCon->Y_REL) < (SchNet->YMIN)) {
                SchNet->YMIN = SchCon->Y_REL;
            }
        }
    }

    /* Channel Output connectors */

    for (ScanCell = ColumnTo->CELL; ScanCell; ScanCell = ScanCell->NEXT) {
        if (ScanCell->TYPE == SCP_CEL_BOX) {
            SchBox = (schbox_list *) ScanCell->ROOT;
            for (ScanCon = SchBox->CON_IN; ScanCon; ScanCon = ScanCon->NEXT) {
                SchNet = ScanCon->NET;
                if (!IsSchNetCanal(SchNet)) {
                    numschnet++;
                    SetSchNetCanal(SchNet);
                    SchNetList = addchain(SchNetList, SchNet);
                    SchNet->YMAX = SchBox->Y + ScanCon->Y_REL;
                    SetSchNetMaxOutput(SchNet);
                    SchNet->YMIN = SchBox->Y + ScanCon->Y_REL;
                    SetSchNetMinOutput(SchNet);
                }
                else if ((SchBox->Y + ScanCon->Y_REL) > (SchNet->YMAX)) {
                    SchNet->YMAX = SchBox->Y + ScanCon->Y_REL;
                    SetSchNetMaxOutput(SchNet);
                }
                else if ((SchBox->Y + ScanCon->Y_REL) < (SchNet->YMIN)) {
                    SchNet->YMIN = SchBox->Y + ScanCon->Y_REL;
                    SetSchNetMinOutput(SchNet);
                }
            }
        }
        else if (ScanCell->TYPE == SCP_CEL_CON) {
            SchCon = (schcon_list *) ScanCell->ROOT;

            SchNet = SchCon->NET;
            if (!IsSchNetCanal(SchNet)) {
                numschnet++;
                SetSchNetCanal(SchNet);
                SchNetList = addchain(SchNetList, SchNet);
                SchNet->YMAX = SchCon->Y_REL;
                SetSchNetMaxOutput(SchNet);
                SchNet->YMIN = SchCon->Y_REL;
                SetSchNetMinOutput(SchNet);
            }
            if ((SchCon->Y_REL) > (SchNet->YMAX)) {
                SchNet->YMAX = SchCon->Y_REL;
                SetSchNetMaxOutput(SchNet);
            }
            if ((SchCon->Y_REL) < (SchNet->YMIN)) {
                SchNet->YMIN = SchCon->Y_REL;
                SetSchNetMinOutput(SchNet);
            }
        }
    }

    /* build table of SchNets and sort */

    SchNetTable = (schnet_list **) mbkalloc(numschnet * sizeof(schnet_list *));
    for (ptchain = SchNetList; ptchain; ptchain = ptchain->NEXT) {
        SchNetTable[i++] = (schnet_list *) ptchain->DATA;
    }
    qsort(SchNetTable, numschnet, sizeof(schnet_list *), SchNetCompare);

    /* Allocate a channel number to each SchNet */

    channel = Xdep + 4;
    for (i = 0; i < numschnet; i++) {
        if ((SchNetTable[i])->YMAX == (SchNetTable[i])->YMIN) {
            (SchNetTable[i])->CANAL = channel;
            continue;
        }
        if (CheckNetIntersection(CurChannelNets, SchNetTable[i])) {
            channel += 2;
            if (CurChannelNets)
                freechain(CurChannelNets);
            CurChannelNets = addchain(NULL, SchNetTable[i]);
        }
        else
            CurChannelNets = addchain(CurChannelNets, SchNetTable[i]);
        (SchNetTable[i])->CANAL = channel;
    }
    mbkfree(SchNetTable);

    /* Draw horizontal wires from input column to vertical wire */

    for (ScanCell = ColumnFrom->CELL; ScanCell; ScanCell = ScanCell->NEXT) {
        if (ScanCell->TYPE == SCP_CEL_BOX) {
            SchBox = ScanCell->ROOT;

            for (ScanCon = SchBox->CON_OUT; ScanCon; ScanCon = ScanCon->NEXT) {
                SchNet = ScanCon->NET;
                SchWire = addschwir(Figure, SchNet);
                SchWire->X = SchBox->X + ScanCon->X_REL;
                SchWire->DX = SchNet->CANAL;
                SchWire->Y = SchBox->Y + ScanCon->Y_REL;
                SchWire->DY = SchBox->Y + ScanCon->Y_REL;
            }
        }
        else if (ScanCell->TYPE == SCP_CEL_CON) {
            SchCon = (schcon_list *) ScanCell->ROOT;
            SchNet = SchCon->NET;
            SchWire = addschwir(Figure, SchNet);
            SchWire->X = SchCon->X_REL + 2;
            SchWire->DX = SchNet->CANAL;
            SchWire->Y = SchCon->Y_REL;
            SchWire->DY = SchCon->Y_REL;
        }
    }

    /* Draw horizontal wires from vertical wire to output column */

    for (ScanCell = ColumnTo->CELL; ScanCell; ScanCell = ScanCell->NEXT) {
        if (ScanCell->TYPE == SCP_CEL_BOX) {
            SchBox = ScanCell->ROOT;
            SchBox->X = channel + 6 + (ColumnTo->MAX_DX_CELL - SchBox->DX) / 2;

            for (ScanCon = SchBox->CON_IN; ScanCon; ScanCon = ScanCon->NEXT) {
                SchNet = ScanCon->NET;
                SchWire = addschwir(Figure, SchNet);
                SchWire->X = SchBox->X + ScanCon->X_REL;
                SchWire->DX = SchNet->CANAL;
                SchWire->Y = SchBox->Y + ScanCon->Y_REL;
                SchWire->DY = SchBox->Y + ScanCon->Y_REL;
            }
        }
        else if (ScanCell->TYPE == SCP_CEL_CON) {
            SchCon = (schcon_list *) ScanCell->ROOT;
            SchCon->X_REL = channel + 6;

            SchNet = SchCon->NET;
            SchWire = addschwir(Figure, SchNet);
            SchWire->X = SchCon->X_REL - 2;
            SchWire->DX = SchNet->CANAL;
            SchWire->Y = SchCon->Y_REL;
            SchWire->DY = SchCon->Y_REL;
        }
    }

    /* Draw vertical wires */

    for (ptchain = SchNetList; ptchain; ptchain = ptchain->NEXT) {
        SchNet = (schnet_list *) ptchain->DATA;
        ClearSchNetCanal(SchNet);
        ClearSchNetDir(SchNet);
        if (SchNet->YMIN == SchNet->YMAX)
            continue;
        SchWire = addschwir(Figure, SchNet);
        SchWire->X = SchNet->CANAL;
        SchWire->DX = SchNet->CANAL;
        SchWire->Y = SchNet->YMIN;
        SchWire->DY = SchNet->YMAX;
    }
    freechain(SchNetList);

    return (channel + 6);
}

static int 
SchNetCompare(net1, net2)
    schnet_list   **net1;
    schnet_list   **net2;
{
    long            type1, type2;

    type1 = GetSchNetDirType(*net1);
    type2 = GetSchNetDirType(*net2);

    if (type1 > type2)
        return (1);
    if (type1 < type2)
        return (-1);

    switch (type1) {
    case 0:
        return (((*net1)->YMAX - (*net1)->YMIN) - ((*net2)->YMAX - (*net2)->YMIN));
        break;
    case SCH_NET_MINDIR:
        return ((*net1)->YMIN - (*net2)->YMIN);
        break;
    case SCH_NET_MAXDIR:
        return ((*net2)->YMAX - (*net1)->YMAX);
        break;
    case SCH_NET_MAXDIR | SCH_NET_MINDIR:
        return (((*net2)->YMAX - (*net2)->YMIN) - ((*net1)->YMAX - (*net1)->YMIN));
        break;
    }
    return 0;
}

static int 
CheckNetIntersection(NetList, Net)
    chain_list     *NetList;
    schnet_list    *Net;
{
    chain_list     *ptchain;
    schnet_list    *CmpNet;

    for (ptchain = NetList; ptchain; ptchain = ptchain->NEXT) {
        CmpNet = (schnet_list *) ptchain->DATA;
        if (Net->YMIN > CmpNet->YMAX || Net->YMAX < CmpNet->YMIN)
            continue;
        else
            break;
    }
    if (ptchain != NULL)
        return 1;
    else
        return 0;
}

/*------------------------------------------------------------\
|                                                             |
|                      Function  Loc Route                    |
|                                                             |
\------------------------------------------------------------*/
static schfig_list *
loc_route(Figure)
    schfig_list    *Figure;
{
    scpcol_list    *Column_from;
    scpcol_list    *Column_to;

    long            Xdep = 0;


    for (Column_from = HEAD_SCPCOL;
         Column_from->NEXT != (scpcol_list *) 0;
         Column_from = Column_from->NEXT) {
        Column_to = Column_from->NEXT;
        Xdep = loc_routage_canal(Figure, Column_from, Column_to, Xdep);
    }
    return (Figure);
}


/*------------------------------------------------------------\
|                                                             |
|                Function  Loc Add Transparence               |
|                                                             |
\------------------------------------------------------------*/
static schbox_list *
loc_add_transparence(Figure, Net)
    schfig_list    *Figure;
    schnet_list    *Net;
{
    schbox_list    *Box;
    schcon_list    *Con_in;
    schcon_list    *Con_out;


    Box = addschbox(Figure, "Transparence");
    SetSchBoxTransparence(Box);

    Con_out = addschboxconout(Figure, Box, "T_out");
    Con_in = addschboxconin(Figure, Box, "T_in");

    addschnetcon(Net, Con_out);
    addschnetcon(Net, Con_in);

    return (Box);
}
/*------------------------------------------------------------\
|                                                             |
|                   Function  Loc Test Trans                  |
|                                                             |
\------------------------------------------------------------*/
static int 
loc_test_trans(Net, Column)
    schnet_list    *Net;
    scpcol_list    *Column;
{
    scpcel_list    *ScpCel;
    schcon_list    *SchCon;
    schbox_list    *SchBox;
    chain_list     *ChainCon;


    for (ChainCon = Net->CON_NET;
         ChainCon != (chain_list *) 0;
         ChainCon = ChainCon->NEXT) {
        SchCon = (schcon_list *) ChainCon->DATA;
        if (IsSchConInternal(SchCon)) {
            SchBox = (schbox_list *) SchCon->ROOT;
            if (IsSchBoxTransparence(SchBox)) {
                ScpCel = (scpcel_list *) SchBox->USER;

                if (ScpCel->COL == Column) {
                    return (0);
                }
            }
        }
    }
    return (1);
}


/*------------------------------------------------------------\
|                                                             |
|               Function  Loc Placement Initial               |
|                                                             |
\------------------------------------------------------------*/
static schfig_list *
loc_placement_initial(Figure)
    schfig_list    *Figure;
{

    int             cpt_transparence;

    int             cpt;
    schcon_list    *Con;
    schcon_list    *Con2;
    chain_list     *Chain;

    schbox_list    *Box;
    schbox_list    *Box2;
    schbox_list    *Box4;
    scpcol_list    *Column2;
    scpcol_list    *Column;
    scpcel_list    *Cell;
    scpcel_list    *Cell2;

    schbox_list    *BoxBox4;
    scpcel_list    *BoxCell2;
    schcon_list    *BoxCon;
    schcon_list    *BoxCon2;
    chain_list     *BoxChain;


    /* connecteurs externes de sortie dans la premiere colonne */

    Column2 = addscpcol();
    Column = addscpcol();


    for (Con = Figure->CON_OUT;
         Con != (schcon_list *) 0;
         Con = Con->NEXT) {
        addschcontoscpcel(Con, addscpcel(Column2));

        for (Chain = Con->NET->CON_NET;
             Chain != (chain_list *) 0;
             Chain = Chain->NEXT) {
            if (loc_test_trans(Con->NET, Column)) {
                Box2 = loc_add_transparence(Figure, Con->NET);
                SetSchBoxTaged(Box2);
                Cell2 = addscpcel(Column);
                addschboxtoscpcel(Box2, Cell2);
            }
        }
    }

    while (cpt_transparence != Column->NUMBER_CELL) {

        Column2 = Column;
        Column = addscpcol();

        cpt = Column2->NUMBER_CELL;
        cpt_transparence = 0;

        for (Cell = Column2->CELL;
             Cell != (scpcel_list *) 0;
             Cell = Cell->NEXT) {
            if (Cell->TYPE == SCP_CEL_BOX) {
                Box = (schbox_list *) Cell->ROOT;
                SetSchBoxReTaged(Box);
            }
        }

        for (Cell = Column2->LAST_CELL;
             cpt != 0;
             Cell = (scpcel_list *) Cell->PREV) {
            cpt--;
            if (Cell->TYPE == SCP_CEL_BOX) {
                Box = (schbox_list *) Cell->ROOT;

                for (Con = Box->CON_IN;
                     Con != (schcon_list *) 0;
                     Con = Con->NEXT) {
                    for (Chain = Con->NET->CON_NET;
                         Chain != (chain_list *) 0;
                         Chain = Chain->NEXT) {
                        Con2 = (schcon_list *) Chain->DATA;

                        if (Con != Con2) {
                            if (IsSchConIn(Con2) && IsSchConInternal(Con2)
                                && !IsSchBoxTaged((schbox_list *) Con2->ROOT)
                                && loc_test_trans(Con->NET, Column)) {
                                Box4 = loc_add_transparence(Figure, Con2->NET);
                                Box4->NAME = namealloc("In_int_not_tag");
                                SetSchBoxTaged(Box4);
                                Cell2 = addscpcel(Column);
                                addschboxtoscpcel(Box4, Cell2);
                                cpt_transparence++;
                            }

                            if (IsSchConOut(Con2) && IsSchConInternal(Con2)) {
                                Box4 = (schbox_list *) Con2->ROOT;
                                if (!IsSchBoxTaged(Box4)) {
                                    SetSchBoxTaged(Box4);
                                    Cell2 = addscpcel(Column);
                                    addschboxtoscpcel(Box4, Cell2);

                                    for (BoxCon = Box4->CON_IN;
                                         BoxCon != (schcon_list *) 0;
                                         BoxCon = BoxCon->NEXT) {
                                        for (BoxChain = BoxCon->NET->CON_NET;
                                             BoxChain != (chain_list *) 0;
                                             BoxChain = BoxChain->NEXT) {
                                            BoxCon2 = (schcon_list *) BoxChain->DATA;

                                            if (BoxCon != BoxCon2) {
                                                if (IsSchConInternal(BoxCon2)
                                                    && IsSchBoxTaged((schbox_list *) BoxCon2->ROOT)
                                                    && IsSchBoxReTaged((schbox_list *) BoxCon2->ROOT)
                                                    && loc_test_trans(BoxCon->NET, Column)) {
                                                    BoxBox4 = loc_add_transparence(Figure, BoxCon2->NET);
                                                    BoxBox4->NAME = namealloc("IN__Int_Tag");
                                                    SetSchBoxTaged(BoxBox4);
                                                    BoxCell2 = addscpcel(Column);
                                                    addschboxtoscpcel(BoxBox4, BoxCell2);
                                                    cpt_transparence++;
                                                }
                                            }
                                        }
                                    }
                                    for (BoxCon = Box4->CON_OUT;
                                         BoxCon != (schcon_list *) 0;
                                         BoxCon = BoxCon->NEXT) {
                                        for (BoxChain = BoxCon->NET->CON_NET;
                                             BoxChain != (chain_list *) 0;
                                             BoxChain = BoxChain->NEXT) {
                                            BoxCon2 = (schcon_list *) BoxChain->DATA;

                                            if (BoxCon != BoxCon2) {
                                                if (IsSchConIn(BoxCon2) && IsSchConInternal(BoxCon2)
                                                    && !IsSchBoxReTaged((schbox_list *) BoxCon2->ROOT)
                                                    && loc_test_trans(BoxCon->NET, Column)) {
                                                    BoxBox4 = loc_add_transparence(Figure, BoxCon2->NET);
                                                    BoxBox4->NAME = namealloc("OUT__In_Int_Tag");
                                                    SetSchBoxTaged(BoxBox4);
                                                    BoxCell2 = addscpcel(Column);
                                                    addschboxtoscpcel(BoxBox4, BoxCell2);
                                                    cpt_transparence++;
                                                }
                                            }
                                        }
                                    }
                                }
                            }

                            /* relier un coonecteur d'entree */

                            if (IsSchConIn(Con2) && IsSchConExternal(Con2)
                                && loc_test_trans(Con->NET, Column)) {
                                Box4 = loc_add_transparence(Figure, Con2->NET);
                                SetSchBoxTaged(Box4);
                                Cell2 = addscpcel(Column);
                                addschboxtoscpcel(Box4, Cell2);
                                cpt_transparence++;
                            }
                        }
                    }
                }                /* for */
            }
        }
    }
    Column2 = Column;
    Column = addscpcol();
    cpt = Column2->NUMBER_CELL;

    for (Cell = Column2->LAST_CELL;
         cpt != 0;
         Cell = (scpcel_list *) Cell->PREV) {
        cpt--;
        if (Cell->TYPE == SCP_CEL_BOX) {
            Box = (schbox_list *) Cell->ROOT;
            Con2 = Box->CON_IN;
            for (Chain = Con2->NET->CON_NET;
                 Chain != (chain_list *) 0;
                 Chain = Chain->NEXT) {
                Con = (schcon_list *) Chain->DATA;
                if (IsSchConExternal(Con) && IsSchConIn(Con) && !IsSchConTaged(Con)) {
                    Cell2 = addscpcel(Column);
                    addschcontoscpcel(Con, Cell2);
                    SetSchConTaged(Con);
                }
            }
        }
    }
    return (Figure);
}
/*------------------------------------------------------------\
|                                                             |
|                    Function  Loc Build Box                  |
|                                                             |
\------------------------------------------------------------*/
schfig_list    *
loc_buildbox(Figure)
    schfig_list    *Figure;
{
    schbox_list    *Box;
    schcon_list    *Con;
    long            Y_cont1;
    long            Y_cont2;


    for (Box = Figure->BOX;
         Box != (schbox_list *) 0;
         Box = Box->NEXT) {
        if (IsSchBoxTransparence(Box)) {
            Box->DY = 2;
            Y_cont1 = 1;
            Y_cont2 = 1;
        }
        else if (Box->NUMBER_IN >= Box->NUMBER_OUT) {
            Box->DY = 4 * (Box->NUMBER_IN);
            Y_cont1 = Box->DY - 2;
            Y_cont2 = Box->DY - (Box->NUMBER_IN - Box->NUMBER_OUT) * 2 - 2;
        }
        else {
            Box->DY = 4 * (Box->NUMBER_OUT);
            Y_cont1 = Box->DY - (Box->NUMBER_OUT - Box->NUMBER_IN) * 2 - 2;
            Y_cont2 = Box->DY - 2;
        }

        if (IsSchBoxCluster(Box)) {
            Box->DX = BOX_SIZE_MAX;
        }
        else {
            if (IsSchBoxTransparence(Box)) {
                Box->DX = 0;
            }
            else {
                if (Box->DY >= BOX_SIZE_MAX) {
                    Box->DX = BOX_SIZE_MAX;
                }
                else {
                    if (Box->DY <= BOX_SIZE_MIN) {
                        Box->DX = BOX_SIZE_MIN;
                    }
                    else {
                        Box->DX = Box->DY;
                    }
                }
            }
        }
        for (Con = Box->CON_IN;
             Con != (schcon_list *) 0;
             Con = Con->NEXT) {
            Con->X_REL = -2;
            Con->Y_REL = Y_cont1;
            Y_cont1 = Y_cont1 - 4;
        }
        for (Con = Box->CON_OUT;
             Con != (schcon_list *) 0;
             Con = Con->NEXT) {
            Con->X_REL = Box->DX + 2;
            Con->Y_REL = Y_cont2;
            Y_cont2 = Y_cont2 - 4;
        }
    }
    return (Figure);
}

/*------------------------------------------------------------\
|                                                             |
|                    Function  Loc Clean All                  |
|                                                             |
\------------------------------------------------------------*/
static void 
loc_clean_all(Figure)
    schfig_list    *Figure;
{
    schbox_list    *ScanBox;


    for (ScanBox = Figure->BOX;
         ScanBox != (schbox_list *) 0;
         ScanBox = ScanBox->NEXT) {
        if (IsSchBoxCluster(ScanBox)) {
            loc_destroycluster(Figure, ScanBox);
        }
        if (IsSchBoxTransparence(ScanBox)) {
            loc_destroytransparence(Figure, ScanBox);
        }
    }
    delscpcol();
}

/*------------------------------------------------------------\
|                                                             |
|                         Function  Build                     |
|                                                             |
\------------------------------------------------------------*/
void 
buildschfig(Figure)
    schfig_list    *Figure;
{
# ifdef DEBUG
    fprintf(stdout, "--- Detection des clusters ---\n");
    fflush(stdout);
# endif
    loc_buildcluster(Figure);
# ifdef DEBUG
    fprintf(stdout, "----- Placement  initial -----\n");
    fflush(stdout);
# endif
    loc_placement_initial(Figure);
# ifdef DEBUG
    fprintf(stdout, "--- Fabrication des boites ---\n");
    fflush(stdout);
    fflush(stdout);
# endif
    loc_buildbox(Figure);
# ifdef DEBUG
    fprintf(stdout, "---------  Placement ---------\n");
    fflush(stdout);
    fflush(stdout);
# endif
    loc_buildplace(Figure);
# ifdef DEBUG
    fprintf(stdout, "----------  Routage ----------\n");
    fflush(stdout);
    fflush(stdout);
# endif
    loc_route(Figure);
# ifdef DEBUG
    fprintf(stdout, "---------- Nettoyage  --------\n");
    fflush(stdout);
# endif
    loc_clean_all(Figure);
}
