/*------------------------------------------------------------\
|                                                             |
| Tool    :                     SYF                           |
|                                                             |
| File    :                  syf_comp.c                       |
|                                                             |
| Authors :                  C. Sarwary                       |
|                     Modified by Jacomme Ludovic             |
|                                                             |
| Date    :                   01.11.94                        |
|                                                             |
\------------------------------------------------------------*/

/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include FSM_H
# include FVH_H
# include FTL_H


# include <stdio.h>
# include <stdlib.h>
# include "syf_fsm.h"
# include "syf_error.h"
# include "syf_comp.h"
# include "kis_parse.h"
# include "fbh_parse.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         SyfRemoveStable                     |
|                                                             |
\------------------------------------------------------------*/

static void SyfRemoveStable( Expr )

  ablexpr *Expr;
{
  if ( ! ABL_ATOM( Expr ) )
  {
    if ( ABL_OPER( Expr ) == ABL_STABLE )
    {
      freechain( ABL_CADR( Expr ) );
      freechain( ABL_CDR( Expr ) );
      freechain( ABL_CAR( Expr ) );

      ABL_CDR( Expr ) = (ablexpr *)0;
      ABL_ATOM_VALUE( Expr ) = getablatomzero();

      return;
    }

    while ( ( Expr = ABL_CDR( Expr ) ) != (ablexpr *)0 )
    {
      SyfRemoveStable( ABL_CAR( Expr ) );
    }
  }
}


/*------------------------------------------------------------\
|                                                             |
|                          SyfCompile                         |
|                                                             |
\------------------------------------------------------------*/

fsmfig_list *SyfCompile( FileName, FlagScan, FlagSynopsys )

  char *FileName;
  char  FlagScan;
  char  FlagSynopsys;
{
  fsmfig_list *FsmFigure;
  syfinfo     *SyfInfo;

  if ( FSM_IN == FSM_VHDL_FORMAT )
  {
    FsmFigure = SyfFbhParse( FileName, FlagScan );
  }
  else
  {
    FsmFigure = SyfKissParse( FileName, FlagScan );
  }
 
  if ( FlagSynopsys )
  {
    SyfInfo = FSM_SYF_INFO( FsmFigure );
    SyfRemoveStable( FsmFigure->CLOCK_ABL );
  }

  return( FsmFigure );
}
