/*------------------------------------------------------------\
|                                                             |
| Tool    :                     VASY                          |
|                                                             |
| File    :                 vasy_debug.h                      |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   25.08.97                        |
|                                                             |
\------------------------------------------------------------*/

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/

# ifndef VASY_DEBUG_H
# define VASY_DEBUG_H

# include <time.h>

/*------------------------------------------------------------\
|                                                             |
|                         Debug Mode                          |
|                                                             |
\------------------------------------------------------------*/

# define VASY_DEBUG_LEVEL0     0x01
# define VASY_DEBUG_LEVEL1     0x03
# define VASY_DEBUG_LEVEL2     0x07

/*------------------------------------------------------------\
|                                                             |
|                         Debug Option                        |
|                                                             |
\------------------------------------------------------------*/

# define VASY_DEBUG_NO_RED_INST  0x01
# define VASY_DEBUG_STATISTICS   0x02
# define VASY_DEBUG_DRIVE_STDOUT 0x04

/*------------------------------------------------------------\
|                                                             |
|                           Chrono                            |
|                                                             |
\------------------------------------------------------------*/

# define VASY_DEBUG_MAX_CHRONO   10

/*------------------------------------------------------------\
|                                                             |
|                            Macro                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Save Macro                         |
|                                                             |
\------------------------------------------------------------*/

# define VasyDebugSaveVpnFig( F ) \
                                 \
  (VasyDebugLocSaveVpnFig( (F), __FILE__, __LINE__ ))

/*------------------------------------------------------------\
|                                                             |
|                         Debug Macro                         |
|                                                             |
\------------------------------------------------------------*/

# define IsVasyDebugLevel0()  (VasyDebugSearchItem( __FILE__, VASY_DEBUG_LEVEL0))
# define IsVasyDebugLevel1()  (VasyDebugSearchItem( __FILE__, VASY_DEBUG_LEVEL1))
# define IsVasyDebugLevel2()  (VasyDebugSearchItem( __FILE__, VASY_DEBUG_LEVEL2))

/*------------------------------------------------------------\
|                                                             |
|                         Print Macro                         |
|                                                             |
\------------------------------------------------------------*/

# define VasyPrintf   VasyDebugPrint( __FILE__, __LINE__ ); fprintf 

/*------------------------------------------------------------\
|                                                             |
|                       Debug Option Macro                    |
|                                                             |
\------------------------------------------------------------*/

# define IsVasyDebugNoRedInst()    (VasyDebugOption & VASY_DEBUG_NO_RED_INST)
# define SetVasyDebugNoRedInst()   (VasyDebugOption |= VASY_DEBUG_NO_RED_INST)
# define IsVasyDebugStatistics()   (VasyDebugOption & VASY_DEBUG_STATISTICS)
# define SetVasyDebugStatistics()  (VasyDebugOption |= VASY_DEBUG_STATISTICS)
# define IsVasyDebugDriveStdout()  (VasyDebugOption & VASY_DEBUG_DRIVE_STDOUT)
# define SetVasyDebugDriveStdout() (VasyDebugOption |= VASY_DEBUG_DRIVE_STDOUT)

/*------------------------------------------------------------\
|                                                             |
|                          Time Macro                         |
|                                                             |
\------------------------------------------------------------*/

# define VasyDebugStartChrono(N) (void)time(&VasyDebugChrono[(N)])
# define VasyDebugReadChrono(N)  (time((void *)0) - VasyDebugChrono[(N)])

/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  extern int    VasyDebugFlag;
  extern int    VasyDebugOption;
  extern time_t VasyDebugChrono[ VASY_DEBUG_MAX_CHRONO ];

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/

  extern int  VasyDebugSearchItem();
  extern int  VasyDebugAddItem();

  extern int  VasyDebugSetOption();

  extern void VasyDebugLocSaveVpnFig();
  extern void VasyDebugPrint();

# endif
