/*------------------------------------------------------------\
|                                                             |
| Tool    :                     VASY                          |
|                                                             |
| File    :                 vasy_drvvex.c                     |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   25.08.97                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include VEX_H
# include VPN_H
# include VTL_H
# include RTN_H

# include "vasy_debug.h"
# include "vasy_error.h"
# include "vasy_simprtl.h"
# include "vasy_shared.h"
# include "vasy_drvvex.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      Private variables                      |
|                                                             |
\------------------------------------------------------------*/

  static FILE        *VasyFile         = (FILE        *)0;
  static rtlfig_list *VasyFigure       = (rtlfig_list *)0; 
  static long         VasyDriveLength  = 0;
  static int          VasyVerilog      = 0;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                     VasyDriveVexFile                        |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveVexFile( String, Length )

  char *String;
  int   Length;
{
  VasyDriveLength += Length;

  if ( VasyDriveLength >= VASY_DRIVE_VEX_LENGTH )
  {
    fprintf( VasyFile, "\n" );
    VasyDriveLength = 0;
  }

  if ( *String )
  {
    fwrite( String, sizeof(char), Length, VasyFile );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                     VasyDriveVexAtom                        |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveVexAtom( Expr )

  vexexpr *Expr;
{
  rtldecl_list *RtlDeclar;
  vexexpr      *RtlAtom;
  char         *String;
  char          Buffer[32];
  int           Index;
  int           Length;

  String = GetVexAtomValue( Expr );
  Length = strlen( String );

  if ( ( VasyVerilog               ) &&
       ( IsVexAtomLiteral ( Expr ) ) )
  {
    sprintf( Buffer, "%d'b", Expr->WIDTH );
    VasyDriveVexFile( Buffer, strlen( Buffer ) );

    for ( Index = 1; Index < Length - 1; Index++ )
    {
      if ( ( String[ Index ] == '-' ) ||
           ( String[ Index ] == 'd' ) ) VasyDriveVexFile( "?", 1 );
      else                              VasyDriveVexFile( String + Index, 1 );
    }
  }
  else
  {
    VasyDriveVexFile( String, Length );
  }

  if ( (   IsVexAtomVector( Expr )  ) &&
       ( ! IsVexAtomLiteral( Expr ) ) )
  {
    if ( Expr->LEFT == Expr->RIGHT )
    {
      if ( VasyVerilog ) sprintf( Buffer, "[%d]", Expr->LEFT );
      else               sprintf( Buffer, "(%d)", Expr->LEFT );
    }
    else
    {
      RtlDeclar = searchrtldecl( VasyFigure, String );
      RtlAtom   = RtlDeclar->VEX_ATOM;

      if ( ( Expr->LEFT  == RtlAtom->LEFT   ) &&
           ( Expr->RIGHT == RtlAtom->RIGHT  ) ) return;

      if ( VasyVerilog )
      {
        sprintf( Buffer, "[%d:%d]", Expr->LEFT, Expr->RIGHT );
      }
      else
      if ( Expr->LEFT <  Expr->RIGHT )
      {
        sprintf( Buffer, "(%d to %d)", Expr->LEFT, Expr->RIGHT );
      }
      else
      {
        sprintf( Buffer, "(%d downto %d)", Expr->LEFT, Expr->RIGHT );
      }
    }

    VasyDriveVexFile( Buffer, strlen( Buffer ) );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                    VasyDriveVexExpr                         |
|                                                             |
\------------------------------------------------------------*/

static void loc_VasyDriveVexExpr( Expr )

  vexexpr *Expr;
{
  chain_list *ScanOper;
  long        Oper;
  char       *String;
  int         Length;

  if ( IsVexNodeAtom( Expr ) )
  {
    VasyDriveVexAtom( Expr );
  }
  else
  if ( IsVexNodeOper( Expr ) )
  {
    Oper = GetVexOperValue( Expr );

    if ( VasyVerilog ) String = getvexoperverilogname( Oper );
    else               String = getvexoperuppername( Oper );

    if ( String == (char *)0 )
    {
      VasyError( VASY_ERROR_IN_DRVVEX, Oper );
    }

    Length = strlen( String );

    if ( isvexunaryoper( Oper ) )
    {
      if ( ( Oper == VEX_EVENT  ) ||
           ( Oper == VEX_DRIVER ) )
      {
        VasyDriveVexFile( "(", 1 );

        loc_VasyDriveVexExpr( GetVexOperand( Expr->OPERAND ) );

        VasyDriveVexFile( "'", 1 );
        VasyDriveVexFile( String, Length );
      }
      else
      if ( Oper == VEX_NEG )
      {
        VasyDriveVexFile( "(-", 2 );
        loc_VasyDriveVexExpr( GetVexOperand( Expr->OPERAND ) );
      }
      else
      {
        VasyDriveVexFile( String, Length );
        VasyDriveVexFile( "(", 1 );
        loc_VasyDriveVexExpr( GetVexOperand( Expr->OPERAND ) );
      }
    }
    else
    {
      if ( isvexpositiveoper( Oper ) )
      {
        if ( ( VasyVerilog        ) &&
             ( Oper == VEX_CONCAT ) ) VasyDriveVexFile( "{", 1 );
        else                          VasyDriveVexFile( "(", 1 );
      }
      else
      {
        if ( VasyVerilog )
        {
          VasyDriveVexFile( "~(", 2 );
          String = getvexoperverilogname( getvexnotoper( Oper ) );
        }
        else
        {
          VasyDriveVexFile( "NOT(", 4 );
          String = getvexoperuppername( getvexnotoper( Oper ) );
        }

        Length = strlen( String );
      }

      for ( ScanOper  = Expr->OPERAND;
            ScanOper != (chain_list *)0;
            ScanOper  = ScanOper->NEXT )
      {
        loc_VasyDriveVexExpr( GetVexOperand( ScanOper ) );
        
        if ( ScanOper->NEXT != (chain_list *)0 )
        {
          VasyDriveVexFile( " ", 1 );
          VasyDriveVexFile( String, Length );
          VasyDriveVexFile( " ", 1 );
        }
      }
    }

    if ( ( VasyVerilog        ) &&
         ( Oper == VEX_CONCAT ) ) VasyDriveVexFile( "}", 1 );
    else                          VasyDriveVexFile( ")", 1 );
  }
  else
  {
    String = GetVexFuncValue( Expr );
    Length = strlen( String );

    VasyDriveVexFile( String, Length );
    VasyDriveVexFile( "(", 1 );

    for ( ScanOper  = Expr->OPERAND;
          ScanOper != (chain_list *)0;
          ScanOper  = ScanOper->NEXT )
    {
      loc_VasyDriveVexExpr( GetVexOperand( ScanOper ) );
       
      if ( ScanOper->NEXT != (chain_list *)0 )
      {
        VasyDriveVexFile( ",", 1 );
      }
    }

    VasyDriveVexFile( ")", 1 );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                    VasyDriveVexExpr                         |
|                                                             |
\------------------------------------------------------------*/

void VasyDriveVexExpr( File, RtlFigure, Expr, Verilog )

  FILE        *File;
  rtlfig_list *RtlFigure;
  vexexpr     *Expr;
  int          Verilog;
{
  if ( Expr != (vexexpr *)0 )
  {
    VasyFigure      = RtlFigure;
    VasyFile        = File;
    VasyDriveLength = 0;
    VasyVerilog     = Verilog;

    loc_VasyDriveVexExpr( Expr );
  }
}
