/*------------------------------------------------------------\
|                                                             |
| Tool    :                    VASY                           |
|                                                             |
| File    :                 vasy_shared.c                     |
|                                                             |
| Author  :                 Jacomme Ludovic                   |
|                                                             |
| Date    :                   25.08.97                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include VEX_H
# include VPN_H
# include VTL_H

# include "vasy_error.h"
# include "vasy_shared.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  char *VASY_SYM_TYPE[ VASY_MAX_SYM_TYPE ] =
  {
    "UNKNOWN",
    "UNUSED",
    "CONSTANT",
    "COMBINATORIAL",
    "TRISTATE",
    "REGISTER"
  };

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      VasyAddVpnProcInfo                     |
|                                                             |
\------------------------------------------------------------*/

vasyprocinfo *VasyAddVpnProcInfo( VpnProc )

   vpnproc_list *VpnProc;
{
  vasyprocinfo *ProcInfo;

  ProcInfo = (vasyprocinfo *)autallocheap( sizeof( vasyprocinfo ) );

  SetVasyVpnProcInfo( VpnProc, ProcInfo );

  return( ProcInfo );
}

/*------------------------------------------------------------\
|                                                             |
|                      VasyDelVpnProcInfo                     |
|                                                             |
\------------------------------------------------------------*/

void VasyDelVpnProcInfo( VpnProc )

   vpnproc_list *VpnProc;
{
  vasyprocinfo   *ProcInfo;
  vasybivex_list *ProcBiVex;
  vasybivex_list *DelProcBiVex;

  ProcInfo = GetVasyVpnProcInfo( VpnProc );

  ProcBiVex = ProcInfo->BIVEX;

  while ( ProcBiVex != (vasybivex_list *)0 )
  {
    freevexexpr( ProcBiVex->VEX_COND );
    freevexexpr( ProcBiVex->VEX_DATA );

    DelProcBiVex = ProcBiVex;
    ProcBiVex    = ProcBiVex->NEXT;

    autfreeheap( DelProcBiVex, sizeof( vasybivex_list ) );
  }

  autfreeheap( ProcInfo, sizeof( vasyprocinfo ) );

  SetVasyVpnProcInfo( VpnProc, (vasyprocinfo *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                      VasyAddVpnActInfo                      |
|                                                             |
\------------------------------------------------------------*/

vasyactinfo *VasyAddVpnActInfo( VpnAction )

   vpnact_list *VpnAction;
{
  vasyactinfo    *ActInfo;
  vexexpr       *AsgAtom;
  chain_list   **Support;
  chain_list   **Literal;
  chain_list   **Event;
  vasysyminfo    *SymInfo;
  unsigned char *Flags;

  AsgAtom = VpnAction->VEX_ATOM;

  ActInfo = (vasyactinfo    *)autallocheap( sizeof( vasyactinfo ) );

  Support = (chain_list   **)autallocblock( sizeof( chain_list *) * AsgAtom->WIDTH );
  Literal = (chain_list   **)autallocblock( sizeof( chain_list *) * AsgAtom->WIDTH );
  Event   = (chain_list   **)autallocblock( sizeof( chain_list *) * AsgAtom->WIDTH );
  Flags   = (unsigned char *)autallocblock( sizeof( char        ) * AsgAtom->WIDTH );
  SymInfo = (vasysyminfo    *)autallocblock( sizeof( vasysyminfo  ) * AsgAtom->WIDTH );

  ActInfo->SUPPORT  = Support;
  ActInfo->LITERAL  = Literal;
  ActInfo->EVENT    = Event;
  ActInfo->FLAGS    = Flags;
  ActInfo->SYM_INFO = SymInfo;

  SetVasyVpnActInfo( VpnAction, ActInfo );

  return( ActInfo );
}

/*------------------------------------------------------------\
|                                                             |
|                      VasyDelVpnActInfo                      |
|                                                             |
\------------------------------------------------------------*/

void VasyDelVpnActInfo( VpnAction )

   vpnact_list *VpnAction;
{
  vasyactinfo     *ActInfo;
  vexexpr        *AsgAtom;
  chain_list    **Support;
  chain_list    **Literal;
  chain_list    **Event;
  unsigned char  *Flags;
  vasysyminfo     *SymInfo;
  vasybivex_list *SymBiVex;
  vasybivex_list *DelSymBiVex;
  int             Index;

  ActInfo = GetVasyVpnActInfo( VpnAction );

  AsgAtom = VpnAction->VEX_ATOM;
  Support = ActInfo->SUPPORT;
  Literal = ActInfo->LITERAL;
  Event   = ActInfo->EVENT;
  Flags   = ActInfo->FLAGS;
  SymInfo = ActInfo->SYM_INFO;

  for ( Index = 0; Index < AsgAtom->WIDTH; Index++ )
  {
    freechain( Support[ Index ] );
    freechain( Literal[ Index ] );

    SymBiVex = SymInfo[ Index ].BIVEX;

    while ( SymBiVex != (vasybivex_list *)0 )
    {
      freevexexpr( SymBiVex->VEX_COND );
      freevexexpr( SymBiVex->VEX_DATA );

      DelSymBiVex = SymBiVex;
      SymBiVex    = SymBiVex->NEXT;

      autfreeheap( DelSymBiVex, sizeof( vasybivex_list ) );
    }

    freevexexpr( SymInfo[ Index ].VEX_DATA );
  }

  autfreeblock( Support );
  autfreeblock( Literal );
  autfreeblock( Event   );
  autfreeblock( Flags   );
  autfreeblock( SymInfo );

  autfreeheap( ActInfo, sizeof( vasyactinfo ) );

  SetVasyVpnActInfo( VpnAction, (vasyactinfo *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                    VasyAddVpnWaitInfo                       |
|                                                             |
\------------------------------------------------------------*/

vasywaitinfo *VasyAddVpnWaitInfo( VpnTrans )

   vpntrans_list *VpnTrans;
{
  vasywaitinfo  *WaitInfo;

  WaitInfo = (vasywaitinfo *)autallocheap( sizeof( vasywaitinfo ) );
  SetVasyVpnWaitInfo( VpnTrans, WaitInfo );

  return( WaitInfo );
}

/*------------------------------------------------------------\
|                                                             |
|                     VasyDelVpnWaitInfo                      |
|                                                             |
\------------------------------------------------------------*/

void VasyDelVpnWaitInfo( VpnTrans )

   vpntrans_list *VpnTrans;
{
  vasywaitinfo  *WaitInfo;

  WaitInfo = GetVasyVpnWaitInfo( VpnTrans );
  autfreeheap( WaitInfo, sizeof( vasywaitinfo ) );

  SetVasyVpnWaitInfo( VpnTrans, (vasywaitinfo *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                    VasyAddVpnTransInfo                      |
|                                                             |
\------------------------------------------------------------*/

vasytransinfo *VasyAddVpnTransInfo( VpnTrans )

   vpntrans_list *VpnTrans;
{
  vasytransinfo  *TransInfo;

  TransInfo = (vasytransinfo *)autallocheap( sizeof( vasytransinfo ) );

  TransInfo->HASH_ASSIGN  = createauthtable( 100 );
  TransInfo->HASH_BITVEC  = createauthtable( 100 );
  TransInfo->HASH_SUPPORT = createauth2table( 100 );
  TransInfo->HASH_LITERAL = createauth2table( 100 );
  TransInfo->HASH_EVENT   = createauth2table( 100 );

  SetVasyVpnTransInfo( VpnTrans, TransInfo );

  return( TransInfo );
}

/*------------------------------------------------------------\
|                                                             |
|                     VasyDelVpnTransInfo                     |
|                                                             |
\------------------------------------------------------------*/

void VasyDelVpnTransInfo( VpnTrans )

   vpntrans_list *VpnTrans;
{
  vasytransinfo  *TransInfo;

  TransInfo = GetVasyVpnTransInfo( VpnTrans );

  destroyauthtable( TransInfo->HASH_ASSIGN  );
  destroyauthtable( TransInfo->HASH_BITVEC  );
  destroyauth2table( TransInfo->HASH_SUPPORT );
  destroyauth2table( TransInfo->HASH_LITERAL );
  destroyauth2table( TransInfo->HASH_EVENT   );

  autfreeheap( TransInfo, sizeof( vasytransinfo ) );

  SetVasyVpnTransInfo( VpnTrans, (vasytransinfo *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                    VasyAddVpnSymBiVex                       |
|                                                             |
\------------------------------------------------------------*/

vasybivex_list *VasyAddVpnSymBiVex( SymInfo, VexCond, VexData, Type )

  vasysyminfo *SymInfo;
  vexexpr    *VexCond;
  vexexpr    *VexData;
  int         Type;
{
  vasybivex_list *SymBiVex;

  SymBiVex = (vasybivex_list *)autallocheap( sizeof( vasybivex_list ) );

  SymBiVex->VEX_COND = VexCond;
  SymBiVex->VEX_DATA = VexData;
  SymBiVex->TYPE     = Type;

  SymBiVex->NEXT = SymInfo->BIVEX;
  SymInfo->BIVEX = SymBiVex;

  return( SymBiVex );
}

/*------------------------------------------------------------\
|                                                             |
|                    VasyAddVpnProcBiVex                      |
|                                                             |
\------------------------------------------------------------*/

vasybivex_list *VasyAddVpnProcBiVex( ProcInfo, VexCond, VexData, Type )

  vasyprocinfo *ProcInfo;
  vexexpr     *VexCond;
  vexexpr     *VexData;
  int          Type;
{
  vasybivex_list *SymBiVex;

  SymBiVex = (vasybivex_list *)autallocheap( sizeof( vasybivex_list ) );

  SymBiVex->VEX_COND = VexCond;
  SymBiVex->VEX_DATA = VexData;
  SymBiVex->TYPE     = Type;

  SymBiVex->NEXT  = ProcInfo->BIVEX;
  ProcInfo->BIVEX = SymBiVex;

  return( SymBiVex );
}
