/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Vbh                           |
|                                                             |
| File    :                   vbhview.c                       |
|                                                             |
| Date    :                   03.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include VEX_H
# include VBH_H

# include <stdio.h>
# include <string.h>

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        View  Functions                      |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                    Local Get Vbh Type                       |
|                                                             |
\------------------------------------------------------------*/

static char *loc_getvbhtypename( Type )

    vbtyp_list *Type;
{
  static char NotDefined[] = "not defined !";

  if ( Type == (vbtyp_list *)0 )
  {
    return( NotDefined );
  }

  return( Type->NAME );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Ifs                       |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbifs( ScanIfs  )

  vbifs_list  *ScanIfs;
{
  fprintf( stdout, "[%ld] ", ScanIfs->LINE );
  fprintf( stdout, "IF (" );
  viewvexexprbound( ScanIfs->CND );
  fprintf( stdout, ")\nTHEN\n" );

  if ( ScanIfs->CNDTRUE != (ptype_list *)0 )
  {
    vbh_viewvbinst( ScanIfs->CNDTRUE );
  }

  if ( ScanIfs->CNDFALSE != (ptype_list *)0 )
  {
    fprintf( stdout, "ELSE\n" );
    vbh_viewvbinst( ScanIfs->CNDFALSE );
  }
  fprintf( stdout, "END IF\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Var                       |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbvar( ScanVar  )

  vbvar_list  *ScanVar;
{
  fprintf( stdout, "[%ld] ", ScanVar->LINE );
  viewvexexprbound( ScanVar->TARGET );
  fprintf( stdout, " TYPE %s ", loc_getvbhtypename( ScanVar->TYPE ) );
  fprintf( stdout, " := " );
  viewvexexprbound( ScanVar->VEX );
  fprintf( stdout, "\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Asg                       |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbasg( ScanAsg  )

  vbasg_list  *ScanAsg;
{
  fprintf( stdout, "[%ld] ", ScanAsg->LINE );
  viewvexexprbound( ScanAsg->TARGET );
  fprintf( stdout, " TYPE %s ", loc_getvbhtypename( ScanAsg->TYPE ) );
  fprintf( stdout, " <= " );
  viewvexexprbound( ScanAsg->VEX );
  fprintf( stdout, "\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Case                      |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbcas( ScanCase  )

  vbcas_list  *ScanCase;
{
  vbcho_list    *ScanChoice;
  int            Offset;

  fprintf( stdout, "[%ld] ", ScanCase->LINE );
  fprintf( stdout, "CASE (" );
  viewvexexprbound( ScanCase->VEX );
  fprintf( stdout, ") IS\n" );

  for ( Offset = 0; Offset < ScanCase->SIZE; Offset++ )
  {
    ScanChoice = &ScanCase->CHOICE[ Offset ];
    fprintf( stdout, "WHEN %s => ", ScanChoice->VALUE );

    vbh_viewvbinst( ScanChoice->INSTRUCTION );
  }

  fprintf( stdout, "END CASE\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Wait                      |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbwas( ScanWait  )

  vbwas_list  *ScanWait;
{
  chain_list *ScanSens;

  fprintf( stdout, "[%ld] ", ScanWait->LINE );
  fprintf( stdout, "WAIT" );

  if ( ScanWait->SEN != (chain_list *)0 )
  {
    fprintf( stdout, " ON " );
    for ( ScanSens  = ScanWait->SEN;
          ScanSens != (chain_list *)0;
          ScanSens  = ScanSens->NEXT )
    {
      fprintf( stdout, "%s", (char *)ScanSens->DATA );
      if ( ScanSens->NEXT != (chain_list *)0 )
      {
        fprintf( stdout, ", " );
      }
    }
  }

  if ( ScanWait->CND != (vexexpr *)0 )
  {
    fprintf( stdout, " UNTIL " );
    viewvexexprbound( ScanWait->CND );
  }
  fprintf( stdout, "\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh For                       |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbfor( ScanFor )

  vbfor_list  *ScanFor;
{
  fprintf( stdout, "[%ld] ", ScanFor->LINE );
  fprintf( stdout, "%s : ", ScanFor->LABEL );

  fprintf( stdout, "FOR " );
  viewvexexprbound( ScanFor->VARIABLE );
  fprintf( stdout, "IN " );

  viewvexexprbound( ScanFor->LEFT );
  if ( ScanFor->UP ) fprintf( stdout, " TO "     );
  else               fprintf( stdout, " DOWNTO " );
  viewvexexprbound( ScanFor->RIGHT );

  fprintf( stdout, "\nLOOP\n" );
  
  vbh_viewvbinst( ScanFor->INSTRUCTION );

  fprintf( stdout, "END LOOP %s\n", ScanFor->LABEL );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Loop                      |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvblop( ScanLoop )

  vblop_list  *ScanLoop;
{
  fprintf( stdout, "[%ld] ", ScanLoop->LINE );
  fprintf( stdout, "%s : ", ScanLoop->LABEL );

  fprintf( stdout, "LOOP\n" );
  
  vbh_viewvbinst( ScanLoop->INSTRUCTION );

  fprintf( stdout, "END LOOP %s\n", ScanLoop->LABEL );
  fprintf( stdout, "\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh While                     |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbwhi( ScanWhile )

  vbwhi_list  *ScanWhile;
{
  fprintf( stdout, "[%ld] ", ScanWhile->LINE );
  fprintf( stdout, "%s : ", ScanWhile->LABEL );

  fprintf( stdout, "WHILE ( " );
  viewvexexprbound( ScanWhile->CND );
  fprintf( stdout, " )\n" );
  fprintf( stdout, "LOOP\n" );
  
  vbh_viewvbinst( ScanWhile->INSTRUCTION );

  fprintf( stdout, "END LOOP %s\n", ScanWhile->LABEL );
  fprintf( stdout, "\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Aggregate                 |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbagr( ScanAggreg )

  vbagr_list  *ScanAggreg;
{
  fprintf( stdout, "[%ld] ", ScanAggreg->LINE );
  fprintf( stdout, "AGGREGATE (\n" );
  vbh_viewvbinst( ScanAggreg->ASSIGN );
  fprintf( stdout, ")\n\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Exit                      |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbext( ScanExit  )

  vbext_list  *ScanExit;
{
  fprintf( stdout, "[%ld] ", ScanExit->LINE );
  fprintf( stdout, "EXIT" );

  if ( ScanExit->LABEL != (char *)0 )
  {
    fprintf( stdout, " %s", ScanExit->LABEL );
  }

  if ( ScanExit->CND != (vexexpr *)0 )
  {
    fprintf( stdout," WHEN ( " );
    viewvexexprbound( ScanExit->CND );
    fprintf( stdout, " )\n" );
  }
  else
  {
    fprintf( stdout, "\n" );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Next                      |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbnxt( ScanNext  )

  vbnxt_list  *ScanNext;
{
  fprintf( stdout, "[%ld] ", ScanNext->LINE );
  fprintf( stdout, "NEXT" );

  if ( ScanNext->LABEL != (char *)0 )
  {
    fprintf( stdout, " %s", ScanNext->LABEL );
  }

  if ( ScanNext->CND != (vexexpr *)0 )
  {
    fprintf( stdout," WHEN ( " );
    viewvexexprbound( ScanNext->CND );
    fprintf( stdout, " )\n" );
  }
  else
  {
    fprintf( stdout, "\n" );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Return                    |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbret( ScanReturn  )

  vbret_list  *ScanReturn;
{
  fprintf( stdout, "[%ld] ", ScanReturn->LINE );
  fprintf( stdout, "RETURN " );

  viewvexexprbound( ScanReturn->RET );

  fprintf( stdout, "\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Call                      |
|                                                             |
\------------------------------------------------------------*/

static void loc_viewvbcal( ScanCall  )

  vbcal_list  *ScanCall;
{
  fprintf( stdout, "[%ld] ", ScanCall->LINE );
  fprintf( stdout, "CALL " );

  viewvexexprbound( ScanCall->CALL );

  fprintf( stdout, "\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                    Local View Vbh Instruction               |
|                                                             |
\------------------------------------------------------------*/

void vbh_viewvbinst( Instruction )

  ptype_list *Instruction;
{
  while ( Instruction != (ptype_list *)0 )
  {
    switch ( Instruction->TYPE )
    {
      case VBH_BEIFS : loc_viewvbifs( Instruction->DATA );
      break;

      case VBH_BEASG : loc_viewvbasg( Instruction->DATA );
      break;

      case VBH_BEVAR : loc_viewvbvar( Instruction->DATA );
      break;

      case VBH_BECAS : loc_viewvbcas( Instruction->DATA );
      break;

      case VBH_BEWAS : loc_viewvbwas( Instruction->DATA );
      break;

      case VBH_BEFOR : loc_viewvbfor( Instruction->DATA );
      break;

      case VBH_BEWHI : loc_viewvbwhi( Instruction->DATA );
      break;

      case VBH_BELOP : loc_viewvblop( Instruction->DATA );
      break;

      case VBH_BENXT : loc_viewvbnxt( Instruction->DATA );
      break;

      case VBH_BEEXT : loc_viewvbext( Instruction->DATA );
      break;

      case VBH_BERET : loc_viewvbret( Instruction->DATA );
      break;

      case VBH_BECAL : loc_viewvbcal( Instruction->DATA );
      break;

      case VBH_BEAGR : loc_viewvbagr( Instruction->DATA );
      break;
    }

    Instruction = Instruction->NEXT;
  }
}

/*------------------------------------------------------------\
|                                                             |
|                         View Vbh Type                       |
|                                                             |
\------------------------------------------------------------*/

void vbh_viewvbtyp( BeTyp )

  vbtyp_list *BeTyp;
{
  int Index;

  fprintf( stdout, " > [%ld]\n", BeTyp->LINE );
  fprintf( stdout, "   NAME   : %s\n",      BeTyp->NAME  ); 
  fprintf( stdout, "   LEFT   : %d\n", (int)BeTyp->LEFT  ); 
  fprintf( stdout, "   RIGHT  : %d\n", (int)BeTyp->RIGHT ); 
  fprintf( stdout, "   VALUE  : \n" );

  for ( Index = 0; Index < BeTyp->SIZE; Index++ )
  {
    fprintf( stdout, "%s ", BeTyp->VALUE[ Index ] );
  }

  if ( BeTyp->SIZE != 0 ) fprintf( stdout, "\n" );

  fprintf( stdout, "   SIZE   : %d\n", BeTyp->SIZE   );
  fprintf( stdout, "   BYTE   : %d\n", BeTyp->BYTE   );
  fprintf( stdout, "   CLASS  : %c\n", BeTyp->CLASS  );
  fprintf( stdout, "   INDEX  : %d\n", BeTyp->INDEX  );

  if ( BeTyp->BASE != (vbtyp_list *)0 )
  {
    fprintf( stdout, "   BASE   : %s\n", BeTyp->BASE->NAME );
  }
  else
  {
    fprintf( stdout, "   BASE   : none\n" );
  }

  if ( BeTyp->RESOLV != (vbfun_list *)0 )
  {
    fprintf( stdout, "   RESOLV : %s\n", BeTyp->RESOLV->NAME );
  }
  else
  {
    fprintf( stdout, "   RESOLV : none\n" );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                         View Vbh Figure                     |
|                                                             |
\------------------------------------------------------------*/

void vbh_viewvbfig( Figure )

  vbfig_list *Figure;
{
  vbaux_list *BeAux;
  vbatr_list *BeAtr;
  vbcst_list *BeCst;
  vbpor_list *BePor;
  vbpcs_list *BePcs;
  vbfun_list *BeFun;
  vbarg_list *BePar;
  vbvar_list *BeVar;
  vbtyp_list *BeTyp;
  ptype_list *ScanPtype;

  fprintf( stdout, "NAME: %s\n", Figure->NAME );
  fprintf( stdout, "FLAG: %d\n", Figure->FLAG );

  fprintf( stdout, "BETYP\n" );

  for ( BeTyp  = Figure->BETYP;
        BeTyp != (vbtyp_list *)0;
        BeTyp  = BeTyp->NEXT )
  {
    vbh_viewvbtyp( BeTyp );
  }

  fprintf( stdout, "BEPOR\n" );

  for ( BePor  = Figure->BEPOR;
        BePor != (vbpor_list *)0;
        BePor  = BePor->NEXT )
  {
    fprintf( stdout, "  > [%ld]", BePor->LINE );
    viewvexexprbound( BePor->TARGET );
    fprintf( stdout, " TYPE %s ", loc_getvbhtypename( BePor->TYPE ) );
    fprintf( stdout, " DIR  %c ", BePor->DIR );
    fprintf( stdout, " KIND %d ", BePor->KIND );
    if ( BePor->VEX != (vexexpr *)0 )
    {
      fprintf( stdout, " := " );
      viewvexexprbound( BePor->VEX );
    }
    fprintf( stdout, "\n" );
  }

  fprintf( stdout, "BEATR\n" );

  for ( BeAtr  = Figure->BEATR;
        BeAtr != (vbatr_list *)0;
        BeAtr  = BeAtr->NEXT )
  {
    fprintf( stdout, "  > [%ld]", BeAtr->LINE );
    fprintf( stdout, " NAME   %s ", BeAtr->NAME   );
    fprintf( stdout, " ENTITY %s ", BeAtr->ENTITY );
    fprintf( stdout, " TYPE   %d ", BeAtr->TYPE   );
    fprintf( stdout, " VALUE  %s ", BeAtr->VALUE  );

    fprintf( stdout, "\n" );
  }

  fprintf( stdout, "BEAUX\n" );

  for ( BeAux  = Figure->BEAUX;
        BeAux != (vbaux_list *)0;
        BeAux  = BeAux->NEXT )
  {
    fprintf( stdout, "  > [%ld]", BeAux->LINE );
    viewvexexprbound( BeAux->TARGET );
    fprintf( stdout, " TYPE %s ", loc_getvbhtypename( BeAux->TYPE ) );
    fprintf( stdout, " KIND %d ", (int)BeAux->KIND );
    if ( BeAux->VEX != (vexexpr *)0 )
    {
      fprintf( stdout, " := " );
      viewvexexprbound( BeAux->VEX );
    }
    fprintf( stdout, "\n" );
  }

  fprintf( stdout, "BECST\n" );

  for ( BeCst  = Figure->BECST;
        BeCst != (vbcst_list *)0;
        BeCst  = BeCst->NEXT )
  {
    fprintf( stdout, "  > [%ld]", BeCst->LINE );
    viewvexexprbound( BeCst->TARGET );
    fprintf( stdout, " TYPE %s ", loc_getvbhtypename( BeCst->TYPE ) );
    if ( BeCst->VEX != (vexexpr *)0 )
    {
      fprintf( stdout, " := " );
      viewvexexprbound( BeCst->VEX );
    }
    fprintf( stdout, "\n" );
  }

  fprintf( stdout, "BEFUN\n" );
 
  for ( BeFun  = Figure->BEFUN;
        BeFun != (vbfun_list *)0;
        BeFun  = BeFun->NEXT )
  {
    fprintf( stdout, "  > [%ld]\n", BeFun->LINE );
    fprintf( stdout, "    NAME : %s\n", BeFun->NAME );
    fprintf( stdout, "    TYPE : %x\n", (int)BeFun->TYPE );

    fprintf( stdout, "ARGUMENT\n" );

    for ( BePar  = BeFun->ARGUMENT;
          BePar != (vbarg_list *)0;
          BePar  = BePar->NEXT )
    {
      fprintf( stdout, "  > [%ld]\n", BePar->LINE );
      viewvexexprbound( BePar->TARGET );
      fprintf( stdout, " TYPE  %s ", loc_getvbhtypename( BePar->TYPE ) );
      fprintf( stdout, " DIR   %c ", BePar->DIR );
      fprintf( stdout, " CLASS %d ", BePar->CLASS );
      fprintf( stdout, "\n" );
    }

    if ( BeFun->RETURN != (vbarg_list *)0 )
    {
      BePar = BeFun->RETURN;

      fprintf( stdout, "  > [%ld]\n", BePar->LINE );
      fprintf( stdout, "RETURN\n" );
     
      fprintf( stdout, "  < " );
      viewvexexprbound( BePar->TARGET );
      fprintf( stdout, " TYPE %s ", loc_getvbhtypename( BePar->TYPE ) );
      fprintf( stdout, "\n" );
    }

    fprintf( stdout, "VARIABLE\n" );

    for ( ScanPtype  = BeFun->VARIABLE;
          ScanPtype != (ptype_list *)0;
          ScanPtype  = ScanPtype->NEXT )
    {
      BeVar = (vbvar_list *)ScanPtype->DATA;

      fprintf( stdout, "  > [%ld]\n", BeVar->LINE );
      fprintf( stdout, " TYPE %s ", loc_getvbhtypename( BeVar->TYPE ) );
      viewvexexprbound( BeVar->TARGET );
      if ( BeVar->VEX != (vexexpr *)0 )
      {
        fprintf( stdout, " := " );
        viewvexexprbound( BeVar->VEX );
      }
      fprintf( stdout, "\n" );
    }

    fprintf( stdout, "INSTRUCTION\n" );

    vbh_viewvbinst( BeFun->INSTRUCTION );
  }

  fprintf( stdout, "BEPCS\n" );
 
  for ( BePcs  = Figure->BEPCS;
        BePcs != (vbpcs_list *)0;
        BePcs  = BePcs->NEXT )
  {
    fprintf( stdout, "  > [%ld]\n", BePcs->LINE );

    if ( BePcs->LABEL != (char *)0 )
    {
      fprintf( stdout, "LABEL : %s\n", BePcs->LABEL );
    }

    fprintf( stdout, "TYPE  : %x\n", (int)BePcs->TYPE );

    fprintf( stdout, "VARIABLE\n" );

    for ( ScanPtype  = BePcs->VARIABLE;
          ScanPtype != (ptype_list *)0;
          ScanPtype  = ScanPtype->NEXT )
    {
      BeVar = (vbvar_list *)ScanPtype->DATA;

      fprintf( stdout, "  > [%ld]\n", BeVar->LINE );
      fprintf( stdout, " TYPE %s ", loc_getvbhtypename( BeVar->TYPE ) );
      viewvexexprbound( BeVar->TARGET );
      if ( BeVar->VEX != (vexexpr *)0 )
      {
        fprintf( stdout, " := " );
        viewvexexprbound( BeVar->VEX );
      }
      fprintf( stdout, "\n" );
    }

    fprintf( stdout, "INSTRUCTION\n" );

    vbh_viewvbinst( BePcs->INSTRUCTION );
  }
}
