/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Vex                           |
|                                                             |
| File    :                  vexcreate.c                      |
|                                                             |
| Date    :                   03.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include VEX_H

# include <stdio.h>
# include "vexcreate.h"
# include "vexerror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Varivexes                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        Create Functions                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        Create Vex Atom Vector               |
|                                                             |
\------------------------------------------------------------*/

vexexpr *createvexatomvec( Name, Left, Right )

  char *Name;
  short Left;
  short Right;
{
  vexexpr  *Node;
  short     Width;

  Name = namealloc( Name );
  Node = allocvexnode();

  SetVexNodeAtom( Node );
  SetVexAtomValue( Node, Name );
  SetVexAtomVector( Node );

  if ( ( Left  < 0 ) ||
       ( Right < 0 ) )
  {
    Node->WIDTH = 1;
    SetVexNodeVarWidth( Node );
  }
  else
  {
    if ( Left < Right )
    {
      Width = 1 + Right - Left;
    }
    else
    {
      Width = 1 + Left  - Right;
      SetVexAtomDown( Node );
    }

    Node->LEFT  = Left;
    Node->RIGHT = Right;
    Node->WIDTH = Width;
  }

  if ( Name[ 0 ] == '"' )
  {
    SetVexAtomLiteral( Node );
  }

  return( Node );
}

/*------------------------------------------------------------\
|                                                             |
|                        Create Vex Atom Literal              |
|                                                             |
\------------------------------------------------------------*/

vexexpr *createvexatomveclit( Literal, Width  )

  unsigned char Literal;
  short         Width;
{
  vexexpr *Node;
  char    *Name;
  char     Buffer[ 512 ];
  char     Cote;

  if ( Width > 1 ) Cote = '\"';
  else             Cote = '\'';


 Buffer[ 0         ] = Cote;
 Buffer[ Width + 1 ] = Cote;
 Buffer[ Width + 2 ] = '\0';

  memset( Buffer + 1, Literal, Width );

  Name = namealloc( Buffer );
  Node = allocvexnode();

  SetVexNodeAtom( Node );
  SetVexAtomValue( Node, Name );
  SetVexAtomLiteral( Node );

  Node->LEFT  = Width - 1;
  Node->RIGHT = 0;
  Node->WIDTH = Width;

  if ( Width > 1 ) SetVexAtomVector( Node );
  else             SetVexAtomBit( Node );

  return( Node );
}

/*------------------------------------------------------------\
|                                                             |
|                        Create Vex Atom Bit                  |
|                                                             |
\------------------------------------------------------------*/

vexexpr *createvexatombit( Name )

  char *Name;
{
  vexexpr *Node;

  Name = namealloc( Name );
  Node = allocvexnode();

  SetVexNodeAtom( Node );
  SetVexAtomValue( Node, Name );
  SetVexAtomBit( Node );

  Node->WIDTH = 1;

  if ( Name[ 0 ] == '\'' )
  {
    SetVexAtomLiteral( Node );
  }

  return( Node );
}

/*------------------------------------------------------------\
|                                                             |
|                      Create Vex Oper                        |
|                                                             |
\------------------------------------------------------------*/

vexexpr *createvexoper( Oper, Width )

  long  Oper;
  short Width;
{
  vexexpr *Node;

  if ( ( Oper < 0                 ) ||
       ( Oper >= VEX_MAX_OPERATOR ) )
  {
    vexerror( VEX_OPERATOR_ERROR, Oper );
  }

  Node = allocvexnode();

  if ( Width <= 0 )
  {
    Width = 1;
    SetVexNodeVarWidth( Node );
  }

  Node->WIDTH = Width;

  SetVexNodeOper( Node );
  SetVexOperValue( Node, Oper );

  return( Node );
}

/*------------------------------------------------------------\
|                                                             |
|                        Create Vex Bin Expr                  |
|                                                             |
\------------------------------------------------------------*/

vexexpr *createvexbinexpr( Oper, Width, Expr1, Expr2 )

  long      Oper;
  short     Width;
  vexexpr  *Expr1;
  vexexpr  *Expr2;
{
  vexexpr *Expr;

  if ( ! isvexbinaryoper( Oper ) )
  {
    vexerror( VEX_OPERATOR_ERROR, Oper );
  }

  Expr = createvexoper( Oper, Width );

  addvexhexpr( Expr, Expr2 );
  addvexhexpr( Expr, Expr1 );

  return( Expr );
}

/*------------------------------------------------------------\
|                                                             |
|                      Create Vex Unary Expr                  |
|                                                             |
\------------------------------------------------------------*/

vexexpr *createvexunaryexpr( Oper, Width, Expr )

  long      Oper;
  short     Width;
  vexexpr  *Expr;
{
  vexexpr *Expr1;

  if ( ! isvexunaryoper( Oper ) )
  {
    vexerror( VEX_NOT_UNARY_ERROR, Oper );
  }

  Expr1 = createvexoper( Oper, Width );
  addvexhexpr( Expr1, Expr );

  return( Expr1 );
}

/*------------------------------------------------------------\
|                                                             |
|                   Create Vex Ternary Expr                   |
|                                                             |
\------------------------------------------------------------*/

vexexpr *createvexternaryexpr( Oper, Width, Expr1, Expr2, Expr3 )

  long      Oper;
  short     Width;
  vexexpr  *Expr1;
  vexexpr  *Expr2;
  vexexpr  *Expr3;
{
  vexexpr *Expr;

  if ( ! isvexternaryoper( Oper ) )
  {
    vexerror( VEX_OPERATOR_ERROR, Oper );
  }

  Expr = createvexoper( Oper, Width );

  addvexhexpr( Expr, Expr3 );
  addvexhexpr( Expr, Expr2 );
  addvexhexpr( Expr, Expr1 );

  return( Expr );
}

/*------------------------------------------------------------\
|                                                             |
|                      Create Vex Func                        |
|                                                             |
\------------------------------------------------------------*/

vexexpr *createvexfunc( Func, Width )

  char  *Func;
  short  Width;
{
  vexexpr *Node;

  Node = allocvexnode();

  if ( Width <= 0 )
  {
    Width = 1;
    SetVexNodeVarWidth( Node );
  }

  Node->WIDTH = Width;

  SetVexNodeFunc( Node );
  SetVexFuncValue( Node, Func );

  return( Node );
}

/*------------------------------------------------------------\
|                                                             |
|                   Create Vex Atom Long Int                  |
|                                                             |
\------------------------------------------------------------*/

vexexpr *createvexatomlong( Integer, Width, Signed )

  long  Integer;
  short Width;
  short Signed;
{
  vexexpr       *Node;
  char           Buffer[ 40 ];
  short          Index;
  unsigned long  Mask;

  Node = allocvexnode();

  if ( Width == 0 )
  {
    Width = 32;
    Mask  = 1L << 31;

    if ( ( Integer & Mask ) &&
         ( Signed         ) )
    {
      while ( ( Integer & Mask ) != 0 )
      {
        Width--; Mask = Mask >> 1;
      }

      if ( Mask == 0 ) Mask = 0x1;
      else             Mask = Mask << 1;

      Width++;
    }
    else
    if ( Integer != 0 )
    {
      while ( ( Integer & Mask ) == 0 )
      {
        Width--; Mask = Mask >> 1;
      }

      if ( Signed )
      { 
        Mask = Mask << 1; Width++;
      }
    }
    else
    {
      Width = 1; Mask = 0x1;
    }
  }
  else
  {
    Mask = 1L << (Width - 1);
  }

  Node->WIDTH = Width;
  Node->LEFT  = Width - 1;
  Node->RIGHT = 0;

  Index = 1;

  do
  {
    if ( Integer & Mask ) Buffer[ Index ] = '1';
    else                  Buffer[ Index ] = '0';

    Index = Index + 1;
    Mask  = Mask >> 1;
  }
  while ( Mask != 0 );

  if ( Width == 1 )
  {
    Buffer[ 0       ] = '\'';
    Buffer[ Index++ ] = '\'';

    SetVexAtomBit( Node );
  }
  else
  {
    Buffer[ 0       ] = '"';
    Buffer[ Index++ ] = '"';

    SetVexAtomVector( Node );
  }

  Buffer[ Index ] = '\0';

  if ( Signed ) SetVexNodeSigned( Node );

  SetVexNodeAtom( Node );
  SetVexAtomValue( Node, namealloc( Buffer ) );
  SetVexAtomLiteral( Node );
  
  return( Node );
}
