/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include MUT_H 
#include MLO_H
#include "genscan.h"
#include "utile.h"
#include "scan_path.h"
#include "environment.h"


/**********************************************************************************************/
/*                                                                                            */
/*     this function gets from an input text file the order of the ffs in the scan path       */
/*            and constructs the scan chain                                                   */
/*                                                                                            */
/**********************************************************************************************/

chain_list *get_scan_order( filename, ptr_mem)
char *filename;
chain_list *ptr_mem;
{
	FILE *order_file;
	char s[255];
	int found,Line = 0;
	chain_list *ptr,*ptr1,*scan_order_chain = NULL;
	
    if (filename!=NULL)
	{
		order_file = fopen(filename,"r");
		if (order_file == NULL)
		{
			fprintf(stderr,"ERROR : file %s not found \n",filename);
			exit(1);
		}
		
		fscanf(order_file,"%s",s);

		while (!feof(order_file))
		{
			Line++;
		if (strlen(s) >255)
		{
			fprintf(stderr,"ERROR : instance name too long, line %d in file %s",Line,filename);
			exit(1);
		}
			if (strlen(s) > 0)
			{
				found = 0;
				for (ptr = ptr_mem; ptr ; ptr = ptr->NEXT)
				{
					if ( namealloc(s) == ((loins_list *)(ptr->DATA))->INSNAME)
					{
                                           for (ptr1 = scan_order_chain; ptr1 != NULL; ptr1 = ptr1->NEXT)
                                               if (((loins_list *)(ptr1->DATA))->INSNAME == ((loins_list *)(ptr->DATA))->INSNAME)
						{
						fprintf(stderr,"ERROR : duplicate instance %s line %d in file %s\n",((loins_list *)(ptr->DATA))->INSNAME,Line,filename);
						exit(1);
						}

						scan_order_chain = addchain(scan_order_chain,(void*)(((loins_list *)(ptr->DATA))));
						found = 1;
						break;
						
					}
				}
				if (!found)
				{
					fprintf(stderr,"ERROR : instance %s line %d in file %s not found \n",s,Line,filename);
					exit(1);
				}
			fscanf(order_file,"%s",s);
				
			}	
			else
			{
				printf("error in file %s line %d\n",filename,Line);
				exit(1);
			}
			
		}
		scan_order_chain = reverse(scan_order_chain);
	}
	else
	{ 
		
		for (ptr = ptr_mem; ptr ; ptr = ptr->NEXT)
		{
			scan_order_chain = addchain(scan_order_chain,(void*)(((loins_list *)(ptr->DATA))));
				

		}
		
		
	}
	return(scan_order_chain);
	
}


/*******************************************************************************************************************/
/*                                                                                                                 */
/*                   This function ( get_scan ) Gets a pointer to the scan chain of the netlist                    */
/*                                                                                                                 */
/*******************************************************************************************************************/

int _atoi(str)
char * str;
{
    char* st=str; 
	if (*str=='\0') return CVR_ERROR;
	while (*str!='\0')
	{
		if (!isdigit(*str++)) return CVR_ERROR;
	}
	return (atoi(st));
}

int get_ff_index(pti)
loins_list *pti;
{
	char *name = namealloc(pti->INSNAME);
    char *char_index;
	
	char_index = (char *)rindex(name,'_');
    if (char_index != NULL) 
	{
	char *pt;

	/* strdup() n'existe pas sur dec
        char_index = (strdup(++char_index));
        */

	pt = (char *)mbkalloc(strlen(char_index + 1) +1);
	strcpy(pt,char_index + 1);
        char_index = pt ;

		return(_atoi(char_index));
	}
	
	else
	{
		return(CVR_ERROR);
	}
	
}

chain_list *get_scan(ptr_fig,scan_in,scan_out)
lofig_list *ptr_fig;
losig_list *scan_in, *scan_out;
{
    loins_list *ptr;
    locon_list *ptr1;
    losig_list *ptr2;
	
    int in_scan_chain = 0;    /* true if the cell is in the scan chain */
    int not_found = 0 ;       /* true if a cell is not in the scan chain */
    int scan_nbr = mem_nbr;	  /* number of Dff in the scan chain */
    int exchange = 0 ;  /**/
    int not_found_scan_path = 0;
    int non_complete_scan_path = 0;
    chain_list *ptr_scan = NULL;
	
	/* pointer to the intenal signals of the scan Chain */
	
	losig_list  *scan_signal = scan_in;

	/* netlists containing SCAN_DFF only */
	
	do 
	{
		for(ptr = ptr_fig->LOINS ; ptr ; ptr = ptr->NEXT)
		{
			 if (ptr->FIGNAME == SCAN_FF)
			{
	        	printf("cell : %s , %s\n",ptr->INSNAME,ptr->FIGNAME);
				
				if (get_con(ptr,SCAN_FF_SCIN)-> SIG == scan_signal)
				{
					ptr_scan = addchain(ptr_scan,(void *)(ptr));
					in_scan_chain = 1;
				        ptr2 = get_con(ptr,SCAN_FF_OUT)->SIG;
				}
			}
		
			if (in_scan_chain) 
			{ 
				scan_signal = ptr2;
				nbr_scan++;
				break;
			}
		}
		
		if (in_scan_chain) {in_scan_chain = 0;}
		else not_found = 1;
		
	}	
	while ((scan_signal != scan_out) && (!not_found));
	
	
	if ((scan_signal != scan_out) || not_found) 
	{
	   not_found_scan_path = 1;
           fprintf(stderr,"ERROR : Path not found between %s and %s \n",scan_in,scan_out);	
	   exit(1);
	}
            ptr_scan = reverse(ptr_scan);
            display_loins_list(ptr_scan);
	    return(ptr_scan);
	
}

/***********************************************************************************************/
/*                                                                                             */
/*           gets a pointer to the list of sequential elements of the circuit                  */
/*                                                                                             */
/***********************************************************************************************/

chain_list *get_mem_cells(ptfig)
lofig_list *ptfig;
{
    struct chain *ptr = NULL;
	struct loins *pti;
	for(pti = ptfig->LOINS; pti ; pti = pti->NEXT)
	{
		
		if (is_tstate(pti))
		{
			printf(" tristate gates are not allowed \n");
			exit(1);
		}	
	
		if (is_mem(pti)) 
		{ 
			ptr = addchain(ptr,(void *)pti);
			mem_nbr++; 
			
		}
		
	}
	
	return(ptr);
}

/*********************************************************************************************************************/
/*                                                                                                                   */
/*                              scan insert function                                                                 */
/*                                                                                                                   */
/*                                                                                                                   */
/*********************************************************************************************************************/

void insert_scan(ptfig,comb_circuit,scanorder_filename,Fixed_Scout)
lofig_list *ptfig;
int comb_circuit;
char *scanorder_filename;
int Fixed_Scout;
{
	losig_list *scan_signal;
	locon_list *ptr_con, *ptr1_con;
	loins_list *ptr_ins;
	chain_list *ptr,*ptr1,*ptr2,*ptr3,*scan_order,*AuxInsCon= NULL;
	int _exchange;
	int found;
	int number_of_connectors = 0;  
	char *insname;  
	
	if (comb_circuit)
	{
		fprintf(stderr,"ERROR : Can't insert scan path in combinatory circuit !\n");
		fprintf(stderr,"Check environment variables for Dff models\n");
		exit(1);
	}
	else /* sequential circuit */
	{
		
		if (scan_in_sig == NULL)
			scan_in_sig = addlocon(ptfig,scan_in,addlosig(ptfig, get_index() ,addchain(NULL,(void *)scan_in),EXTERNAL),IN);
		else
		{
                  if (scan_in_sig->SIG == NULL) 
                      scan_in_sig->SIG  = addlosig(ptfig, get_index() ,addchain(NULL ,(void *)scan_in),EXTERNAL); 
		  else 
			if (Fixed_Scout == 0)
			{
			fprintf(stderr,"WARNING : signal %s already connected to internal logic\n ",scan_in);
			fprintf(stderr,"WARNING : removing logic connected between %s and %s \n", scan_in, scan_out);
			scan_signal = scan_in_sig->SIG;
			while (scan_signal != scan_out_sig->SIG )
			{found =0;
 			for (ptr_ins=(loins_list *)(ptfig->LOINS); ptr_ins != NULL; ptr_ins = ptr_ins->NEXT)
   			     {
				for (ptr_con=(locon_list *)(ptr_ins->LOCON); ptr_con != NULL; ptr_con = ptr_con->NEXT)
                		if (scan_signal == (losig_list *)(ptr_con->SIG))
                		{
			/* Verifier si l'instance n'est pas un buffer ou un inverseur */

					 for (ptr_con=(locon_list *)(ptr_ins->LOCON); ptr_con != NULL; ptr_con = ptr_con->NEXT)
					 if (ptr_con->DIRECTION == OUT)
				  	 { 
                			  	 fprintf(stderr,"\nWARNING : removing instance  %s\n",ptr_ins->INSNAME);
						 dellosig(ptfig,((losig_list *)(scan_signal))->INDEX);
						 scan_signal = (losig_list *)(ptr_con->SIG);
						 number_of_connectors = 0;
						 for (ptr1_con=(locon_list *)(ptr_ins->LOCON); ptr1_con != NULL; ptr1_con = ptr1_con->NEXT) number_of_connectors++;
						 if (number_of_connectors > 4) 
						 {
						 fprintf(stderr,"ERROR : model %s is not allowd to be used between %s and %s", ptr_ins->FIGNAME,scan_in_sig->NAME,scan_out_sig->NAME);	
						 exit(1);
						}
					 	 delloins(ptfig,ptr_ins->INSNAME);
						 found =1;
						 break;
					 }
                		}
				
			     }
				if (found == 0)
				{ fprintf(stderr,"ERROR: no path between %s and %s\n",scan_in,scan_out);
				exit(1);
				}
			  }
			dellosig(ptfig,((losig_list *)(scan_out_sig->SIG))->INDEX);
                        scan_in_sig->SIG  = addlosig(ptfig, get_index() ,addchain(NULL ,(void *)scan_in),EXTERNAL); 
			scan_out_sig->SIG = NULL;
			fprintf(stderr,"WARNING : removing internal signal connected to  %s\n",scan_out_sig->NAME);	
		        }	
			else
			{
 			for (ptr_ins=(loins_list *)(ptfig->LOINS); ptr_ins != NULL; ptr_ins = ptr_ins->NEXT)
   			     {
				for (ptr_con=(locon_list *)(ptr_ins->LOCON); ptr_con != NULL; ptr_con = ptr_con->NEXT)
                		if (scan_in_sig->SIG == (losig_list *)(ptr_con->SIG))
                		{
				 AuxInsCon = addchain(AuxInsCon,ptr_con);
				break; 
				}
			     }

			}
		}
		
		
		if (test_sig == NULL )
			test_sig = addlocon(ptfig,test,addlosig(ptfig, get_index() ,addchain(NULL,(void *)namealloc("test")),EXTERNAL),IN);
		else
		{
		 if (test_sig->SIG == NULL) 
                      test_sig->SIG  = addlosig(ptfig, get_index() ,addchain(NULL ,(void *)namealloc("test")), EXTERNAL);
		else
			fprintf(stderr,"WARNING : signal %s already connected to internal logic\n ",test);
		}

        /*    Insert scan-path   */
		
		fprintf(stdout,"\nInserting scan_path  .......\n\n");
		        
			scan_order = get_scan_order(scanorder_filename,mem_ptr);

			ptr1 = scan_order;
					
			if(scanorder_filename == NULL)
			do 
			{
				_exchange = 0;
								
				for(ptr = ptr1 ; ptr->NEXT ;ptr = ptr->NEXT)
				{
					if (  get_ff_index((loins_list *)(ptr->DATA)) >get_ff_index((loins_list *)(ptr->NEXT->DATA)))
					{
						ptr->NEXT->NEXT = addchain(ptr->NEXT->NEXT,(void *)((loins_list *)(ptr->DATA)));
                        ptr1 = delchain (ptr1, ptr);
						_exchange = 1;
						break;						                        
					}

				}
			}
			while (_exchange );
		   
		
		scan_signal = scan_in_sig->SIG;
		
		/*display_loins_list(ptr1);*/

		for(ptr = ptr1 ; ptr ; ptr = ptr->NEXT)
		{
			if (((loins_list *)(ptr->DATA))->FIGNAME == D_FF)
			{
				ptr3 = NULL;
				ptr3 = addchain(ptr3,(void *)(vss_sig->SIG));
				ptr3 = addchain(ptr3,(void *)(vdd_sig->SIG));
				for (ptr_con = (locon_list *)(getlofig(SCAN_FF,'A')->LOCON); ptr_con != NULL;ptr_con = ptr_con->NEXT)
					{ 
						if (ptr_con->NAME == SCAN_FF_IN)
				ptr3 = addchain(ptr3,(void *)(get_con((loins_list *)(ptr->DATA),FF_IN)->SIG));
						else if (ptr_con->NAME == SCAN_FF_SCIN)
				ptr3 = addchain(ptr3,(void *)(scan_signal));
						else if (ptr_con->NAME == SCAN_FF_SE)
				ptr3 = addchain(ptr3,(void *)(test_sig->SIG));
						else if (ptr_con->NAME == SCAN_FF_CLOCK)
				ptr3 = addchain(ptr3,(void *)(get_con((loins_list *)(ptr->DATA),FF_CLOCK)->SIG));
						else if (ptr_con->NAME == SCAN_FF_OUT)
				ptr3 = addchain(ptr3,(void *)(get_con((loins_list *)(ptr->DATA),FF_OUT)->SIG));
					}
				ptr3 = reverse(ptr3);
				scan_signal = get_con((loins_list *)(ptr->DATA),FF_OUT)->SIG;
				
				/* scan_ff instantiating */
                
               			insname = ((loins_list *)(ptr->DATA))->INSNAME;
				delloins(ptfig,((loins_list *)(ptr->DATA))->INSNAME);
				fprintf(stdout," instence %s is inserted in scan-path\n",insname);
				addloins(ptfig,insname,getlofig(SCAN_FF,'A' ), ptr3);
				freechain(ptr3);
				
			}
			else
			{
			fprintf(stderr,"ERROR : Unknown model of instance %s\n",((loins_list *)(ptr->DATA))->INSNAME);
			exit(1);
			}
	   		
		} 
        freechain(ptr1);
	
	if ( AuxInsCon == NULL )
	{
	for (ptr_con=(locon_list *)(ptfig->LOCON); ptr_con != NULL; ptr_con = ptr_con->NEXT)
	{
		if (scan_signal == (losig_list *)(ptr_con->SIG))	
		{
		fprintf(stderr,"\nWARNING : ScanOut connector is %s\n",ptr_con->NAME);
        	fprintf(stdout,"\ndone ! inserting scan-path\n");
		return;
		} 
	}
			
	if (scan_out_sig == NULL)
	  scan_out_sig = addlocon(ptfig,scan_out,NULL,INOUT);
	else
	{
		if( scan_out_sig->SIG != NULL)
		{
		fprintf(stderr,"ERROR : signal %s already connected to an output\n ",scan_out);
		exit(1);
		}
	}
	scan_signal->TYPE = EXTERNAL;  
	scan_out_sig->SIG = scan_signal;
	scan_out_sig->DIRECTION = INOUT;
	/*get_special_signals(ptfig);
	mem_ptr = get_mem_cells(ptfig);	*/
	}
	else
	{
	((losig_list *)((locon_list *)AuxInsCon->DATA)->SIG) = scan_signal; 
		
	}

        fprintf(stdout,"\n done ! inserting scan-path\n");
	}
}
